Require Import VST.floyd.proofauto.
Local Open Scope logic.
(*Model-level representation of master and clock*)

Definition Time : Type := Z.
Parameter ClockID : Type.
Definition ClockRepTp:Type := ClockID * Time * val.

Record ClockAPD := {
  Crep: ClockRepTp -> val -> mpred;
(*Not needed but sensible emreging property, ie pattern-level history invariant?
 Crep_wf: forall i1 tv1 c1 i2 tv2 c2, (Crep (i1,tv1) c1 * Crep (i2,tv2) c2) |-- !!(i1=i2 <-> c1=c2)*)
}.

(*MyTime, and for each potential clock the most recent sync time*)
Definition MasterRepTp:Type := Time * (ClockID -> option Time).

Definition emptyMaster: (ClockID -> option Time) := fun i => None.

Definition newMaster: MasterRepTp:= (0, emptyMaster).

Definition sync (T:MasterRepTp):MasterRepTp :=
  match T with (t,M) =>
    (t, fun s => if zeq s t then S (M s) else M s)
  end.

Definition kill (T:MasterRepTp) (i:nat):MasterRepTp :=
  match T with (t,L) =>
    (t, fun j => if beq_nat i j then None else L j)
  end.


Definition newEntry (T:MasterRepTp) (C:ClockRepTp) (U:MasterRepTp) (m:val):Prop :=
  match T with (t, L) =>
    exists i, L i = None /\ U = upd T i /\ C = (t,i,m)
  end.

(*Master can hand out a token of friendship (whose type is secret to master).
  But note that neither MToken nor MFriend show up in any ASI specs of 
  neither Master nor Clock, nor in the type ClockRepTp . Instead, we'll roll
  holding the friendship token into the definition of Crep (so so the
  clock APD instance will be parametric in a Master APD instance), and model
  the handing over of tokens using the entailments Mrep_connect (unifies BarnettNaumann's attch with connect)
  and Mrep_detach (corresponds to BN's detach). These entailments play the role of BN's verification-only functions.
  TODO: can the types ClockID and MToken be unified?*)
Record MasterAPD := {
  Mrep: MasterRepTp -> val -> mpred;
  Mrep_ptrnull: forall T m, Mrep T m |-- !!(is_pointer_or_null m);

  Mrep_range: forall t L m, Mrep (t,L) m |-- !!(0 <= t <= Int.max_signed);

  Mrep_finite: forall t L m, Mrep (t,L) m |-- !!exists i, L i = None;

  MToken: ClockID -> Type;
  (*A frienship token that Master hands to a Clock when the Clock connects,
    modeling the 'attach'; Needs to be presented for detaching*)
  MFriend: MToken -> val -> mpred;

  (*alternative: initialize i to 0 in L*)
  Mrep_connect: forall i t L m, Mrep T m |-- EX q, MFriend (q i) m * (Mrep (up q T) m)%logic;

  Mrep_detach: forall q T m, (MFriend q m * Mrep T m)%logic |-- Mrep (down q T) m;

  Mrep_wellsynced: forall t M m, Mrep (t,M) m |-- !!(forall s, t < s -> M s = 0);
}.

