Require Export VST.floyd.proofauto.

Lemma eqb_calling_convention_true c1 c2:
  true = eqb_calling_convention c1 c2 <-> c1=c2.
Proof.
  destruct c1; destruct c2.
  unfold eqb_calling_convention; simpl. split; intros.
+ apply andb_true_eq in H; destruct H.
  apply andb_true_eq in H0; destruct H0.
  symmetry in H; apply eqb_prop in H.
  symmetry in H0; apply eqb_prop in H0.
  symmetry in H1; apply eqb_prop in H1. subst. trivial.
+ inv H. rewrite 3 Raux.eqb_true; trivial.
Qed.

Fixpoint Forallb2 {A B} (f: A -> B -> bool) (l1 : list A) (l2: list B):bool :=
  match l1, l2 with
    nil, nil => true
  | (a1::t1), (a2::t2) => f a1 a2 && Forallb2 f t1 t2
  | _, _ => false
  end.

Lemma Forallb2_char {A B f}: forall l1 l2, @Forallb2 A B f l1 l2 = true <-> Forall2 (fun a b => f a b = true) l1 l2.
Proof.
induction l1; destruct l2; simpl; split; intros; auto; try discriminate.
inv H. inv H. 
apply andb_true_iff in H; destruct H. constructor; [ | apply IHl1]; trivial.
inv H. apply andb_true_iff; split; [ | apply IHl1]; trivial.
Qed.
 
Definition eqb_type_list:= Forallb2 eqb_type.

Lemma eqb_type_list_refl {l}: eqb_type_list l l = true.
Proof.
  induction l; simpl; trivial.
  apply andb_true_iff. split. apply eqb_type_refl. apply IHl.
Qed. 

Lemma eqb_type_list_char {l1 l2}: eqb_type_list l1 l2 = true <-> l1 =l2.
Proof.
  revert l2. induction l1; destruct l2; simpl; split; intros; trivial; try solve [inv H].
  apply andb_true_iff in H; destruct H. apply eqb_type_spec in H; subst. f_equal. apply IHl1. apply H0.
  inv H. apply andb_true_iff; split. apply eqb_type_refl. apply IHl1; trivial.
Qed.

Definition eqb_typesig (s t:compcert_rmaps.typesig):=
 andb (eqb_type (snd s) (snd t)) (eqb_type_list (fst s) (fst t)).

Lemma eqb_typesig_char s t: true = eqb_typesig s t <-> s=t.
Proof. unfold eqb_typesig. split; intros.
+ symmetry in H. rewrite andb_true_iff, eqb_type_spec, eqb_type_list_char in H; destruct H.
  destruct s; destruct t; simpl in *; subst; trivial.
+  subst. rewrite eqb_type_refl, eqb_type_list_refl; trivial.
Qed. 

Definition binary_intersection (phi psi:funspec): option funspec.
  destruct phi as [f c A1 P1 Q1 P1_ne Q1_ne].
  destruct psi as [f2 c2 A2 P2 Q2 P2_ne Q2_ne]. 
(*  destruct (eq_dec f f2); [subst f2 | apply None].*)
  remember (eqb_typesig f f2) as y.
  destruct y; [ | apply None]. apply eqb_typesig_char in Heqy; subst f2.
(*  destruct (eq_dec c c2); [subst c2 | apply None].*)
  remember (eqb_calling_convention c c2) as x.
  destruct x; [ | apply None].
  apply eqb_calling_convention_true in Heqx; subst c2.
  remember (binarySUMArgs P1 P2) as P.
  remember (binarySUM Q1 Q2) as Q.
  apply Some. apply (mk_funspec f c _ P Q).
  subst P; apply (binarySUMArgs_ne P1_ne P2_ne).
  subst Q; apply (binarySUM_ne Q1_ne Q2_ne).
Defined.

Lemma BIS phi psi: binary_intersection phi psi = seplog.binary_intersection phi psi.
destruct phi; destruct psi; simpl.
remember (eqb_typesig t t0).  destruct b; simpl.
+ apply eqb_typesig_char in Heqb. subst t0.
  remember (eqb_calling_convention c c0). destruct b; simpl.
  - apply eqb_calling_convention_true in Heqb; subst.
    rewrite 2 if_true; trivial.
  - rewrite if_true, if_false; trivial. intros N.
    apply eqb_calling_convention_true in N. congruence.
+ rewrite if_false; trivial. intros N. apply eqb_typesig_char in N. congruence. 
Qed.


Require Export MasterClockModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Clock. (*needed for function and type names *)

Global Open Scope funspec_scope.
Definition MasterTP:type:=  Tstruct _master noattr.
Definition ClockTP:type:=  Tstruct _clock noattr.

Section ClockASI.
  Variable M: MallocFreeAPD.
  Variable MS: MasterAPD.
  Variable CL: ClockAPD.

  Definition newClock_spec :=
    DECLARE _newClock
    WITH m:val, T: MasterRepTp, gv: globals
    PRE [tptr MasterTP ] 
    PROP() 
    PARAMS (m) GLOBALS (gv) 
    SEP(Mrep MS T m; mem_mgr M gv)
    POST[ tptr ClockTP ]
    EX c: val, EX C, EX U,
     PROP( newEntry T C U m)
     LOCAL(temp ret_temp c)
     SEP(Crep CL C c; Mrep MS U m; mem_mgr M gv).

  Definition sync_spec :=
    DECLARE _Sync
    WITH c:val, t: Time, i: ClockID, m:val, mt:Time, L:ClockID -> option Time, u:Time
    PRE [ tptr ClockTP ]
     PROP(L i <> None)
     PARAMS (c) GLOBALS ()
     SEP(Crep CL (t,i,m) c; Mrep MS (mt,L,u) m)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(Crep CL (mt,i,m) c; Mrep MS (upd (mt,L,u) i) m).

  Definition freeClock_spec:=
    DECLARE _freeClock 
    WITH gv: globals, c:val, t: Time, i: ClockID, m:val, T:MasterRepTp
    PRE [ tptr ClockTP ]
     PROP()
     PARAMS (c) GLOBALS (gv)
     SEP(Crep CL (t,i,m) c; Mrep MS T m; mem_mgr M gv)
    POST[ Tvoid ]
      PROP()
      LOCAL()
      SEP (Mrep MS (kill T i) m; mem_mgr M gv).

  Definition switchMaster_spec:=
    DECLARE _switchMaster 
    WITH c:val, t:Time, i:ClockID, mold:val, T:MasterRepTp, mnew:val, R:MasterRepTp
    PRE [ tptr ClockTP, tptr MasterTP]
     PROP()
     PARAMS (c;mnew) GLOBALS ()
     SEP(Crep CL (t,i,mold) c; Mrep MS T mold; Mrep MS R mnew)
    POST[ Tvoid ] EX C, EX U,
      PROP(newEntry R C U mnew)
      LOCAL()
      SEP (Crep CL C c; Mrep MS (kill T i) mold; Mrep MS U mnew).

  (*premits a purged clock to reconnect, but it'll receive
    a (likely) new ID*)
  Definition reconnect_spec:=
    DECLARE _switchMaster 
    WITH c:val, t:Time, i:ClockID, m:val, R:MasterRepTp
    PRE [ tptr ClockTP, tptr MasterTP]
     PROP()
     PARAMS (c;m) GLOBALS ()
     SEP(Crep CL (t,i,m) c; Mrep MS R m)
    POST[ Tvoid ] EX C, EX U,
      PROP(newEntry R C U m)
      LOCAL()
      SEP (Crep CL C c; Mrep MS U m(*; EX q, MFriend MS i q m*)).

  Definition switchM_spec: funspec.
  Proof.
    remember (binary_intersection (snd switchMaster_spec) (snd reconnect_spec)) as phi.
    destruct phi. apply f. simpl in Heqphi; congruence.
  Defined.

  Definition ClockASI:funspecs := [ freeClock_spec; newClock_spec; sync_spec; 
        (*switchMaster_spec*)(_switchMaster,switchM_spec) ].

End ClockASI.
