Require Export VST.floyd.proofauto.
Require Export MasterClockModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Master. (*needed for function and type names *)

Global Open Scope funspec_scope.
Definition MasterTP:type:=  Tstruct _master noattr.
Section MasterASI.
  Variable M: MallocFreeAPD.
  Variable MS: MasterAPD.

  Definition newMaster_spec :=
    DECLARE _newMaster
    WITH gv: globals, u:Time
    PRE [ ] PROP(0<u) PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr MasterTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(Mrep MS (newMaster u) p; mem_mgr M gv).

  (*In presence of timeout, ticking must involve purging*)
  Definition tick_spec :=
    DECLARE _Tick
    WITH m:val, n: Z, t:Time, L:ClockID -> option Time, u:Time
    PRE [ tptr MasterTP, tint]
     PROP(t <= t + n <= Int.max_signed)
     PARAMS (m;Vint (Int.repr n)) GLOBALS ()
     SEP(Mrep MS (t,L,u) m)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(Mrep MS (purge (t+n,L,u)) m).

  Definition getTime_spec :=
    DECLARE _getTime
    WITH m:val, T:MasterRepTp, i:nat
    PRE [ tptr MasterTP ]
     PROP((snd (fst T)) i <> None) (*ensures that clock with index i did register, and has not been purged*)
     PARAMS (m) GLOBALS ()
     SEP(Mrep MS T m)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr (fst (fst T)))))
      SEP(Mrep MS (upd T i) m).

  Definition reset_spec :=
    DECLARE _Reset
    WITH m:val, t:Time, u:Time
    PRE [ tptr MasterTP  ]
     PROP ()
     PARAMS (m) GLOBALS ()
     SEP (Mrep MS (t,emptyMaster,u) m)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (Mrep MS (newMaster u) m).

  Definition freeMaster_spec:=
    DECLARE _freeMaster 
    WITH gv: globals, m:val, T:MasterRepTp
    PRE [ tptr MasterTP ]
     PROP()
     PARAMS (m) GLOBALS (gv)
     SEP(Mrep MS T m; mem_mgr M gv)
    POST[ Tvoid ]
      PROP()
      LOCAL()
      SEP (mem_mgr M gv).

  Definition MasterASI:funspecs := [ freeMaster_spec; newMaster_spec; tick_spec; getTime_spec; reset_spec].

End MasterASI.
