Require Import VST.floyd.proofauto.
Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import surelyMalloc.
Global Open Scope funspec_scope.

Local Open Scope assert.

Section SM_ASI.
Variable M:MallocFreeAPD.

Definition surely_malloc_numericspec :=
 DECLARE _surely_malloc
   WITH n:Z, gv: globals
   PRE [ size_t ]
       PROP (0 <= n <= Ptrofs.max_unsigned)
       PARAMS (Vptrofs (Ptrofs.repr n))
       GLOBALS (gv)
       SEP (mem_mgr M gv)
    POST [ tptr tvoid ] EX p:_,
       PROP ()
       LOCAL (temp ret_temp p)
       SEP (mem_mgr M gv; malloc_token' M Ews n p; memory_block Ews n p).

Definition surely_malloc_spec {cs:compspecs} :=
  DECLARE _surely_malloc
   WITH t:type, gv: globals (*, cs:compspecs -- universe inconsistency*)
   PRE [ tuint ]
       PROP (0 <= @sizeof cs t <= Int.max_unsigned;
                @complete_legal_cosu_type (@cenv_cs cs) t = true;
                @natural_aligned cs natural_alignment t = true)
       PARAMS (Vint (Int.repr (@sizeof cs t)))
       GLOBALS (gv)
       SEP (mem_mgr M gv)
    POST [ tptr tvoid ] EX p:_,
       PROP ()
       LOCAL (temp ret_temp p)
       SEP (mem_mgr M gv; malloc_token M Ews t p; data_at_ Ews t p).

Definition SM_ASI : funspecs := [@surely_malloc_numericspec].

Lemma surelymalloc_spec_sub:
 forall {cs: compspecs}, 
   funspec_sub (snd surely_malloc_numericspec) (snd (@surely_malloc_spec cs)).
Proof.
do_funspec_sub. destruct w as [t gv]. clear H.
Exists (sizeof t, gv) emp. simpl; entailer!.
intros tau ? ?. Exists (eval_id ret_temp tau).
entailer!.
unfold malloc_token.
assert_PROP (field_compatible t [] (eval_id ret_temp tau)).
{ entailer!.
  apply malloc_compatible_field_compatible; auto. }
entailer!.
rewrite memory_block_data_at_; auto.
Qed.

Lemma natural_aligned_TarrayD {cs} t z a:
  @natural_aligned cs natural_alignment (Tarray t z a) = true ->
  @natural_aligned cs natural_alignment t = true.
Proof.
  intros. apply andb_true_iff. apply andb_true_iff in H; destruct H.
  simpl in *. split; trivial.
  unfold is_aligned, is_aligned_aux in *. simpl in *.
  apply andb_true_iff in H0. destruct H0. apply andb_true_iff in H0. destruct H0.
  apply andb_true_iff; split; trivial.
Qed.

Lemma ha_sub {cs cs'} (HSUB: ha_env_cs_sub (@ha_env_cs cs) (@ha_env_cs cs')): forall t
    (L: @complete_legal_cosu_type (@cenv_cs cs) t = true),
    hardware_alignof (@ha_env_cs cs) t = hardware_alignof (@ha_env_cs cs') t.
Proof.
  induction t; simpl; intros; auto.
  + remember ((@cenv_cs cs) ! i) as q. destruct q; [symmetry in Heqq | discriminate].
    specialize (HSUB i); red in HSUB.
    assert (exists ha : Z, (@ha_env_cs cs) ! i = @Some Z ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct H as [ha HA]. rewrite HA in *.
    rewrite HSUB; trivial.
  + remember ((@cenv_cs cs) ! i) as q. destruct q; [symmetry in Heqq | discriminate].
    specialize (HSUB i); red in HSUB.
    assert (exists ha : Z, (@ha_env_cs cs) ! i = @Some Z ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct H as [ha HA]. rewrite HA in *.
    rewrite HSUB; trivial.
Qed.
  
Lemma natural_aligned_cspecs_sub {cs cs'} (C: cspecs_sub cs cs'): forall t,
    @complete_legal_cosu_type (@cenv_cs cs) t = true ->
    @natural_aligned cs natural_alignment t = true ->
    @natural_aligned cs' natural_alignment t = true.
Proof.
  induction t; simpl; intros; auto.
  + specialize (IHt H (natural_aligned_TarrayD _ _ _ H0)). 
    apply andb_true_iff in H0; destruct H0. apply andb_true_iff in IHt; destruct IHt.
    apply andb_true_iff. simpl in *. split; trivial. 
    unfold is_aligned, is_aligned_aux in *. simpl in *.
    rewrite andb_true_r in *.
    apply andb_true_iff in H1; destruct H1.
    apply andb_true_iff. split; trivial.
    specialize (complete_legal_cosu_type_complete_type _ _ H); intros.
    destruct C as [C1 [C2 C3]].
    rewrite (@expr_lemmas4.cenv_sub_sizeof _ _ C1); trivial.
    rewrite <- (ha_sub C2); trivial.
  + unfold natural_aligned in *; simpl in *.
    remember ((@cenv_cs cs) ! i) as q; destruct q; [ symmetry in Heqq | discriminate].
    destruct C as [CEsub [HAsub LAsub]].
    specialize (HAsub i); red in HAsub.
    assert (HA: exists ha : Z, (@ha_env_cs cs) ! i = Some ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct HA as [ha HA]. rewrite HA in *.
    rewrite HAsub.
    apply andb_true_iff in H0; destruct H0.
    rewrite H0; simpl. 
    unfold is_aligned, is_aligned_aux in *.
    rewrite andb_true_r in *.
    specialize (LAsub i); red in LAsub.
    assert (LA: exists la, (@la_env_cs cs) ! i = Some la).
    { apply (@la_env_cs_complete cs i). eexists; eassumption. }
    unfold legal_alignas_type, legal_alignas_obs in *.
    destruct LA as [la LA]. rewrite LA in *. rewrite LAsub in *; trivial.
  + (*same proof as in previous case*)
    unfold natural_aligned in *; simpl in *.
    remember ((@cenv_cs cs) ! i) as q; destruct q; [ symmetry in Heqq | discriminate].
    destruct C as [CEsub [HAsub LAsub]].
    specialize (HAsub i); red in HAsub.
    assert (HA: exists ha : Z, (@ha_env_cs cs) ! i = Some ha).
    { apply (@ha_env_cs_complete cs i). eexists; eassumption. }
    destruct HA as [ha HA]. rewrite HA in *.
    rewrite HAsub.
    apply andb_true_iff in H0; destruct H0.
    rewrite H0; simpl. 
    unfold is_aligned, is_aligned_aux in *.
    rewrite andb_true_r in *.
    specialize (LAsub i); red in LAsub.
    assert (LA: exists la, (@la_env_cs cs) ! i = Some la).
    { apply (@la_env_cs_complete cs i). eexists; eassumption. }
    unfold legal_alignas_type, legal_alignas_obs in *.
    destruct LA as [la LA]. rewrite LA in *. rewrite LAsub in *; trivial.
Qed.

Lemma cspecs_sub_cs_preserve_type {cs1 cs2} (CSUB:cspecs_sub cs1 cs2) t x i
      (T: t = field_type i (co_members (get_co x))) coeq
  (C: coeq = cs_preserve_env cs1 cs2):
  cs_preserve_type cs1 cs2 coeq t = true.
Proof.
Abort.
(*
Lemma malloc_spec_cenv_sub_inv cs1 cs2 t (CSUB:cspecs_sub cs1 cs2):
  funspec_sub (snd (@malloc_spec cs2 t)) (snd (@malloc_spec cs1 t)).
Proof.
  intros.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intros gv. 
  Exists gv emp.
  rewrite !insert_SEP. 
  apply andp_right.
  { entailer!. rewrite (natural_aligned_cspecs_sub CSUB _ H0 H1).
    destruct CSUB as [CEsub [HAsub LAsub]].
    specialize (@semax.cenv_sub_complete_legal_cosu_type _ _ CEsub _ H0); intros. 
    specialize (complete_legal_cosu_type_complete_type _ _ H0); intros.
    specialize (complete_legal_cosu_type_complete_type _ _ H2); intros.
    rewrite (expr_lemmas4.cenv_sub_sizeof CEsub) in *; trivial.
    split; [ trivial | split3]; trivial. }
  apply prop_right.
  Intros p. Exists p.
  rewrite !insert_SEP.
  entailer!.
  destruct (eq_dec p nullval). trivial.
Abort. 
*)
(*
Lemma malloc_spec_cenv_sub cs1 cs2 t (CSUB:cspecs_sub cs1 cs2):
  funspec_sub (snd (@malloc_spec cs1 t)) (snd (@malloc_spec cs2 t)).
Proof.
  intros.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intros gv. 
  Exists gv emp.
  rewrite !insert_SEP. 
  apply andp_right.
  { entailer!. (* rewrite (natural_aligned_cspecs_sub CSUB _ H0 H1).*)
    destruct CSUB as [CEsub [HAsub LAsub]].
    specialize (@semax.cenv_sub_complete_legal_cosu_type _ _ CEsub).  _ H1); intros. 
    specialize (complete_legal_cosu_type_complete_type _ _ H0); intros.
    specialize (complete_legal_cosu_type_complete_type _ _ H2); intros.
    rewrite (expr_lemmas4.cenv_sub_sizeof CEsub) in *; trivial.
    split; [ trivial | split3]; trivial. }
  apply prop_right.
  Intros p. Exists p.
  rewrite !insert_SEP.
  entailer!.
  destruct (eq_dec p nullval). trivial.
*)
End SM_ASI.