Require Import VST.floyd.proofauto.
Require Import spec_stdlib. (*for malloc_token*)

Require Import MasterClockModel. (*the semantic constructions*)

Require Import Clock.
Notation MasterTP := (Tstruct _master noattr).
Notation ClockTP := (Tstruct _clock noattr).

Instance CCompSpecs : compspecs. make_compspecs prog. Defined.

Section ClockAPDef.
Variable M: MallocFreeAPD.
Variable MS: MasterAPD.

Definition ClockRepTp':Type := option Time * ClockID * val.
Definition Clock_rep' (C:ClockRepTp') (c: val): mpred :=
  match C with (topt,i,m) =>
   (malloc_token M Ews ClockTP c * (EX q, MFriend MS i q m) *
    match topt with
      Some t => @data_at CCompSpecs Ews ClockTP (Vint (Int.repr t), m) c
    | None => @data_at CCompSpecs Ews ClockTP (Vundef, m) c
    end)%logic
  end.

Definition Clock_rep (C:ClockRepTp) (c: val): mpred :=
  match C with (t,i,m) => Clock_rep' (Some t,i,m) c end.

Opaque Clock_rep'.
Opaque Clock_rep.

Definition ClockAP := Build_ClockAPD Clock_rep.

End ClockAPDef.

Require Import spec_surelyMalloc.
Require Import spec_Master.
Require Import spec_Clock.
Require Import VST.floyd.VSU.

Section ClockVSU.
Variable M: MallocFreeAPD.
Variable MS: MasterAPD.

  Notation CL := (ClockAP M MS). (*instantiate the ClockAP defined above to M*)

  Definition CLVprog : varspecs. mk_varspecs prog. Defined.

  Definition ClockImports := [free_spec' M; getTime_spec MS] ++ SM_ASI M.

  Definition sync_privatespec :=
    DECLARE _Sync
    WITH c:val, topt: option Time, i: ClockID, m:val, mt:Time, L:ClockID -> option Time, u:Time
    PRE [ tptr ClockTP ]
     PROP(L i <> None)
     PARAMS (c) GLOBALS ()
     SEP(Clock_rep' M MS (topt,i,m) c; Mrep MS (mt,L,u) m)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(Clock_rep' M MS (Some mt,i,m) c; Mrep MS (upd (mt,L,u) i) m).

  Definition clock_internal_specs: funspecs := (*ClockASI M MS CL.*)
    [ freeClock_spec M MS CL; newClock_spec M MS CL; sync_privatespec; (*switchMaster_spec MS CL*)(_switchMaster,switchM_spec MS CL)].

  Definition CLGprog:funspecs := ClockImports ++ clock_internal_specs.

  Lemma newClock_verif:
    semax_body CLVprog CLGprog f_newClock (newClock_spec M MS CL).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (ClockTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros c.
    forward. simpl.
    destruct T as [[t L] u]; simpl.
    sep_apply (Mrep_connect MS). Intros j q.
    remember (upd (t, L,u) j) as T. simpl in HeqT.
    destruct T as [[newt newL] newu]; inv HeqT.
    forward_call (c, @None Time, j, m, t, fun j0 : nat => if (j =? j0)%nat then Some t else L j0, u); simpl.
    { Exists q; cancel. }
    { rewrite Nat.eqb_refl; congruence. }
    forward.
    Exists c (t,j,m) (upd (t, L,u) j). simpl; entailer!.
    + exists j; eauto.
    + Exists x; cancel. apply derives_refl'. f_equal. f_equal. f_equal.
      extensionality. destruct (beq_nat j x0); trivial.
  Qed.

  Lemma freeClock_verif:
    semax_body CLVprog CLGprog f_freeClock (freeClock_spec M MS CL).
  Proof.
    start_function. simpl; Intros q.
    assert_PROP (isptr c) as cPtr by entailer!.
    destruct c; inv cPtr.
    forward_call (sizeof ClockTP, Vptr b i0, gv).
    { unfold malloc_token; simpl.
      destruct t; sep_apply data_at_memory_block; simpl; entailer!. }
    entailer!. apply MS.
  Qed.

  Lemma switchMaster_verif:
    semax_body CLVprog CLGprog f_switchMaster (switchMaster_spec MS CL).
  Proof.
    start_function. simpl; Intros q.
    forward.
    destruct R as [[r L] u].
    sep_apply (Mrep_connect MS); Intros j qnew.
    remember (upd (r, L,u) j) as R.
    destruct R as [[rr LL] uu]. inv HeqR.
    forward_call (c, Some t, j, mnew, r, fun j0 : nat => if (j =? j0)%nat then Some r else L j0, u).
    { simpl. Exists qnew. cancel. }
    { rewrite Nat.eqb_refl; congruence. }
    simpl fst. Exists (r, j, mnew) (upd (upd (r, L,u) j) j).
    entailer!.
    + red. exists j; split3; trivial.
      unfold upd. f_equal. f_equal. extensionality k.
      destruct ((j =? k)%nat); trivial.
    + apply MS.
  Qed.

  Lemma reconnect_verif:
    semax_body CLVprog CLGprog f_switchMaster (reconnect_spec MS CL).
  Proof.
    start_function. simpl; Intros q.
    forward.
    destruct R as [[r L] u].
    sep_apply (Mrep_connect MS); Intros j qnew.
    remember (upd (r, L,u) j) as R.
    destruct R as [[rr LL] uu]. inv HeqR.
    forward_call (c, Some t, j, m, r, fun j0 : nat => if (j =? j0)%nat then @Some Time r else L j0, u).
    { simpl. Exists qnew. cancel. }
    { rewrite Nat.eqb_refl; congruence. }
    Exists (r, j, m) (upd (r, fun j0 : nat => if (j =? j0)%nat then @Some Time r else L j0, u) j).
    entailer!.
    + red. simpl. exists j; split3; trivial.
      f_equal. f_equal. extensionality k.
      remember ((j =? k)%nat). destruct b; trivial.
    + apply MS. (* Exists q; trivial.*)
  Qed.

  Lemma switchbody: semax_body CLVprog CLGprog f_switchMaster (_switchMaster, switchM_spec MS CL).
  Proof.
    eapply (semax_body_binaryintersection _ _ _ _ switchMaster_verif reconnect_verif).
    rewrite <- BIS; reflexivity.
  Qed.

  Lemma Sync_privateverif:
    semax_body CLVprog CLGprog f_Sync sync_privatespec.
  Proof.
    start_function. simpl; Intros q.
    assert_PROP (is_pointer_or_null m) as mP.
    { sep_apply (Mrep_ptrnull MS); entailer!. }
    destruct topt. 
    + forward.
      forward_call (m, (mt,L,u), i).
      forward. entailer!. simpl. Exists q; cancel.
    + forward.
      forward_call (m, (mt,L,u), i).
      forward. entailer!. simpl; Exists q; cancel.
  Qed.

  Lemma sync_funspec_sub: funspec_sub (snd sync_privatespec) (snd (sync_spec MS CL)).
  Proof. do_funspec_sub. destruct w as [[[[[[c t] i] m] s] L] u]; simpl in *; Intros.
      subst. Exists (c, Some t, i, m, s,L,u) emp; simpl; entailer!.
  Qed.

  Goal (*Lemma Sync_verif:*)
    semax_body CLVprog CLGprog f_Sync (sync_spec MS CL).
  Proof.
    eapply semax_body_funspec_sub.
    + apply Sync_privateverif.
    + apply sync_funspec_sub.
    + LNR_tac.
  Qed.

  Definition ClockComponent: @Component NullExtension.Espec CLVprog _ 
      nil ClockImports prog (ClockASI M MS CL) emp clock_internal_specs.
  Proof. 
    mkComponent.
  + solve_SF_internal newClock_verif.
  + solve_SF_internal Sync_privateverif.
  + solve_SF_internal freeClock_verif.
  + (*solve_SF_internal switchMaster_verif.*)solve_SF_internal switchbody.
  + apply sync_funspec_sub.
  Qed.

  Definition ClockVSU: @VSU NullExtension.Espec CLVprog _ 
      nil ClockImports prog (ClockASI M MS CL) emp.
  Proof. eexists; apply ClockComponent. Qed.

End ClockVSU.
Definition Clock_CanVSU M MS:= VSU_to_CanonicalVSU  (ClockVSU M MS).