Require Import VST.floyd.proofauto.
Require Import spec_stdlib. (*for malloc_token*)

Require Import MasterClockModel. (*the semantic constructions*)

Require Import Master.
Notation MasterTP := (Tstruct _master noattr).

Instance MCompSpecs : compspecs. make_compspecs prog. Defined.

Section MasterAPDef.
Variable M: MallocFreeAPD.

Definition Master_rep (T:MasterRepTp) (m: val): mpred :=
  match T with
    (t,L,u) => EX maxID,
             !!(0<u /\ 0 <= t <= Int.max_signed /\
               (forall i ti, L i = Some ti -> (i <= maxID)%nat /\ t-u<= ti <= t))  &&
             (malloc_token M Ews MasterTP m *
             @data_at MCompSpecs Ews MasterTP (Vint (Int.repr t)) m)%logic
  end.

Lemma MrepPtrnull T m: Master_rep T m |-- !!(is_pointer_or_null m).
Proof. destruct T as [[t L] u]; simpl; Intros maxID. entailer!. Qed.

Lemma MrepRange t L u m: Master_rep  (t,L,u) m |-- !!(0<u /\ 0 <= t <= Int.max_signed).
Proof. simpl; Intros maxID; entailer!. Qed.

Lemma MrepFinite t L u m: Master_rep (t,L,u) m |-- !!(exists i, L i = None).
Proof. simpl; Intros maxID. entailer!. exists (S maxID).
  specialize (H1 (S maxID)). destruct (L (S maxID)); trivial.
  destruct (H1 _ (eq_refl _)). lia. 
Qed.

Lemma MrepFriend t L u m: Master_rep (t,L,u) m |-- !!(forall i ti, L i = Some ti -> t-u <= ti <= t).
Proof. simpl; Intros maxID; entailer!. intros ? ? Ti. apply (H1 _ _ Ti). Qed.

Definition Master_Token:Type := nat.

Definition Master_Friend (i:ClockID) (k:Master_Token) (m: val): mpred := (*!!(L i <> None)*)!!((i <= k)%nat) && emp.

Lemma MrepFriend_emp i q m: Master_Friend i q m |-- emp.
Proof. unfold Master_Friend; entailer!. Qed.

Lemma MrepConnect t L u m: Master_rep (t,L,u) m |-- EX i q, !!(L i = None) && Master_Friend i q m * Master_rep (upd (t,L,u) i) m.
Proof.
  assert_PROP(exists i, L i = None). apply MrepFinite. destruct H as [i Li].
  simpl; Intros maxID. unfold Master_Friend.
  Exists i (Nat.max maxID i) (Nat.max maxID i); entailer!.
  intros j tj Tj. remember (beq_nat i j); destruct b.
    - apply beq_nat_eq in Heqb. inv Tj. lia.
    - destruct (H1 _ _ Tj); lia.
Qed.

Lemma MrepDetach: forall i q T m, (Master_Friend i q m * Master_rep T m)%logic |-- Master_rep (kill T i) m.
Proof. intros. destruct T as [[t L] u]; unfold Master_Friend; simpl; Intros maxID.
  Exists maxID; entailer!.
  intros j tj Tj. remember (beq_nat i j); destruct b; inv Tj. auto.
Qed.

Lemma MrepRely T m: (Master_rep T m)%logic |-- Master_rep (purge T) m.
Proof. destruct T as [[t L] u]; simpl; Intros maxID.
  Exists maxID; entailer!.
  intros j tj Tj. specialize (H1 j). destruct (L j); [|inv Tj].
  destruct (zlt (t0 + u) t); inv Tj. apply H1; trivial.
Qed.

Opaque Master_rep.

Definition MasterAP := Build_MasterAPD _ MrepPtrnull MrepRange _ _ MrepFriend_emp MrepConnect MrepDetach MrepRely MrepFriend.

End MasterAPDef.

Require Import spec_surelyMalloc.
Require Import spec_Master.
Require Import VST.floyd.VSU.

Section MasterVSU.
Variable M: MallocFreeAPD.

  Notation MS := (MasterAP M). (*Instantiate the MasterAP defined above to M*)

  Definition MSVprog : varspecs. mk_varspecs prog. Defined.

  Definition MasterImports := [free_spec' M] ++ SM_ASI M.

  Definition master_internal_specs: funspecs := MasterASI M MS.

  Definition MSGprog:funspecs := MasterImports ++ master_internal_specs.

  Lemma newMaster_verif:
    semax_body MSVprog MSGprog f_newMaster (newMaster_spec M MS).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (MasterTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros m.
    do 2 forward. Exists m; entailer!. simpl.
    Exists O; entailer!. unfold emptyMaster; congruence.
  Qed.

  Lemma freeMaster_verif:
    semax_body MSVprog MSGprog f_freeMaster (freeMaster_spec M MS).
  Proof.
    start_function. destruct T as [[t L] u]. simpl; Intros maxID.
    assert_PROP (isptr m) as mPtr by entailer!.
    destruct m; inv mPtr.
    forward_call (sizeof MasterTP, Vptr b i, gv).
    { sep_apply data_at_memory_block; unfold malloc_token. simpl; entailer!. }
    entailer!.
  Qed.

  Lemma Tick_verif:
    semax_body MSVprog MSGprog f_Tick (tick_spec MS).
  Proof.
    start_function. rename H into Bound. simpl; Intros maxID.
    rename H into U; rename H0 into MR; rename H1 into TI.
    forward. forward.
    entailer!. simpl. Exists maxID. entailer!.
    intros. specialize (TI i). destruct (L i); [ | inv H2].
    destruct (TI _ (eq_refl _)); clear TI H H0 H1 PNm.
    destruct (zlt (t0 + u) (t + n)); inv H2. lia.
  Qed.

  Lemma getTime_verif:
    semax_body MSVprog MSGprog f_getTime (getTime_spec MS).
  Proof. start_function. rename H into Ifresh.
    destruct T as [[t L] u]; simpl in Ifresh. simpl; Intros maxID.
    rename H into U; rename H0 into MR; rename H1 into TI.
    do 2 forward.
    simpl. Exists maxID. entailer!.
    intros j tj Tj. remember (beq_nat i j); destruct b; eauto.
    apply beq_nat_eq in Heqb. inv Tj. specialize (TI j).
    destruct (L j); [ destruct (TI _ (eq_refl _)); lia | congruence].
  Qed.

  Lemma Reset_verif: semax_body MSVprog MSGprog f_Reset (reset_spec MS).
  Proof.
    start_function. simpl; Intros maxID.
    rename H into MR; rename H0 into TI.
    forward.
    entailer!; simpl. Exists maxID. entailer!. unfold emptyMaster; congruence.
  Qed.

  Definition MasterComponent: @Component NullExtension.Espec MSVprog _ 
      nil MasterImports prog (MasterASI M MS) emp master_internal_specs.
  Proof. 
    mkComponent.
  + solve_SF_internal newMaster_verif.
  + solve_SF_internal Tick_verif.
  + solve_SF_internal getTime_verif.
  + solve_SF_internal Reset_verif.
  + solve_SF_internal freeMaster_verif.
  Qed.

  Definition MasterVSU: @VSU NullExtension.Espec MSVprog _ 
      nil MasterImports prog (MasterASI M MS) emp.
  Proof. eexists; apply MasterComponent. Qed.

End MasterVSU.
Definition Master_CanVSU M := VSU_to_CanonicalVSU  (MasterVSU M).