Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import VST.veric.initial_world.

Require Import VST.floyd.linking.
Require Import spec_stdlib.
Require Import verif_stdlib.
Require Import spec_surelyMalloc.
Require Import verif_surelyMalloc.

Require Import spec_Master.
Require Import spec_Clock.
Require Import verif_Master.
Require Import verif_Clock.

Section CORE_VSU.
Variable M: MallocFreeAPD.

Definition MS := MasterAP M.

Definition CL := ClockAP M MS. 

Definition mc_prog': Clight.program := 
  match link_progs Master.prog Clock.prog with
    Errors.OK p => p
  | _ => Master.prog (*arbitrary*)
  end.
Definition mc_prog := ltac:
  (let x := eval hnf in mc_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance mc_cs : compspecs. make_compspecs mc_prog. Defined.

Definition mc_Vprog:varspecs. mk_varspecs mc_prog. Defined.

Definition mc_Imports:funspecs := [free_spec' M; surely_malloc_numericspec M].

Definition mc_Exports:funspecs := G_merge (MasterASI M MS) (ClockASI M MS CL).

Definition MC_VSU:
@VSU NullExtension.Espec
   mc_Vprog mc_cs nil mc_Imports mc_prog mc_Exports emp.
Proof.
  VSULink_tac (MasterVSU M) (ClockVSU M MS).
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

Definition sm_mc_prog': Clight.program := 
  match link_progs surelyMalloc.prog mc_prog with
    Errors.OK p => p
  | _ => surelyMalloc.prog (*arbitrary*)
  end.
Definition sm_mc_prog := ltac:
  (let x := eval hnf in sm_mc_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance sm_mc_cs : compspecs. make_compspecs sm_mc_prog. Defined.

Definition sm_mc_Vprog:varspecs. mk_varspecs sm_mc_prog. Defined.

Definition sm_mc_Imports:funspecs := (*MF_ASI.*) [free_spec' M; malloc_spec' M; exit_spec].

Definition sm_mc_Exports:funspecs := G_merge mc_Exports (SM_ASI M).

Definition SM_MC_VSU:
@VSU NullExtension.Espec
   sm_mc_Vprog sm_mc_cs nil sm_mc_Imports sm_mc_prog sm_mc_Exports emp.
Proof.
  VSULink_tac MC_VSU (SMVSU M).
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

Definition core_prog': Clight.program := 
  match link_progs stdlib.prog sm_mc_prog with
    Errors.OK p => p
  | _ => stdlib.prog (*arbitrary*)
  end.
Definition coreprog := ltac:
  (let x := eval hnf in core_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance coreCS : compspecs. make_compspecs coreprog. Defined.

Definition coreVprog:varspecs. mk_varspecs coreprog. Defined.

Definition coreImports:funspecs := nil.

Definition coreExports:funspecs := G_merge (MallocFreeASI M) sm_mc_Exports.
Definition coreBuiltins:funspecs := (MallocFreeASI M). (*i.e. MF_E*)

Definition Core_InitPred (gv: globals): mpred:= emp.

Definition Core_VSU:
@VSU NullExtension.Espec coreVprog coreCS coreBuiltins coreImports coreprog coreExports Core_InitPred.
Proof.
  VSULink_tac (MallocFreeVSU M) SM_MC_VSU.
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

Definition Core_CanVSU: @CanonicalVSU NullExtension.Espec
   coreVprog coreCS coreBuiltins coreImports coreprog coreExports Core_InitPred.
Proof. apply (VSU_to_CanonicalVSU Core_VSU). Qed.

End CORE_VSU.