Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import spec_Master.
Require Import spec_Clock.
Require surelyMalloc.
Require Import main.
Require Import spec_main.

Section MainInstASI.
  Variable M: MallocFreeAPD.
  Variable G: list (globals -> mpred).

(*Specification of "module-instantiated main" - we permit an arbitrary G here
 (eventually, we'll roll M (or better: memmrg M gv) into G, too.*)
Definition main_inst_spec:=
 DECLARE _main
 WITH gv: globals
 PRE [ ] (*(main_preinst tt G gv)*)
    PROP ()
    PARAMS () GLOBALS (gv)
    SEP (fold_right (fun g p => g gv * p) emp G; has_ext tt)
 POST[ tint ]
    PROP()
    LOCAL(temp ret_temp (Vint (Int.repr 6)))
    SEP(mem_mgr M gv; has_ext tt).
End MainInstASI.

Definition linked_prog : Clight.program :=
 ltac: (linking.link_progs_list [
   stdlib.prog; surelyMalloc.prog; Master.prog; Clock.prog; main.prog]).

Instance LinkedCompSpecs : compspecs. make_compspecs linked_prog. Defined.

Definition LinkedVprog : varspecs. mk_varspecs linked_prog. Defined.

(*Instantiating main_spec with linked_prog rather than main.prog ensures that all 
  gv's are present after we do start_function in body_main*)
Notation mainspec M := (main_spec M linked_prog).

(*We still verify the body of main abstractly in OP and SP*)
Section MAIN.
  Variable M: MallocFreeAPD.
  Variable MS: MasterAPD.
  Variable CL: ClockAPD.
  Definition MS_ASI:= MasterASI M MS.
  Definition CL_ASI:= ClockASI M MS CL.

  Definition MainImports := [ newMaster_spec M MS; tick_spec MS; reset_spec MS;
   newClock_spec M MS CL; sync_spec MS CL; freeClock_spec M MS CL; freeMaster_spec M MS].

  Definition MyInitPred: list (globals -> mpred) := [].
  Definition maininstspec := (main_inst_spec M MyInitPred).

  Definition main_internal_specs: funspecs := [mainspec M].

  Definition MainGprog:funspecs := MainImports ++ main_internal_specs.

  Variable u:Time.
  Variable timeoutPos: (*0<u.*)10<=u.
  (* Instance MainCompSpecs : compspecs. make_compspecs main.prog. Defined.
     Definition MainVprog : varspecs. mk_varspecs main.prog. Defined.*)

  (*As explained in Pile, the fact the we use linked_prog in main_spec
    generally means we need to use LinkedCompSpecs and LinkedVprog in body_main.
    In this particular, case, we could just as easily use MainCompSpecs and MainVprog, too*)
(*  Lemma body_main: semax_body MainVprog MainGprog f_main (mainspec M).*)
  Lemma body_main: semax_body LinkedVprog MainGprog f_main maininstspec.
  Proof.
  start_function.
  sep_apply (spec_stdlib.make_mem_mgr M gv). 

  (*newMaster - now needs a timeout*)
  forward_call (gv,u). lia. Intros m. unfold newMaster; simpl.

  (*Tick (m, 5*)
  forward_call (m, 5, 0, emptyMaster, u);  simpl. rep_lia.

  (*Reset(m)*)
  forward_call (m, 5, u). { unfold emptyMaster; cancel. }

  (*New Clock(m)*)
  forward_call. Intros x. destruct x as [[c1 C1] [t L]].
  simpl in H. destruct H as [ID1 [HID1 [X Y]]]; subst C1. inv X. simpl.

  (*Tick(m,7)*)
  forward_call (m, 7, 0, fun j : nat => if (ID1 =? j)%nat then @Some Time 0 else emptyMaster j,u); simpl.
  { rep_lia. }

  remember (fun i : ClockID =>
   match (if (ID1 =? i)%nat then @Some Time 0 else emptyMaster i) with
   | Some s => if zlt (s + u) 7 then None else @Some Time s
   | None => None
   end) as T.
  (*Sync(c1)*)
  forward_call (c1, 0, ID1, m, 7, T, u); simpl.
  { subst T; rewrite Nat.eqb_refl, if_false. congruence. lia. }

  (*Tick m, 3*)
  forward_call (m, 3, 7, fun j : nat => if (ID1 =? j)%nat then @Some Time 7 else T j, u); simpl.
  { rep_lia. }

  (*newClock(m)*)
  forward_call. Intros x. destruct x as [[c2 C2] [[t L] uu]].
  simpl in H. destruct H as [ID2 [HID2 [X Y]]]; subst C2. inv X. simpl.

  remember ((ID1 =? ID2)%nat); destruct b. rewrite if_false in HID2. congruence. lia.

  (*Sync(c1)*)
  forward_call (c1, 7, ID1, m, 10,fun j : nat =>
   if (ID2 =? j)%nat
   then @Some Time 10
   else
    match
      (if (ID1 =? j)%nat
       then @Some Time 7
       else
        match (if (ID1 =? j)%nat then @Some Time 0 else emptyMaster j) with
        | Some s => if zlt (s + u) 7 then None else @Some Time s
        | None => None
        end)
    with
    | Some s => if zlt (s + u) 10 then None else @Some Time s
    | None => None
    end, u); simpl.
  { rewrite NPeano.Nat.eqb_sym, <- Heqb, Nat.eqb_refl, if_false. congruence. lia. }

  (*Reset(m)*)
  forward_call (m, 10, u).
  { unfold emptyMaster in *. simpl in *. 
  Admitted. (*Indeed, rightly fails here*)

End MAIN.

Require Import verif_core.
Require Import VST.veric.initial_world.

(*Finally, we assert existence of a mallocfree library.*)
Parameter M: MallocFreeAPD.

Lemma tc_VG: tycontext_subVG LinkedVprog (MainGprog M (MS M) (CL M))
  LinkedVprog (mainspec M :: coreExports M).
Proof. split.
          * intros i. red. rewrite 2 semax_prog.make_context_g_char, 2 make_tycontext_s_find_id by LNR_tac.
            remember (find_id i (MainGprog M (MS M) (CL M))) as w.
            destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
            repeat (if_tac in Heqw; [ subst i; inv Heqw; reflexivity |]).
            congruence.
          * intros i; red. rewrite 2 make_tycontext_s_find_id.
            remember (find_id i (MainGprog M (MS M) (CL M))) as w.
            destruct w; [symmetry in Heqw | trivial]. simpl in Heqw.
            repeat (if_tac in Heqw; [ subst i; inv Heqw;
                                      eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                                    | ]).
            congruence.
Qed.

Definition ExtIDtable : PTree.t unit :=
  ltac:(let x := constr:(fold_right (fun i t => PTree.set i tt t) (PTree.empty _) (ExtIDs linked_prog))
           in let x := eval compute in x
           in exact x).

Definition in_ExtIDs (ia: ident*funspec) : bool :=
 match PTree.get (fst ia) ExtIDtable with Some _ => true | None => false end.

Definition LinkedSYS_pre:funspecs :=
   filter in_ExtIDs (augment_funspecs linked_prog (MallocFreeASI M)).

Definition LinkedSYS:funspecs := 
  ltac: 
    (let x := eval hnf in LinkedSYS_pre in
     let x := eval simpl in x in 
     (*let x := eval compute in x in leaving this in leads Coq to timeout/crash*)
       exact x).

Lemma ApplibSys_in_LinkedSys: forall i phi, find_id i (MallocFreeASI M) = Some phi -> find_id i LinkedSYS = Some phi.
  Proof. (*Boilerplate proof*)
    intros. specialize (find_id_In_map_fst _ _ _ H); intros.
    simpl in H0. repeat (destruct H0 as [HO | H0]; [ subst i; inv H; reflexivity |]). contradiction.
  Qed. 

Lemma LinkedSYS_External: forall i, In i (map fst LinkedSYS) ->
      exists ef ts t cc, find_id i (prog_defs linked_prog) = Some (Gfun (External ef ts t cc)).
  Proof. (*Boilerplate proof*)
    intros. cbv in H. 
    repeat (destruct H as [H | H];
      [ subst; try solve [do 4 eexists; split; reflexivity ]
      | ]).
    contradiction.
  Qed.

Lemma LinkedSYS_vacuous i phi: find_id i LinkedSYS = Some phi -> find_id i (coreBuiltins M) = None ->
        exists ef argsig retsig cc, 
           phi = vacuous_funspec (External ef argsig retsig cc) /\ 
           find_id i (prog_funct coreprog) = Some (External ef argsig retsig cc) /\
           ef_sig ef = {| sig_args := typlist_of_typelist argsig;
                          sig_res := rettype_of_type retsig;
                          sig_cc := cc_of_fundef (External ef argsig retsig cc) |}.
  Proof. intros. specialize (find_id_In_map_fst _ _ _ H); intros.
    cbv in H1.
    Time repeat (destruct H1 as [H1 | H1]; 
      [ subst; inv H; try solve [do 4 eexists; split3; reflexivity]
      | ]). (*3s*)
    inv H0. inv H0. inv H0. contradiction.
  Qed.

Lemma disjoint_Vprog_linkedfuncts: 
      list_disjoint (map fst LinkedVprog) (map fst (prog_funct linked_prog)).
Proof.
  intros x y X Y ?; subst x; cbv in X; apply assoclists.find_id_None_iff in Y; [ trivial | clear H Y];
  repeat (destruct X as [X | X]; [ subst y; cbv; reflexivity |]); contradiction.
Qed.

Definition Imports:funspecs:=nil.
(*
Lemma CSSUB: cspecs_sub MainCompSpecs LinkedCompSpecs.
Proof.
  split3.
+ intros i; red; remember ((@cenv_cs MainCompSpecs) ! i) as w; destruct w; 
   [symmetry in Heqw; simpl in Heqw; rewrite PTree.gleaf in Heqw; congruence | trivial].
+ intros i. red. remember ((@ha_env_cs MainCompSpecs) ! i) as w. destruct w; [symmetry in Heqw | trivial].
  simpl in Heqw. rewrite PTree.gleaf in Heqw. congruence.
+ intros i. red. remember ((@la_env_cs MainCompSpecs) ! i) as w. destruct w; [symmetry in Heqw | trivial].
  simpl in Heqw. rewrite PTree.gleaf in Heqw. congruence.
Qed.*)

Lemma main_sub: funspec_sub (snd (main_inst_spec M MyInitPred))
                             (snd (mainspec M)).
Proof.
  do_funspec_sub. unfold main_pre; simpl; Intros; subst. clear. 
  Exists (initialize.genviron2globals g) emp. entailer!.
  + intros; entailer!.
Qed.

Require Import VST.floyd.VSU_addmain.

Parameter u:Time.
Parameter timeoutPos: (*0<u.*)10<=u.

Definition MasterClock_VSU: @LinkedProgVSU NullExtension.Espec LinkedVprog LinkedCompSpecs
      LinkedSYS Imports linked_prog [mainspec M] Core_InitPred.
Proof.
 VSUAddMain_tac (Core_CanVSU M).
   + apply disjoint_Vprog_linkedfuncts.
   + apply LinkedSYS_External.
   + eapply semax_body_subsumption.
       * eapply semax_body_funspec_sub. 
         - apply (body_main M (MS M) (CL M) u). apply timeoutPos.
         - apply main_sub.
         - LNR_tac.
       * apply tycontext_sub_i99. apply tc_VG.
   + apply LinkedSYS_vacuous.
Qed.

Lemma prog_correct:
  exists G, 
 @semax_prog NullExtension.Espec LinkedCompSpecs linked_prog tt LinkedVprog G.
Proof.
(*boilerplate: same proof script as in VSUpile, connpool*)
  destruct MasterClock_VSU as [G Comp MAIN]. exists G. 
  assert (DomG: map fst G = map fst (prog_funct linked_prog)).
  { destruct Comp. unfold Comp_G in *. rewrite CC_canonical.
    cbv; reflexivity. }
  Time prove_linked_semax_prog. (*7s*)
  all: rewrite augment_funspecs_eq by trivial.
  apply (@Canonical_semax_func _ _ _ _ _ _ _ _ Comp); cbv; reflexivity.
  destruct MAIN as [post [MainG MainExp]]. inv MainExp. rewrite MainG; eexists; reflexivity.
Qed.