Require Import VST.floyd.VSU.
Require Import stdlib.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Subject.
Require Import Observer. (*needed for defining Vprog *)
Require Import spec_Observer.

Require Import verif_Observer. (*get access to the definition of ObserverREP etc*)
Opaque Observer_rep.
Opaque Notify_rep.
(*Opaque ValPre.*)
 
Ltac do_compute_expr_warning::=idtac.

Section Observer_VSU.
  Variable M: MallocFreeAPD.
  Variable SP: SubjectAPD.
  Notation S_rep := (Subjectrep SP).
  Notation RegPre := (RegisterPre SP).
  Notation GetPre := (GetPre SP).
  Notation GetPost := (GetPost SP).
  Notation S_ptrnull:= (Subjectrep_ptrnull SP).

  Notation OP:= (ObserverAP M). (*Instantiate the ObserverAP definined in verif_observer to M*)  

  Definition subject_ASI := SubjectASI M SP OP.

  Definition ObserverImports:= 
           [free_spec' M _free] ++ SM_ASI M ++ [register_spec M SP OP; get_smallspec SP].

  Definition obs_internal_specs: funspecs :=
    [newObserver_spec M OP SP; notify_smallspec OP SP; val_spec OP; freeObserver_spec M OP].

  Definition OVprog : varspecs. mk_varspecs prog. Defined.

  Definition OGprog:funspecs := ObserverImports ++ obs_internal_specs.

  Lemma val_verif:
    semax_body OVprog OGprog f_val (val_spec OP).
  Proof. start_function. simpl. (*
    assert_PROP (Int.min_signed <= v1 <= Int.max_signed /\ Int.min_signed <= v2 <= Int.max_signed).
    { sep_apply Observer_rep_Range. entailer!. }*)
    rewrite Observer_rep_char. Intros u1 u2 c1 c2. inv H0. do 2 forward. forward.
    simpl. rewrite Observer_rep_char. Exists u1 u2 c1 c2. entailer!. 
  Qed.

  Lemma freeObserver_verif:
    semax_body OVprog OGprog f_freeObserver (freeObserver_spec M OP).
  Proof.
   start_function. simpl. rewrite Observer_rep_char; Intros u1 u2 c1 c2; inv H.
   rewrite data_at_isptr; unfold malloc_token; Intros.
   destruct o; try solve [inv Po]. 
   forward_call (sizeof ObserverTP, Vptr b i, gv).
   { sep_apply data_at_memory_block. simpl. cancel. }
   forward.
  Qed.
 
  Lemma notify_verif: semax_body OVprog OGprog f_notify (notify_smallspec OP SP).
  Proof.
    start_function. simpl. rewrite Notify_rep_char. Intros u1 u2 bb1 bb2 v1 v2. symmetry in H; inv H.
    sep_apply (GetPre_ptrnull SP S1 s1); Intros. sep_apply (GetPre_ptrnull SP S2 s2); Intros.
    assert_PROP (Int.min_signed <= snd S1 <= Int.max_signed) by (sep_apply (GetPre_Range SP S1); entailer!).
    assert_PROP (Int.min_signed <= snd S2 <= Int.max_signed) by (sep_apply (GetPre_Range SP S2); entailer!).
    forward. forward_call. (*get 1*) forward.
    forward. forward_call. (*get 2*) forward.
    forward. cancel; simpl. rewrite Observer_rep_char.
    Exists s1 s2. Exists (snd S1) (snd S2). entailer!.
  Qed.
(*
  Lemma notify_verif: semax_body OVprog OGprog f_notify (notify_smallspec OP SP).
  Proof.
    start_function. simpl. rewrite NotifyPre_char; Intros s1 s2 c1 c2 v1 v2; simpl. subst; simpl in H.
    do 3 (destruct subjects; [ inv H |]); [ simpl | inv H].
    destruct p as [s1 U1]. destruct p0 as [s2 U2]; simpl.
    sep_apply (GetPre_ptrnull SP U1); Intros.
    sep_apply (GetPre_ptrnull SP U2); Intros.
    forward.
    forward_call.
    do 2 forward. 
    forward_call.
    do 2 forward. simpl. rewrite Observer_rep_char. Exists s1 s2 (snd U1) (snd U2). entailer!.
  Qed.
*)
(*
Lemma malloc_token_share_join
     : forall (sh1 sh2 sh : Share.t) (t : type) (p : val),
       sepalg.join sh1 sh2 sh -> (malloc_token M sh1 t p * malloc_token M sh2 t p)%logic = malloc_token M sh t p.
Admitted.
Lemma Esw_split: exists sh1 sh2,  sepalg.join sh1 sh2 Ews /\ readable_share sh1.
Proof. unfold sepalg.join. simpl. unfold Ews. Search extern_retainer. Search Share.Rsh. Share.Join_ba.
Admitted.*)

  Lemma newObserver_verif: semax_body OVprog OGprog f_newObserver (newObserver_spec M OP SP).
  Proof.
    start_function. simpl; Intros. rename H into Distinct.
    forward_call surelymalloc_spec_sub (ObserverTP, gv). 
    { simpl. split3; trivial. rep_lia. }
    Intros o. 
    forward. forward. simpl.
    sep_apply (RegisterPre_GetPrePost SP (obs2, v2) s2); Intros.
    assert_PROP (Int.min_signed <= v2 <= Int.max_signed) as Rng2 by (sep_apply (GetPre_Range SP (obs2, v2)); entailer!).
   
    forward_call (o, (*s1,*) obs1, obs2, s1, s2, s1, s2, v1, v2, @None Z, @None Z, gv). 
    { simpl. rewrite Notify_rep_char. Exists s1 s2 (@None Z) (@None Z) Vundef Vundef.
      entailer!. }
    sep_apply wand_frame_elim''. rewrite 2 if_true by reflexivity. simpl.
    sep_apply (Subject_GetPrePost SP (o :: obs1, v1) s1); Intros.
    assert_PROP (Int.min_signed <= v1 <= Int.max_signed) as Rng1 by (sep_apply (GetPre_Range SP (o::obs1, v1)); entailer!).
   
    forward_call (o, (*s2,*) obs2, o::obs1, s2, s1, s1, s2, v2, v1, Some v1, Some v2, gv). 
    { simpl. rewrite Notify_rep_char.
      Exists s1 s2 (Some v1) (Some v2) (Vint (Int.repr v1)) (Vint (Int.repr v2)).
      entailer!. rewrite Observer_rep_char. Intros u1 u2 z1 z2. inv H. cancel. }
    sep_apply wand_frame_elim''.
    forward.
    Exists o. simpl. entailer!.
    rewrite 2 if_false by trivial. unfold newObserver. simpl. cancel. 
  Qed.

  Definition ObserverComponent: @Component NullExtension.Espec OVprog _ 
      nil ObserverImports  prog (ObserverASI M OP SP) emp obs_internal_specs.
  Proof.
    mkComponent.
  + solve_SF_internal newObserver_verif.
  + solve_SF_internal notify_verif.
  + solve_SF_internal val_verif.
  + solve_SF_internal freeObserver_verif.
  Qed.

  Definition ObserverVSU: @VSU NullExtension.Espec OVprog _ 
      nil ObserverImports  prog (ObserverASI M OP SP) emp.
  Proof. eexists; apply ObserverComponent. Qed.

End Observer_VSU.
(*Check ObserverVSU. : forall SP : SubjectPredicates,
       VSU [] (ObserverImports SP) prog (ObserverASI ObserverPreds SP)*)

Definition Observer_CanVSU M SP := VSU_to_CanonicalVSU  (ObserverVSU M SP).