Require Import VST.floyd.VSU.
Require Import stdlib.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Observer.
Require Import Subject. (*needed for defining Vprog *)
Require Import spec_Subject.

Require Import verif_Subject. (*get access to the definition of SubjectREP etc*)

(*Maybe these opacity declarations can be eliminaed?*)
Opaque listrep.
Opaque Subject_rep.
Opaque GetPrePost.

Section Subject_VSU.
Variable M: MallocFreeAPD.

Definition add_spec :=
  DECLARE _add
   WITH n:val, o:val, os:list val, gv: globals
   PRE [ tptr NodeTP, tptr ObserverTP ]
       PROP (isptr o)
       PARAMS (n; o) GLOBALS (gv)
       SEP (mem_mgr M gv; listrep M os n)
    POST [ tptr NodeTP ] EX p:_, 
       PROP ()
       LOCAL (temp ret_temp p)
       SEP (mem_mgr M gv; listrep M (o::os) p).

Lemma Fold_length' {A P} states l:
    @Fold A P states l =
    !!(length states = length l) && Fold P states l.
Proof. apply pred_ext; [ apply andp_right; [ apply Fold_length | trivial] | entailer!]. Qed.

  Variable OP: ObserverAPD. 
  Notation O_rep := (Observerrep OP).
  Notation NotRep := (Notifyrep OP).

  Notation SP := (SubjectAP M). (*Instantiate the SubjectAP definined in verif_subject to M*)

  Notation S_rep := (Subjectrep SP).
  Notation RegPre := (RegisterPre SP).
  Notation GtPre := (GetPre SP).
  Notation GtPost := (GetPost SP).
  Notation UpdPre := (UpdatePre SP).

  Definition SVprog : varspecs. mk_varspecs prog. Defined.

  Definition SubjectImports :=
    [free_spec' M _free] ++ SM_ASI M ++ [notify_smallspec OP SP].

  Definition subject_internal_specs: funspecs :=
    [add_spec; newSubject_spec M SP; register_spec M SP OP; update_spec SP OP;
     get_smallspec SP; detachfirst_spec M SP; freeSubject_spec M SP].

  Definition SGprog:funspecs := SubjectImports ++ subject_internal_specs.

  Lemma add_verif:
    semax_body SVprog SGprog f_add add_spec.
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (NodeTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros p. 
    do 3 forward.
    Exists p. entailer!. apply listrep_fold.  trivial. 
  Qed.

  Lemma newSubject_verif:
    semax_body SVprog SGprog f_newSubject (newSubject_spec M SP).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (SubjectTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros s. 
    do 3 forward. Exists s; entailer!. simpl.
    rewrite <- Subject_rep_newSubject. cancel.
  Qed.

  Lemma freeSubject_verif:
    semax_body SVprog SGprog f_freeSubject (freeSubject_spec M SP).
  Proof.
    start_function. destruct S as [O c]. simpl.
    rewrite Subject_rep_char; Intros o.
    rewrite listrep_ptrnull; Intros.
    forward.
    forward_if.
    { apply denote_tc_test_eq_split. do 3 apply sepcon_valid_pointer1. apply listrep_valid_pointer. 
      apply valid_pointer_null. }
    { destruct O. rewrite listrep_nil; Intros. subst; contradiction.
      forward. Exists 1. simpl. entailer!. discriminate. rewrite Subject_rep_char.
      Exists o. entailer!. }
    { subst. destruct O.
      + rewrite listrep_nil; Intros.
        unfold malloc_token; rewrite data_at_isptr; Intros. destruct s; try inv Ps.
        forward_call (sizeof SubjectTP, Vptr b i, gv).
        { sep_apply data_at_memory_block; simpl; cancel. }
        forward. Exists 0; simpl; entailer!.
      + sep_apply listrep_unfold; Intros n. rewrite data_at_isptr; Intros. inv Pnullval. }
  Qed.

  Lemma detachfirst_verif:
    semax_body SVprog SGprog f_detachfirst (detachfirst_spec M SP).
  Proof.
    start_function. destruct S as [O c]. simpl.
    rewrite Subject_rep_char; Intros o. rename H into RngC.
    rewrite listrep_ptrnull; Intros.
    forward.
    forward_if.
    { apply denote_tc_test_eq_split. do 3 apply sepcon_valid_pointer1. apply listrep_valid_pointer. 
      apply valid_pointer_null. }
    { subst. forward.
      Exists nullval; simpl; entailer!. rewrite if_true; trivial.
      rewrite Subject_rep_char.
      Exists nullval. entailer.
      destruct O. entailer!.
      sep_apply listrep_unfold; Intros n. rewrite data_at_isptr; Intros. inv Pnullval. }
    { destruct O.
      + rewrite listrep_nil; Intros. subst; inv H.
      + sep_apply listrep_unfold; Intros n. rewrite listrep_ptrnull; Intros.
        forward. forward. forward.
        forward_call (sizeof NodeTP, o, gv).
        { rewrite if_false. 2: intros N; subst; inv H.
          unfold malloc_token at 1; Intros; sep_apply data_at_memory_block; cancel. }
        forward. Exists v; entailer. rewrite if_false. 2: intros Q; subst; inv Pv.
        Exists O. simpl. rewrite Subject_rep_char. Exists n. entailer!. }
  Qed.

  Lemma get_verif:
    semax_body SVprog SGprog f_get (get_smallspec SP).
  Proof. start_function. simpl. rewrite GetPrePost_char; Intros. do 2 forward. simpl.
      rewrite GetPrePost_char. entailer!. Qed.

  Lemma register_verif: semax_body SVprog SGprog f_registr (register_spec M SP OP).
  Proof.
    start_function. rename H into Distinct. rename H0 into Disjunction. simpl.
    rewrite Subject_rep_char.
    Intros p. rename H into Rng1. simpl.
    sep_apply Notifyrep_isptr; rewrite listrep_ptrnull; Intros.
 destruct Disjunction.
 + symmetry in H; inv H.
    forward.
    forward_call. 
    Intros r. forward. forward.
    forward_call (o, s1, s2, b1, b2, (obs1, n1), (obs2, n2) ).
    { simpl. unfold_data_at (data_at _ _ _ _). rewrite GetPrePost_char. simpl. entailer!. }
    entailer!.
    simpl. rewrite Subject_rep_char, GetPrePost_char.
    Exists r; unfold_data_at (data_at _ _ _ _); entailer!. 
    rewrite 2 if_true; trivial.
 + symmetry in H; inv H.
    forward.
    forward_call. 
    Intros r. forward. forward.
    forward_call (o, s1, s2, b1, b2, (obs2, n2), (obs1, n1) ).
    { simpl. unfold_data_at (data_at _ _ _ _). rewrite GetPrePost_char. simpl. entailer!. }
    entailer!.
    simpl. rewrite Subject_rep_char, GetPrePost_char.
    Exists r; unfold_data_at (data_at _ _ _ _); entailer!. 
    rewrite 2 if_false; trivial.
  Qed. 

  Fixpoint spine (l: list (val*val)) s z: mpred :=
    (match l with
      nil => !!(s=z /\ is_pointer_or_null s) && emp
    | (o,n)::os => !!(isptr o) && @data_at SCompSpecs Ews NodeTP (o,n) s * @malloc_token SCompSpecs M Ews NodeTP s *
                   spine os n z
    end)%logic.

  Lemma spine_nil p (P:is_pointer_or_null p): spine [] p p = emp.
  Proof. apply pred_ext; simpl; entailer!. Qed.

  Lemma spine_nilD o p : spine [] o p |-- !!(o=p) && emp.
  Proof. simpl; Intros; subst. entailer!. Qed.
  
  Lemma spine_ptr_or_null: forall l s z,
      spine l s z = !!(is_pointer_or_null s) && spine l s z.
  Proof.
    induction l; simpl; intros; apply pred_ext; try solve [entailer!].
    destruct a. rewrite IHl; clear IHl; Intros. entailer!.
  Qed.
  
  Lemma spine_ptr_or_null_precise: forall l s z,
      spine l s z = !!(match l with nil => s=z | _ => isptr s end) && spine l s z.
  Proof.
    induction l; simpl; intros; apply pred_ext; try solve [entailer!].
    destruct a. rewrite IHl; clear IHl; Intros. entailer!.
  Qed.
  
  Lemma spine_lst: forall l s z,
      spine l s z =
      !!(match l with
           nil => (s = z)
         | _ => exists ls o, l=ls++[(o,z)] end)
        && spine l s z.
  Proof.
    induction l; simpl; intros; apply pred_ext; try solve [entailer!].
    specialize (@exists_last _  (a::l)); intros.
    destruct X as [ll [[u1 u2] U]]. congruence. rewrite U. 
    destruct a. rewrite IHl; clear IHl; Intros.
    destruct l.
    + destruct H; subst. destruct ll. inv U; simpl; Intros; subst. entailer!.
      exists nil, u1; reflexivity. 
      inv U. elim (app_cons_not_nil ll nil _ H1).
    + destruct ll; inv U. destruct H as [ls [o HH]]. rewrite ! HH.
      entailer!. split.
      - exists ((v, v0) ::ls), o. 
        rewrite <- 2 app_comm_cons, <- HH, H2; trivial.
      - exists ls, o; split; trivial.
  Qed.

  Lemma spine_cons o n l s z (P: isptr o):
    (data_at Ews NodeTP (o, n) s * malloc_token M Ews NodeTP s * spine l n z)%logic
    |-- spine ((o,n)::l) s z.
  Proof. simpl. entailer!. Qed.
  
  Lemma spine_consD o n l s z:
    (spine ((o,n)::l) s z |-- !!(isptr o) && data_at Ews NodeTP (o, n) s * malloc_token M Ews NodeTP s * spine l n z)%logic.
  Proof. simpl; trivial. Qed.

  Lemma spine_app: forall l1 l2 s u z,
     (spine l1 s u * spine l2 u z |-- spine (l1++l2) s z)%logic.
  Proof. induction l1; simpl; intros; Intros. subst; cancel.
         destruct a. cancel; trivial. 
  Qed. 

  Lemma spine_listrep: forall l s, listrep M l s = EX sp, !!(map fst sp =l) && spine sp s nullval.
  Proof.
    induction l; simpl; intros; apply pred_ext.
    + Intros; subst. Exists (@nil (val*val)). rewrite listrep_nil. simpl; entailer!.
    + Intros sp. destruct sp; [ | inv H].
      simpl; Intros; subst. rewrite listrep_null; trivial.
    + sep_apply listrep_unfold. Intros n. rewrite IHl; clear IHl.
      Intros sp; subst. Exists ((a,n)::sp). simpl; entailer!. 
    + Intros sp. destruct sp; inv H. destruct p; simpl; Intros.
      eapply derives_trans. 2: eapply listrep_fold; trivial. rewrite IHl; clear IHl; cancel.
      Exists sp; entailer!. 
  Qed. 

  Lemma spine_validptr: forall sp p, isptr p -> spine sp p nullval |-- valid_pointer p.
  Proof.
    induction sp; simpl; intros; Intros; subst. apply valid_pointer_null.
    destruct a. entailer!.
  Qed.

  Opaque spine.

  Lemma In_remove {A dec} s a l: In s l -> a<>s -> In s (@remove A dec a l).
   Proof. induction l; simpl; intros. contradiction.
     destruct H; subst. rewrite if_false by trivial. left; trivial.
     specialize (IHl H H0). destruct (dec a a0). trivial. right; trivial.
  Qed.

  Lemma updateNtfys_app SUBJ subjects l2: forall l1, 
        updateNtfys SUBJ subjects (l1++l2) = updateNtfys SUBJ subjects l1 ++ updateNtfys SUBJ subjects l2.
  Proof. induction l1; simpl; trivial. destruct a as [[s1 b1] [s2 b2]]. f_equal; trivial. Qed.

  Lemma updateNtfys_length SUBJ subjects: forall l, length (updateNtfys SUBJ subjects l) = length l.
  Proof. induction l; simpl; trivial. f_equal; trivial. Qed.

  Lemma lookup_app s L2 l2: forall L1 l1, length L1 = length l1 -> 
        lookup s (L1++L2) (l1++l2) = match lookup s L1 l1 with Some x => Some x | None => lookup s L2 l2 end.
  Proof. induction L1; destruct l1; simpl; intros. trivial. congruence. congruence. inv H.
    rewrite IHL1; clear IHL1; trivial.
    destruct (Val.eq s v); trivial. 
  Qed.

  Lemma lookup_None s: forall L l, (~In s l) -> lookup s L l = None.
  Proof. induction L; destruct l; simpl; intros; trivial.
     destruct (Val.eq s v). subst; elim H. left; trivial. apply IHL. intros N. apply H. right; trivial.
  Qed.

  Lemma nodup_app_elim {A}: forall l1 l2, @NoDup A (l1++l2) -> NoDup l1 /\ NoDup l2 /\ list_disjoint l1 l2.
  Proof. induction l1; simpl; intros. split3; simpl; trivial. constructor. red; do 2 intros. inv H0.
  inv H.  destruct (IHl1 _ H3) as [? [? ?]]; clear IHl1. split3; trivial; simpl.
   apply NoDup_cons; trivial. intros N. elim H2. apply in_or_app. left; trivial.
   intros ? ? ? ? ?; subst. inv H4. elim H2; apply in_or_app. right; trivial. apply (H1 y y); trivial.
  Qed.

  Lemma update_verif: semax_body SVprog SGprog f_update (update_spec SP OP).
  Proof. start_function. Intros z.
    destruct H as [N Hsubject]. simpl.
    rewrite (@Fold_length' NotifyRepTp); Intros. rename H into LenOBS.
    rewrite Subject_rep_char; Intros o. rename H into RngZ. rewrite listrep_ptrnull; Intros.
    do 2 forward. rewrite spine_listrep. Intros sp. subst observers.
    rewrite (@Fold_length' SubjectRepTp); Intros. rename H into LenSUBS.
    forward_while 
      (EX p, EX sp1, EX sp2,
       PROP (sp=sp1++sp2 )
            LOCAL (temp _nd p; temp _s s; temp _n (Vint (Int.repr n)))
            SEP (spine sp1 o p; spine sp2 p nullval; malloc_token M Ews SubjectTP s;
                 data_at Ews SubjectTP (o, Vint (Int.repr n)) s;
                 Subjects (Subject_rep M) SUBJ subjects;
                 Notifiers NotRep (updateNtfys ((map fst sp,n)::SUBJ) (s::subjects) (firstn (length sp1) OBS)) (map fst sp1);
                 Notifiers NotRep (skipn (length sp1) OBS) (map fst sp2)(*
                 Observers O_rep s (list_repeat (length sp1) n) (map fst sp1);
                 Notifiers NotPre s (skipn (length sp1) vals) (map fst sp2)*)))%assert.
    { Exists o (@nil (val*val)) sp. entailer!. rewrite spine_nil; trivial. simpl; cancel. }
    { entailer.
      red in PNp. destruct p; try solve [contradiction].
      + destruct Archi.ptr64. contradiction. subst; entailer!.
      + apply denote_tc_test_eq_split; entailer.
        do 5 apply sepcon_valid_pointer1. apply sepcon_valid_pointer2.
        apply spine_validptr. trivial. } 
    { subst sp. rewrite (@Fold_length' NotifyRepTp) with (l:=map fst sp2); Intros.
      destruct sp2; [ sep_apply spine_nilD; Intros; subst; contradiction | simpl].
      destruct p0 as [obs nxt]; Intros. simpl. simpl in H.
      remember (skipn (Datatypes.length sp1) OBS) as w.
      destruct w. inv H. destruct n0 as [[s1 b1] [s2 b2]]; simpl. sep_apply spine_consD; Intros. (* sep_apply NotifyRep_isptr; Intros.*)
      forward.
      sep_apply Notifyrep_DistinctSubjects; Intros. rename H0 into Distinct.
      destruct (Val.eq s1 s).
      { (*cases1=s*) subst s1.
        (*establish GPre U2 s2 for the U2 from SUBJ*)
        sep_apply (Fold_In_split (Subject_rep M) SUBJ subjects s2).
        { subst; clear R2. apply nodup_In. apply In_remove. 2: congruence. 
          clear - Heqw. assert (In (s, b1, (s2, b2)) OBS). eapply (@skipn_In NotifyRepTp) with (Datatypes.length sp1). rewrite <- Heqw. left; trivial.
          clear - H.  induction OBS; simpl in *; trivial.
          destruct H; subst. right; left; trivial.
          destruct a.  destruct p. destruct p0. do 2 right. auto. }
        Intros U2 SUBJ1 SUBJ2 subs1 subs2.
        sep_apply Subject_GetPrePost; Intros.

        forward_call (obs, s, s2, b1, b2, (map fst (sp1 ++ (obs, nxt) :: sp2),n), U2).
        { simpl. rewrite (GetPrePost_char M (map fst (sp1 ++ (obs, nxt) :: sp2),n)).
          unfold_data_at (data_at Ews SubjectTP _ _). entailer!. }
        simpl. rewrite (spine_ptr_or_null sp2); Intros.
        forward.
        Exists (nxt, sp1++[(obs, nxt)], sp2).
        entailer!.
        + rewrite <- app_assoc; trivial.
        + sep_apply wand_frame_elim''.
          sep_apply (Fold_combine (Subject_rep M) U2 s2); rewrite H2; cancel. 
          rewrite GetPrePost_char; Intros. unfold_data_at (data_at Ews SubjectTP _ _); cancel.
          specialize (spine_cons obs nxt nil p nxt Pobs); rewrite spine_nil; trivial.
          intros XX; sep_apply XX; clear XX.
          sep_apply (spine_app sp1 [(obs, nxt)] o p nxt); cancel.
          (*rewrite H2, (Fold_app (Subject_rep M)); trivial. simpl; cancel.*)
          rewrite ! map_app, app_length. simpl.
          remember ((map fst sp1 ++ obs :: map fst sp2, n) :: SUBJ1 ++ U2 :: SUBJ2) as ww. 
          remember (s :: subs1 ++ s2 :: subs2) as uu.
          rewrite <- (firstn_skipn (Datatypes.length sp1) OBS). simpl.
          apply sepcon_derives.
          - rewrite <- Heqw, <- firstn_app, firstn_app1, firstn_firstn. 2: lia.
            2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite skipn_app2.
            2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite firstn_length_le, minus_diag. simpl.
            2:{ rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite updateNtfys_app, Fold_app.
            2:{ rewrite updateNtfys_length. rewrite firstn_length_le, map_length. lia. 
                rewrite LenOBS, map_app, app_length, map_length; lia. }
            cancel. simpl. destruct (Val.eq s s); [ clear e | elim n0; trivial].
            destruct (Val.eq s2 s); [ elim Distinct; subst; trivial | clear n0].
            destruct (Val.eq s2 s2); [ clear e | elim n0; trivial].
            assert (Hn: lookup s ww uu = Some n).
            { subst. simpl. destruct (Val.eq s s); [ trivial | elim n0; trivial]. }
            rewrite Hn.
            assert (Hn2: lookup s2 ww uu = Some (snd U2)).
            { subst. simpl. destruct (Val.eq s2 s); [ elim Distinct; subst; trivial | clear n0].
              rewrite lookup_app; trivial. simpl. destruct (Val.eq s2 s2); [ clear e | elim n0; trivial].
              clear - H2. rewrite lookup_None; trivial. intros N. 

              specialize (NoDup_nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS))). 
              rewrite H2; clear H2. intros. apply NoDup_remove in H; destruct H.
              apply H0. apply in_or_app; left; trivial. }
            rewrite Hn2, Observer_Notify; cancel.
        - apply derives_refl'. f_equal. rewrite <- Heqw, <- skipn_skipn, skipn_app1.
          2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
          rewrite skipn_firstn, minus_diag; trivial. }
      rename n0 into S1S. destruct (Val.eq s2 s).
      { (*case s2=s*) subst s2. clear S1S.
        (*establish GPre U1 s1 for the U1 from SUBJ*)
        sep_apply (Fold_In_split (Subject_rep M) SUBJ subjects s1).
        { subst; clear R2. apply nodup_In. apply In_remove. 2: congruence. 
          clear - Heqw. assert (In (s1, b1, (s, b2)) OBS). eapply (@skipn_In NotifyRepTp) with (Datatypes.length sp1). rewrite <- Heqw. left; trivial.
          clear - H.  induction OBS; simpl in *; trivial.
          destruct H; subst. left; trivial.
          destruct a.  destruct p. destruct p0. do 2 right. auto. }
        Intros U1 SUBJ1 SUBJ2 subs1 subs2.
        sep_apply Subject_GetPrePost; Intros.

        forward_call (obs, s1, s, b1, b2, U1, (map fst (sp1 ++ (obs, nxt) :: sp2),n)).
        { simpl. rewrite (GetPrePost_char M (map fst (sp1 ++ (obs, nxt) :: sp2),n)).
          unfold_data_at (data_at Ews SubjectTP _ _). entailer!. }
        simpl. rewrite (spine_ptr_or_null sp2); Intros.
        forward.
        Exists (nxt, sp1++[(obs, nxt)], sp2).
        entailer!.
        + rewrite <- app_assoc; trivial.
        + sep_apply wand_frame_elim''.
          sep_apply (Fold_combine (Subject_rep M) U1 s1); rewrite H2; cancel. 
          rewrite GetPrePost_char; Intros. unfold_data_at (data_at Ews SubjectTP _ _); cancel.
          specialize (spine_cons obs nxt nil p nxt Pobs); rewrite spine_nil; trivial.
          intros XX; sep_apply XX; clear XX.
          sep_apply (spine_app sp1 [(obs, nxt)] o p nxt); cancel.
          (*rewrite H2, (Fold_app (Subject_rep M)); trivial. simpl; cancel.*)
          rewrite ! map_app, app_length. simpl.
          remember ((map fst sp1 ++ obs :: map fst sp2, n) :: SUBJ1 ++ U1 :: SUBJ2) as ww. 
          remember (s :: subs1 ++ s1 :: subs2) as uu.
          rewrite <- (firstn_skipn (Datatypes.length sp1) OBS). simpl.
          apply sepcon_derives.
          - rewrite <- Heqw, <- firstn_app, firstn_app1, firstn_firstn. 2: lia.
            2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite skipn_app2.
            2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite firstn_length_le, minus_diag. simpl.
            2:{ rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite updateNtfys_app, Fold_app.
            2:{ rewrite updateNtfys_length. rewrite firstn_length_le, map_length. lia. 
                rewrite LenOBS, map_app, app_length, map_length; lia. }
            cancel. simpl. destruct (Val.eq s1 s1); [ clear e | elim n0; trivial].
            destruct (Val.eq s s1); [ elim Distinct; subst; trivial | clear n0].
            destruct (Val.eq s s); [ clear e | elim n0; trivial].
            assert (Hn1: lookup s1 ww uu = Some (snd U1)).
            { subst. simpl. destruct (Val.eq s1 s); [ elim Distinct; trivial | clear n0]. 
              rewrite lookup_app; trivial. simpl. destruct (Val.eq s1 s1); [ clear e | elim n0; trivial].
              clear - H2. rewrite lookup_None; trivial. intros N. 

              specialize (NoDup_nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS))). 
              rewrite H2; clear H2. intros. apply NoDup_remove in H; destruct H.
              apply H0. apply in_or_app; left; trivial. }
            rewrite Hn1.
            assert (Hn: lookup s ww uu = Some n).
            { subst. simpl. destruct (Val.eq s s); [ trivial | elim n0; trivial]. }
            rewrite Hn, Observer_Notify; cancel.
        - apply derives_refl'. f_equal. rewrite <- Heqw, <- skipn_skipn, skipn_app1.
          2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
          rewrite skipn_firstn, minus_diag; trivial. }

      { (*case s2<>s<>s1 - should not really occur but...*) rename n0 into S2S.
        (*establish GPre U1 s1 for the U1 from SUBJ*)
        sep_apply (Fold_In_split (Subject_rep M) SUBJ subjects s1).
        { subst; clear R2. apply nodup_In. apply In_remove. 2: congruence. 
          clear - Heqw. assert (In (s1, b1, (s2, b2)) OBS). eapply (@skipn_In NotifyRepTp) with (Datatypes.length sp1). rewrite <- Heqw. left; trivial.
          clear - H.  induction OBS; simpl in *; trivial.
          destruct H; subst. left; trivial.
          destruct a.  destruct p. destruct p0. do 2 right. auto. }
        Intros U1 SUBJ1 SUBJ2 subs1 subs2.
        sep_apply Subject_GetPrePost; Intros.

        assert (In s2 subs1 \/ In s2 subs2).
        { assert (In s2 (subs1 ++ s1 :: subs2)).
          { rewrite <- H2; clear H2; subst subjects.
            apply nodup_In. apply In_remove. 2: congruence.
            clear - Heqw. assert (In (s1, b1, (s2, b2)) OBS). eapply (@skipn_In NotifyRepTp) with (Datatypes.length sp1). rewrite <- Heqw. left; trivial.
            clear - H.  induction OBS; simpl in *; trivial.
            destruct H; subst. right; left; trivial.
            destruct a. destruct p. destruct p0. do 2 right. auto. }
          apply in_app_or in H3. destruct H3. left; trivial. inv H3. congruence. right; trivial. } 
        rewrite (Fold_length' SUBJ1 subs1); Intros. rename H0 into L1.
        rewrite (Fold_length' SUBJ2 subs2); Intros. rename H0 into L2.
        destruct H3 as [HS2 | HS2].
        * sep_apply (Fold_In_split (Subject_rep M) SUBJ1 subs1 s2).
          Intros U2 SUBJ_21 SUBJ_22 subs_21 subs_22.

          sep_apply Subject_GetPrePost; Intros.
          forward_call (obs, s1, s2, b1, b2, U1, U2).
          simpl. rewrite (spine_ptr_or_null sp2); Intros.
          forward.
          Exists (nxt, sp1++[(obs, nxt)], sp2).
          entailer!.
          + rewrite <- app_assoc; trivial.
          + sep_apply wand_frame_elim''. sep_apply wand_frame_elim''.
            sep_apply (Fold_combine (Subject_rep M) U2 s2).
            sep_apply (Fold_combine (Subject_rep M) U1 s1); rewrite H2; cancel. 
(*            rewrite GetPrePost_char; Intros. unfold_data_at (data_at Ews SubjectTP _ _); cancel.*)
            specialize (spine_cons obs nxt nil p nxt Pobs); rewrite spine_nil; trivial.
            intros XX; sep_apply XX; clear XX.
            sep_apply (spine_app sp1 [(obs, nxt)] o p nxt); cancel.
            (*rewrite H2, (Fold_app (Subject_rep M)); trivial. simpl; cancel.*)
            rewrite ! map_app, app_length. simpl.
            remember ((map fst sp1 ++ obs :: map fst sp2, n) :: (SUBJ_21 ++ U2 :: SUBJ_22) ++ U1 :: SUBJ2) as ww. 
            remember (s :: (subs_21 ++ s2 :: subs_22) ++ s1 :: subs2)  as uu.
            rewrite <- (firstn_skipn (Datatypes.length sp1) OBS). simpl.
            apply sepcon_derives.
            - rewrite <- Heqw, <- firstn_app, firstn_app1, firstn_firstn. 2: lia.
              2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
              rewrite skipn_app2.
              2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
              rewrite firstn_length_le, minus_diag. simpl.
              2:{ rewrite LenOBS, map_app, app_length, map_length; lia. }
              rewrite updateNtfys_app, Fold_app.
              2:{ rewrite updateNtfys_length. rewrite firstn_length_le, map_length. lia. 
                  rewrite LenOBS, map_app, app_length, map_length; lia. }
              cancel. simpl. destruct (Val.eq s1 s1); [ clear e | elim n0; trivial].
              destruct (Val.eq s2 s1); [ elim Distinct; subst; trivial | clear n0].
              destruct (Val.eq s2 s2); [ clear e | elim n0; trivial].
              assert (Hn1: lookup s1 ww uu = Some (snd U1)).
              { subst. simpl. destruct (Val.eq s1 s); [ congruence | ].
                rewrite lookup_app; trivial. simpl.
                destruct (Val.eq s1 s1); [ clear e | congruence].
                clear - H2. rewrite lookup_None; trivial. intros N. 

                specialize (NoDup_nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS))). 
                rewrite H2; clear H2. intros. apply NoDup_remove in H; destruct H.
                apply H0. apply in_or_app; left; trivial. }
              rewrite Hn1.
              assert (Hn2: lookup s2 ww uu = Some (snd U2)).
              { subst. simpl. destruct (Val.eq s2 s); [ congruence | ].
                rewrite lookup_app; trivial. simpl.
                destruct (Val.eq s2 s1); [ congruence |].
                rewrite lookup_app; trivial. simpl.
                destruct (Val.eq s2 s2); [ clear e | congruence].
                clear - H2. rewrite lookup_None; trivial. intros N. 

                specialize (NoDup_nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS))). 
                rewrite H2; clear H2. intros. apply NoDup_remove in H; destruct H. clear H0.
                apply nodup_app_elim in H. destruct H as [? [? ?]].
                apply NoDup_remove in H; destruct H. clear H.
                apply H2. apply in_or_app; left; trivial. }
              rewrite Hn2, Observer_Notify; cancel.
          - apply derives_refl'. f_equal. rewrite <- Heqw, <- skipn_skipn, skipn_app1.
            2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite skipn_firstn, minus_diag; trivial. 
        * sep_apply (Fold_In_split (Subject_rep M) SUBJ2 subs2 s2).
          Intros U2 SUBJ_21 SUBJ_22 subs_21 subs_22.

          sep_apply Subject_GetPrePost; Intros.
          forward_call (obs, s1, s2, b1, b2, U1, U2).
          simpl. rewrite (spine_ptr_or_null sp2); Intros.
          forward.
          Exists (nxt, sp1++[(obs, nxt)], sp2).
          entailer!.
          + rewrite <- app_assoc; trivial.
          + sep_apply wand_frame_elim''. sep_apply wand_frame_elim''.
            sep_apply (Fold_combine (Subject_rep M) U2 s2).
            sep_apply (Fold_combine (Subject_rep M) U1 s1 SUBJ1); rewrite  H2; cancel.
(*            rewrite GetPrePost_char; Intros. unfold_data_at (data_at Ews SubjectTP _ _); cancel.*)
            specialize (spine_cons obs nxt nil p nxt Pobs); rewrite spine_nil; trivial.
            intros XX; sep_apply XX; clear XX.
            sep_apply (spine_app sp1 [(obs, nxt)] o p nxt); cancel.
            (*rewrite H2, (Fold_app (Subject_rep M)); trivial. simpl; cancel.*)
            rewrite ! map_app, app_length. simpl.
            remember ((map fst sp1 ++ obs :: map fst sp2, n) :: SUBJ1 ++ U1 :: SUBJ_21 ++ U2 :: SUBJ_22) as ww. 
            remember (s :: subs1 ++ s1 :: subs_21 ++ s2 :: subs_22)  as uu.
            rewrite <- (firstn_skipn (Datatypes.length sp1) OBS). simpl.
            apply sepcon_derives.
            - rewrite <- Heqw, <- firstn_app, firstn_app1, firstn_firstn. 2: lia.
              2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
              rewrite skipn_app2.
              2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
              rewrite firstn_length_le, minus_diag. simpl.
              2:{ rewrite LenOBS, map_app, app_length, map_length; lia. }
              rewrite updateNtfys_app, Fold_app.
              2:{ rewrite updateNtfys_length. rewrite firstn_length_le, map_length. lia. 
                  rewrite LenOBS, map_app, app_length, map_length; lia. }
              cancel. simpl. destruct (Val.eq s1 s1); [ clear e | elim n0; trivial].
              destruct (Val.eq s2 s1); [ elim Distinct; subst; trivial | clear n0].
              destruct (Val.eq s2 s2); [ clear e | elim n0; trivial].
              assert (Hn1: lookup s1 ww uu = Some (snd U1)).
              { subst. simpl. destruct (Val.eq s1 s); [ congruence | ].
                rewrite lookup_app; trivial. simpl.
                destruct (Val.eq s1 s1); [ clear e | congruence].
                clear - H2. rewrite lookup_None; trivial. intros N. 

                specialize (NoDup_nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS))). 
                rewrite H2; clear H2. intros. apply NoDup_remove in H; destruct H.
                apply H0. apply in_or_app; left; trivial. }
              rewrite Hn1.
              assert (Hn2: lookup s2 ww uu = Some (snd U2)).
              { subst. simpl. destruct (Val.eq s2 s); [ congruence | ].
                rewrite lookup_app; trivial. simpl.
                destruct (Val.eq s2 s1); [ congruence |].
                rewrite lookup_app; trivial. simpl.
                destruct (Val.eq s2 s2); [ clear e | congruence].
                clear - H2. rewrite 2 lookup_None; trivial; intros N. 

                + specialize (NoDup_nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS))). 
                  rewrite H2; clear H2. intros. apply NoDup_remove in H; destruct H. clear H0.
                  apply nodup_app_elim in H. destruct H as [? [? ?]].
                  apply NoDup_remove in H0; destruct H0. clear H0.
                  apply H2. apply in_or_app; left; trivial.
                + specialize (NoDup_nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS))). 
                  rewrite H2; clear H2. intros. apply NoDup_remove in H; destruct H. clear H0. rewrite app_assoc in H.
                  apply nodup_app_elim in H. destruct H as [? [? ?]]. clear - H1 N.
                  elim (H1 s2 s2). apply in_or_app; left; trivial. left; trivial. trivial. }
              rewrite Hn2, Observer_Notify; cancel.
          - apply derives_refl'. f_equal. rewrite <- Heqw, <- skipn_skipn, skipn_app1.
            2:{ rewrite firstn_length_le. lia. rewrite LenOBS, map_app, app_length, map_length; lia. }
            rewrite skipn_firstn, minus_diag; trivial. }
    }
    forward. entailer!.
    simpl S_rep; rewrite Subject_rep_char. entailer!. Exists o. cancel.
    rewrite (spine_ptr_or_null_precise sp2); Intros.
    destruct sp2; simpl; Intros; [ clear H | contradiction]. rewrite spine_nil; [|trivial].
    rewrite app_nil_r,(*. map_length,*) spine_listrep.
    Exists sp1. entailer!.
    rewrite app_nil_r, map_length in LenOBS. rewrite <- LenOBS, skipn_all, firstn_all. simpl. cancel.
  Qed.

  Definition SubjectComponent: @Component NullExtension.Espec SVprog _ 
      nil SubjectImports prog (SubjectASI M SP OP) emp subject_internal_specs.
  Proof. 
    mkComponent.
  + solve_SF_internal add_verif.
  + solve_SF_internal newSubject_verif.
  + solve_SF_internal register_verif.
  + solve_SF_internal update_verif.
  + solve_SF_internal get_verif.
  + solve_SF_internal detachfirst_verif.
  + solve_SF_internal freeSubject_verif. 
  Qed.

  Definition SubjectVSU: @VSU NullExtension.Espec SVprog _ 
      nil SubjectImports prog (SubjectASI M SP OP) emp.
  Proof. eexists; apply SubjectComponent. Qed.

End Subject_VSU.
(*Check SubjectVSU. : forall OP : ObserverPredicates,
       VSU [] (SubjectImports OP) prog (SubjectASI SubjectPreds OP)*)

Definition Subject_CanVSU M OP := VSU_to_CanonicalVSU  (SubjectVSU M OP). 