Require Import VST.floyd.proofauto.

(*Model-level representation of subjects and observers*)

(*Not sufficient, due to information loss in the update->notify->get callback
Definition SubjectRepTp:Type := (list (val*val)) * Z. (* list of (pointers to) observers, current value) *)*)

(* list of (pointers to) nodes of observers, current value) *)
Definition SubjectRepTp:Type := (list val) * Z.

Definition newSubject: SubjectRepTp:= (nil, 0).

(*Model-level representation of observers*)
Definition ObserverRepTp:Type := (val * Z) * (val * Z). (*2 links to subjects, each with the observer's local belief*)

Definition newObserver (Sub1 Sub2: val * SubjectRepTp): ObserverRepTp:= ((fst Sub1, snd(snd Sub1)), (fst Sub2, snd(snd Sub2))).

Record SubjectAPD := {
  Subjectrep: SubjectRepTp -> val -> mpred;
  RegisterPre: SubjectRepTp -> val -> mpred;
  UpdatePre: SubjectRepTp -> val -> mpred;
  GetPre: SubjectRepTp -> val -> mpred;
  GetPost: SubjectRepTp -> val -> mpred;
  Subject_Register: forall S s, Subjectrep S s |-- RegisterPre S s;
  Subject_Update: forall S s, Subjectrep S s |-- UpdatePre S s;
  Subject_GetPrePost: forall S s, Subjectrep S s |-- (GetPre S s * (GetPost S s -* Subjectrep S s))%logic;
  RegisterPre_GetPrePost: forall S s, RegisterPre S s |-- (GetPre S s * (GetPost S s -* RegisterPre S s))%logic;
  GetPre_ptrnull: forall S s, GetPre S s |-- !!(is_pointer_or_null s);
    (*could add ptrnull lemmas for the other residual predicates, too*)
  GetPost_ptrnull: forall S s, GetPost S s |-- !!(is_pointer_or_null s);
  RegPre_ptrnull: forall S s, RegisterPre S s |-- !!(is_pointer_or_null s);
  UpdPre_ptrnull: forall S s, UpdatePre S s |-- !!(is_pointer_or_null s);(*
  Register_validptr: forall S s, RegisterPre S s |-- valid_pointer s;
  GetPost_validptr S s: GetPost S s |-- valid_pointer s*) 
  GetPre_Range: forall S s, GetPre S s |-- !!(Int.min_signed <= snd S <= Int.max_signed);
  Subject_conflict: forall S1 S2 s1 s2, Subjectrep S1 s1 * Subjectrep S2 s2 |-- !!(s1<>s2)
}. (*
Lemma Subjectrep_validptr A S s: Subjectrep A S s |-- valid_pointer s.
Proof. sep_apply Subject_Register. sep_apply Register_validptr. entailer!. Qed.
*)
Lemma Subjectrep_ptrnull A S s: Subjectrep A S s |-- !!(is_pointer_or_null s).
Proof. sep_apply Subject_GetPrePost. sep_apply GetPre_ptrnull. entailer!. Qed.

Definition NotifyRepTp:Type := (val * option Z) * (val * option Z). (*pair of ptrs to subjects, each with optional cached value*)

Record ObserverAPD := {
    Observerrep: ObserverRepTp -> val -> mpred;
    Notifyrep: NotifyRepTp -> val -> mpred;
    Notifyrep_isptr: forall N o, Notifyrep N o |-- !!(isptr o);
    ValPre: ObserverRepTp -> val -> mpred;
    (*Observer_Notify: forall O o, Observerrep O o |-- NotifyPre O o;*)
    Observer_Notify: forall s1 s2 v1 v2 o, Notifyrep ((s1, Some v1),(s2, Some v2)) o = Observerrep ((s1,v1),(s2,v2)) o;
    Observer_Val: forall O o, Observerrep O o |-- ValPre O o;
    Notifyrep_DistinctSubjects: forall N o, Notifyrep N o |-- !!(match N with ((s1,v1),(s2,v2)) => s1<>s2 end)
  (*  Observerrep_Range: forall O o, Observerrep O o |-- !!(match O with ((s1,v1),(s2,v2)) => Int.min_signed <= v1 <= Int.max_signed /\ Int.min_signed <= v2 <= Int.max_signed end)*)
}.

Lemma Observerrep_isptr A O o: Observerrep A O o |-- !!(isptr o).
Proof. destruct O. destruct p; destruct p0. rewrite <- Observer_Notify. sep_apply Notifyrep_isptr. entailer!. Qed.

Lemma Observerrep_DistinctSubjects A O o: Observerrep A O o |-- !!(match O with ((s1,v1),(s2,v2)) => s1<>s2 end).
Proof. destruct O. destruct p; destruct p0. rewrite <- Observer_Notify. sep_apply Notifyrep_DistinctSubjects. entailer!. Qed.

