Require Export VST.floyd.proofauto.
Require Export SubjectObserverModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Observer. (*needed for function and type names *)

Global Open Scope funspec_scope.
Definition ObserverTP:type:=  Tstruct _observer noattr.
Definition SubjectTP:type:=  Tstruct _subject noattr.

Section ObserverASI.
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.

  Notation S_rep := (Subjectrep SP).
  Notation R_Pre := (RegisterPre SP). (*
  Notation U_Pre := (UpdatePre SP).*)
  Notation G_Pre := (GetPre SP).
  Notation G_Post := (GetPost SP).
  Notation O_rep := (Observerrep OP).
  Notation N_rep:= (Notifyrep OP).
  Notation V_Pre:= (ValPre OP).

Fixpoint Subjectlist (SPred:SubjectRepTp -> val -> mpred) (l: list (val * SubjectRepTp)) :=
    match l with
      nil => emp
    | (s, SP)::l' => (SPred SP s * Subjectlist SPred l')%logic
    end.

  Definition newObserver_spec :=
    DECLARE _newObserver
    WITH gv: globals, (*subjects:list (val * SubjectRepTp),*) s1:val, s2:val, obs1:list val, obs2:list val, v1:Z, v2:Z
    PRE [ tptr SubjectTP, tptr SubjectTP]
      PROP(s1<>s2 (*follows in principle from use of sepcon in Subjectlist*)(*[s1;s2] = map fst subjects; list_norepet (map fst subjects)*)) (*later: C code should take a list of subject pointers*)
      PARAMS (s1;s2) GLOBALS (gv)
      SEP(Subjectlist R_Pre [(s1,(obs1,v1));(s2,(obs2,v2))]; mem_mgr M gv)
    POST[ tptr ObserverTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(O_rep (newObserver (*subjects*)(s1,(obs1,v1)) (s2,(obs2,v2))) p; 
         (*Subjectlist S_rep (map (fun x => (fst x, (p::fst (snd x), snd (snd x)))) subjects);*)
         Subjectlist S_rep [(s1,(p::obs1,v1)); (s2,(p::obs2,v2))];
         mem_mgr M gv).

  Definition notify_smallspec :=
    DECLARE _notify
    WITH o:val, s1:val, s2:val, b1: option Z, b2:option Z (*c:Z, s:val*),S1: SubjectRepTp, S2: SubjectRepTp(*, oldState:NotifyRepTp*)
    PRE [ tptr ObserverTP ]
     PROP((*s=s1 \/ s=s2*))
     PARAMS (o) GLOBALS ()
     SEP(N_rep ((s1,b1),(s2,b2)) o; G_Pre S1 s1; G_Pre S2 s2)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(O_rep ((s1,(*if Val.eq s1 s then Some (snd S1) else b1*)snd S1),
                 (s2,(*if Val.eq s2 s then Some (snd S) else b2)*)snd S2)) o; G_Post S1 s1; G_Post S2 s2). (*
  Definition notify_smallspec :=
    DECLARE _notify
    WITH o:val, s:val, subjects:list (val * SubjectRepTp), oldState:NotifyRepTp
    PRE [ tptr ObserverTP, tptr SubjectTP  ]
     PROP(map fst oldState = map fst subjects)
     PARAMS (o;s) GLOBALS ()
     SEP(N_rep oldState o; Subjectlist G_Pre subjects)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(N_rep (map (fun sS => match sS with (u, v) => (u, if Val.eq u s then Some(snd S) else v) end) oldState) o; Subjectlist G_Post subjects).*)


  Definition val_spec :=
    DECLARE _val
    WITH o:val, (*Ostate: ObserverRepTp*)s1:val, s2:val, v1: Z, v2: Z
    PRE [ tptr ObserverTP ]
     PROP(Int.min_signed <= v1 + v2 <= Int.max_signed)
     PARAMS (o) GLOBALS ()
     SEP(V_Pre ((s1,v1),(s2,v2)) o)
    POST[ tint ]
      PROP()
      RETURN (Vint (Int.repr (v1+v2)))
      SEP(O_rep ((s1,v1),(s2,v2)) o).

  Definition freeObserver_spec:=
    DECLARE _freeObserver 
    WITH gv: globals, o:val, O:ObserverRepTp
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS (gv)
     SEP(O_rep O o; mem_mgr M gv)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(mem_mgr M gv). 

  Definition ObserverASI:funspecs := [ newObserver_spec; notify_smallspec; val_spec; freeObserver_spec].

  Definition notify_largespec :=
    DECLARE _notify
    WITH o:val, s1:val, s2:val, b1: option Z, b2:option Z (*c:Z, s:val,*),S1: SubjectRepTp, S2: SubjectRepTp(*, oldState:NotifyRepTp*)
    PRE [ tptr ObserverTP ]
     PROP((*s=s1 \/ s=s2*))
     PARAMS (o) GLOBALS ()
     SEP(N_rep ((s1,b1),(s2,b2)) o; S_rep S1 s1; S_rep S2 s2)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(O_rep ((s1,(*if Val.eq s1 s then Some (snd S1) else b1*)snd S1),
                 (s2,(*if Val.eq s2 s then Some (snd S) else b2)*)snd S2)) o; S_rep S1 s1; S_rep S2 s2).

Lemma Notify_sub: funspec_sub (snd notify_smallspec) (snd notify_largespec).
Proof.
  do_funspec_sub. clear H. Exists w.
  destruct w as [[[[[[o s1] s2] b1] b2] U1] U2]. simpl; entailer.
  Exists ((G_Post U1 s1 -* S_rep U1 s1) * (G_Post U2 s2 -* S_rep U2 s2))%logic.
  simpl; entailer!.
  { intros. cancel. 
    apply derives_trans with (((G_Post U1 s1 -* S_rep U1 s1) * G_Post U1 s1) * ((G_Post U2 s2 -* S_rep U2 s2) * G_Post U2 s2))%logic. cancel.
    apply sepcon_derives;  apply wand_frame_elim''. }
  do 2 sep_apply Subject_GetPrePost. cancel. 
Qed.
(* 
  Definition notify_largespec :=
    DECLARE _notify
    WITH o:val, s1:val, s2:val, b1: option Z, b2:option Z (*c:Z, *), s:val, S: SubjectRepTp(*, oldState:NotifyRepTp*)
    PRE [ tptr ObserverTP ]
     PROP(s=s1 \/ s=s2)
     PARAMS (o) GLOBALS ()
     SEP(N_rep ((s1,b1),(s2,b2)) o; S_rep S s)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(N_rep ((s1,if Val.eq s1 s then Some (snd S) else b1),
                 (s2,if Val.eq s2 s then Some (snd S) else b2)) o; S_rep S s).

Lemma Notify_sub: funspec_sub (snd notify_smallspec) (snd notify_largespec).
Proof.
  do_funspec_sub. clear H. Exists w.
  destruct w as [[[[[[o s1] s2] b1] b2] s] U]. simpl; entailer.
  Exists (G_Post U s -* S_rep U s).
  simpl; entailer!.
  { intros. cancel. apply wand_frame_elim''. }
  sep_apply Subject_GetPrePost. cancel. 
Qed.*)
(*
  Definition notify_largespec :=
    DECLARE _notify
    WITH o:val, c:Z, subjects:list (val * SubjectRepTp), oldState:ObserverRepTp
    PRE [ tptr ObserverTP ]
     PROP(map fst oldState = map fst subjects)
     PARAMS (o) GLOBALS ()
     SEP(N_Pre oldState o; Subjectlist S_rep subjects)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(O_rep (map (fun sS => (fst sS, snd (snd sS))) subjects) o; Subjectlist S_rep subjects).

Lemma Notify_sub: funspec_sub (snd notify_smallspec) (snd notify_largespec).
Proof.
  do_funspec_sub. clear H.
  destruct w as [[[o c] subjects] oldState]. simpl; entailer.
  Exists (((o, c), subjects), oldState) (Subjectlist (fun SP s => G_Post SP s -* S_rep SP s) subjects).
  simpl; entailer!.
  { intros. cancel. clear. induction subjects; simpl. cancel.
    destruct a as [s X]. sep_apply IHsubjects; clear IHsubjects. cancel. apply wand_frame_elim''. }
  clear. induction subjects; simpl. cancel.
    destruct a as [s X]. sep_apply IHsubjects; clear IHsubjects. sep_apply Subject_GetPrePost. cancel. 
Qed.*)
End ObserverASI.