Require Export VST.floyd.proofauto.
Require Export SubjectObserverModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Subject. (*needed for function and type names *)

Global Open Scope funspec_scope.
Definition SubjectTP:type:=  Tstruct _subject noattr.
Definition ObserverTP:type:=  Tstruct _observer noattr.
(*
Fixpoint Observers (OPred:ObserverRepTp -> val -> mpred) spointers (vals: list Z) (l:list val) :=
    match vals, l with
      nil, nil => emp
    | v::vs, o::os => (OPred (spointers,v) o * Observers OPred spointers vs os)%logic
    | _, _ => FF
    end.
Check Observers.

Lemma Observers_app OPred s: forall vals1 l1 vals2 l2 (L: length vals1 = length l1),
      Observers OPred s (vals1 ++ vals2) (l1++l2) = 
      (Observers OPred s vals1 l1 * Observers OPred s vals2 l2)%logic.
  Proof.
    induction vals1; intros; destruct l1; simpl; inv L.
    + rewrite emp_sepcon; trivial.
    + rewrite IHvals1; [ rewrite sepcon_assoc |]; trivial.
  Qed.*)
Fixpoint Fold {A} (Pred:A -> val -> mpred) (States: list A) (l:list val) :=
    match States, l with
      nil, nil => emp
    | x::xs, o::os => (Pred x o * Fold Pred xs os)%logic
    | _, _ => FF
    end.
Lemma Fold_app {A} Pred: forall states1 l1 states2 l2 (L: length states1 = length l1),
      @Fold A Pred (states1 ++ states2) (l1++l2) = 
      (Fold Pred states1 l1 * Fold Pred states2 l2)%logic.
  Proof.
    induction states1; intros; destruct l1; simpl; inv L.
    + rewrite emp_sepcon; trivial.
    + rewrite IHstates1; [ rewrite sepcon_assoc |]; trivial.
  Qed.
Lemma Fold_length {A} Pred: forall states l,
      @Fold A Pred states l |-- !!(length states = length l).
  Proof. induction states; intros; destruct l; simpl; entailer!. Qed.

Lemma Fold_split1 {A} Pred: forall states1 states2 l, 
      @Fold A Pred (states1++states2) l |--
      EX l1, EX l2, !!(l1++l2 = l /\ length states1 = length l1) &&
                     Fold Pred states1 l1 * Fold Pred states2 l2.
Proof. intros. assert_PROP (length (states1++states2) = length l) as L. sep_apply (@Fold_length A); entailer!.
  rewrite app_length in L.
  Exists (firstn (length states1) l) (skipn (length states1) l).
  rewrite firstn_length_le, (firstn_skipn (length states1)) by lia. entailer!.
  rewrite <- Fold_app.
  + apply derives_refl'. f_equal. rewrite (firstn_skipn (length states1)); trivial.
  + rewrite firstn_length_le; trivial; lia.
Qed.
Lemma Fold_split2 {A} Pred: forall states l1 l2, 
      @Fold A Pred states (l1++l2) |--
      EX states1, EX states2, !!(states1++states2 = states /\ length states1 = length l1) &&
                                Fold Pred states1 l1 * Fold Pred states2 l2.
Proof. intros. assert_PROP (length states = length (l1++l2)) as L. sep_apply (@Fold_length A); entailer!.
  rewrite <- (firstn_skipn (length l1) states).
  Exists (firstn (length l1) states) (skipn (length l1) states).
  rewrite Fold_app; rewrite firstn_length_le; trivial; try (rewrite L, app_length; lia).
  entailer!.
Qed.

Lemma Fold_In_split {A} Pred states l x (X:In x l):
      @Fold A Pred states l |-- 
          EX s states1 states2 l1 l2, !!(length states1=length l1 /\ states=states1++s::states2/\l=l1++x::l2) &&
          Pred s x * @Fold A Pred states1 l1 * @Fold A Pred states2 l2.
Proof. 
  assert_PROP (length states = length l) as L by apply Fold_length.
  destruct (in_split _ _ X) as [l1 [l2 ?]]; subst.
  sep_apply (@Fold_split2 A). Intros states1 states2; subst.
  rewrite 2 app_length in L. simpl in L.
  assert (L2: Datatypes.length states2 = S (Datatypes.length l2)) by lia.
  destruct states2; inv L2. 
  Exists a states1 states2 l1 l2. simpl. entailer!.
Qed. 

Lemma Fold_combine {A} P s x: forall states1 states2 l1 l2,
      P s x * @Fold A P states1 l1 * Fold P states2 l2 |-- Fold P (states1++s::states2) (l1++x::l2).
Proof. induction states1; intros; destruct l1; simpl; entailer!. trivial. Qed.

Notation Observers := (@Fold ObserverRepTp).
Notation Notifiers := (@Fold NotifyRepTp).
Notation Subjects := (@Fold SubjectRepTp).

(*Definition updateO (s:val)(n:Z): ObserverRepTp -> ObserverRepTp := map (fun x => match x with (u,v) => (u, if Val.eq u s then n else v) end).*)

Definition updateO (s:val)(n:Z) (x:ObserverRepTp): ObserverRepTp := 
      match x with ((s1,b1),(s2,b2)) => ((s1, if Val.eq s1 s then n else b1), 
                                         (s2, if Val.eq s2 s then n else b2)) end.

Definition updateN (s:val)(n:Z) (x:NotifyRepTp): NotifyRepTp :=
      match x with ((s1,b1),(s2,b2)) => ((s1, if Val.eq s1 s then Some n else b1), 
                                         (s2, if Val.eq s2 s then Some n else b2)) end.

Require Import compcert.common.Values.
Section SubjectASI.
  Variable M: MallocFreeAPD.
  Variable SP: SubjectAPD.
  Variable OP: ObserverAPD.

  Notation S_rep := (Subjectrep SP).
  Notation R_Pre := (RegisterPre SP).
  Notation U_Pre := (UpdatePre SP).
  Notation G_Pre := (GetPre SP).
  Notation G_Post := (GetPost SP).
  Notation O_rep := (Observerrep OP).
  Notation N_rep:= (Notifyrep OP).

  Definition newSubject_spec :=
    DECLARE _newSubject
    WITH gv: globals
    PRE [ ] PROP() PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr SubjectTP ]
    EX p: val,
     PROP() RETURN (p) SEP(S_rep newSubject p; mem_mgr M gv).

  Definition get_smallspec :=
    DECLARE _get
    WITH s:val, S: SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(G_Pre S s)
    POST[ tint ]
      PROP() RETURN (Vint (Int.repr (snd S))) SEP(G_Post S s).

(*
  Definition update_spec2 :=
    DECLARE _update
    WITH s:val, n: Z, observers: list val, OldState1: NotifyRepTp, OldState2: NotifyRepTp,
         (*s11:val, s12:val, s21:val, s22:val, b11:option Z, b12:option Z, b21:option Z, b22:option Z, subjects1:list val, subjects2:list val,SUBJ: list SubjectRepTp*)
         x11: val * option Z, x12: val * option Z, x21: val * option Z, x22: val * option Z, 
         b:option Z, X:list (val * option Z), Y:list (val * option Z)
    PRE [ tptr SubjectTP, tint ]
     PROP(Int.min_signed <= n <= Int.max_signed /\
          OldState1 = (x11,x12) /\ OldState2=(x21,x22) /\
          [x11; x12; x21; x22] = X ++(s,b)::Y /\
          list_norepet (map fst (X ++(s,b)::Y)))
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(EX z, U_Pre (observers,z) s;
         Subjects G_Pre (map snd (X++Y)) (map fst (X+Y));
         Notifiers N_rep [OldState1;OldState2] observers)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (observers,n) s; Subjects G_Post (map snd (X++Y)) (map fst (X+Y));
          Notifiers N_rep (map (updateN s n) [OldState1;OldState2]) observers).
*)
 Fixpoint subjects_ofNtfys (l:list NotifyRepTp):list val :=
  match l with
    nil => nil
  | (h::t) => match h with ((s1,b1),(s2,b2)) => s1::s2::subjects_ofNtfys t end
  end.

 Fixpoint lookup (u:val) (SUBS: list SubjectRepTp) (subs:list val): option Z :=
    match SUBS, subs with
      | (SubH::SubT), (subh::subt) => if Val.eq u subh then Some(snd SubH) else lookup u SubT subt
      | _ , _ => None
    end.

  Fixpoint updateNtfys SUBJ subjects (OBS: list NotifyRepTp): list NotifyRepTp :=
  match OBS with
     nil => nil
  | (h::t) => match h with ((s1,b1),(s2,b2)) => 
               match lookup s1 SUBJ subjects, lookup s2 SUBJ subjects with
                 Some n1, Some n2 => updateN s1 n1 (updateN s2 n2 h)
               | _, _ => h
               end
             end :: updateNtfys SUBJ subjects t
  end.

  Definition update_spec :=
    DECLARE _update
    WITH s:val, n: Z, observers: list val, OBS: list NotifyRepTp,
         subjects:list val, SUBJ: list SubjectRepTp
    PRE [ tptr SubjectTP, tint ]
     PROP(Int.min_signed <= n <= Int.max_signed /\ 
          nodup Val.eq (remove Val.eq s (subjects_ofNtfys OBS)) = subjects)
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(EX z, U_Pre (observers,z) s;
         Subjects S_rep SUBJ subjects;
         Notifiers N_rep OBS observers)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (observers,n) s; Subjects S_rep SUBJ subjects;
          Notifiers N_rep (updateNtfys ((observers,n)::SUBJ) (s::subjects) OBS) observers).

  Definition register_spec :=
    DECLARE _registr
    WITH o:val, obs1:list val, obs2:list val, u1:val, u2:val, s1:val, s2:val, n1:Z, n2:Z, b1:option Z, b2:option Z, gv:globals
    PRE [ tptr SubjectTP, tptr ObserverTP  ]
     PROP (s1<>s2; ((s1,s2) = (u1,u2) \/ (s1,s2)=(u2,u1)))
     PARAMS (u1; o) GLOBALS (gv)
     SEP (R_Pre (obs1,n1) u1; G_Pre (obs2,n2) u2; mem_mgr M gv; N_rep ((s1,b1),(s2,b2)) o)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (S_rep (o::obs1, n1) u1; G_Post (obs2,n2) u2; mem_mgr M gv; O_rep ((s1, if Val.eq s1 u1 then n1 else n2),
                                                                            (s2, if Val.eq s1 u1 then n2 else n1)) o).

  Definition freeSubject_spec:=
    DECLARE _freeSubject 
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv)
     SEP(S_rep S s; mem_mgr M gv)
    POST[ tint ] EX r,
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr r)))
      SEP(if zeq r 0 then !!(fst S = @nil val)&&emp 
          else !!(r=1 /\ fst S <> @nil val) && S_rep S s; mem_mgr M gv).

  Definition detachfirst_spec:=
    DECLARE _detachfirst 
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv)
     SEP(S_rep S s; mem_mgr M gv)
    POST[ tptr ObserverTP ] EX r,
      PROP()
      LOCAL(temp ret_temp r)
      SEP(if Val.eq r nullval 
          then !!(fst S = @nil val) && S_rep S s 
          else EX O, !!(fst S=r::O) && S_rep (O, snd S) s; mem_mgr M gv).

  Definition SubjectASI:funspecs := [ detachfirst_spec; freeSubject_spec; newSubject_spec; register_spec; update_spec; get_smallspec].

  Definition get_largespec :=
    DECLARE _get
    WITH s:val, S: SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(S_rep S s)
    POST[ tint ]
      PROP() RETURN (Vint (Int.repr (snd S))) SEP(S_rep S s).

Lemma Get_sub: funspec_sub (snd get_smallspec) (snd get_largespec).
Proof.
  do_funspec_sub. Exists w. destruct w as [s [l n]]; clear H.
  Exists (GetPost SP (l, n) s -* Subjectrep SP (l, n) s).
  simpl; entailer!.
  { intros. rewrite <- H0. entailer!.
    apply wand_frame_elim''. }
  rewrite sepcon_comm. apply (Subject_GetPrePost SP (l, n)).
Qed.

Definition OrepTp2NRepTp (x:ObserverRepTp):NotifyRepTp.
Proof. destruct x as [[s1 c1] [s2 c2]]. apply ((s1, Some c1),(s2, Some c2)). Defined.

(*Motivated by a reviewer*)
Lemma Observers_Orep_NtfPre: forall ostates l,
      Observers O_rep ostates l |-- Notifiers N_rep (map OrepTp2NRepTp ostates) l.
Proof. unfold Notifiers.
 induction ostates; simpl; intros.
  + trivial.
  + destruct l; trivial. sep_apply (IHostates l). clear. cancel. unfold OrepTp2NRepTp.
    destruct a as [[s1 c1] [s2 c2]]. rewrite <- Observer_Notify. trivial.
Qed.

Lemma Observers_Orep_NtfPre': forall ostates l, 
      Observers O_rep ostates l = Notifiers N_rep (map OrepTp2NRepTp ostates) l.
Proof.
 induction ostates; simpl; intros.
  + trivial.
  + destruct l; trivial. rewrite (IHostates l). clear. f_equal. unfold OrepTp2NRepTp.
    destruct a as [[s1 c1] [s2 c2]]. rewrite <- Observer_Notify. trivial.
Qed.

(*Motivated by a reviewer*)
 Fixpoint subjects_ofOBS (l:list ObserverRepTp):list val :=
  match l with
    nil => nil
  | (h::t) => match h with ((s1,b1),(s2,b2)) => s1::s2::subjects_ofOBS t end
  end.
  Fixpoint updateObs SUBJ subjects (OBS: list ObserverRepTp): list ObserverRepTp :=
  match OBS with
     nil => nil
  | (h::t) => match h with ((s1,b1),(s2,b2)) => 
               match lookup s1 SUBJ subjects, lookup s2 SUBJ subjects with
                 Some n1, Some n2 => updateO s1 n1 (updateO s2 n2 h)
               | _, _ => h
               end
             end :: updateObs SUBJ subjects t
  end.
  Definition update_Obs_spec :=
    DECLARE _update
    WITH s:val, n: Z, observers: list val, OBS: list ObserverRepTp,
         subjects:list val, SUBJ: list SubjectRepTp
    PRE [ tptr SubjectTP, tint ]
     PROP(Int.min_signed <= n <= Int.max_signed /\ 
          nodup Val.eq (remove Val.eq s (subjects_ofOBS OBS)) = subjects)
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(EX z, U_Pre (observers,z) s;
         Subjects S_rep SUBJ subjects;
         Observers O_rep OBS observers)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (observers,n) s; Subjects S_rep SUBJ subjects;
          Observers O_rep (updateObs ((observers,n)::SUBJ) (s::subjects) OBS) observers).
 (* Definition update_Obs_spec :=
    DECLARE _update
    WITH s:val, n: Z, observers:list val, OldOStates: list ObserverRepTp
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(EX z, U_Pre (observers,z) s;
         Observers O_rep OldOStates observers)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (observers,n) s; Observers O_rep (map (updateO s n) OldOStates) observers).*)
(*Definition update_Obs_spec :=
    DECLARE _update
    WITH s:val, n: Z, observers:list val, OldOStates: list ObserverRepTp
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(EX z, U_Pre (observers,z) s;
         Observers O_rep OldOStates observers)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (observers,n) s; Observers O_rep (map (updateN s n) OldOStates) observers).*)
Lemma subjects_of: forall Obs, subjects_ofNtfys (map OrepTp2NRepTp Obs) = subjects_ofOBS Obs. 
Proof. induction Obs; simpl. trivial.
  destruct a. destruct p. destruct p0. simpl. rewrite IHObs. trivial.
Qed.

Lemma update_eq: forall Obs a b, updateNtfys b a (map OrepTp2NRepTp Obs) = map OrepTp2NRepTp (updateObs b a Obs).
Proof. induction Obs; simpl; intros; trivial.
  destruct a; simpl. destruct p. destruct p0. f_equal; trivial.
  destruct (lookup v b a0); simpl; trivial.
  destruct (lookup v0 b a0); simpl; trivial. f_equal; f_equal.
  + destruct (Val.eq v v); subst; simpl; trivial. 
    destruct (Val.eq v v0); subst; simpl; trivial.
  + destruct (Val.eq v0 v); subst; simpl; trivial. 
    destruct (Val.eq v0 v0); subst; simpl; trivial.
Qed.

Lemma Upd_sub: funspec_sub (snd update_spec) (snd update_Obs_spec).
Proof.
  do_funspec_sub. destruct w as [[[[[s n] l] Obs] subject] SUBS]; clear H.
  Exists (((((s, n), l), map OrepTp2NRepTp Obs), subject), SUBS) emp.
  rewrite subjects_of, update_eq. simpl. entailer!.
  + intros. rewrite Observers_Orep_NtfPre'. cancel.
  + apply Observers_Orep_NtfPre. 
Qed.

End SubjectASI.
