Require Import spec_stdlib. (*for malloc_token*)
Require Import Observer.
Require Import spec_Observer.

Instance ObserverCompSpecs : compspecs. make_compspecs prog. Defined.

Section ObserverAP.
Variable M: MallocFreeAPD.
(*
Definition comparable s := EX sh n, (*!!(readable_share sh) &&*) malloc_token' M sh n s.

Lemma comparable_validptr s: comparable s |-- valid_pointer s.
Proof. unfold comparable. Intros sh n. apply malloc_token'_valid_pointer. Qed. 

Lemma comparable_isptr s: comparable s |-- !! (isptr s) && comparable s.
Proof. unfold comparable. Intros sh n. Exists sh n. sep_apply malloc_token'_local_facts. Intros. 
red in H. destruct s; try contradiction.  entailer!. Qed.
 *)
Definition Observer_rep (O:ObserverRepTp) (o: val): mpred :=
  EX s1 s2 c1 c2,
  !!(O = ((s1,c1), (s2,c2)) /\ Int.min_signed <= c1 <= Int.max_signed /\ Int.min_signed <= c2 <= Int.max_signed /\ s1<>s2) &&
  @malloc_token ObserverCompSpecs M Ews ObserverTP o * (*comparable s1 * comparable s2 **)
  @data_at ObserverCompSpecs Ews ObserverTP (s1, (s2, (Vint (Int.repr c1), Vint (Int.repr c2)))) o.

Definition Observer_rep_char O o: Observer_rep O o = 
  EX s1 s2 c1 c2,
  !!(O = ((s1,c1), (s2,c2)) /\ Int.min_signed <= c1 <= Int.max_signed /\ Int.min_signed <= c2 <= Int.max_signed /\ s1<>s2) &&
  @malloc_token ObserverCompSpecs M Ews ObserverTP o * (*comparable s1 * comparable s2 **)
  @data_at ObserverCompSpecs Ews ObserverTP (s1, (s2, (Vint (Int.repr c1), Vint (Int.repr c2)))) o.
Proof. reflexivity. Qed.

Definition Notify_rep (N:NotifyRepTp) (o:val): mpred :=
  EX s1 s2 b1 b2,
  !!(N = ((s1,b1), (s2,b2)) /\ s1<>s2) &&
  (malloc_token M Ews ObserverTP o * (*comparable s1 * comparable s2 **)
   EX v1 v2, !!((match b1 with None => v1=Vundef | Some z1 => Int.min_signed <= z1 <= Int.max_signed /\ v1=Vint (Int.repr z1) end) /\ 
                (match b2 with None => v2=Vundef | Some z2 => Int.min_signed <= z2 <= Int.max_signed /\ v2=Vint (Int.repr z2) end)) && 
   @data_at ObserverCompSpecs Ews ObserverTP (s1, (s2, (v1, v2))) o)%logic.

(*We instantiate ValPre by Observer_rep*)

Lemma  Observer_Notify s1 s2 v1 v2 o: Notify_rep ((s1, Some v1),(s2, Some v2)) o = Observer_rep ((s1,v1),(s2,v2)) o.
Proof.
  unfold Observer_rep, Notify_rep. f_equal. extensionality u1. f_equal. extensionality u2. 
  apply pred_ext. 
  + Intros b1 b2. inv H. normalize. Exists v1 v2. entailer!.
  + Intros c1 c2. inv H. Exists (Some c1) (Some c2) (Vint (Int.repr c1)) (Vint (Int.repr c2)).
    entailer!.
Qed.

Lemma Notify_rep_char N o:
      Notify_rep N o =
  EX s1 s2 b1 b2,
  !!(N = ((s1,b1), (s2,b2)) /\ s1<>s2) &&
  (malloc_token M Ews ObserverTP o * (*comparable s1 * comparable s2 **)
   EX v1 v2, !!((match b1 with None => v1=Vundef | Some z1 => Int.min_signed <= z1 <= Int.max_signed /\ v1=Vint (Int.repr z1) end) /\ 
                (match b2 with None => v2=Vundef | Some z2 => Int.min_signed <= z2 <= Int.max_signed /\ v2=Vint (Int.repr z2) end)) && 
   @data_at ObserverCompSpecs Ews ObserverTP (s1, (s2, (v1, v2))) o)%logic.
Proof. reflexivity. Qed.

Lemma Observer_rep_idem O o: Observer_rep O o |-- Observer_rep O o. Proof. trivial. Qed. 
  
Lemma Notify_rep_isptr N o: Notify_rep N o |-- !!(isptr o).
Proof. unfold Notify_rep. Intros s1 s2 c1 c2 v1 v2.  simpl; entailer!. Qed.

Lemma Observer_rep_Range O o: Observer_rep O o |-- !!(match O with ((s1,v1),(s2,v2)) => Int.min_signed <= v1 <= Int.max_signed /\ Int.min_signed <= v2 <= Int.max_signed end).
Proof. rewrite Observer_rep_char. Intros s1 s2 c1 c2. subst. entailer!. Qed.

Lemma Notify_rep_DistinctSubjects N o: Notify_rep N o |-- !!(match N with ((s1,v1),(s2,v2)) => s1<>s2 end).
Proof. unfold Notify_rep. Intros s1 s2 c1 c2 v1 v2. subst. entailer!. Qed.

Opaque Observer_rep.
Opaque Notify_rep.
(*
Lemma Notify_sub SR GPre GPost (H: forall S s, SR S s |-- GPre S s * (GPost S s -* SR S s)):
  funspec_sub (snd (notify_smallspec GPre GPost Observer_rep NotifyPre))
              (snd (notify_largespec SR Observer_rep NotifyPre)).
Proof.
  apply NDsubsume_subsume.
  { simpl; intros; split; trivial. }
  split3; trivial. intros [[[o s] c] S]. simpl; intros rho. 
  apply andp_left2. specialize (H S s); rewrite sepcon_comm in H. 
  Exists (((o, s), c), S) (GPost S s -* SR S s).
  unfold PROPx, LOCALx, SEPx; simpl. unfold local, liftx, lift1, lift; simpl.
  entailer!; [intros | trivial].
  cancel. rewrite sepcon_comm. apply wand_frame_elim. 
Qed.*)

Definition ObserverAP: ObserverAPD:=
  Build_ObserverAPD _ (*Observer_rep_isptr*) _ Notify_rep_isptr Observer_rep
                       Observer_Notify Observer_rep_idem Notify_rep_DistinctSubjects.
End ObserverAP.