Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*For MallocFreeAPD.*)
Require Import spec_Subject.
Require Import spec_Observer.
Require Import main.

(*We verify the body of main abstractly in M, OP and SP.
  We just provide the semax_body statement, no VSU.
  The spec of main is not VST standard spec, but a variant
  that can be instantiated to contain the appropriate initialization
  predidates, given suitable implementations (ie VSUs) of the code modules
  and the library.*)
Section MAIN_Verif.
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.
  Definition S_ASI:= SubjectASI M SP OP.
  Definition O_ASI:= ObserverASI M OP SP.

  (*Specification of "module-instantiated main" - we permit an arbitrary G here too.*)
  Definition main_inst_spec G:=
  DECLARE _main
  WITH gv: globals
  PRE [ ]
    PROP ()
    PARAMS () GLOBALS (gv)
    SEP (fold_right (fun g p => g gv * p) emp G; has_ext tt)
  POST[ tint ]
    PROP()
    LOCAL(temp ret_temp (Vint (Int.repr 14)))
    SEP((*mem_mgr M gv; *)TT;has_ext tt).

  Definition MainImports := [ newObserver_spec M OP SP; val_spec OP; 
   freeObserver_spec M OP; newSubject_spec M SP; update_spec SP OP;
   detachfirst_spec M SP; freeSubject_spec M SP].

  Definition MainInitPred: list (globals -> mpred) := nil.
  Definition maininstspec := (main_inst_spec [mem_mgr M]).

  (*Definition main_internal_specs: funspecs := [mainspec (*M*)].*)

  Definition MainVprog : varspecs. mk_varspecs main.prog. Defined.
  Definition MainGprog:funspecs := MainImports(* ++ main_internal_specs*).

  Instance MainCompSpecs : compspecs. make_compspecs main.prog. Defined.
  Lemma body_main: semax_body MainVprog MainGprog f_main maininstspec.
  Proof.
  start_function. simpl; Intros. 
  (*sep_apply (spec_stdlib.make_mem_mgr M gv). *)

  (*s1 = newSubject *)
  forward_call. Intros s1. simpl. (*nothing happens as SD is abstract*)

  (*s2= newSubject*)
  forward_call. Intros s2. simpl. (*nothing happens as SD is abstract*)
  sep_apply Subject_conflict; Intros. rename H into SubjectsDistinct.

  (*o = newObserver*s1,s2*)
  do 2 sep_apply Subject_Register. unfold newSubject. simpl.
  forward_call (gv,s1, s2, @nil val, @nil val, 0, 0). { simpl; cancel. }
  Intros o1.
  simpl. (*only the definitions of newSubject and Subjectrep (both model functions!) are simplified*) 

  (*update(s1,2)*)
  sep_apply Subject_Update. simpl.

  forward_call (s1, 2, [o1], [(s1, Some 0, (s2, Some 0))], [s2], [([o1], 0)]).
  { simpl. Exists 0. cancel. unfold newObserver. simpl. rewrite <- Observer_Notify. cancel. }
  { simpl. rewrite if_true by trivial. rewrite if_false by congruence. simpl. split; trivial. rep_lia. }
  simpl.
  destruct (Val.eq s1 s1); [ clear e | congruence]. 
  destruct (Val.eq s2 s2); [ clear e | congruence].
  destruct (Val.eq s2 s1); [ congruence | clear n].
  Intros.
  rewrite Observer_Notify.

  (*r1=val(o)*)
  sep_apply Observer_Val.
  forward_call. rep_lia.

  (*update(s1,5)*)
  sep_apply Subject_Update. simpl.
  forward_call (s1, 5, [o1], [(s1, Some 2, (s2, Some 0))], [s2], [([o1], 0)]).
  { simpl. Exists 2. cancel. rewrite <- Observer_Notify. cancel. }
  { simpl. rewrite if_true by trivial. rewrite if_false by congruence. simpl. split; trivial. rep_lia. }
  simpl.
  destruct (Val.eq s1 s1); [ clear e | congruence]. 
  destruct (Val.eq s2 s2); [ clear e | congruence].
  destruct (Val.eq s2 s1); [ congruence | clear n].
  Intros.
  rewrite Observer_Notify.

  (*update(s2,7)*)
  sep_apply (Subject_Update SP ([o1], 0)). simpl.
  forward_call (s2, 7, [o1], [(s1, Some 5, (s2, Some 0))], [s1], [([o1], 5)]).
  { simpl. Exists 0. cancel. rewrite <- Observer_Notify. cancel. }
  { simpl. rewrite if_false by congruence. rewrite if_true by trivial. simpl. split; trivial. rep_lia. }
  simpl.
  destruct (Val.eq s1 s1); [ clear e | congruence]. 
  destruct (Val.eq s2 s2); [ clear e | congruence].
  destruct (Val.eq s2 s1); [ congruence | clear n].
  destruct (Val.eq s1 s2); [ congruence | clear n].
  Intros.
  rewrite Observer_Notify.
  
  (*r2=val(o)*)
  sep_apply Observer_Val.
  forward_call. rep_lia.

  (*p1=detachfirst(s1)*)
  forward_call. Intros p1. simpl. if_tac; Intros z. inv H0; deadvars.

  (*p2=detachfirst(s2)*)
  forward_call. Intros o1. simpl. if_tac; Intros z. inv H1; deadvars.

  (*freeSubject(s1)*)
  forward_call; Intros v1. simpl. if_tac; Intros; [ subst; simpl | congruence].

  (*freeSubject(s2)*)
  forward_call; Intros v2. simpl. if_tac; Intros; [ subst; simpl | congruence].

  (*freeObserver(p1)*)
  forward_call (gv, o1, (s1, 5, (s2, 7))).

  forward.
  cancel.
  Qed.

End MAIN_Verif.