Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import spec_Subject.
Require Import spec_Observer.
Require surelyMalloc.
Require Import main.
Require Import spec_main.

Instance CompSpecs : compspecs. make_compspecs main.prog. Defined.

Section MainVSU.
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.
  Definition S_ASI:= SubjectASI M SP OP.
  Definition O_ASI:= ObserverASI M OP SP.

  Definition MainImports := [ newObserver_spec M OP SP; val_spec OP; 
   freeObserver_spec M OP; newSubject_spec M SP; update_spec SP OP;
   detachfirst_spec M SP; freeSubject_spec M SP].

  Definition main_internal_specs: funspecs :=
    [main_spec M prog].

  Definition MainVprog : varspecs. mk_varspecs (*linked_prog*)main.prog. Defined.

  Definition MainGprog:funspecs := MainImports ++ main_internal_specs.

Lemma body_main: semax_body MainVprog MainGprog f_main (main_spec M (*linked_prog*)main.prog).
Proof.
start_function.
sep_apply (spec_stdlib.make_mem_mgr M gv). 
(*sep_apply (make_apile gv).
generalize (make_onepile gv).
assert (change_composite_env spec_onepile.CompSpecs CompSpecs).
make_cs_preserve spec_onepile.CompSpecs CompSpecs.
change_compspecs CompSpecs.
 *)
forward_call. Intros s. simpl. (*nothing happens as SD is abstract*)
sep_apply Subject_Register. 
forward_call (gv,s, newSubject). Intros o1. simpl. (*only the definition of newSubject is simplified*) 
sep_apply Subject_Update.
sep_apply Observer_Notify.
forward_call (s, 2, o1::fst newSubject). 
{ Exists (snd newSubject). simpl. cancel.
  Exists [0].
  simpl. unfold newObserver; simpl. cancel. }

simpl. (*again just the definition of newSubject is simplified*)
sep_apply Subject_Register. 
forward_call (gv,s, ([o1], 2)). Intros o2. simpl.

sep_apply Subject_Update.
sep_apply Observer_Notify.
sep_apply Observer_Notify.
forward_call (s, 3, [o2;o1]). 
{ Exists 2. cancel.
  Exists [2;2]. 
  simpl. unfold newObserver; simpl. cancel. }

simpl. Intros.
sep_apply Observer_Val.
sep_apply Observer_Val.
forward_call (o1,s,3).
forward_call (o2,s,3).
forward.
forward_call. Intros p1. simpl. if_tac; Intros z. inv H0; deadvars.
forward_call. Intros p2. simpl. if_tac; Intros z. inv H1; deadvars.
forward_call; Intros v. simpl. if_tac; Intros; [ subst; simpl | congruence].
subst; simpl. 
forward_call (gv, p1, (s,3)).
forward_call (gv, p2, (s,3)).
forward.
Qed.

Definition MainComponent: @Component NullExtension.Espec MainVprog CompSpecs 
      nil MainImports main.prog [main_spec M main.prog] emp main_internal_specs.
  Proof. 
    mkComponent. clear; solve_SF_internal body_main.
  Qed.

Definition MainCanComponent:=Comp_to_CanComp MainComponent. 

Definition MainVSU: @VSU NullExtension.Espec MainVprog CompSpecs 
      nil MainImports main.prog [main_spec M main.prog] emp.
  Proof. eexists; apply MainComponent. Qed.

Definition MainProgVSU: @LinkedProgVSU NullExtension.Espec MainVprog CompSpecs 
      nil MainImports main.prog [main_spec M main.prog] emp.
Proof. 
 eexists. apply MainCanComponent. 
 destruct MainCanComponent as [G [HG1 HG2]].
 eexists; split. apply CCM_main. reflexivity.
Qed.
End MainVSU.