Require Import VST.floyd.VSU.
Require Import stdlib.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Subject.
Require Import Observer. (*needed for defining Vprog *)
Require Import spec_Observer.

Require Import verif_Observer. (*get access to the definition of ObserverREP etc*)
Opaque Observer_rep.
Opaque NotifyPre.
(*Opaque ValPre.*)

Section Observer_VSU.
  Variable M: MallocFreeAPD.
  Variable SP: SubjectAPD.
  Notation S_rep := (Subjectrep SP).
  Notation RegPre := (RegisterPre SP).
  Notation GetPre := (GetPre SP).
  Notation GetPost := (GetPost SP).
  Notation S_ptrnull:= (Subjectrep_ptrnull SP).

  Notation OP:= (ObserverAP M). (*Instantiate the ObserverAP definined in verif_observer to M*)  

  Definition subject_ASI := SubjectASI M SP OP.

  Definition ObserverImports:= 
           [free_spec' M _free] ++ SM_ASI M ++ [register_spec M SP OP; get_smallspec SP].

  Definition obs_internal_specs: funspecs :=
    [newObserver_spec M OP SP; notify_smallspec OP SP; val_spec OP; freeObserver_spec M OP].

  Definition OVprog : varspecs. mk_varspecs prog. Defined.

  Definition OGprog:funspecs := ObserverImports ++ obs_internal_specs.

  Lemma val_verif:
    semax_body OVprog OGprog f_val (val_spec OP).
  Proof. start_function. simpl. rewrite Observer_rep_char; Intros. do 2 forward. Qed.

  Lemma freeObserver_verif:
    semax_body OVprog OGprog f_freeObserver (freeObserver_spec M OP).
  Proof.
   start_function. simpl. rewrite Observer_rep_char; Intros.
   rewrite data_at_isptr; unfold malloc_token; Intros.
   destruct o; try solve [inv Po]. 
   forward_call (sizeof ObserverTP, Vptr b i, gv).
   { sep_apply data_at_memory_block. simpl. cancel. }
   forward.
  Qed.

  Lemma notify_verif: semax_body OVprog OGprog f_notify (notify_smallspec OP SP).
  Proof.
    start_function. simpl. rewrite NotifyPre_char; Intros v.
    sep_apply GetPre_ptrnull; Intros.
    forward.
    forward_call.
    do 2 forward. simpl. rewrite Observer_rep_char. cancel. 
  Qed.
(*  Lemma notify_verif:
    semax_body OVprog OGprog f_notify (notify_spec S_rep GetPre Observer_rep (spec_Observer.NotifyPre OD)).
  Proof.
    start_function. unfold spec_Observer.NotifyPre, OD, NotifyPre; simpl. Intros v.
    rewrite GetPre_ptrnull; Intros.
    forward.
    forward_call.
    do 2 forward. unfold Observer_rep. cancel.
  Qed.*)

  Lemma newObserver_verif: semax_body OVprog OGprog f_newObserver (newObserver_spec M OP SP).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (ObserverTP, gv). 
    { simpl. split3; trivial. rep_lia. }
    Intros o. 
    forward. simpl.
    destruct S as [l n].
    forward_call (s, l, n, o, gv).
    {  Exists 0. simpl. rewrite NotifyPre_char. Exists Vundef. cancel. }
    forward.
    Exists o; simpl. entailer!.
  Qed.

  Definition ObserverComponent: @Component NullExtension.Espec OVprog _ 
      nil ObserverImports  prog (ObserverASI M OP SP) emp obs_internal_specs.
  Proof.
    mkComponent.
  + solve_SF_internal newObserver_verif.
  + solve_SF_internal notify_verif.
  + solve_SF_internal val_verif.
  + solve_SF_internal freeObserver_verif.
  Qed.

  Definition ObserverVSU: @VSU NullExtension.Espec OVprog _ 
      nil ObserverImports  prog (ObserverASI M OP SP) emp.
  Proof. eexists; apply ObserverComponent. Qed.

End Observer_VSU.
(*Check ObserverVSU. : forall SP : SubjectPredicates,
       VSU [] (ObserverImports SP) prog (ObserverASI ObserverPreds SP)*)

Definition Observer_CanVSU M SP := VSU_to_CanonicalVSU  (ObserverVSU M SP).