Require Import VST.floyd.VSU.
Require Import stdlib.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Observer.
Require Import Subject. (*needed for defining Vprog *)
Require Import spec_Subject.

Require Import verif_Subject. (*get access to the definition of SubjectREP etc*)

(*Maybe these opacity declarations can be eliminaed?*)
Opaque listrep.
Opaque Subject_rep.
Opaque GetPrePost.

Section Subject_VSU.
Variable M: MallocFreeAPD.

Definition add_spec :=
  DECLARE _add
   WITH n:val, o:val, os:list val, gv: globals
   PRE [ tptr NodeTP, tptr ObserverTP ]
       PROP (isptr o)
       PARAMS (n; o) GLOBALS (gv)
       SEP (mem_mgr M gv; listrep M os n)
    POST [ tptr NodeTP ] EX p:_, 
       PROP ()
       LOCAL (temp ret_temp p)
       SEP (mem_mgr M gv; listrep M (o::os) p).

Lemma Observers_length O_rep s: forall vals l,
    Observers O_rep s vals l =
    !!(length vals = length l) && Observers O_rep s vals l.
Proof.
  induction vals; destruct l; simpl; apply pred_ext; entailer.
  rewrite IHvals. Intros. rewrite H. entailer!.
Qed.

  Variable OP: ObserverAPD. 
  Notation O_rep := (Observerrep OP).
  Notation NotPre := (NotifyPre OP).

  Notation SP := (SubjectAP M). (*Instantiate the SubjectAP definined in verif_subject to M*)

  Notation S_rep := (Subjectrep SP).
  Notation RegPre := (RegisterPre SP).
  Notation GtPre := (GetPre SP).
  Notation GtPost := (GetPost SP).
  Notation UpdPre := (UpdatePre SP).

  Definition SVprog : varspecs. mk_varspecs prog. Defined.

  Definition SubjectImports :=
    [free_spec' M _free] ++ SM_ASI M ++ [notify_smallspec OP SP].

  Definition subject_internal_specs: funspecs :=
    [add_spec; newSubject_spec M SP; register_spec M SP OP; update_spec SP OP;
     get_smallspec SP; detachfirst_spec M SP; freeSubject_spec M SP].

  Definition SGprog:funspecs := SubjectImports ++ subject_internal_specs.

  Lemma add_verif:
    semax_body SVprog SGprog f_add add_spec.
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (NodeTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros p. 
    do 3 forward.
    Exists p. entailer!. apply listrep_fold.  trivial. 
  Qed.

  Lemma newSubject_verif:
    semax_body SVprog SGprog f_newSubject (newSubject_spec M SP).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (SubjectTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros s. 
    do 3 forward. Exists s; entailer!. simpl.
    rewrite <- Subject_rep_newSubject. cancel.
  Qed.

  Lemma freeSubject_verif:
    semax_body SVprog SGprog f_freeSubject (freeSubject_spec M SP).
  Proof.
    start_function. destruct S as [O c]. simpl.
    rewrite Subject_rep_char; Intros o.
    rewrite listrep_ptrnull; Intros.
    forward.
    forward_if.
    { apply denote_tc_test_eq_split. do 3 apply sepcon_valid_pointer1. apply listrep_valid_pointer. 
      apply valid_pointer_null. }
    { destruct O. rewrite listrep_nil; Intros. subst; contradiction.
      forward. Exists 1. simpl. entailer!. discriminate. rewrite Subject_rep_char.
      Exists o. cancel. }
    { subst. destruct O.
      + rewrite listrep_nil; Intros.
        unfold malloc_token; rewrite data_at_isptr; Intros. destruct s; try inv Ps.
        forward_call (sizeof SubjectTP, Vptr b i, gv).
        { sep_apply data_at_memory_block; simpl; cancel. }
        forward. Exists 0; simpl; entailer!.
      + sep_apply listrep_unfold; Intros n. rewrite data_at_isptr; Intros. inv Pnullval. }
  Qed.

  Lemma detachfirst_verif:
    semax_body SVprog SGprog f_detachfirst (detachfirst_spec M SP).
  Proof.
    start_function. destruct S as [O c]. simpl.
    rewrite Subject_rep_char; Intros o.
    rewrite listrep_ptrnull; Intros.
    forward.
    forward_if.
    { apply denote_tc_test_eq_split. do 3 apply sepcon_valid_pointer1. apply listrep_valid_pointer. 
      apply valid_pointer_null. }
    { subst. forward.
      Exists nullval; simpl; entailer!. rewrite if_true; trivial.
      rewrite Subject_rep_char.
      Exists nullval. entailer.
      destruct O. entailer!.
      sep_apply listrep_unfold; Intros n. rewrite data_at_isptr; Intros. inv Pnullval. }
    { destruct O.
      + rewrite listrep_nil; Intros. subst; inv H.
      + sep_apply listrep_unfold; Intros n. rewrite listrep_ptrnull; Intros.
        forward. forward. forward.
        forward_call (sizeof NodeTP, o, gv).
        { rewrite if_false. 2: intros N; subst; inv H.
          unfold malloc_token at 1; Intros; sep_apply data_at_memory_block; cancel. }
        forward. Exists v; entailer. rewrite if_false. 2: intros Q; subst; inv Pv.
        Exists O. simpl. rewrite Subject_rep_char. Exists n. entailer!. }
  Qed.

  Lemma get_verif:
    semax_body SVprog SGprog f_get (get_smallspec SP).
  Proof. start_function. simpl. rewrite GetPrePost_char; Intros. do 2 forward. Qed.

  Lemma register_verif: semax_body SVprog SGprog f_registr (register_spec M SP OP).
  Proof.
    start_function. Intros z. simpl. rewrite Subject_rep_char.
    Intros p. sep_apply NotifyPre_isptr; rewrite listrep_ptrnull; Intros.
    forward.
    forward_call. 
    Intros r. forward. forward.
    forward_call (o, s, z, (o::l,n)).
    { simpl. unfold_data_at (data_at _ _ _ _). rewrite GetPrePost_char. cancel. }
    forward.
    simpl. rewrite Subject_rep_char, GetPrePost_char. Exists r; unfold_data_at (data_at _ _ _ _); cancel. 
  Qed.

  Fixpoint spine (l: list (val*val)) s z: mpred :=
    (match l with
      nil => !!(s=z /\ is_pointer_or_null s) && emp
    | (o,n)::os => !!(isptr o) && @data_at SCompSpecs Ews NodeTP (o,n) s * @malloc_token SCompSpecs M Ews NodeTP s *
                   spine os n z
    end)%logic.

  Lemma spine_nil p (P:is_pointer_or_null p): spine [] p p = emp.
  Proof. apply pred_ext; simpl; entailer!. Qed.

  Lemma spine_nilD o p : spine [] o p |-- !!(o=p) && emp.
  Proof. simpl; Intros; subst. entailer!. Qed.
  
  Lemma spine_ptr_or_null: forall l s z,
      spine l s z = !!(is_pointer_or_null s) && spine l s z.
  Proof.
    induction l; simpl; intros; apply pred_ext; try solve [entailer!].
    destruct a. rewrite IHl; clear IHl; Intros. entailer!.
  Qed.
  
  Lemma spine_ptr_or_null_precise: forall l s z,
      spine l s z = !!(match l with nil => s=z | _ => isptr s end) && spine l s z.
  Proof.
    induction l; simpl; intros; apply pred_ext; try solve [entailer!].
    destruct a. rewrite IHl; clear IHl; Intros. entailer!.
  Qed.
  
  Lemma spine_lst: forall l s z,
      spine l s z =
      !!(match l with
           nil => (s = z)
         | _ => exists ls o, l=ls++[(o,z)] end)
        && spine l s z.
  Proof.
    induction l; simpl; intros; apply pred_ext; try solve [entailer!].
    specialize (@exists_last _  (a::l)); intros.
    destruct X as [ll [[u1 u2] U]]. congruence. rewrite U. 
    destruct a. rewrite IHl; clear IHl; Intros.
    destruct l.
    + destruct H; subst. destruct ll. inv U; simpl; Intros; subst. entailer!.
      exists nil, u1; reflexivity. 
      inv U. elim (app_cons_not_nil ll nil _ H1).
    + destruct ll; inv U. destruct H as [ls [o HH]]. rewrite ! HH.
      entailer!. split.
      - exists ((v, v0) ::ls), o. 
        rewrite <- 2 app_comm_cons, <- HH, H2; trivial.
      - exists ls, o; split; trivial.
  Qed.

  Lemma spine_cons o n l s z (P: isptr o):
    (data_at Ews NodeTP (o, n) s * malloc_token M Ews NodeTP s * spine l n z)%logic
    |-- spine ((o,n)::l) s z.
  Proof. simpl. entailer!. Qed.
  
  Lemma spine_consD o n l s z:
    (spine ((o,n)::l) s z |-- !!(isptr o) && data_at Ews NodeTP (o, n) s * malloc_token M Ews NodeTP s * spine l n z)%logic.
  Proof. simpl; trivial. Qed.

  Lemma spine_app: forall l1 l2 s u z,
     (spine l1 s u * spine l2 u z |-- spine (l1++l2) s z)%logic.
  Proof. induction l1; simpl; intros; Intros. subst; cancel.
         destruct a. cancel; trivial. 
  Qed. 

  Lemma spine_listrep: forall l s, listrep M l s = EX sp, !!(map fst sp =l) && spine sp s nullval.
  Proof.
    induction l; simpl; intros; apply pred_ext.
    + Intros; subst. Exists (@nil (val*val)). rewrite listrep_nil. simpl; entailer!.
    + Intros sp. destruct sp; [ | inv H].
      simpl; Intros; subst. rewrite listrep_null; trivial.
    + sep_apply listrep_unfold. Intros n. rewrite IHl; clear IHl.
      Intros sp; subst. Exists ((a,n)::sp). simpl; entailer!. 
    + Intros sp. destruct sp; inv H. destruct p; simpl; Intros.
      eapply derives_trans. 2: eapply listrep_fold; trivial. rewrite IHl; clear IHl; cancel.
      Exists sp; entailer!. 
  Qed. 

  Lemma spine_validptr: forall sp p, isptr p -> spine sp p nullval |-- valid_pointer p.
  Proof.
    induction sp; simpl; intros; Intros; subst. apply valid_pointer_null.
    destruct a. entailer!.
  Qed.

  Opaque spine.

  Lemma update_verif: semax_body SVprog SGprog f_update (update_spec SP OP).
  Proof.
    start_function. Intros z vals.
    simpl. rewrite Subject_rep_char; Intros o. rewrite listrep_ptrnull; Intros.
    do 2 forward. rewrite spine_listrep. Intros sp.
    forward_while 
      (EX p, EX sp1, EX sp2,
       PROP (sp=sp1++sp2 )
            LOCAL (temp _nd p; temp _s s; temp _n (Vint (Int.repr n)))
            SEP (spine sp1 o p; spine sp2 p nullval; malloc_token M Ews SubjectTP s;
                 data_at Ews SubjectTP (o, Vint (Int.repr n)) s;
                 Observers O_rep s (list_repeat (length sp1) n) (map fst sp1);
                 Observers NotPre s (skipn (length sp1) vals) (map fst sp2)))%assert.
    { Exists o (@nil (val*val)) sp. entailer!. rewrite spine_nil; trivial. simpl; cancel.  }
    { entailer.
      red in PNp. destruct p; try solve [contradiction].
      + destruct Archi.ptr64. contradiction. subst; entailer!.
      + 
        apply denote_tc_test_eq_split; entailer.
        do 4 apply sepcon_valid_pointer1. apply sepcon_valid_pointer2.
        apply spine_validptr. trivial. } 
    { subst sp. rewrite Observers_length with (l:=map fst sp2); Intros.
      destruct sp2; [ sep_apply spine_nilD; Intros; subst; contradiction | simpl].
      destruct p0 as [obs nxt]; Intros. simpl.
      remember (skipn (Datatypes.length sp1) vals) as w.
      destruct w. inv H0. simpl. sep_apply spine_consD; Intros. sep_apply NotifyPre_isptr; Intros.
      forward.

      forward_call (obs, s, z0, (obs::map fst sp2, n)).
      { simpl. rewrite GetPrePost_char.
        unfold_data_at (data_at Ews SubjectTP _ _). cancel. } 
      simpl. rewrite (spine_ptr_or_null sp2); Intros.
      forward.
      Exists (nxt, sp1++[(obs, nxt)], sp2).
      entailer!.
      + rewrite <- app_assoc; trivial.
      + rewrite GetPrePost_char. unfold_data_at (data_at Ews SubjectTP _ _); cancel.
        rewrite map_app, app_length, <- list_repeat_app, Observers_app.
        2: rewrite length_list_repeat, map_length; trivial.
        simpl; cancel.
        apply sepcon_derives.
        - eapply derives_trans. 2: apply spine_app with (u:=p).
          cancel. eapply derives_trans. 2: apply spine_cons; trivial. rewrite spine_nil; [ cancel | trivial]. 
        - apply derives_refl'. f_equal. clear - Heqw.
          rewrite plus_comm. simpl.
          generalize dependent vals. generalize dependent w. generalize dependent z0.
          induction sp1; simpl; intros.
          * subst; trivial. 
          * destruct vals. inv Heqw. apply (IHsp1 _ _ _ Heqw). }
    forward.
    simpl S_rep; rewrite Subject_rep_char. entailer!. simpl. Exists o. cancel.
    rewrite (spine_ptr_or_null_precise sp2); Intros.
    destruct sp2; simpl; Intros; [ clear H | contradiction]. rewrite spine_nil; [|trivial].
    rewrite app_nil_r, map_length, spine_listrep.
    Exists sp1. entailer!.
    clear.
    generalize dependent vals.
    induction sp1; destruct vals; simpl; trivial; entailer!.
  Qed.
Lemma update_Obs_verif: semax_body SVprog SGprog f_update (update_Obs_spec SP OP).
  Proof.
    start_function. Intros z vals.
    replace_SEP 1 (Observers NotPre s vals l).
    { entailer!. generalize dependent l; clear. induction vals; simpl; intros.
      destruct l; simpl; trivial.
      destruct l; simpl; trivial. sep_apply Observer_Notify. cancel. }
    simpl. rewrite Subject_rep_char; Intros o. rewrite listrep_ptrnull; Intros.
    do 2 forward. rewrite spine_listrep. Intros sp.
    forward_while 
      (EX p, EX sp1, EX sp2,
       PROP (sp=sp1++sp2 )
            LOCAL (temp _nd p; temp _s s; temp _n (Vint (Int.repr n)))
            SEP (spine sp1 o p; spine sp2 p nullval; malloc_token M Ews SubjectTP s;
                 data_at Ews SubjectTP (o, Vint (Int.repr n)) s;
                 Observers O_rep s (list_repeat (length sp1) n) (map fst sp1);
                 Observers NotPre s (skipn (length sp1) vals) (map fst sp2)))%assert.
    { Exists o (@nil (val*val)) sp. entailer!. rewrite spine_nil; trivial. simpl; cancel.  }
    { entailer.
      red in PNp. destruct p; try solve [contradiction].
      + destruct Archi.ptr64. contradiction. subst; entailer!.
      + 
        apply denote_tc_test_eq_split; entailer.
        do 4 apply sepcon_valid_pointer1. apply sepcon_valid_pointer2.
        apply spine_validptr. trivial. } 
    { subst sp. rewrite Observers_length with (l:=map fst sp2); Intros.
      destruct sp2; [ sep_apply spine_nilD; Intros; subst; contradiction | simpl].
      destruct p0 as [obs nxt]; Intros. simpl.
      remember (skipn (Datatypes.length sp1) vals) as w.
      destruct w. inv H0. simpl. sep_apply spine_consD; Intros. sep_apply NotifyPre_isptr; Intros.
      forward.

      forward_call (obs, s, z0, (obs::map fst sp2, n)).
      { simpl. rewrite GetPrePost_char.
        unfold_data_at (data_at Ews SubjectTP _ _). cancel. } 
      simpl. rewrite (spine_ptr_or_null sp2); Intros.
      forward.
      Exists (nxt, sp1++[(obs, nxt)], sp2).
      entailer!.
      + rewrite <- app_assoc; trivial.
      + rewrite GetPrePost_char. unfold_data_at (data_at Ews SubjectTP _ _); cancel.
        rewrite map_app, app_length, <- list_repeat_app, Observers_app.
        2: rewrite length_list_repeat, map_length; trivial.
        simpl; cancel.
        apply sepcon_derives.
        - eapply derives_trans. 2: apply spine_app with (u:=p).
          cancel. eapply derives_trans. 2: apply spine_cons; trivial. rewrite spine_nil; [ cancel | trivial]. 
        - apply derives_refl'. f_equal. clear - Heqw.
          rewrite plus_comm. simpl.
          generalize dependent vals. generalize dependent w. generalize dependent z0.
          induction sp1; simpl; intros.
          * subst; trivial. 
          * destruct vals. inv Heqw. apply (IHsp1 _ _ _ Heqw). }
    forward.
    simpl S_rep; rewrite Subject_rep_char. entailer!. simpl. Exists o. cancel.
    rewrite (spine_ptr_or_null_precise sp2); Intros.
    destruct sp2; simpl; Intros; [ clear H | contradiction]. rewrite spine_nil; [|trivial].
    rewrite app_nil_r, map_length, spine_listrep.
    Exists sp1. entailer!.
    clear.
    generalize dependent vals.
    induction sp1; destruct vals; simpl; trivial; entailer!.
  Qed.
  Definition SubjectComponent: @Component NullExtension.Espec SVprog _ 
      nil SubjectImports prog (SubjectASI M SP OP) emp subject_internal_specs.
  Proof. 
    mkComponent.
  + solve_SF_internal add_verif.
  + solve_SF_internal newSubject_verif.
  + solve_SF_internal register_verif.
  + solve_SF_internal update_verif.
  + solve_SF_internal get_verif.
  + solve_SF_internal detachfirst_verif.
  + solve_SF_internal freeSubject_verif. 
  Qed.

  Definition SubjectVSU: @VSU NullExtension.Espec SVprog _ 
      nil SubjectImports prog (SubjectASI M SP OP) emp.
  Proof. eexists; apply SubjectComponent. Qed.

End Subject_VSU.
(*Check SubjectVSU. : forall OP : ObserverPredicates,
       VSU [] (SubjectImports OP) prog (SubjectASI SubjectPreds OP)*)

Definition Subject_CanVSU M OP := VSU_to_CanonicalVSU  (SubjectVSU M OP).