Require Import VST.floyd.proofauto.
Require Import VST.veric.initial_world.

Global Open Scope funspec_scope.
Definition Pr_star (Pr:val -> val -> Z -> mpred)
    (O: list val) s v:mpred :=
    fold_right (fun o P => Pr o s v * P)%logic emp O.

Record AggregateAPD := {
   Sub: val -> list val -> Z -> mpred;
   Obs: val -> val -> Z -> mpred
}.

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Subject. (*needed for function and type names *)
Require Import Observer. (*needed for function and type names *)

Definition SubjectTP:type:=  Tstruct _subject noattr.
Definition ObserverTP:type:=  Tstruct _observer noattr.

Section AggregateASI.
  Variable M: MallocFreeAPD.
  Variable AP: AggregateAPD.
  Notation Sub := (Sub AP).
  Notation Obs := (Obs AP).

  Definition SubObs (s:val) (O:list val) (v:Z) :mpred := 
       (Sub s O v * Pr_star Obs O s v)%logic.

  Definition Obs_ (o s:val) : mpred := EX v, Obs o s v.

  Definition SubObs_ (s:val) (O:list val):mpred := EX v, SubObs s O v.

  (*Differences to Parkinson's spec: gv and mem_mgr*)
  Definition newS_spec :=
    DECLARE _newSubject
    WITH gv: globals
    PRE [ ] PROP() PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr SubjectTP ]
    EX s: val,
     PROP()
     LOCAL (temp ret_temp s)
     SEP(SubObs_ s nil; mem_mgr M gv).

  (*Differences to Parkinson's spec: gv and mem_mgr*)
  Definition reg_spec :=
    DECLARE _registr
    WITH s:val, O:list val, v: Z, o:val, gv:globals
    PRE [ tptr SubjectTP, tptr ObserverTP ]
     PROP ()
     PARAMS (s; o) GLOBALS (gv)
     SEP (Sub s O v; Obs_ o s; mem_mgr M gv)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (Sub s (o::O) v; Obs o s v; mem_mgr M gv).

  Definition upd_spec :=
    DECLARE _update
    WITH s:val, n: Z, O:list val
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s;Vint (Int.repr n)) GLOBALS () 
     SEP(SubObs_ s O)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(SubObs s O n).

  Definition get_spec :=
    DECLARE _get
    WITH s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS () 
     SEP(Sub s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(Sub s O v).
  
  (*Differences to Parkinson's spec: gv and mem_mgr*)
  Definition newO_spec :=
    DECLARE _newObserver
    WITH gv: globals, s:val, O:list val, v:Z
    PRE [ tptr SubjectTP ]
      PROP()
      PARAMS (s) GLOBALS (gv) 
      SEP(SubObs s O v; mem_mgr M gv)
    POST[ tptr ObserverTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(SubObs s (p::O) v; mem_mgr M gv).

  Definition notify_spec :=
    DECLARE _notify
    WITH o:val, s:val, O: list val, v:Z
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS () 
     SEP(Sub s O v; Obs_ o s)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(Sub s O v; Obs o s v).
  
  Definition val_spec :=
    DECLARE _val
    WITH o:val, s:val, O:list val, v: Z
    PRE [ tptr ObserverTP ]
     PROP(In o O)
     PARAMS (o) GLOBALS () 
     SEP(SubObs s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(SubObs s O v).

  Definition freeO_spec:=
    DECLARE _freeObserver 
    WITH gv: globals, o:val, s:val
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS (gv) 
     SEP(Obs_ o s; mem_mgr M gv)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(mem_mgr M gv).

  Definition freeS_spec:=
    DECLARE _freeSubject
    WITH gv: globals, s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv) 
     SEP(Sub s O v; mem_mgr M gv)
    POST[ tint ] EX r,
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr r)))
      SEP(match O with
            nil => !!(r=0) && emp
          | (h::t) => !!(r=1) && Sub s O v
          end; mem_mgr M gv).

  Definition det_spec:=
    DECLARE _detachfirst 
    WITH gv: globals, s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv) 
     SEP(Sub s O v; mem_mgr M gv)
    POST[ tptr ObserverTP ] EX r,
      PROP()
      LOCAL(temp ret_temp r)
      SEP(match O with
            nil => !!(r=nullval) && Sub s O v
          | (h::t) => !!(r=h) && Sub s t v
          end; mem_mgr M gv).

  Definition AggregateASI:funspecs := 
     [ det_spec; freeS_spec;
        newS_spec; reg_spec; upd_spec; get_spec; 
        newO_spec; notify_spec; val_spec; freeO_spec].

End AggregateASI.

Definition ASI_sub (ASI1 ASI2: funspecs) :=
  map fst ASI1 = map fst ASI2 /\
  Forall2 funspec_sub (map snd ASI1) (map snd ASI2).

Require Import spec_Subject.
Require Import spec_Observer.

Section MakeAggregateAPD.
  Variable SP: SubjectAPD.
  Variable OP: ObserverAPD.

  Notation S_rep := (Subjectrep SP).
  Notation O_rep := (Observerrep OP).

  Definition Sub' (s:val) (O: list val) (v:Z):mpred := S_rep (O,v) s.
  Definition Obs' (o s:val)(v:Z):mpred := O_rep (s,v) o.

  Program Definition AggregateAP:AggregateAPD := 
    Build_AggregateAPD Sub' Obs'.
End MakeAggregateAPD.

Import VST.msl.seplog.

Section SubjectObserver_ASIsub.
  Variable M: MallocFreeAPD.
  Variable SP: SubjectAPD.
  Variable OP: ObserverAPD.
  Notation AP := (AggregateAP SP OP).

  Notation LowASI:= (SubjectASI M SP OP ++ ObserverASI M OP SP).
  Notation HighASI:= (AggregateASI M AP).

  Lemma LowHigh_IDs: map fst LowASI = map fst HighASI.
  Proof. reflexivity. Qed.

  Notation Obs_star := (Pr_star (Obs AP)).

  Lemma Obs_star_char: forall O s v,
      Obs_star O s v = Observers (Observerrep OP) s (list_repeat (length O) v) O.
  Proof.
    unfold Obs_star.
    induction O; simpl; intros. reflexivity.
    rewrite <- IHO. unfold Obs'. trivial.
  Qed.

  Lemma Pr_star_app {P s v}: forall L M, 
  Pr_star P (L ++ M) s v = (Pr_star P L s v * Pr_star P M s v)%logic.
  Proof.
   induction L; simpl; intros. rewrite emp_sepcon; trivial.
   rewrite IHL; rewrite sepcon_assoc; trivial.
  Qed.

  Ltac rep := first [ solve [constructor] | constructor; [ | rep] ].

  Ltac go_sub := 
      apply NDsubsume_subsume;
      [ split; reflexivity
      | split; [ split; reflexivity 
               | intros par [g args]; simpl in par; simpl; 
                 unfold LAMBDAx, GLOBALSx, PROPx, LOCALx, SEPx, argsassert2assert, local, liftx, lift1, lift; simpl; entailer!]].

  Lemma LowHigh_ASI_sub: ASI_sub LowASI HighASI.
  Proof.
    split. reflexivity. rep; go_sub.
    + (*detach*)
      destruct par as (((gv, s), O), v). simpl in *; subst. 
      Exists (gv, s, (O,v)) emp; rewrite sepcon_emp; normalize.
      apply andp_right.
      - apply prop_right. split; [ simpl; intros; normalize | intuition].
        Exists x; entailer!.
        if_tac; subst; normalize; apply derives_refl; reflexivity.
      - apply sepcon_derives; trivial. apply derives_refl; reflexivity. 
    + (*freeS*) 
      destruct par as (((gv, s), O), v). normalize. simpl in *; subst.
      Exists (gv, s, (O,v)) emp; rewrite sepcon_emp; normalize.
      apply andp_right.
      - apply prop_right. split; [ simpl; intros; normalize | intuition]. 
        Exists x; entailer!.
        if_tac; subst; normalize.
        destruct O; simpl; [ congruence | entailer!].
      - apply sepcon_derives; trivial. apply derives_refl; reflexivity. 
    + (*newS*)
      rename par into gv.
      Exists gv emp; rewrite sepcon_emp; normalize.
      apply andp_right; trivial.
      apply prop_right. split; [ simpl; intros; normalize | intuition].
      Exists x; entailer!.
      Exists 0. unfold SubObs; simpl. unfold Sub'.
      rewrite seplog.sepcon_emp. apply derives_refl.
    + (*register*)
      destruct par as ((((s,O),v), o), gv); simpl in *; subst; normalize.
      Exists ((((s,O),v), o), gv) emp; rewrite sepcon_emp; normalize.
      unfold Sub', Obs_; sep_apply Subject_Register; Intros z.
      Exists z. apply andp_right.
      - apply prop_right. split; [ simpl; intros; normalize | intuition]. cancel.
      - cancel. apply Observer_Notify.
    + (*update*)
      Exists par emp. destruct par as ((s,n),l); rewrite sepcon_emp. 
      normalize. simpl in *; subst.
      unfold SubObs_; Intros z. Exists z; entailer!.
      - intros. apply sepcon_derives. apply derives_refl.
        rewrite <- Obs_star_char. apply derives_refl.
      - unfold SubObs; simpl; unfold Sub'. sep_apply Subject_Update.
        Exists (list_repeat (Datatypes.length l) z). cancel. clear.
        induction l; simpl; [trivial | intros].
        eapply sepcon_derives; [ apply Observer_Notify | apply IHl].
    + (*get*)
      destruct par as ((s,O),v); rewrite sepcon_emp. 
      normalize. simpl in *; subst. unfold Sub'.
      sep_apply Subject_GetPrePost.
      Exists (s, (O, v)) (GetPost SP (O, v) s -* Subjectrep SP (O, v) s).
      entailer!.
      intros. apply wand_frame_elim'' (*(GetPost SP (O, v) s) (Subjectrep SP (O, v) s))*).
    + (*newObserver*)
      destruct par as (((gv, s), O),v); rewrite sepcon_emp. 
      normalize. simpl in *; subst. 
      Exists (gv, s, (O, v)) (Obs_star O s v). entailer!.
      - intros. Exists x0; entailer!. unfold SubObs, Sub'; simpl; cancel.
        apply derives_refl.
      - unfold SubObs; simpl; unfold Sub'. sep_apply Subject_Register. cancel.
    + (*notify*)
      destruct par as (((o, s), O), v); rewrite sepcon_emp. 
      normalize. simpl in *; subst. 
      unfold Obs_, Sub'; Intros z; simpl. unfold Obs'. 
      sep_apply Subject_GetPrePost. sep_apply Observer_Notify.
      Exists (o, s, z, (O, v)) (GetPost SP (O, v) s -* Subjectrep SP (O, v) s).
      entailer!.
      simpl; intros. cancel.
      apply wand_frame_elim'' (*(GetPost SP (O, v) s) (Subjectrep SP (O, v) s))*).
    + (*val*)
      destruct par as (((o, s) ,O),v); rewrite sepcon_emp.
      normalize. simpl in *; subst. unfold SubObs; simpl.
      apply in_split in H0. destruct H0 as [O1 [O2 HO]]. subst O.
      rewrite Pr_star_app.
      Exists (o, s, v) ((Obs_star O1 s v * Obs_star O2 s v * Sub AP s (O1++o::O2) v)%logic). 
      entailer!.
      - intros. rewrite ! Obs_star_char. simpl. cancel.
      - simpl. cancel. apply Observer_Val.
    + (*freeO*)
      destruct par as ((gv, o), s); rewrite sepcon_emp. 
      normalize. simpl in *; subst.
      unfold Obs_; Intros z. 
      Exists (gv, o, (s, z)) emp. entailer!. apply derives_refl.
  Qed.
End SubjectObserver_ASIsub.

(*Exploiting SubjectObserver_ASIsub, we next set up a ApplicationVSU
  where the export sepcs are aggregate specs. The construction is
  similar to VSU_Application.v*)
Require Import VST.floyd.VSU.
Require Import VST.floyd.linking.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Subject.
Require Import spec_Observer.
Require Import verif_Subject.
Require Import verif_Observer.
Require Import SubjectComponent.
Require Import ObserverComponent.
Require Import verif_surelyMalloc.

Section Application_AggVSU.
Variable M: MallocFreeAPD.

Definition Subject := SubjectAP M.
Definition Observer := ObserverAP M. 

(****** 1. Link subject.prog with observer.prog, as in  verif_core_SM_linkedlate*)
Definition subj_obs_prog': Clight.program := 
  match link_progs Subject.prog Observer.prog with
    Errors.OK p => p
  | _ => Subject.prog (*arbitrary*)
  end.
Definition subj_obs_prog := ltac:
  (let x := eval hnf in subj_obs_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).
Instance subj_obs_cs : compspecs. make_compspecs subj_obs_prog. Defined.

Definition subj_obs_Vprog:varspecs. mk_varspecs subj_obs_prog. Defined.

Definition subj_obs_Imports:funspecs := 
[ASI_malloc.free_spec' M _free; surely_malloc_numericspec M].

(*This is LowASI above*)
Definition subj_obs_Exports:funspecs := (SubjectASI M Subject Observer) ++ (ObserverASI M Observer Subject).

Definition SubjObs_VSU:
@VSU NullExtension.Espec
   subj_obs_Vprog subj_obs_cs nil subj_obs_Imports subj_obs_prog subj_obs_Exports emp.
Proof.
  VSULink_tac (SubjectVSU M Observer) (ObserverVSU M Subject).
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

(****** 2. Lift the VSU to the aggregate exports  *****)
Definition subj_obs_AggExports:funspecs := AggregateASI M (AggregateAP Subject Observer).

Definition SubjObs_AggVSU:
@VSU NullExtension.Espec
   subj_obs_Vprog subj_obs_cs nil subj_obs_Imports subj_obs_prog subj_obs_AggExports emp.
Proof.
  eapply VSU_Exports_sub1.
    apply SubjObs_VSU.
    cbv; reflexivity.
    apply LowHigh_ASI_sub.
Qed.

(****** 3. link surely_malloc****)
(* Note that the import contains notify_smallspec Observer (SubjectAP M)*)

Definition application_prog': Clight.program := 
  match link_progs surelyMalloc.prog subj_obs_prog with
    Errors.OK p => p
  | _ => surelyMalloc.prog (*arbitrary*)
  end.

Definition application_prog := ltac:
  (let x := eval hnf in application_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance AppCs : compspecs. make_compspecs application_prog. Defined.

Definition AppVprog:varspecs. mk_varspecs application_prog. Defined.

Definition AppImports:funspecs := [malloc_spec' M _malloc; exit_spec; free_spec' M _free].

Definition AppAggExports:funspecs := G_merge (SM_ASI M) subj_obs_AggExports.

Definition Application_AggVSU:
@VSU NullExtension.Espec
   AppVprog AppCs nil AppImports
   application_prog AppAggExports emp.
Proof.
  VSULink_tac (SMVSU M) SubjObs_AggVSU.
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.
End Application_AggVSU.

(**** Connect application program with stlib+mm  -- as in VSU_AppLib.v***************)
Require Import MMlib_prog.
Require Import verif_stdlib.
Require Import ASI_mmap0.
Section Applib_AggVSU.

Definition applib': Clight.program := 
  match link_progs mmlib_prog application_prog with
    Errors.OK p => p
  | _ => mmlib_prog (*arbitrary*)
  end.
Definition applib := ltac:
  (let x := eval hnf in applib' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance ApplibCS : compspecs. make_compspecs applib. Defined.

Definition ApplibVprog:varspecs. mk_varspecs applib. Defined.

Definition ApplibImports:funspecs := nil.

Definition ApplibAggExports:funspecs := G_merge mmlib_Exports (AppAggExports M).

Definition ApplibSys:funspecs := [exit_spec; munmap_spec mmap0._munmap].

Definition ApplibInitPred (gv: globals): mpred:= mem_mgr M gv.

Definition Applib_AggVSU:
@VSU NullExtension.Espec ApplibVprog ApplibCS ApplibSys 
     ApplibImports applib ApplibAggExports ApplibInitPred.
Proof.
  VSULink_tac MMlib_VSU (Application_AggVSU M).
  extensionality gv; simpl. rewrite sepcon_emp; trivial.
Qed.

Definition Applib_AggCanVSU: @CanonicalVSU NullExtension.Espec
   ApplibVprog ApplibCS ApplibSys 
   ApplibImports applib ApplibAggExports ApplibInitPred.
Proof. apply (VSU_to_CanonicalVSU Applib_AggVSU). Qed.

End Applib_AggVSU.

(******** We now reverify main - again abstractly in M, SP, OP -- wrt the aggregate specs*****)
Require Import main.
Section MAIN_Verif. (*see verif_main.v*)
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.
  Notation AP := (AggregateAP SP OP).

  Definition main_inst_spec G:=
  DECLARE _main
  WITH gv: globals
  PRE [ ]
    PROP ()
    PARAMS () GLOBALS (gv)
    SEP (fold_right (fun g p => g gv * p) emp G; has_ext tt)
  POST[ tint ]
    PROP()
    LOCAL(temp ret_temp (Vint (Int.repr 6)))
    SEP((*mem_mgr M gv; *)TT;has_ext tt).

  Definition MainImports :=
     [ det_spec M AP; freeS_spec M AP;
        newS_spec M AP;  upd_spec AP; 
        newO_spec M AP; AggregatePredicates.val_spec AP; freeO_spec M AP].

  Definition MyInitPred: list (globals -> mpred) := [].
  Definition maininstspec := (main_inst_spec [mem_mgr M]).

  Definition MainVprog : varspecs. mk_varspecs main.prog. Defined.
  Definition MainGprog:funspecs := MainImports(* ++ main_internal_specs*).

  Instance MainCompSpecs : compspecs. make_compspecs main.prog. Defined.
  Lemma body_main: semax_body MainVprog MainGprog f_main maininstspec.
  Proof.
  start_function. simpl; Intros.
(*  sep_apply (spec_stdlib.make_mem_mgr M gv). *)
  forward_call. Intros s. simpl. (*nothing happens as SD is abstract*)
  unfold SubObs_; Intros v.
  forward_call (gv,s, @nil val, v). Intros o1.
  forward_call (s, 2, [o1]).
  { unfold SubObs_. Exists v; cancel. }
  clear v.
  forward_call; Intros o2.
  forward_call (s, 3, [o2;o1]).
  { unfold SubObs_. Exists 2; cancel. }
  forward_call.
  { right; left; trivial. }
  forward_call.
  { left; trivial. }
  forward.
  forward_call (gv, s, [o2; o1], 3).
  { unfold SubObs. cancel. }
  Intros p1; subst o2.
  forward_call (gv, s, [o1], 3).
  Intros p2; subst o1.
  forward_call. Intros v; subst.
  forward_call (gv, p1, s).
  { unfold Obs_; simpl. Exists 3. cancel. }
  forward_call (gv, p2, s).
  { unfold Obs_; simpl. Exists 3. cancel. }
  forward.
  cancel.
  Qed.

End MAIN_Verif.

(******** Finally, connect verif_main with Applib_Agg, as in verif_linked_main **)

Definition linked_prog': Clight.program := 
  match link_progs applib main.prog with
    Errors.OK p => p
  | _ => applib (*arbitrary*)
  end.
Definition linked_prog := ltac:
  (let x := eval hnf in linked_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Definition ExtIDtable : PTree.t unit :=
  ltac:(let x := constr:(fold_right (fun i t => PTree.set i tt t) (PTree.empty _) (ExtIDs linked_prog))
           in let x := eval compute in x
           in exact x).

Definition in_ExtIDs (ia: ident*funspec) : bool :=
 match PTree.get (fst ia) ExtIDtable with Some _ => true | None => false end.

Definition LinkedSYS_pre:funspecs :=
   filter in_ExtIDs (augment_funspecs linked_prog ApplibSys).

Definition LinkedSYS:funspecs := 
  ltac: 
    (let x := eval hnf in LinkedSYS_pre in
     let x := eval simpl in x in 
     (*let x := eval compute in x in leaving this in leads Coq to timeout/crash*)
       exact x).

Lemma ApplibSys_in_LinkedSys: forall i phi, find_id i ApplibSys = Some phi -> find_id i LinkedSYS = Some phi.
  Proof. (*Boilerplate proof*)
    intros. specialize (find_id_In_map_fst _ _ _ H); intros.
    simpl in H0. repeat (destruct H0 as [HO | H0]; [ subst i; inv H; reflexivity |]). contradiction.
  Qed. 

Lemma LinkedSYS_External: forall i, In i (map fst LinkedSYS) ->
      exists ef ts t cc, find_id i (prog_defs linked_prog) = Some (Gfun (External ef ts t cc)).
  Proof. (*Boilerplate proof*)
    intros. cbv in H. 
    repeat (destruct H as [H | H];
      [ subst; try solve [do 4 eexists; split; reflexivity ]
      | ]).
    contradiction.
  Qed.

Lemma LinkedSYS_vacuous i phi: find_id i LinkedSYS = Some phi -> find_id i ApplibSys = None ->
        exists ef argsig retsig cc, 
           phi = vacuous_funspec (External ef argsig retsig cc) /\ 
           find_id i (prog_funct applib) = Some (External ef argsig retsig cc) /\
           ef_sig ef = {| sig_args := typlist_of_typelist argsig;
                          sig_res := rettype_of_type retsig;
                          sig_cc := cc_of_fundef (External ef argsig retsig cc) |}.
  Proof. (*Boilerplate proof*)
    intros. unfold find_id in H. simpl in H. 
    (*Time*) repeat (if_tac in H;
    [ inv H; try inv H0; try solve [do 4 eexists; split3; reflexivity]
    | clear H1 ]); discriminate.
(*Old proof:
    intros. specialize (find_id_In_map_fst _ _ _ H); intros.
    cbv in H1.
    Time repeat (destruct H1 as [H1 | H1]; 
      [ subst; inv H; try solve [do 4 eexists; split3; reflexivity]
      | ]); try inv H0; try contradiction. (*3s*)*)
  Qed.

Definition LinkedVprog : varspecs. mk_varspecs linked_prog. Defined.

Lemma disjoint_Vprog_linkedfuncts: 
      list_disjoint (map fst LinkedVprog) (map fst (prog_funct linked_prog)).
Proof. (*Boilerplate proof*)
  intros x y X Y ?; subst x; cbv in X; apply assoclists.find_id_None_iff in Y; [ trivial | clear H Y];
  repeat (destruct X as [X | X]; [ subst y; cbv; reflexivity |]); contradiction.
Qed.

Instance LinkedCompSpecs : compspecs. make_compspecs linked_prog. Defined.

Lemma CSSUB: cspecs_sub MainCompSpecs LinkedCompSpecs.
Proof. (*Boilerplate proof*)
  split3.
+ intros i; red; remember ((@cenv_cs MainCompSpecs) ! i) as w; destruct w; 
   [symmetry in Heqw; simpl in Heqw; rewrite PTree.gleaf in Heqw; congruence | trivial].
+ intros i. red. remember ((@ha_env_cs MainCompSpecs) ! i) as w. destruct w; [symmetry in Heqw | trivial].
  simpl in Heqw. rewrite PTree.gleaf in Heqw. congruence.
+ intros i. red. remember ((@la_env_cs MainCompSpecs) ! i) as w. destruct w; [symmetry in Heqw | trivial].
  simpl in Heqw. rewrite PTree.gleaf in Heqw. congruence.
Qed.

Require Import spec_main.
Notation mainspec:= (main_spec linked_prog). (*VST's standard spec for main*)

Lemma main_sub: funspec_sub (snd (main_inst_spec [mem_mgr M])) (snd (main_spec linked_prog)).
Proof. (*boilerplate*)
  do_funspec_sub. unfold main_pre; simpl; Intros; subst. clear. 
  Exists (initialize.genviron2globals g) emp. entailer!.
  + intros; entailer!.
  + specialize (globals_ok_genviron2globals g); intros G.
    Exists emptyResvec. eapply derives_trans. 2: apply VSU_malloc.create_memmgr_R. 
    unfold InitGPred. simpl. entailer. clear.
    unfold globvar2pred; simpl.
    sep_apply mapsto_zeros_isptr; entailer!.
Qed.

Lemma tc_VG: tycontext_subVG MainVprog (MainGprog M (Observer M) (Subject M)) 
                             LinkedVprog (mainspec::ApplibAggExports).
Proof. (*Boilerplate proof*)
  split; intros i; red.
  * rewrite 2 semax_prog.make_context_g_char, 2 make_tycontext_s_find_id by LNR_tac.
    remember (find_id i (MainGprog M (Observer M) (Subject M))) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; reflexivity |]).
    congruence.
  * rewrite 2 make_tycontext_s_find_id.
    remember (find_id i (MainGprog M (Observer M) (Subject M))) as w.
    destruct w; [symmetry in Heqw | trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.

Require Import VST.floyd.VSU_addmain.

Definition SO_VSU: @LinkedProgVSU NullExtension.Espec LinkedVprog LinkedCompSpecs
      LinkedSYS nil linked_prog [mainspec] ApplibInitPred.
Proof.
 VSUAddMain_tac Applib_AggCanVSU.
   + apply disjoint_Vprog_linkedfuncts.
   + apply LinkedSYS_External. 
   + eapply semax_body_subsumption.
     * eapply semax_body_funspec_sub.
       - apply (@semax_body_cenv_sub MainCompSpecs _ CSSUB).
         2: apply (body_main M (Observer M) (Subject M)).
         repeat constructor; eauto.
       - apply main_sub. 
       - LNR_tac.
     * apply tycontext_sub_i99. apply tc_VG.
   + apply LinkedSYS_vacuous.
Qed.

Lemma prog_correct:
  exists G, 
 @semax_prog NullExtension.Espec LinkedCompSpecs linked_prog tt LinkedVprog G.
Proof.
(*boilerplate: same proof script as in VSUpile, connpool*)
  destruct SO_VSU as [G Comp MAIN]. exists G. 
  assert (DomG: map fst G = map fst (prog_funct linked_prog)).
  { destruct Comp. unfold Comp_G in *. rewrite CC_canonical.
    cbv; reflexivity. }
  Time prove_linked_semax_prog. (*7s*)
  all: rewrite augment_funspecs_eq by trivial.
  apply (@Canonical_semax_func _ _ _ _ _ _ _ _ Comp); cbv; reflexivity.
  destruct MAIN as [post [MainG MainExp]]. inv MainExp. rewrite MainG; eexists; reflexivity.
Qed.

(*Print Assumptions prog_correct.*)