Require Import VST.floyd.proofauto.
Require Import spec_stdlib. (*for malloc_token*)

Global Open Scope funspec_scope.

Require Import Subject.
Instance SCompSpecs : compspecs. make_compspecs prog. Defined.

Require Import spec_Subject.

Require Import Observer.
Instance OCompSpecs : compspecs. make_compspecs prog. Defined.

Require Import spec_Observer.

Definition NodeTP:type:= Tstruct _node noattr.

Section AggregateASI.
  Variable M: MallocFreeAPD.

Fixpoint lst (l:val) (O: list val): mpred :=
  match O with
    nil => !!(l=nullval) && emp
  | o::os => EX n, !!(isptr o) && @data_at SCompSpecs Ews NodeTP (o,n) l *
                   @malloc_token SCompSpecs M Ews NodeTP l * lst n os
  end.

Definition Sub (s:val) (O: list val) (v:Z) : mpred :=
  EX FL, EX fl, EX l, 
     lst fl FL * lst l O * @malloc_token SCompSpecs M Ews SubjectTP s *
     @data_at SCompSpecs Ews SubjectTP (fl, (l, Vint (Int.repr v))) s.

Definition Obs (o s:val)(v:Z) : mpred :=
  (@malloc_token OCompSpecs M Ews ObserverTP o * 
   @data_at OCompSpecs Ews ObserverTP (s, Vint (Int.repr v)) o)%logic.

Definition Obs_ (o s:val) : mpred := EX v, Obs o s v.

Definition Obs_star (O: list val) s v:mpred :=
  fold_right (fun o P => Obs o s v * P)%logic emp O.

Definition SubObs (s:val) (O:list val) (v:Z):mpred :=
  (Sub s O v * Obs_star O s v)%logic.
  
Definition SubObs_ (s:val) (O:list val):mpred := EX v, SubObs s O v.

(*Differences to Parkinson's spec: gv and mem_mgr*)
Definition newS_spec :=
    DECLARE _newSubject
    WITH gv: globals
    PRE [ ] PROP() PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr SubjectTP ]
    EX s: val,
     PROP()
     LOCAL(temp ret_temp s)
     SEP(SubObs_ s nil; mem_mgr M gv).

(*Differences to Parkinson's spec: gv and mem_mgr*)
Definition reg_spec :=
    DECLARE _registr
    WITH s:val, O:list val, v: Z, o:val, gv:globals
    PRE [ tptr SubjectTP, tptr ObserverTP ]
     PROP ()
     PARAMS (s; o) GLOBALS (gv)
     SEP (Sub s O v; Obs_ o s; mem_mgr M gv)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (Sub s (o::O) v; Obs o s v; mem_mgr M gv).

Definition upd_spec :=
    DECLARE _update
    WITH s:val, n: Z, O:list val
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(SubObs_ s O)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(SubObs s O n).

Definition get_spec :=
    DECLARE _get
    WITH s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(Sub s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(Sub s O v).

(*Differences to Parkinson's spec: gv and mem_mgr*)
Definition newO_spec :=
    DECLARE _newObserver
    WITH gv: globals, s:val, O:list val, v:Z
    PRE [ tptr SubjectTP ]
      PROP()
      PARAMS (s) GLOBALS (gv)
      SEP(SubObs s O v; mem_mgr M gv)
    POST[ tptr ObserverTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(SubObs s (p::O) v; mem_mgr M gv).

Definition notify_spec :=
    DECLARE _notify
    WITH o:val, s:val, O: list val, v:Z
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS ()
     SEP(Sub s O v; Obs_ o s)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(Sub s O v; Obs o s v).

Definition Oval_spec :=
    DECLARE _val
    WITH o:val, s:val, O:list val, v: Z
    PRE [ tptr ObserverTP ]
     PROP(In o O)
     PARAMS (o) GLOBALS ()
     SEP(SubObs s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(SubObs s O v).

End AggregateASI.

Require Import verif_Subject.
Lemma lst_char {M}: forall O l, lst M l O = listrep M O l.
Proof.
  induction O; intros; simpl; trivial.
  apply pred_ext; Intros n; Exists n; entailer!;
  rewrite IHO; apply derives_refl. 
Qed.

Lemma Sub_char {M} s O v: Sub M s O v = Subject_rep M (O,v) s.
Proof.
  apply pred_ext; unfold Sub, Subject_rep; Intros FL fl l; Exists FL fl l;
  rewrite 2 lst_char; cancel; apply derives_refl.
Qed.

Require Import verif_Observer.
Lemma Obs_char {M} o s v: Obs M o s v = Observer_rep M (s,v) o.
Proof. trivial. Qed.

Lemma Obs_star_char {M}: forall O s v,
    Obs_star M O s v = Observers (Observer_rep M) s (list_repeat (length O) v) O.
Proof.
  unfold Obs_star.
  induction O; simpl; intros. reflexivity.
  rewrite <- IHO, Obs_char; trivial.
Qed.

Lemma Obs_star_cons {M} O s v p: (Obs_star M O s v * Observer_rep M (newObserver (O, v) s) p = Obs_star M (p :: O) s v)%logic.
Proof. simpl. apply pred_ext;  cancel; apply derives_refl. Qed.

Lemma Obs_star_app {M s v}: forall L1 L2, 
  Obs_star M (L1 ++ L2) s v = (Obs_star M L1 s v * Obs_star M L2 s v)%logic.
  Proof.
   induction L1; simpl; intros. rewrite emp_sepcon; trivial.
   rewrite IHL1; rewrite sepcon_assoc; trivial.
  Qed.

Import VST.msl.seplog.

  Ltac go_sub := 
      apply NDsubsume_subsume;
      [ split; reflexivity
      | split; [ split; reflexivity 
               | intros par [g args]; simpl in par; simpl; 
                 unfold LAMBDAx, GLOBALSx, PROPx, LOCALx, SEPx, argsassert2assert, local, liftx, lift1, lift; simpl; entailer!]].

Section AggregateConcreteSub.
  Variable M: MallocFreeAPD.
  Notation SP:= (SubjectAP M).
  Notation OP:= (ObserverAP M).

Lemma newS_sub: funspec_sub (snd (newSubject_spec M SP)) (snd (newS_spec M)).
Proof.
  go_sub. rename par into gv. 
  Exists gv emp. entailer!.
  intros. Exists FL. entailer!.
  Exists 0. unfold SubObs. rewrite Sub_char. simpl. Exists x0 fl nullval. entailer!.
Qed.

Lemma register_sub: funspec_sub (snd (register_spec M SP OP)) (snd (reg_spec M)).
Proof.
  go_sub.
  destruct par as ((((s,O),v), o), gv). simpl; normalize.
  Exists ((((s,O),v), o), gv) emp. entailer!.
  + intros. simpl. cancel.  rewrite Sub_char, Obs_char; simpl. cancel.
    Exists x0 fl n. cancel. Exists x1; entailer!.
  + unfold Obs_; Intros z. rewrite Sub_char, Obs_char. sep_apply Observer_NotifyPre.
    Exists z; cancel. apply derives_refl. 
Qed.

Lemma update_sub: funspec_sub (snd (update_spec SP OP)) (snd (upd_spec M)).
Proof.
  go_sub.
  destruct par as ((s,n),l); simpl; normalize.
  Exists ((s,n),l) emp.
  unfold SubObs_; Intros z. entailer!.
  + intros. unfold SubObs. rewrite Obs_star_char; rewrite <- 2 lst_char. cancel.
    unfold Sub. Exists x0 fl o. cancel. apply derives_refl.
  + unfold SubObs, Sub. normalize. rewrite lst_char, Obs_star_char, lst_char.
    Exists (list_repeat (Datatypes.length l) z) z. Exists FL fl x. cancel.
    apply sepcon_derives. apply derives_refl.
    clear. induction l; simpl; trivial.
    rewrite NotifyPre_char. simpl. Exists (Vint (Int.repr z)). cancel.
Qed.

Lemma get_sub: funspec_sub (snd (get_largespec SP)) (snd (get_spec M)).
Proof.
  go_sub. destruct par as ((s,O),v).
  Exists (s, (O, v)) emp. simpl; entailer!.
  - intros. rewrite Sub_char. unfold Subject_rep. Exists fl x0 o; cancel.
  - rewrite Sub_char. apply derives_refl.
Qed.

Lemma newO_sub: funspec_sub (snd (newObserver_spec M OP SP)) (snd (newO_spec M)).
Proof.
  go_sub. destruct par as (((gv, s) ,O),v). simpl; normalize.
  Exists (gv, s, (O, v)) (Obs_star M O s v). entailer!.
  - intros. unfold SubObs. Exists FL. simpl. rewrite Sub_char, Obs_char. simpl. entailer!.
    Exists n fl o x0. entailer!.
  - unfold SubObs. rewrite Sub_char. cancel. 
Qed.

Lemma notify_sub: funspec_sub (snd (notify_largespec OP SP)) (snd (notify_spec M)).
Proof.
  go_sub. destruct par as (((o, s), O), v). simpl; normalize.
  unfold Obs_; Intros z. rewrite Sub_char, Obs_char.
  Exists (o, s, z, (O, v)) emp; entailer!. 
  - intros. cancel. rewrite Obs_char. apply derives_refl.
  - simpl. rewrite NotifyPre_char. Exists (Vint (Int.repr z)). cancel.
Qed.

Lemma val_sub: funspec_sub (snd (val_spec OP)) (snd (Oval_spec M)).
Proof.
  go_sub. destruct par as (((o, s), O),v); simpl; normalize.
  unfold SubObs.
  apply in_split in H0. destruct H0 as [O1 [O2 HO]]. subst O.
  rewrite Obs_star_app. simpl.
  Exists (o, s, v) (Obs_star M O1 s v * Obs_star M O2 s v * Sub M s (O1++o::O2) v)%logic. entailer!.
  - intros. cancel. rewrite Obs_char. apply derives_refl.
  - rewrite Obs_char. apply derives_refl.
Qed.
(*Old proofs:
Lemma newS_sub: funspec_sub (snd (newSubject_spec SubjectPreds)) (snd newS_spec).
Proof.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intro gv. simpl in gv.
  Exists gv emp.
  rewrite !insert_SEP.
  apply andp_right.
  entailer!.
  apply prop_right.
  Intros p. Exists p.
  rewrite !insert_SEP.
  unfold SubObs_; Exists 0. entailer!. 
  unfold SubObs; rewrite Sub_char. simpl; cancel; try apply derives_refl.
Qed.

Lemma register_sub: funspec_sub (snd (register_spec SubjectPreds ObserverPreds)) (snd reg_spec).
Proof.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intro par. simpl in par. 
  Exists par emp. destruct par as ((((s,O),v), o), gv).
  rewrite !insert_SEP.
  apply andp_right.
  { entailer!; clear.
    unfold NotifyPre, Obs_; first [Intros z | normalize]. rewrite Sub_char, Obs_char.
    replace Observer_rep with (Observerrep ObserverPreds) by reflexivity.
    sep_apply (Observer_Notify ObserverPreds (s, x) o).
    Exists x; cancel.
    apply (Subject_Register SubjectPreds (O, v) s). }
  apply prop_right. entailer!.
  rewrite Sub_char, Obs_char.
  replace Observer_rep with (Observerrep ObserverPreds) by reflexivity.
  replace Subject_rep with (Subjectrep SubjectPreds) by reflexivity.
  trivial.
Qed.

Lemma update_sub: funspec_sub (snd (update_spec SubjectPreds ObserverPreds)) (snd upd_spec).
Proof.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intro par. simpl in par. 
  Exists par emp. destruct par as ((s,n),l).
  rewrite !insert_SEP.
  apply andp_right.
  { entailer!; clear.
    unfold SubObs_, SubObs. Intros v. rewrite Sub_char, Obs_star_char.
    replace Subject_rep with (Subjectrep SubjectPreds) by reflexivity.
    sep_apply Subject_Update.
    replace Observer_rep with (Observerrep ObserverPreds) by reflexivity.
    Exists (list_repeat (Datatypes.length l) v) v. cancel.
    induction l. simpl; trivial. simpl.
    replace Observer_rep with (Observerrep ObserverPreds) by reflexivity.
    sep_apply IHl. rewrite NotifyPre_char. cancel.
    Exists (Vint (Int.repr v)). simpl; cancel. } 
  apply prop_right. entailer!.
  unfold SubObs. rewrite Sub_char, Obs_star_char.
  replace Observer_rep with (Observerrep ObserverPreds) by reflexivity.
  replace Subject_rep with (Subjectrep SubjectPreds) by reflexivity.
  trivial.
Qed.

Lemma get_sub: funspec_sub (snd (get_largespec SubjectPreds)) (snd get_spec).
Proof.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intro par. simpl in par. destruct par as ((s,O),v).
  Exists (s, (O, v)) emp. 
  rewrite !insert_SEP.
  apply andp_right; rewrite Sub_char.
  { entailer!. apply derives_refl. }
  apply prop_right; entailer!. apply derives_refl.
Qed.

Lemma newO_sub: funspec_sub (snd (newObserver_spec ObserverPreds SubjectPreds)) (snd newO_spec).
Proof.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intro par. simpl in par. destruct par as (((gv, s) ,O),v).
  Exists (gv, s, (O, v)) (Obs_star O s v). 
  rewrite !insert_SEP.
  apply andp_right.
  { unfold SubObs; rewrite Sub_char; entailer!.
    replace Subject_rep with (Subjectrep SubjectPreds) by reflexivity.
    apply Subject_Register. }
  apply prop_right.
  Intros p. Exists p.
  rewrite !insert_SEP.
  unfold SubObs; rewrite Sub_char; entailer!.
  replace Subject_rep with (Subjectrep SubjectPreds) by reflexivity.
  cancel.
  rewrite Obs_star_cons; trivial.
Qed.

Lemma notify_sub: funspec_sub (snd (notify_largespec ObserverPreds SubjectPreds)) (snd notify_spec).
Proof.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intro par. simpl in par. destruct par as (((o, s), O), v).
  unfold Obs_; Intros c; rewrite Sub_char, Obs_char.
  Exists (o, s, c, (O, v)) emp. 
  rewrite !insert_SEP.
  apply andp_right.
  { entailer!. rewrite NotifyPre_char.
    replace Subject_rep with (Subjectrep SubjectPreds) by reflexivity.
    cancel.
    Exists (Vint (Int.repr c)); simpl; trivial. }
  apply prop_right. rewrite Obs_char; entailer!.
  replace Subject_rep with (Subjectrep SubjectPreds) by reflexivity.
  replace Observer_rep with (Observerrep ObserverPreds) by reflexivity.
  cancel.
Qed.

Lemma val_sub: funspec_sub (snd (val_spec ObserverPreds)) (snd Oval_spec).
Proof.
  apply NDsubsume_subsume.
  split; extensionality x; reflexivity.
  split3; auto.
  intro par. simpl in par. destruct par as (((o, s) ,O),v).
  unfold SubObs.
  eapply derives_trans
  with (EX O1, EX O2,
        PROP (O = O1 ++ o :: O2)
        LOCAL (temp _o o)
        SEP (Sub s O v * Obs_star O1 s v * Obs o s v * Obs_star O2 s v)).
  { entailer!. clear H0. apply in_split in H.
    destruct H as [O1 [O2 HO]]; subst. Exists O1 O2. entailer!.
    rewrite ! Obs_star_char.
    rewrite app_length, <- list_repeat_app, Observers_app.
    simpl; cancel. rewrite Obs_char; apply derives_refl.
    apply length_list_repeat. }
  Intros O1 O2. rewrite Obs_char.
  Exists (o, s, v) (Obs_star O1 s v * Obs_star O2 s v * Sub s O v). 
  rewrite !insert_SEP.
  apply derives_extract_PROP'; intros; subst.
  apply andp_right.
  { old_go_lower. entailer!. } 
  apply prop_right.
  entailer!. 
  rewrite ! Obs_star_char.
  rewrite app_length, <- list_repeat_app, Observers_app.
  simpl; cancel.
  apply length_list_repeat. 
Qed.
*)
End AggregateConcreteSub.