Require Import VST.floyd.proofauto.

Global Open Scope funspec_scope.
Definition Pr_star (Pr:val -> val -> Z -> mpred)
    (O: list val) s v:mpred :=
    fold_right (fun o P => Pr o s v * P)%logic emp O.

(*Despite the existence of the function pointer and its exposure in the signature of notify, 
  we can use a residual predicate just as before*)
Record AggregateAPD := {
   Sub: val -> list val -> Z -> mpred;
   Obs: val -> val -> Z -> mpred;
   GPre: val -> list val -> Z -> mpred;
   GPost: val -> list val -> Z -> mpred;
   Sub_GPrePost: forall s O v, Sub s O v |-- (GPre s O v * (GPost s O v -* Sub s O v))%logic;
}.

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Subject. (*needed for function and type names *)
Require Import Observer. (*needed for function and type names *)

Definition SubjectTP:type:=  Tstruct _subject noattr.
Definition ObserverTP:type:=  Tstruct _observer noattr.

Section AggregateASI.
  Variable M: MallocFreeAPD.
  Variable AP: AggregateAPD.
  Notation Sub := (Sub AP).
  Notation Obs := (Obs AP).

  Definition SubObs (s:val) (O:list val) (v:Z) :mpred := 
       (Sub s O v * Pr_star Obs O s v)%logic.

  Definition Obs_ (o s:val) : mpred := EX v, Obs o s v.

  Definition SubObs_ (s:val) (O:list val):mpred := EX v, SubObs s O v.

  (*Differences to Parkinson's spec: gv and mem_mgr*)
  Definition newS_spec :=
    DECLARE _newSubject
    WITH gv: globals
    PRE [ ] PROP() PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr SubjectTP ]
    EX s: val,
     PROP()
     LOCAL (temp ret_temp s)
     SEP(SubObs_ s nil; mem_mgr M gv).

  (*Differences to Parkinson's spec: gv and mem_mgr*)
  Definition reg_spec :=
    DECLARE _registr
    WITH s:val, O:list val, v: Z, o:val, gv:globals
    PRE [ tptr SubjectTP, tptr ObserverTP ]
     PROP ()
     PARAMS (s; o) GLOBALS (gv)
     SEP (Sub s O v; Obs_ o s; mem_mgr M gv)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (Sub s (o::O) v; Obs o s v; mem_mgr M gv).

  Definition upd_spec :=
    DECLARE _update
    WITH s:val, n: Z, O:list val, gv: globals
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s;Vint (Int.repr n)) GLOBALS (gv) 
     SEP(SubObs_ s O)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(SubObs s O n).

  Definition get_spec :=
    DECLARE _get
    WITH s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS () 
     SEP(Sub s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(Sub s O v).
  
  (*Differences to Parkinson's spec: gv and mem_mgr*)
  Definition newO_spec :=
    DECLARE _newObserver
    WITH gv: globals, s:val, O:list val, v:Z
    PRE [ tptr SubjectTP ]
      PROP()
      PARAMS (s) GLOBALS (gv) 
      SEP(SubObs s O v; mem_mgr M gv)
    POST[ tptr ObserverTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(SubObs s (p::O) v; mem_mgr M gv).

Definition gspec :=
    WITH s:val, O: list val, v: Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(GPre AP s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(GPost AP s O v).

  Definition notify_spec :=
    DECLARE _notify
    WITH o:val, s:val, O: list val, v:Z, fp:val
    PRE [ tptr ObserverTP,
          tptr (Tfunction (Tcons (tptr SubjectTP) Tnil) tint cc_default) ]
     PROP()
     PARAMS (o;fp) GLOBALS ()
     SEP(Sub s O v; Obs_ o s; func_ptr' gspec fp)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(Sub s O v; Obs o s v).
  
  Definition val_spec :=
    DECLARE _val
    WITH o:val, s:val, O:list val, v: Z
    PRE [ tptr ObserverTP ]
     PROP(In o O)
     PARAMS (o) GLOBALS () 
     SEP(SubObs s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(SubObs s O v).

  Definition freeO_spec:=
    DECLARE _freeObserver 
    WITH gv: globals, o:val, s:val
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS (gv) 
     SEP(Obs_ o s; mem_mgr M gv)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(mem_mgr M gv).

  Definition freeS_spec:=
    DECLARE _freeSubject
    WITH gv: globals, s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv) 
     SEP(Sub s O v; mem_mgr M gv)
    POST[ tint ] EX r,
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr r)))
      SEP(match O with
            nil => !!(r=0) && emp
          | (h::t) => !!(r=1) && Sub s O v
          end; mem_mgr M gv).

  Definition det_spec:=
    DECLARE _detachfirst 
    WITH gv: globals, s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv) 
     SEP(Sub s O v; mem_mgr M gv)
    POST[ tptr ObserverTP ] EX r,
      PROP()
      LOCAL(temp ret_temp r)
      SEP(match O with
            nil => !!(r=nullval) && Sub s O v
          | (h::t) => !!(r=h) && Sub s t v
          end; mem_mgr M gv).

  Definition AggregateASI:funspecs := 
     [ det_spec; freeS_spec;
        newS_spec; reg_spec; upd_spec; get_spec; 
        newO_spec; notify_spec; val_spec; freeO_spec].

End AggregateASI.

(*Forced removal of notify from PatternASI reveals we should
  refine ASI_sub a litle, too, and only require funspec-sub on
  the functions that remain exposed in the abstract spec
Definition ASI_sub (ASI1 ASI2: funspecs) :=
  map fst ASI1 = map fst ASI2 /\
  Forall2 funspec_sub (map snd ASI1) (map snd ASI2).*)
Require Import VST.floyd.VSU.
Definition ASI_sub (ASI1 ASI2: funspecs) := VSU.funspecs_sqsub ASI1 ASI2.

Require Import spec_Subject.
Require Import spec_Observer.

Section MakeAggregateAPD.
  Variable SP: SubjectAPD.
  Variable OP: ObserverAPD.

  Notation S_rep := (Subjectrep SP).
  Notation O_rep := (Observerrep OP).

  Definition Sub' (s:val) (O: list val) (v:Z):mpred := S_rep (O,v) s.
  Definition Obs' (o s:val)(v:Z):mpred := O_rep (s,v) o.

  Definition GPre' (s: val) (O: list val) (v:Z): mpred := GetPre SP (O,v) s.
  Definition GPost' (s: val) (O: list val) (v:Z): mpred := GetPost SP (O,v) s.

  Lemma Sub_GPrePost' s O v: Sub' s O v |-- (GPre' s O v * (GPost' s O v -* Sub' s O v))%logic.
  Proof. apply Subject_GetPrePost. Qed.
}

  Program Definition AggregateAP:AggregateAPD := 
    Build_AggregateAPD Sub' Obs' _ _ Sub_GPrePost'.
End MakeAggregateAPD.

Import VST.veric.initial_world.
Import VST.msl.seplog.

Section SubjectObserver_ASIsub.
  Variable M: MallocFreeAPD.
  Variable SP: SubjectAPD.
  Variable OP: ObserverAPD.
  Notation AP := (AggregateAP SP OP).

  Notation LowASI:= (SubjectASI M SP OP ++ ObserverASI M OP SP).
  Notation HighASI:= (AggregateASI M AP).
(*
  Lemma LowHigh_IDs: map fst LowASI = map fst HighASI.
  Proof. reflexivity. Qed.
*)
  Notation Obs_star := (Pr_star (Obs AP)).

  Lemma Obs_star_char: forall O s v,
      Obs_star O s v = Observers (Observerrep OP) s (list_repeat (length O) v) O.
  Proof.
    unfold Obs_star.
    induction O; simpl; intros. reflexivity.
    rewrite <- IHO. unfold Obs'. trivial.
  Qed.

  Lemma Pr_star_app {P s v}: forall L M, 
  Pr_star P (L ++ M) s v = (Pr_star P L s v * Pr_star P M s v)%logic.
  Proof.
   induction L; simpl; intros. rewrite emp_sepcon; trivial.
   rewrite IHL; rewrite sepcon_assoc; trivial.
  Qed.

  Ltac rep := first [ solve [constructor] | constructor; [ | rep] ].

  Ltac go_sub := 
      apply NDsubsume_subsume;
      [ split; reflexivity
      | split; [ split; reflexivity 
               | intros par [g args]; simpl in par; simpl; 
                 unfold LAMBDAx, GLOBALSx, PROPx, LOCALx, SEPx, argsassert2assert, local, liftx, lift1, lift; simpl; entailer!]].

  Lemma LowHigh_ASI_sub: ASI_sub LowASI HighASI.
  Proof. intros i phi2 Phi2.
    specialize (find_id_In_map_fst _ _ _ Phi2); simpl; intros.
    repeat (destruct H as [? | H]); try contradiction; subst; simpl.
    + (*detachfirst*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as (((gv, s), O), v). simpl in *; subst. 
      Exists (gv, s, (O,v)) emp; rewrite sepcon_emp; normalize.
      apply andp_right.
      - apply prop_right. split; [ simpl; intros; normalize | intuition].
        Exists x; entailer!.
        if_tac; subst; normalize; apply derives_refl; reflexivity.
      - apply sepcon_derives; trivial. apply derives_refl; reflexivity. 
    + (*freeS*) 
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as (((gv, s), O), v). normalize. simpl in *; subst.
      Exists (gv, s, (O,v)) emp; rewrite sepcon_emp; normalize.
      apply andp_right.
      - apply prop_right. split; [ simpl; intros; normalize | intuition]. 
        Exists x; entailer!.
        if_tac; subst; normalize.
        destruct O; simpl; [ congruence | entailer!].
      - apply sepcon_derives; trivial. apply derives_refl; reflexivity. 
    + (*newS*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      rename par into gv.
      Exists gv emp; rewrite sepcon_emp; normalize.
      apply andp_right; trivial.
      apply prop_right. split; [ simpl; intros; normalize | intuition].
      Exists x; entailer!.
      Exists 0. unfold SubObs; simpl. unfold Sub'.
      rewrite seplog.sepcon_emp. apply derives_refl.
    + (*register*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as ((((s,O),v), o), gv); simpl in *; subst; normalize.
      Exists ((((s,O),v), o), gv) emp; rewrite sepcon_emp; normalize.
      unfold Sub', Obs_; sep_apply Subject_Register; Intros z.
      Exists z. apply andp_right.
      - apply prop_right. split; [ simpl; intros; normalize | intuition]. cancel.
      - cancel. apply Observer_Notify.
    + (*update*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      Exists par emp. destruct par as (((s,n),l), gv); rewrite sepcon_emp. 
      normalize. simpl in *; subst.
      unfold SubObs_; Intros z. Exists z; entailer!.
      - intros. apply sepcon_derives. apply derives_refl.
        rewrite <- Obs_star_char. apply derives_refl.
      - unfold SubObs; simpl; unfold Sub'. sep_apply Subject_Update.
        Exists (list_repeat (Datatypes.length l) z). cancel. clear.
        induction l; simpl; [trivial | intros].
        eapply sepcon_derives; [ apply Observer_Notify | apply IHl].
    + (*get*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as ((s,O),v); rewrite sepcon_emp. 
      normalize. simpl in *; subst. unfold Sub'.
      sep_apply Subject_GetPrePost.
      Exists (s, (O, v)) (GetPost SP (O, v) s -* Subjectrep SP (O, v) s).
      entailer!.
      intros. apply wand_frame_elim'' (*(GetPost SP (O, v) s) (Subjectrep SP (O, v) s))*).
    + (*newObserver*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as (((gv, s), O),v); rewrite sepcon_emp. 
      normalize. simpl in *; subst. 
      Exists (gv, s, (O, v)) (Obs_star O s v). entailer!.
      - intros. Exists x0; entailer!. unfold SubObs, Sub'; simpl; cancel.
        apply derives_refl.
      - unfold SubObs; simpl; unfold Sub'. sep_apply Subject_Register. cancel.
    + (*notify *)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as ((((o, s), O), v), fp); rewrite sepcon_emp.
      normalize. simpl in *; subst. 
      unfold Obs_, Sub'; Intros z; simpl. unfold Obs'. 
      sep_apply Subject_GetPrePost. sep_apply Observer_Notify. simpl in *.
      Exists (o, s, z, (O, v), fp) (GetPost SP (O, v) s -* Subjectrep SP (O, v) s).
      entailer!.
      simpl; intros. cancel.
      apply wand_frame_elim'' (*(GetPost SP (O, v) s) (Subjectrep SP (O, v) s))*).
      unfold func_ptr'. apply andp_derives; trivial. apply func_ptr_mono. 
      unfold gspec, getspec; simpl. unfold GPre', GPost'.  clear.
      split. split; reflexivity. intros. simpl in x2. destruct gargs as [g args]. 
      simpl; destruct x2 as [s [O v]]; Intros. Exists ts2 (s, O, v) emp.
      entailer!.
    + (*val*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as (((o, s) ,O),v); rewrite sepcon_emp.
      normalize. simpl in *; subst. unfold SubObs; simpl.
      apply in_split in H0. destruct H0 as [O1 [O2 HO]]. subst O.
      rewrite Pr_star_app.
      Exists (o, s, v) ((Obs_star O1 s v * Obs_star O2 s v * Sub AP s (O1++o::O2) v)%logic). 
      entailer!.
      - intros. rewrite ! Obs_star_char. simpl. cancel.
      - simpl. cancel. apply Observer_Val.
    + (*freeO*)
      eexists; split; [ reflexivity | inv Phi2; go_sub].
      destruct par as ((gv, o), s); rewrite sepcon_emp. 
      normalize. simpl in *; subst.
      unfold Obs_; Intros z. 
      Exists (gv, o, (s, z)) emp. entailer!. apply derives_refl.
  Qed.
End SubjectObserver_ASIsub.