Require Import VST.floyd.proofauto.
Require Import spec_stdlib. (*for malloc_token*)

Global Open Scope funspec_scope.

Require Import Subject.
Instance SCompSpecs : compspecs. make_compspecs prog. Defined.

Require Import spec_Subject.

Require Import Observer.
Instance OCompSpecs : compspecs. make_compspecs prog. Defined.

Require Import spec_Observer.

Definition NodeTP:type:= Tstruct _node noattr.

Section AggregateASI.
Variable M: MallocFreeAPD.

Fixpoint lst (l:val) (O: list val): mpred :=
  match O with
    nil => !!(l=nullval) && emp
  | o::os => EX n, !!(isptr o) && @data_at SCompSpecs Ews NodeTP (o,n) l *
                   @malloc_token SCompSpecs M Ews NodeTP l * lst n os
  end.

Definition Sub (s:val) (O: list val) (v:Z) : mpred :=
  EX l:_,
     lst l O * @malloc_token SCompSpecs M Ews SubjectTP s *
     @data_at SCompSpecs Ews SubjectTP (l, Vint (Int.repr v)) s.

Definition Obs (o s:val)(v:Z) : mpred :=
  (@malloc_token OCompSpecs M Ews ObserverTP o * 
   @data_at OCompSpecs Ews ObserverTP (s, Vint (Int.repr v)) o)%logic.

Definition Obs_ (o s:val) : mpred := EX v, Obs o s v.

Definition Obs_star (O: list val) s v:mpred :=
  fold_right (fun o P => Obs o s v * P)%logic emp O.

Definition SubObs (s:val) (O:list val) (v:Z):mpred :=
  (Sub s O v * Obs_star O s v)%logic.
  
Definition SubObs_ (s:val) (O:list val):mpred := EX v, SubObs s O v.

Definition GPrePost (s:val) (O: list val) (v:Z)  := 
  (@field_at SCompSpecs Ews SubjectTP [StructField _value] (Vint (Int.repr v)) s
   * @malloc_token SCompSpecs M Ews SubjectTP s )%logic.

Lemma Sub_GPrePost' s O v: Sub s O v |-- (GPrePost s O v * (GPrePost s O v -* Sub s O v))%logic.
  Proof.
  unfold Sub, GPrePost; Intros l. unfold_data_at (data_at _ _ _ _); cancel.
  apply wand_frame_intro'. Exists l. unfold_data_at (data_at _ _ _ _); cancel.
  Qed.

(*Differences to Parkinson's spec: gv and mem_mgr*)
Definition newS_spec :=
    DECLARE _newSubject
    WITH gv: globals
    PRE [ ] PROP() PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr SubjectTP ]
    EX s: val,
     PROP()
     LOCAL(temp ret_temp s)
     SEP(SubObs_ s nil; mem_mgr M gv).

(*Differences to Parkinson's spec: gv and mem_mgr*)
Definition reg_spec :=
    DECLARE _registr
    WITH s:val, O:list val, v: Z, o:val, gv:globals
    PRE [ tptr SubjectTP, tptr ObserverTP ]
     PROP ()
     PARAMS (s; o) GLOBALS (gv)
     SEP (Sub s O v; Obs_ o s; mem_mgr M gv)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (Sub s (o::O) v; Obs o s v; mem_mgr M gv).

(*now with gv*)
Definition upd_spec :=
    DECLARE _update
    WITH s:val, n: Z, O:list val, gv: globals
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s; Vint (Int.repr n)) GLOBALS (gv)
     SEP(SubObs_ s O)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(SubObs s O n).

Definition get_spec :=
    DECLARE _get
    WITH s:val, O: list val, v:Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(Sub s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(Sub s O v).

(*Differences to Parkinson's spec: gv and mem_mgr*)
Definition newO_spec :=
    DECLARE _newObserver
    WITH gv: globals, s:val, O:list val, v:Z
    PRE [ tptr SubjectTP ]
      PROP()
      PARAMS (s) GLOBALS (gv)
      SEP(SubObs s O v; mem_mgr M gv)
    POST[ tptr ObserverTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(SubObs s (p::O) v; mem_mgr M gv).

Definition gspec :=
    WITH s:val, O: list val, v: Z
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(GPrePost s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(GPrePost s O v).

Definition notify_spec :=
    DECLARE _notify
    WITH o:val, s:val, O: list val, v:Z, fp:val
    PRE [ tptr ObserverTP,
          tptr (Tfunction (Tcons (tptr SubjectTP) Tnil) tint cc_default) ]
     PROP()
     PARAMS (o;fp) GLOBALS ()
     SEP(Sub s O v; Obs_ o s; func_ptr' gspec fp)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(Sub s O v; Obs o s v).

Definition Oval_spec :=
    DECLARE _val
    WITH o:val, s:val, O:list val, v: Z
    PRE [ tptr ObserverTP ]
     PROP(In o O)
     PARAMS (o) GLOBALS ()
     SEP(SubObs s O v)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr v)))
      SEP(SubObs s O v).

End AggregateASI.

Require Import verif_Subject.
Lemma lst_char {M}: forall O l, lst M l O = listrep M O l.
Proof.
  induction O; intros; simpl; trivial.
  apply pred_ext; Intros n; Exists n; entailer!;
  rewrite IHO; apply derives_refl. 
Qed.

Lemma Sub_char {M} s O v: Sub M s O v = Subject_rep M (O,v) s.
Proof.
  apply pred_ext; unfold Sub, Subject_rep; Intros l; Exists l;
  rewrite lst_char; apply derives_refl.
Qed.

Require Import verif_Observer.
Lemma Obs_char {M} o s v: Obs M o s v = Observer_rep M (s,v) o.
Proof. trivial. Qed.

Lemma Obs_star_char {M}: forall O s v,
    Obs_star M O s v = Observers (Observer_rep M) s (list_repeat (length O) v) O.
Proof.
  unfold Obs_star.
  induction O; simpl; intros. reflexivity.
  rewrite <- IHO, Obs_char; trivial.
Qed.

Lemma Obs_star_cons {M} O s v p: (Obs_star M O s v * Observer_rep M (newObserver (O, v) s) p = Obs_star M (p :: O) s v)%logic.
Proof. simpl. apply pred_ext;  cancel; apply derives_refl. Qed.

Lemma Obs_star_app {M s v}: forall L1 L2, 
  Obs_star M (L1 ++ L2) s v = (Obs_star M L1 s v * Obs_star M L2 s v)%logic.
  Proof.
   induction L1; simpl; intros. rewrite emp_sepcon; trivial.
   rewrite IHL1; rewrite sepcon_assoc; trivial.
  Qed.

Import VST.msl.seplog.

  Ltac go_sub := 
      apply NDsubsume_subsume;
      [ split; reflexivity
      | split; [ split; reflexivity 
               | intros par [g args]; simpl in par; simpl; 
                 unfold LAMBDAx, GLOBALSx, PROPx, LOCALx, SEPx, argsassert2assert, local, liftx, lift1, lift; simpl; entailer!]].

Section AggregateConcreteSub.
  Variable M: MallocFreeAPD.
  Notation SP:= (SubjectAP M).
  Notation OP:= (ObserverAP M).

Lemma newS_sub: funspec_sub (snd (newSubject_spec M SP)) (snd (newS_spec M)).
Proof.
  go_sub. rename par into gv. 
  Exists gv emp. entailer!.
  intros. Exists o. entailer!. Exists 0. unfold SubObs; simpl.
  rewrite seplog.sepcon_emp, Sub_char. simpl. Exists nullval; entailer!.
Qed.

Lemma register_sub: funspec_sub (snd (register_spec M SP OP)) (snd (reg_spec M)).
Proof.
  go_sub.
  destruct par as ((((s,O),v), o), gv). simpl; normalize.
  Exists ((((s,O),v), o), gv) emp. entailer!.
  + intros. simpl. cancel.  rewrite Sub_char, Obs_char; simpl.
    Exists x0 n. entailer!.
  + unfold Obs_; Intros z. rewrite Sub_char, Obs_char. sep_apply Observer_NotifyPre.
    Exists z; cancel. apply derives_refl. 
Qed.

Lemma update_sub: funspec_sub (snd (update_spec SP OP)) (snd (upd_spec M)).
Proof.
  go_sub.
  destruct par as (((s,n),l), gv); simpl; normalize. clear H.
  Exists (((s,n),l), gv) emp.
  unfold SubObs_; Intros z. entailer!.
  + intros. unfold SubObs. rewrite Obs_star_char, <- lst_char. cancel.
    unfold Sub. Exists x0. cancel. apply derives_refl.
  + unfold SubObs, Sub. normalize. rewrite lst_char, Obs_star_char.
    Exists (list_repeat (Datatypes.length l) z) z x. cancel.
    apply sepcon_derives. apply derives_refl.
    clear. induction l; simpl; trivial.
    rewrite NotifyPre_char. simpl. Exists (Vint (Int.repr z)). cancel.
Qed.

Lemma get_sub: funspec_sub (snd (get_largespec SP)) (snd (get_spec M)).
Proof.
  go_sub. destruct par as ((s,O),v).
  Exists (s, (O, v)) emp. simpl; entailer!.
  - intros. rewrite Sub_char. Exists x0; cancel.
  - rewrite Sub_char. apply derives_refl.
Qed.

Lemma newO_sub: funspec_sub (snd (newObserver_spec M OP SP)) (snd (newO_spec M)).
Proof.
  go_sub. destruct par as (((gv, s) ,O),v). simpl; normalize.
  Exists (gv, s, (O, v)) (Obs_star M O s v). entailer!.
  - intros. unfold SubObs. Exists o. simpl. rewrite Sub_char, Obs_char. simpl. entailer!.
    Exists n x0. entailer!.
  - unfold SubObs. rewrite Sub_char. cancel. 
Qed.

Lemma notify_sub: funspec_sub (snd (notify_largespec OP SP)) (snd (notify_spec M)).
Proof.
  go_sub. destruct par as ((((o, s), O), v), fp). simpl; normalize. clear H.
  unfold Obs_; Intros z. rewrite Sub_char, Obs_char.
  Exists (o, s, z, (O, v), fp) emp; entailer!. 
  - intros. cancel. rewrite Obs_char. apply derives_refl.
  - simpl. rewrite NotifyPre_char. Exists (Vint (Int.repr z)). cancel.
    unfold func_ptr'. apply andp_derives; trivial. apply func_ptr_mono.
    unfold gspec, getspec. simpl. unfold GPrePost, GetPrePost, GetPre.
    split. split; reflexivity. intros; Intros. clear.
    destruct x2 as [s [O v]]; simpl. Exists (@nil Type).
    Exists (s, O, v) emp. entailer!.
    * intros. apply derives_refl.
    * intros. apply derives_refl.
Qed.

Lemma val_sub: funspec_sub (snd (val_spec OP)) (snd (Oval_spec M)).
Proof.
  go_sub. destruct par as (((o, s), O),v); simpl; normalize.
  unfold SubObs.
  apply in_split in H0. destruct H0 as [O1 [O2 HO]]. subst O.
  rewrite Obs_star_app. simpl.
  Exists (o, s, v) (Obs_star M O1 s v * Obs_star M O2 s v * Sub M s (O1++o::O2) v)%logic. entailer!.
  - intros. cancel. rewrite Obs_char. apply derives_refl.
  - rewrite Obs_char. apply derives_refl.
Qed.

End AggregateConcreteSub.