Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import VST.veric.initial_world.

Require Import VST.floyd.linking.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Subject.
Require Import spec_Observer.
Require Import verif_Subject.
Require Import verif_Observer.
Require Import SubjectComponent.
Require Import ObserverComponent.
Require Import verif_surelyMalloc.

(***** In this file, we link stdlib + ((surelymalloc + subject) + (surelymalloc + observer)),
  so have two copies of the (by then internal function) surelymalloc when we
  link sm_sub with sm_obs), exercising the support of multiple imported 
  libraries with VSU*)

Section Application_VSU.
Variable M: MallocFreeAPD.

Definition Subject := SubjectAP M.
Definition Observer := ObserverAP M. 

(****** 1. A program unit containing Subject.c and surely_malloc****)
(* Note that the import contains notify_smallspec Observer (SubjectAP M)*)

Definition sm_subj_prog': Clight.program := 
  match link_progs surelyMalloc.prog Subject.prog with
    Errors.OK p => p
  | _ => surelyMalloc.prog (*arbitrary*)
  end.

Definition sm_subj_prog := ltac:
  (let x := eval hnf in sm_subj_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance sm_subj_cs : compspecs. make_compspecs sm_subj_prog. Defined.

Definition sm_subj_Vprog:varspecs. mk_varspecs sm_subj_prog. Defined.

(*MF_ASI has different order (TODO: change VSULink_tac tactic so that
  imports need not be list equality but can be extensional equality*)
Definition sm_subj_Imports:funspecs := 
   [free_spec' M Subject._free; notify_smallspec Observer (SubjectAP M); malloc_spec' M surelyMalloc._malloc; exit_spec].

Definition sm_subj_Exports:funspecs := G_merge (SubjectASI M Subject Observer) (SM_ASI M).

Definition SM_Subj_VSU:
@VSU NullExtension.Espec
   sm_subj_Vprog sm_subj_cs nil sm_subj_Imports sm_subj_prog sm_subj_Exports emp.
Proof.
  VSULink_tac (SubjectVSU M Observer)  (SMVSU M).
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

(****** 2. A program unit containing Observer.c and surely_malloc****)
(* We will thus have two copies of surely_malloc. This time, the import
  contains register_spec and get_smallspec.*)

Definition sm_obs_prog': Clight.program := 
  match link_progs surelyMalloc.prog Observer.prog with
    Errors.OK p => p
  | _ => surelyMalloc.prog (*arbitrary*)
  end.

Definition sm_obs_prog := ltac:
  (let x := eval hnf in sm_obs_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance sm_obs_cs : compspecs. make_compspecs sm_obs_prog. Defined.

Definition sm_obs_Vprog:varspecs. mk_varspecs sm_obs_prog. Defined.

Definition sm_obs_Imports:funspecs := [free_spec' M Subject._free; register_spec M Subject (ObserverAP M); 
                                       malloc_spec' M surelyMalloc._malloc; exit_spec].

Definition sm_obs_Exports:funspecs := G_merge (ObserverASI M Observer Subject) (SM_ASI M).

Definition SM_Obs_VSU:
@VSU NullExtension.Espec
   sm_obs_Vprog sm_obs_cs nil sm_obs_Imports sm_obs_prog sm_obs_Exports emp.
Proof.
  VSULink_tac (ObserverVSU M Subject)  (SMVSU M).
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

(****** 3. Let's merge the two VSUs, to obtain a SM_subjObs VSU. ***)
Goal (find_id surelyMalloc._surely_malloc (prog_funct sm_subj_prog) = 
      Some (Internal surelyMalloc.f_surely_malloc)).
reflexivity. Qed.
Goal (find_id surelyMalloc._surely_malloc (prog_funct sm_obs_prog) = 
      Some (Internal surelyMalloc.f_surely_malloc)).
reflexivity. Qed.

Definition application_prog': Clight.program := 
  match link_progs sm_subj_prog sm_obs_prog with
    Errors.OK p => p
  | _ => Subject.prog (*arbitrary*)
  end.
Definition application_prog := ltac:
  (let x := eval hnf in application_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance AppCs : compspecs. make_compspecs application_prog. Defined.

Definition AppVprog:varspecs. mk_varspecs application_prog. Defined.

Definition AppImports:funspecs := [free_spec' M Subject._free; malloc_spec' M surelyMalloc._malloc; exit_spec].

Definition AppExports:funspecs := G_merge sm_subj_Exports sm_obs_Exports.

Definition Application_VSU:
@VSU NullExtension.Espec AppVprog AppCs nil AppImports application_prog AppExports emp.
Proof.
  VSULink_tac SM_Subj_VSU SM_Obs_VSU.
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.
End Application_VSU.