Require Export VST.floyd.proofauto.
Require Import VST.veric.compcert_rmaps.
Require Import VST.veric.rmaps. (*for TypeTree constructors*)
Require Import VST.progs.conclib. (*for approx_idem; This line requires progs and concurrency to be included in the VST_DIRS in Makefile, and in _CoqProject*)

Import FashNotation.

Lemma sepcon_f_equal P Q P' Q': P=P' -> Q=Q' -> (sepcon P Q) = (sepcon P' Q').
Proof. intros; subst; trivial. Qed.

Lemma approx_f_equal n P P':
  P=P' -> compcert_rmaps.RML.R.approx n P = compcert_rmaps.RML.R.approx n P'.
Proof. intros; subst; trivial. Qed.

Lemma func_ptr'_mono fs gs v (FS: funspec_sub fs gs): func_ptr' fs v |-- func_ptr' gs v.
  Proof.  apply (func_ptr_mono _ _ v) in FS. apply andp_derives; trivial. Qed.

Lemma approx_prop_andp {P Q:Prop} n:
  approx n (prop (P /\ Q)) = (approx n (prop P)) && (approx n (prop Q)).
Proof. apply approx_prop_andp. Qed.

Lemma wand_nonexpansive_l: forall P Q n,
  approx n (P -* Q)%logic = approx n (approx n P  -* Q)%logic.
Proof.
  repeat intro.
  apply (nonexpansive_super_non_expansive (fun P => predicates_sl.wand P Q)).
  split; intros ?? Hshift ??????.
  - eapply Hshift; eauto.
    apply necR_level in H1; apply necR_level in H2.
    apply join_level in H3 as [].
    apply (H y0); auto; lia.
  - eapply Hshift; eauto.
    apply necR_level in H1; apply necR_level in H2.
    apply join_level in H3 as [].
    apply (H y0); auto; lia.
Qed.

Lemma wand_nonexpansive_r: forall P Q n,
  approx n (P -* Q)%logic = approx n (P  -* approx n Q)%logic.
Proof.
  repeat intro.
  apply (nonexpansive_super_non_expansive (fun Q => predicates_sl.wand P Q)).
  split; intros ?? Hshift ??????.
  - eapply Hshift in H4; eauto.
    apply necR_level in H1; apply necR_level in H2.
    apply join_level in H3 as [].
    apply (H z); auto; lia.
  - eapply Hshift in H4; eauto.
    apply necR_level in H1; apply necR_level in H2.
    apply join_level in H3 as [].
    apply (H z); auto; lia.
Qed.

Lemma wand_nonexpansive: forall P Q n,
  approx n (P -* Q)%logic = approx n (approx n P  -* approx n Q)%logic.
Proof.
  intros; rewrite wand_nonexpansive_l, wand_nonexpansive_r; reflexivity.
Qed.

Lemma quick_derives_right {P Q }: TT |-- Q -> P |-- Q.
Proof. intros. eapply derives_trans. 2: apply H. apply TT_right. Qed.

Lemma hered_derives {P Q: mpred}: predicates_hered.derives P Q -> derives P Q.
Proof. intros. apply H. Qed.

Require Export SubjectObserverModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Observer. (*needed for function and type names *)
(*Require Import VST.floyd.canon.*)

Global Open Scope funspec_scope.
Definition ObserverTP:type:=  Tstruct _observer noattr.
Definition SubjectTP:type:=  Tstruct _subject noattr.

Section ObserverASI.
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.

  Notation S_rep := (Subjectrep SP).
  Notation R_Pre := (RegisterPre SP). (*
  Notation U_Pre := (UpdatePre SP).*)
(*  Notation G_Pre := (GetPre SP).
  Notation G_Post := (GetPost SP).*)
  Notation O_rep := (Observerrep OP).
  Notation N_Pre:= (NotifyPre OP).
  Notation V_Pre:= (ValPre OP).

  Definition newObserver_spec :=
    DECLARE _newObserver
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
      PROP()
      PARAMS (s) GLOBALS (gv)
      SEP(R_Pre S s; mem_mgr M gv)
    POST[ tptr ObserverTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(O_rep (newObserver S s) p;
           S_rep (p::fst S, snd S) s; mem_mgr M gv).

  Definition W:Type := val * val * Z * SubjectRepTp * val.
  Definition PreType:Type := SubjectRepTp -> val -> mpred.
  Definition PostType:Type := SubjectRepTp -> val -> mpred.
  Definition fpspec (GPre:PreType) (GPost:PostType):funspec :=
            WITH t:val, T: SubjectRepTp
            PRE [ tptr SubjectTP ]
              PROP() PARAMS (t) GLOBALS () SEP(GPre T t)
            POST[ tint ]
              PROP()
              LOCAL(temp ret_temp (Vint (Int.repr (snd T))))
              SEP(GPost T t).
  Lemma fspec_nonexpansive pre post fp n:
        approx n (func_ptr' (fpspec pre post) fp) =
        approx n (func_ptr' (fpspec (fun U u => approx n (pre U u))
                                    (fun U u=> approx n (post U u))) fp).
  Proof. 
    unfold fpspec, PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert; simpl.
    apply approx_Sn_eq_weaken.
    rewrite approx_func_ptr'.
    setoid_rewrite approx_func_ptr' at 2. apply f_equal.
    apply funcptr_f_equal'; trivial. simpl.
    apply semax_prog.funspec_eq.
      * extensionality tss a rho'; destruct a.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
      * extensionality tss a rho'; destruct a.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
  Qed.

(*Should be  Definition notif_smallpre (ts: list Type) (x: W * PreType * PostType) :=
   match x with ((o, s, c, U, fp), GPre, GPost) =>
     PROP (is_pointer_or_null s)
     PARAMS (o;fp) GLOBALS ()
     SEP (GPre U s; N_Pre (s,c) o; func_ptr' (fpspec GPre GPost) fp)
    end.
for some reason, the notation mechnisms does not work here in most recent VST, 
even if one manually imports floys.canon again.
Same error occurs for notif_largepre (line 215 below)*)

  Definition notif_smallpre (ts: list Type) (x: W * PreType * PostType) :=
   match x with ((o, s, c, U, fp), GPre, GPost) =>
     PROP (is_pointer_or_null s)
     (PARAMSx [o;fp] (GLOBALSx []
     (SEPx [GPre U s; N_Pre (s,c) o; func_ptr' (fpspec GPre GPost) fp])))
    end.

  Definition notif_smallpost (ts: list Type) (x: W * PreType * PostType) :=
   match x with ((o, s, c, U, fp), GPre, GPost) =>
      PROP()
      LOCAL()
      SEP(O_rep (s, snd U) o; GPost U s)
   end.

  Definition notif_argType:TypeTree :=
  ProdType (ProdType (ConstType W) 
                     (ArrowType (ConstType SubjectRepTp) 
                                (ArrowType (ConstType val) Mpred)))
           (ArrowType (ConstType SubjectRepTp)
                      (ArrowType (ConstType val) Mpred)).

  Lemma notif_smallpre_nonexpansive:@args_super_non_expansive notif_argType notif_smallpre.
  Proof. repeat intro. 
    destruct x as [[w pre] post]; simpl. simpl in pre, post.
    destruct w as [[[[o s] c] U] fp]; simpl. unfold fpspec.
    unfold PROPx; simpl; rewrite !approx_andp. (*f_equal*)
    unfold LAMBDAx. rewrite !approx_andp. (* f_equal.*)
    unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. (*f_equal.*)
    unfold argsassert2assert. simpl.
    unfold SEPx; simpl. rewrite !sepcon_emp.
    rewrite !approx_sepcon. rewrite approx_idem.
    apply pred_ext. 
    + apply andp_derives; [ apply derives_refl |]. 
      apply andp_derives; [ apply derives_refl |]. 
      apply andp_derives; [ apply derives_refl |]. 
      apply sepcon_derives; [ apply derives_refl |].
      apply sepcon_derives; [ apply derives_refl |].
      apply derives_refl'.
      apply approx_Sn_eq_weaken.
      rewrite approx_func_ptr'.
      setoid_rewrite approx_func_ptr' at 2. apply f_equal.
      apply funcptr_f_equal'; trivial. simpl.
      apply semax_prog.funspec_eq.
      - extensionality tss a rho'; destruct a.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
      - extensionality tss a rho'; destruct a.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
    + apply andp_derives; [ apply derives_refl |]. 
      apply andp_derives; [ apply derives_refl |].
      apply andp_derives; [ apply derives_refl |]. 
      apply sepcon_derives; [ apply derives_refl |].
      apply sepcon_derives; [ apply derives_refl |].
      apply derives_refl'.
      apply approx_Sn_eq_weaken.
      rewrite approx_func_ptr'.
      setoid_rewrite approx_func_ptr' at 2. apply f_equal.
      apply funcptr_f_equal'; trivial. simpl.
      apply semax_prog.funspec_eq.
      - extensionality tss a rho'; destruct a.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
      - extensionality tss a rho'; destruct a.
        rewrite !approx_andp, !approx_sepcon, approx_idem; auto.
  Qed.

  Lemma notif_smallpost_nonexpansive: @super_non_expansive notif_argType notif_smallpost.
  Proof. repeat intro. 
  destruct x as [[w pre] post]; simpl.
  destruct w as [[[[o s] c] U] fp]; simpl.
  unfold PROPx; simpl; rewrite !approx_andp. f_equal.
  unfold LOCALx; simpl. rewrite !approx_andp. f_equal.
  unfold SEPx; simpl. rewrite ! sepcon_emp. 
  rewrite !approx_sepcon. rewrite approx_idem. trivial.
  Qed. 

(*again notation issue
  Definition notif_largepre (ts: list Type) (x: W * PreType * PostType) :=
   match x with ((o, s, c, U, fp), GPre, GPost) =>
     PROP (is_pointer_or_null s)
     PARAMS (o;fp) GLOBALS ()
     SEP (S_rep U s; N_Pre (s,c) o; func_ptr' (fpspec GPre GPost) fp;
          (!ALL S s, (S_rep S s >=> (GPre S s * (*|> *)(GPost S s -* S_rep S s)))) && emp)
    end.*)
  Definition notif_largepre (ts: list Type) (x: W * PreType * PostType) :=
   match x with ((o, s, c, U, fp), GPre, GPost) =>
     PROP (is_pointer_or_null s)
     (PARAMSx [o;fp] (GLOBALSx []
     (SEPx [S_rep U s; N_Pre (s,c) o; func_ptr' (fpspec GPre GPost) fp;
          (!ALL S s, (S_rep S s >=> (GPre S s * (*|> *)(GPost S s -* S_rep S s)))) && emp])))
    end.

  Definition notif_largepost (ts: list Type) (x: W * PreType * PostType) :=
   match x with ((o, s, c, U, fp), GPre, GPost) =>
      PROP()
      LOCAL()
      SEP(O_rep (s, snd U) o; S_rep U s)
   end.

Lemma notif_largepre_nonexpansive:@args_super_non_expansive notif_argType notif_largepre.
Proof. repeat intro. 
  destruct x as [[w pre] post]; simpl. simpl in pre, post.
  destruct w as [[[[o s] c] U] fp]; simpl. (*unfold fpspec.*)
  unfold PROPx; simpl; rewrite !approx_andp. (*f_equal*)
  unfold LAMBDAx. rewrite !approx_andp. (* f_equal.*)
  unfold GLOBALSx, LOCALx; simpl. rewrite !approx_andp. (*f_equal.*)
  unfold argsassert2assert. simpl.
  unfold SEPx; simpl. rewrite !sepcon_emp, approx_prop_andp.
  rewrite !approx_sepcon, !approx_andp.
  apply pred_ext. 
  + apply andp_derives; [ apply derives_refl |]. 
    apply andp_derives; [ apply derives_refl |].
    apply andp_derives; [ apply derives_refl |].
    apply sepcon_derives; [ apply derives_refl |].
    apply sepcon_derives; [ apply derives_refl |].
    apply sepcon_derives.
    - apply derives_refl'; apply fspec_nonexpansive.
    - apply andp_derives; [ | apply derives_refl ].
      apply (approx_derives4 S_rep pre post S_rep n).
  + apply andp_derives; [ apply derives_refl |]. 
    apply andp_derives; [ apply derives_refl |].
    apply andp_derives; [ apply derives_refl |]. 
    apply sepcon_derives; [ apply derives_refl |].
    apply sepcon_derives; [ apply derives_refl |].
    apply sepcon_derives.
    - rewrite <- fspec_nonexpansive; apply derives_refl.
    - apply andp_derives; [| apply derives_refl].
      apply approx_derives4_inv.
Qed.

Lemma notif_largepost_nonexpansive:@super_non_expansive notif_argType notif_largepost.
  Proof. repeat intro. 
  destruct x as [[w pre] post]; simpl.
  destruct w as [[[[o s] c] U] fp]; simpl.
  unfold PROPx; simpl; rewrite !approx_andp. f_equal.
  Qed. 

Definition notifsmallspec:funspec := 
  mk_funspec ([tptr ObserverTP;
              tptr (Tfunction (Tcons (tptr SubjectTP) Tnil) tint cc_default)], tvoid)
             cc_default
             (*(ConstType W)*)_  _ _ 
             notif_smallpre_nonexpansive
             notif_smallpost_nonexpansive.
Definition notiflargespec:funspec := 
  mk_funspec ([tptr ObserverTP;
              tptr (Tfunction (Tcons (tptr SubjectTP) Tnil) tint cc_default)], tvoid)
             cc_default
             (*(ConstType W)*)_  _ _ 
             notif_largepre_nonexpansive
             notif_largepost_nonexpansive.

  Definition notify_smallspec :=
    DECLARE _notify notifsmallspec.

  Definition val_spec :=
    DECLARE _val
    WITH o:val, s:val, c: Z
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS ()
     SEP(V_Pre (s,c) o)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr c)))
      SEP(O_rep (s,c) o).

  Definition freeObserver_spec:=
    DECLARE _freeObserver 
    WITH gv: globals, o:val, O:ObserverRepTp
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS (gv)
     SEP(O_rep O o; mem_mgr M gv)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(mem_mgr M gv). 

  Definition ObserverASI:funspecs := [ newObserver_spec; notify_smallspec; val_spec; freeObserver_spec].

  Definition notify_largespec :=
    DECLARE _notify notiflargespec.

Lemma Notify_subsi: TT |-- seplog.funspec_sub_si (snd notify_smallspec) (snd notify_largespec).
Proof. unfold notify_smallspec, notify_largespec, notifsmallspec, notiflargespec,
  notif_argType, notif_smallpre, notif_smallpost, notif_largepre, notif_largepost.
  simpl snd. (*apply funspec_sub_sub_si. *)unfold seplog.funspec_sub_si.
  apply hered_derives; intros ? _.  split. split; trivial. simpl; intros.
  exists b; clear b. destruct H2. simpl in H2. destruct b1 as [gv args]. exists b0.
  destruct b0 as [[[[[[o s] c] U] fp] gpre] gpost]. 
  exists (((gpost U s -* S_rep U s))%logic)%pred; split.
  + destruct H3 as [PR [AR [LOC SE]]]. simpl in *. unfold argsassert2assert, SEPx in SE.
    simpl in SE. destruct SE as [p [q [Jpq [P Q]]]].
    rewrite <- sepcon_assoc in Q. destruct Q as [q1 [q2 [Jq [Q1 Q2]]]]. rewrite sepcon_emp in Q2.
    destruct Q2 as [IMP Q2].
    assert (Lev: ge (level q2) (level p)).
    { apply join_level in Jq. apply join_level in Jpq. lia. }
    apply (IMP U s p Lev) in P; clear IMP. rewrite sepcon_comm in P.
    destruct P as [t [r [Jtr [T R]]]].
    destruct (join_assoc Jtr Jpq) as [z [Jz1 Jz2]].
    exists t, z; split3; trivial.
    split. apply PR. split. apply AR. split. apply LOC. unfold argsassert2assert, SEPx.
    simpl. exists r, q; split3; trivial. rewrite <- sepcon_assoc.
    exists q1, q2; split3; trivial. apply necR_refl.
  + clear H2 H3. intros. destruct H4 as [? [p [q [Jpq [P Q]]]]].
    destruct Q as [PR [LO SE]]. split3. apply PR. apply LO.
    unfold SEPx; simpl. unfold SEPx in SE; simpl in SE. rewrite sepcon_emp in *.
    destruct SE as [r [t [Jrt [R T]]]]. apply join_comm in Jpq.
    destruct (join_assoc Jrt Jpq) as [g [G1 G2]]. exists r, g; split3; trivial.
    clear - G1 T P. specialize (wand_frame_elim'' (gpost U s) (S_rep U s) g).
    intros. apply H. rewrite sepcon_comm. exists t, p; split3; trivial.
Qed. (*
Lemma Notify_subsi2: TT |-- seplog.funspec_sub_si (snd notify_smallspec) (snd notify_largespec).
Proof. unfold notify_smallspec, notify_largespec, notifsmallspec, notiflargespec,
  notif_argType, notif_smallpre, notif_smallpost, notif_largepre, notif_largepost.
  simpl snd. apply funspec_sub_sub_si.  split. split; trivial. simpl; intros. Intros.
  (*exists b; clear b. destruct H2. simpl in H2. destruct b1 as [gv args]. exists b0.*) Exists ts2.
  (*destruct x2 as [[[[[[o s] c] U] fp] gpre] gpost]. *)
  Exists x2. destruct x2 as [[[[[[o s] c] U] fp] gpre] gpost].
  Exists (((gpost U s -* S_rep U s))%logic)%pred. entailer.
  unfold PROPx, LAMBDAx, GLOBALSx, LOCALx, SEPx, argsassert2assert. simpl; entailer.
  cancel. apply derives_low.
  intros q Q. destruct Q as [q1 [q2 [Jq [Q1 [IMP Q2]]]]].
  rewrite sepcon_assoc in Q1. destruct Q1 as [r1 [r2 [Jr [R1 R2]]]].
  exploit (IMP U s r1); clear IMP.
  { apply join_level in Jq. apply join_level in Jr. lia. }
  apply necR_refl. apply R1.
  intros. split.
  + red. econstructor. exists r1. intros t. simpl.
  apply  simpl in Q2. red in Q2. destruct Q2.
  + destruct H3 as [PR [AR [LOC SE]]]. simpl in *. unfold  simpl., SEPx in SE.
    simpl in SE. destruct SE as [p [q [Jpq [P Q]]]].
    rewrite <- sepcon_assoc in Q. destruct Q as [q1 [q2 [Jq [Q1 Q2]]]]. rewrite sepcon_emp in Q2.
    destruct Q2 as [IMP Q2].
    assert (Lev: ge (level q2) (level p)).
    { apply join_level in Jq. apply join_level in Jpq. lia. }
    apply (IMP U s p Lev) in P; clear IMP. rewrite sepcon_comm in P.
    destruct P as [t [r [Jtr [T R]]]].
    destruct (join_assoc Jtr Jpq) as [z [Jz1 Jz2]].
    exists t, z; split3; trivial. (*later around gpost implication*) 
    split. apply PR. split. apply AR. split. apply LOC. unfold argsassert2assert, SEPx.
    simpl. exists r, q; split3; trivial. rewrite <- sepcon_assoc.
    exists q1, q2; split3; trivial. apply necR_refl.
  + clear H2 H3. intros. destruct H4 as [? [p [q [Jpq [P Q]]]]].
    destruct Q as [PR [LO SE]]. split3. apply PR. apply LO.
    unfold SEPx; simpl. unfold SEPx in SE; simpl in SE. rewrite sepcon_emp in *.
    destruct SE as [r [t [Jrt [R T]]]]. apply join_comm in Jpq.
    destruct (join_assoc Jrt Jpq) as [g [G1 G2]]. exists r, g; split3; trivial.
    clear - G1 T P. specialize (wand_frame_elim'' (gpost U s) (S_rep U s) g).
    intros. apply H. rewrite sepcon_comm. exists t, p; split3; trivial.
Qed*)
End ObserverASI.