Require Export VST.floyd.proofauto.
Require Export SubjectObserverModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Subject. (*needed for function and type names *)

Global Open Scope funspec_scope.
Definition SubjectTP:type:=  Tstruct _subject noattr.
Definition ObserverTP:type:=  Tstruct _observer noattr.

Fixpoint Observers (OPred:ObserverRepTp -> val -> mpred) s (vals: list Z) (l:list val) :=
    match vals, l with
      nil, nil => emp
    | v::vs, o::os => (OPred (s,v) o * Observers OPred s vs os)%logic
    | _, _ => FF
    end.

Lemma Observers_app OPred s: forall vals1 l1 vals2 l2 (L: length vals1 = length l1),
      Observers OPred s (vals1 ++ vals2) (l1++l2) = 
      (Observers OPred s vals1 l1 * Observers OPred s vals2 l2)%logic.
  Proof.
    induction vals1; intros; destruct l1; simpl; inv L.
    + rewrite emp_sepcon; trivial.
    + rewrite IHvals1; [ rewrite sepcon_assoc |]; trivial.
  Qed.

Section SubjectASI.
  Variable M: MallocFreeAPD.
  Variable SP: SubjectAPD.
  Variable OP: ObserverAPD.

  Notation S_rep := (Subjectrep SP).
  Notation R_Pre := (RegisterPre SP).
  Notation U_Pre := (UpdatePre SP).
  Notation G_Pre := (GetPre SP).
  Notation G_Post := (GetPost SP).
  Notation O_rep := (Observerrep OP).
  Notation N_Pre:= (NotifyPre OP).

  Definition newSubject_spec :=
    DECLARE _newSubject
    WITH gv: globals
    PRE [ ] PROP() PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr SubjectTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(S_rep newSubject p; mem_mgr M gv).

  Definition get_smallspec :=
    DECLARE _get
    WITH s:val, S: SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(G_Pre S s)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr (snd S))))
      SEP(G_Post S s).

  (*Spec now needs gv*)
  Definition update_spec :=
    DECLARE _update
    WITH s:val, n: Z, l:list val, gv: globals
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s; Vint (Int.repr n)) GLOBALS (gv)
     SEP(EX z, U_Pre (l,z) s; EX vals, Observers N_Pre s vals l)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (l,n) s; Observers O_rep s (list_repeat (length l) n) l).

  Definition register_spec :=
    DECLARE _registr
    WITH s:val, l:list val, n: Z, o:val, gv:globals
    PRE [ tptr SubjectTP, tptr ObserverTP  ]
     PROP ()
     PARAMS (s; o) GLOBALS (gv)
     SEP (R_Pre (l,n) s; mem_mgr M gv; EX v, N_Pre (s,v) o)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (S_rep (o::l, n) s; mem_mgr M gv; O_rep (s,n) o).

  Definition freeSubject_spec:=
    DECLARE _freeSubject 
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv)
     SEP(S_rep S s; mem_mgr M gv)
    POST[ tint ] EX r,
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr r)))
      SEP(if zeq r 0 then !!(fst S = @nil val)&&emp 
          else !!(r=1 /\ fst S <> @nil val) && S_rep S s; mem_mgr M gv).

  Definition detachfirst_spec:=
    DECLARE _detachfirst 
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv)
     SEP(S_rep S s; mem_mgr M gv)
    POST[ tptr ObserverTP ] EX r,
      PROP()
      LOCAL(temp ret_temp r)
      SEP(if Val.eq r nullval 
          then !!(fst S = @nil val) && S_rep S s 
          else EX O, !!(fst S=r::O) && S_rep (O, snd S) s; mem_mgr M gv).

  Definition SubjectASI:funspecs := [ detachfirst_spec; freeSubject_spec; newSubject_spec; register_spec; update_spec; get_smallspec].

  Definition get_largespec :=
    DECLARE _get
    WITH s:val, S: SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(S_rep S s)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr (snd S))))
      SEP(S_rep S s).
(*
Lemma Get_sub: funspec_sub (snd get_smallspec) (snd get_largespec).
Proof.
  do_funspec_sub. destruct w as [s [l n]]; clear H.
  Exists (s, (l, n)).
  Exists (GetPost SP (l, n) s -* Subjectrep SP (l, n) s).
  simpl; entailer!.
  { intros. rewrite <- H0. entailer!.
    apply wand_frame_elim''. }
  rewrite sepcon_comm. apply (Subject_GetPrePost SP (l, n)).
Qed.*)
Lemma derives_low (P Q:mpred): predicates_hered.derives P Q -> derives P Q. trivial. Qed.

Lemma Get_sub: TT |-- seplog.funspec_sub_si (snd get_smallspec) (snd get_largespec).
Proof. unfold get_smallspec, get_largespec. simpl. 
  simpl snd. apply derives_low; intros ? _.  split. split; trivial. simpl; intros.
  exists b; clear b. destruct H2. simpl in H2. destruct b1 as [gv args]. exists b0.
  destruct b0 as [s [l n]].
  destruct H3 as [PR [AR [LOC SE]]]. unfold argsassert2assert, SEPx in SE.
  simpl in SE.
  destruct SE as [se1 [se2 [JSE [SE1 SE2]]]].
  apply (Subject_GetPrePost SP) in SE2. destruct SE2. simpl in H3. red in H3.
  specialize (H3 (l, n) s). simpl in H3. eapply H3 in SE1; clear H3.
  3: apply ageable.necR_refl.
  2: apply age_sepalg.join_level in JSE; lia.
  rewrite sepcon_comm in SE1. destruct SE1 as [p [q [Jpq [P Q]]]].
  destruct (sepalg.join_assoc Jpq JSE) as [x [X1 X2]].
  exists (((G_Post (l,n) s -* S_rep (l, n) s))%logic)%pred; split.
  + exists p, x; split3; trivial.
    split3; trivial.
    split; trivial. unfold argsassert2assert, SEPx. simpl.
    exists q, se2; split3; trivial.
  + specialize (now_later (G_Post (l, n) s -* S_rep (l, n) s) p); intros X.
    apply X in P; clear X.
    intros. destruct H6 as [? [r [t [Jrt [R T]]]]].
    destruct T as [PRT [LOT SET]]. split3. apply PRT. apply LOT.
    unfold SEPx; simpl. unfold SEPx in SET; simpl in SET. rewrite sepcon_emp in *.
    clear - SET Jrt R. apply (wand_frame_elim'' (G_Post (l,n) s) (S_rep (l,n) s) a'1).
    exists r, t; split3; trivial.
Qed.

End SubjectASI.
