Require Import VST.floyd.proofauto.
Import FashNotation.
Require Import spec_stdlib. (*for malloc_token*)

Require Import SubjectObserverModel. (*the semantic constructions*)

Require Import Subject.
Notation SubjectTP := (Tstruct _subject noattr).
Notation ObserverTP:= (Tstruct _observer noattr).

Instance SCompSpecs : compspecs. make_compspecs prog. Defined.

Section SubjectAP.
Variable M: MallocFreeAPD.

(*Only needed internally*)
Definition NodeTP:type:= Tstruct _node noattr.

Fixpoint listrep (l: list val) s: mpred :=
  match l with
    nil => !!(s=nullval) && emp
  | o::os => EX n, !!(isptr o) &&
                   @data_at SCompSpecs Ews NodeTP (o,n) s * @malloc_token SCompSpecs M Ews NodeTP s *
                   listrep os n
  end.

Lemma listrep_ptrnull l s: listrep l s = !!(is_pointer_or_null s) && listrep l s.
Proof.
  destruct l; simpl; apply pred_ext; try solve [entailer!].
  Intros n. Exists n. entailer!. 
  apply andp_left2; trivial. 
Qed.

Lemma listrep_valid_pointer l s: listrep l s |-- valid_pointer s. 
Proof.
  destruct l; simpl. Intros; subst. entailer.
  Intros n. apply sepcon_valid_pointer1. apply sepcon_valid_pointer2.
  apply malloc_token_valid_pointer. 
Qed.

Lemma listrep_fold o n p os (H:isptr o): 
  (malloc_token M Ews NodeTP p * data_at Ews NodeTP (o, n) p * listrep os n |-- listrep (o :: os) p)%logic.
Proof. simpl. Exists n. entailer!. Qed. 

Lemma listrep_unfold o p os: listrep (o :: os) p |-- EX n, !!(isptr o) && malloc_token M Ews NodeTP p * data_at Ews NodeTP (o, n) p * listrep os n.
Proof. simpl; Intros n; Exists n; entailer!. Qed. 

Lemma listrep_nil s: listrep [] s = !!(s=nullval) && emp.
Proof. simpl; apply pred_ext; entailer!. Qed. 

Lemma listrep_null: listrep [] nullval = emp.
Proof. simpl; apply pred_ext; entailer!. Qed. 

Definition Subject_rep (S:SubjectRepTp) (s: val): mpred :=
  match S with
    (l,v) => EX o, listrep l o * malloc_token M Ews SubjectTP s *
             @data_at SCompSpecs Ews SubjectTP (o, Vint (Int.repr v)) s
  end.

Definition Subject_rep_char l v s: Subject_rep (l,v) s = 
  EX o, listrep l o * @malloc_token SCompSpecs M Ews SubjectTP s *
        @data_at SCompSpecs Ews SubjectTP (o, Vint (Int.repr v)) s.
Proof. simpl; trivial. Qed.

Lemma Subject_rep_newSubject s: (data_at Ews SubjectTP (Vint (Int.repr 0), Vint (Int.repr 0)) s * malloc_token M Ews SubjectTP s)%logic
      = Subject_rep newSubject s.
Proof.
  simpl. apply pred_ext. Exists nullval. entailer!. 
  Intros o. entailer!. 
Qed.

Lemma Subject_rep_ptrnull S s: Subject_rep S s |-- !!(is_pointer_or_null s).
Proof. destruct S as [l v]. unfold Subject_rep; Intros p; entailer!. Qed.

Opaque listrep.

Lemma Subject_rep_idem S s: Subject_rep S s |-- Subject_rep S s. trivial. Qed.

Definition GetPrePost (S:SubjectRepTp) s := 
  (field_at Ews SubjectTP [StructField _value] (Vint (Int.repr (snd S))) s * malloc_token M Ews SubjectTP s)%logic.

Lemma GetPrePost_char S s: GetPrePost S s = 
  (field_at Ews SubjectTP [StructField _value] (Vint (Int.repr (snd S))) s * malloc_token M Ews SubjectTP s)%logic.
Proof. trivial. Qed. 

Lemma GetPrePost_ptrnull S s: GetPrePost S s |-- !!(is_pointer_or_null s).
Proof. destruct S as [l v]; unfold GetPrePost; Intros; entailer!. Qed.

Lemma Subject_GetPrePost' S s: Subject_rep S s |--
                              (GetPrePost S s * (GetPrePost S s -* Subject_rep S s))%logic.
Proof.
  unfold Subject_rep, GetPrePost. destruct S. Intros o.
  unfold_data_at (data_at _ _ _ _); cancel.
  apply wand_frame_intro'. Exists o. unfold_data_at (data_at _ _ _ _); cancel.
Qed.

(*Lemma derives_low (P Q:mpred): predicates_hered.derives P Q -> derives P Q. trivial. Qed.*)

Lemma myderives_emp_unfash_fash P Q: derives P Q  -> derives emp (unfash (fash (imp P Q))).
Proof. apply derives_emp_unfash_fash.  Qed.
Lemma Subject_GetPrePost: 
  (emp |-- (!ALL S s, (Subject_rep S s >=> (GetPrePost S s * (*|> *)(GetPrePost S s -* Subject_rep S s)))) && emp)%logic.
Proof.
  unfold Subject_rep, GetPrePost. (*apply derives_low.
  apply predicates_hered.andp_right; trivial. *)
  apply andp_right; trivial. rewrite unfash_allp. apply allp_right. intros [l z].
  rewrite unfash_allp. apply allp_right. intros v.
  apply myderives_emp_unfash_fash. Intros o. cancel.
  unfold_data_at (data_at _ _ _ _); cancel.
  apply wand_frame_intro'. Exists o. unfold_data_at (data_at _ _ _ _); cancel.
Qed.

Opaque Subject_rep.
Opaque GetPrePost.

Definition SubjectAP :=
 Build_SubjectAPD _ (*Subject_rep_ptrnull*) (*SCompSpecs*)
                     Subject_rep Subject_rep
                     GetPrePost(*Subject_rep *)
                     GetPrePost(*Subject_rep *)
                     Subject_rep_idem Subject_rep_idem Subject_GetPrePost 
                     Subject_rep_ptrnull GetPrePost_ptrnull.

End SubjectAP.