Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*For MallocFreeAPD.*)
Require Import spec_Subject.
Require Import spec_Observer.
Require Import main.

(*We verify the body of main abstractly in M, OP and SP.
  We just provide the semax_body statement, no VSU.
  The spec of main is not VST standard spec, but a variant
  that can be instantiated to contain the appropriate initialization
  predidates, given suitable implementations (ie VSUs) of the code modules
  and the library.*)
Section MAIN_Verif.
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.
  Definition S_ASI:= SubjectASI M SP OP.
  Definition O_ASI:= ObserverASI M OP SP.

  (*Specification of "module-instantiated main" - we permit an arbitrary G here too.*)
  Definition main_inst_spec G:=
  DECLARE _main
  WITH gv: globals
  PRE [ ]
    PROP ()
    PARAMS () GLOBALS (gv)
    SEP (fold_right (fun g p => g gv * p) emp G; has_ext tt)
  POST[ tint ]
    PROP()
    LOCAL(temp ret_temp (Vint (Int.repr 6)))
    SEP((*mem_mgr M gv; *)TT;has_ext tt).

  Definition MainImports := [ newObserver_spec M OP SP; val_spec OP; 
   freeObserver_spec M OP; newSubject_spec M SP; update_spec SP OP;
   detachfirst_spec M SP; freeSubject_spec M SP].

  Definition MainInitPred: list (globals -> mpred) := nil.
  Definition maininstspec := (main_inst_spec [mem_mgr M]).

  (*Definition main_internal_specs: funspecs := [mainspec (*M*)].*)

  Definition MainVprog : varspecs. mk_varspecs main.prog. Defined.
  Definition MainGprog:funspecs := MainImports(* ++ main_internal_specs*).

  Instance MainCompSpecs : compspecs. make_compspecs main.prog. Defined.
  Lemma body_main: semax_body MainVprog MainGprog f_main maininstspec.
  Proof.
  start_function. simpl; Intros. 
  (*sep_apply (spec_stdlib.make_mem_mgr M gv). *)
  forward_call. Intros s. simpl. (*nothing happens as SD is abstract*)
  sep_apply Subject_Register. 
  forward_call (gv,s, newSubject). Intros o1.
  simpl. (*only the definitions of newSubject and Subjectrep (both model functions!) are simplified*) 
  sep_apply Subject_Update.
  sep_apply Observer_Notify.
  forward_call (s, 2, o1::fst newSubject, gv). 
  { Exists (snd newSubject). simpl. cancel.
    Exists [0].
    simpl. unfold newObserver; simpl. cancel. }

  simpl. (*again just the definition of newSubject is simplified*)
  sep_apply Subject_Register. 
  forward_call (gv,s, ([o1], 2)). Intros o2. simpl.

  sep_apply Subject_Update.
  sep_apply Observer_Notify.
  sep_apply Observer_Notify.
  forward_call (s, 3, [o2;o1], gv). 
  { Exists 2. cancel.
    Exists [2;2]. 
    simpl. unfold newObserver; simpl. cancel. }

  simpl. Intros.
  sep_apply Observer_Val.
  sep_apply Observer_Val.
  forward_call (o1,s,3).
  forward_call (o2,s,3).
  forward.
  forward_call. Intros p1. simpl. if_tac; Intros z. inv H0; deadvars.
  forward_call. Intros p2. simpl. if_tac; Intros z. inv H1; deadvars.
  forward_call; Intros v. simpl. if_tac; Intros; [ subst; simpl | congruence].
  subst; simpl. 
  forward_call (gv, p1, (s,3)).
  forward_call (gv, p2, (s,3)).
  forward.
  cancel.
  Qed.

End MAIN_Verif.