Require Export VST.floyd.proofauto.
Require Export SubjectObserverModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Observer. (*needed for function and type names *)

Global Open Scope funspec_scope.
Definition ObserverTP:type:=  Tstruct _observer noattr.
Definition SubjectTP:type:=  Tstruct _subject noattr.

Section ObserverASI.
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.

  Notation S_rep := (Subjectrep SP).
  Notation R_Pre := (RegisterPre SP). (*
  Notation U_Pre := (UpdatePre SP).*)
  Notation G_Pre := (GetPre SP).
  Notation G_Post := (GetPost SP).
  Notation O_rep := (Observerrep OP).
  Notation N_Pre:= (NotifyPre OP).
  Notation V_Pre:= (ValPre OP).

  Definition newObserver_spec :=
    DECLARE _newObserver
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
      PROP()
      PARAMS (s) GLOBALS (gv)
      SEP(R_Pre S s; mem_mgr M gv)
    POST[ tptr ObserverTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(O_rep (newObserver S s) p;
           S_rep (p::fst S, snd S) s; mem_mgr M gv).

  (*ie the heart of the concrete get_smallspec (later, will generalize of Gpre/Gpost, maybe*)
  Definition getspec :=
    WITH s:val, S: SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(G_Pre S s)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr (snd S))))
      SEP(G_Post S s).

  Definition notify_smallspec :=
    DECLARE _notify
    WITH o:val, s:val, c:Z, S:SubjectRepTp, fp:val
    PRE [ tptr ObserverTP,
          tptr (Tfunction (Tcons (tptr SubjectTP) Tnil) tint cc_default) ]
     PROP()
     PARAMS (o;fp) GLOBALS ()
     SEP(N_Pre (s,c) o; G_Pre S s; func_ptr' getspec fp)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(O_rep (s, snd S) o; G_Post S s).

  Definition val_spec :=
    DECLARE _val
    WITH o:val, s:val, c: Z
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS ()
     SEP(V_Pre (s,c) o)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr c)))
      SEP(O_rep (s,c) o).

  Definition freeObserver_spec:=
    DECLARE _freeObserver 
    WITH gv: globals, o:val, O:ObserverRepTp
    PRE [ tptr ObserverTP ]
     PROP()
     PARAMS (o) GLOBALS (gv)
     SEP(O_rep O o; mem_mgr M gv)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(mem_mgr M gv). 

  Definition ObserverASI:funspecs := [ newObserver_spec; notify_smallspec; val_spec; freeObserver_spec].

  Definition notify_largespec :=
    DECLARE _notify
    WITH o:val, s:val, c:Z, S:SubjectRepTp, fp:val
    PRE [ tptr ObserverTP,
          tptr (Tfunction (Tcons (tptr SubjectTP) Tnil) tint cc_default)  ]
     PROP()
     PARAMS (o; fp) GLOBALS ()
     SEP(N_Pre (s,c) o; S_rep S s; func_ptr' getspec fp)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(O_rep (s, snd S) o; S_rep S s).

Lemma Notify_sub: funspec_sub (snd notify_smallspec) (snd notify_largespec).
Proof.
  do_funspec_sub. clear H.
  destruct w as [[[[o s] c] S] fp]. simpl; entailer.
  Exists ((((o, s), c), S), fp) (G_Post S s -* S_rep S s).
  simpl; entailer!.
  { intros. cancel. apply wand_frame_elim''. }
  rewrite sepcon_comm. apply Subject_GetPrePost.
Qed.
End ObserverASI.