Require Import spec_stdlib. (*for malloc_token*)
Require Import Observer.
Require Import spec_Observer.

Instance ObserverCompSpecs : compspecs. make_compspecs prog. Defined.

Section ObserverAP.
Variable M: MallocFreeAPD.

Definition Observer_rep (O:ObserverRepTp) (o: val): mpred :=
  (match O with
    (s,c) => @malloc_token ObserverCompSpecs M Ews ObserverTP o *
    @data_at ObserverCompSpecs Ews ObserverTP (s, Vint (Int.repr c)) o
  end)%logic.

Definition Observer_rep_char O o: Observer_rep O o = 
    (malloc_token M Ews ObserverTP o *
     @data_at ObserverCompSpecs Ews ObserverTP (fst O, Vint (Int.repr (snd O))) o)%logic.
Proof. destruct O; reflexivity. Qed.

Definition NotifyPre (O:ObserverRepTp) (o:val): mpred :=
  (malloc_token M Ews ObserverTP o *
   EX v, @data_at ObserverCompSpecs Ews ObserverTP (fst O, v) o)%logic.

(*We instantiate ValPre by Observer_rep*)

Lemma Observer_NotifyPre O o: Observer_rep O o |-- NotifyPre O o.
Proof.
  unfold Observer_rep, NotifyPre; destruct O as [s v]; simpl. Exists (Vint (Int.repr v)); trivial.
Qed.

Lemma NotifyPre_char O o:  NotifyPre O o = (malloc_token M Ews ObserverTP o * EX v, @data_at ObserverCompSpecs Ews ObserverTP (fst O, v) o)%logic. 
Proof. reflexivity. Qed.

Lemma Observer_rep_idem O o: Observer_rep O o |-- Observer_rep O o. Proof. trivial. Qed. 
  
Lemma NotifyPre_isptr O o: NotifyPre O o |-- !!(isptr o).
Proof. unfold NotifyPre; Intros v; entailer!. Qed.

Opaque Observer_rep.
Opaque NotifyPre.
(*
Lemma Notify_sub SR GPre GPost (H: forall S s, SR S s |-- GPre S s * (GPost S s -* SR S s)):
  funspec_sub (snd (notify_smallspec GPre GPost Observer_rep NotifyPre))
              (snd (notify_largespec SR Observer_rep NotifyPre)).
Proof.
  apply NDsubsume_subsume.
  { simpl; intros; split; trivial. }
  split3; trivial. intros [[[o s] c] S]. simpl; intros rho. 
  apply andp_left2. specialize (H S s); rewrite sepcon_comm in H. 
  Exists (((o, s), c), S) (GPost S s -* SR S s).
  unfold PROPx, LOCALx, SEPx; simpl. unfold local, liftx, lift1, lift; simpl.
  entailer!; [intros | trivial].
  cancel. rewrite sepcon_comm. apply wand_frame_elim. 
Qed.*)

Definition ObserverAP:=
  Build_ObserverAPD _ (*Observer_rep_isptr*) _ NotifyPre_isptr Observer_rep
                       Observer_NotifyPre Observer_rep_idem.
End ObserverAP.