Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import VST.veric.initial_world.
Require Import VST.floyd.linking.
Require Import main.
Require Import VSU_Application. (*Maybe instead Require Export it in AppLib*)
Require Import VSU_AppLib.

Definition linked_prog': Clight.program := 
  match link_progs applib main.prog with
    Errors.OK p => p
  | _ => applib (*arbitrary*)
  end.
Definition linked_prog := ltac:
  (let x := eval hnf in linked_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Definition ExtIDtable : PTree.t unit :=
  ltac:(let x := constr:(fold_right (fun i t => PTree.set i tt t) (PTree.empty _) (ExtIDs linked_prog))
           in let x := eval compute in x
           in exact x).

Definition in_ExtIDs (ia: ident*funspec) : bool :=
 match PTree.get (fst ia) ExtIDtable with Some _ => true | None => false end.

Definition LinkedSYS_pre:funspecs :=
   filter in_ExtIDs (augment_funspecs linked_prog ApplibSys).

Definition LinkedSYS:funspecs := 
  ltac: 
    (let x := eval hnf in LinkedSYS_pre in
     let x := eval simpl in x in 
     (*let x := eval compute in x in leaving this in leads Coq to timeout/crash*)
       exact x).

Lemma ApplibSys_in_LinkedSys: forall i phi, find_id i ApplibSys = Some phi -> find_id i LinkedSYS = Some phi.
  Proof. (*Boilerplate proof*)
    intros. specialize (find_id_In_map_fst _ _ _ H); intros.
    simpl in H0. repeat (destruct H0 as [HO | H0]; [ subst i; inv H; reflexivity |]). contradiction.
  Qed. 

Lemma LinkedSYS_External: forall i, In i (map fst LinkedSYS) ->
      exists ef ts t cc, find_id i (prog_defs linked_prog) = Some (Gfun (External ef ts t cc)).
  Proof. (*Boilerplate proof*)
    intros. cbv in H. 
    repeat (destruct H as [H | H];
      [ subst; try solve [do 4 eexists; split; reflexivity ]
      | ]).
    contradiction.
  Qed.

Lemma LinkedSYS_vacuous i phi: find_id i LinkedSYS = Some phi -> find_id i ApplibSys = None ->
        exists ef argsig retsig cc, 
           phi = vacuous_funspec (External ef argsig retsig cc) /\ 
           find_id i (prog_funct applib) = Some (External ef argsig retsig cc) /\
           ef_sig ef = {| sig_args := typlist_of_typelist argsig;
                          sig_res := rettype_of_type retsig;
                          sig_cc := cc_of_fundef (External ef argsig retsig cc) |}.
  Proof. (*Boilerplate proof*)
    intros. unfold find_id in H. simpl in H. 
    (*Time*) repeat (if_tac in H;
    [ inv H; try inv H0; try solve [do 4 eexists; split3; reflexivity]
    | clear H1 ]); discriminate. 
(*Old proof
    intros. specialize (find_id_In_map_fst _ _ _ H); intros.
    cbv in H1.
    Time repeat (destruct H1 as [H1 | H1]; 
      [ subst; inv H; try solve [do 4 eexists; split3; reflexivity]
      | ]); try inv H0; try contradiction. (*3s*)*)
  Qed.

Definition LinkedVprog : varspecs. mk_varspecs linked_prog. Defined.

Lemma disjoint_Vprog_linkedfuncts: 
      list_disjoint (map fst LinkedVprog) (map fst (prog_funct linked_prog)).
Proof. (*Boilerplate proof*)
  intros x y X Y ?; subst x; cbv in X; apply assoclists.find_id_None_iff in Y; [ trivial | clear H Y];
  repeat (destruct X as [X | X]; [ subst y; cbv; reflexivity |]); contradiction.
Qed.

Instance LinkedCompSpecs : compspecs. make_compspecs linked_prog. Defined.

Require Import verif_main.

Lemma CSSUB: cspecs_sub MainCompSpecs LinkedCompSpecs.
Proof. (*Boilerplate proof*)
  split3.
+ intros i; red; remember ((@cenv_cs MainCompSpecs) ! i) as w; destruct w; 
   [symmetry in Heqw; simpl in Heqw; rewrite PTree.gleaf in Heqw; congruence | trivial].
+ intros i. red. remember ((@ha_env_cs MainCompSpecs) ! i) as w. destruct w; [symmetry in Heqw | trivial].
  simpl in Heqw. rewrite PTree.gleaf in Heqw. congruence.
+ intros i. red. remember ((@la_env_cs MainCompSpecs) ! i) as w. destruct w; [symmetry in Heqw | trivial].
  simpl in Heqw. rewrite PTree.gleaf in Heqw. congruence.
Qed.

Require Import spec_main.
Notation mainspec:= (main_spec linked_prog). (*VST's standard spec for main*)

Notation M:= verif_stdlib.M. (* The concrete MallocFreeADP*)

Require Import ASI_malloc. (*for mem_mgr*) 
Lemma main_sub: funspec_sub (snd (main_inst_spec [mem_mgr M])) (snd (main_spec linked_prog)).
Proof. (*boilerplate*)
  do_funspec_sub. unfold main_pre; simpl; Intros; subst. clear. 
  Exists (initialize.genviron2globals g) emp. entailer!.
  + intros; entailer!.
  + specialize (globals_ok_genviron2globals g); intros G.
    Exists emptyResvec. eapply derives_trans. 2: apply VSU_malloc.create_memmgr_R. 
    unfold InitGPred. simpl. entailer. clear.
    unfold globvar2pred; simpl.
    sep_apply mapsto_zeros_isptr; entailer!.
Qed.

Lemma tc_VG: tycontext_subVG MainVprog (MainGprog M (Observer M) (Subject M)) 
                             LinkedVprog (mainspec::ApplibExports).
Proof. (*Boilerplate proof*)
  split; intros i; red.
  * rewrite 2 semax_prog.make_context_g_char, 2 make_tycontext_s_find_id by LNR_tac.
    remember (find_id i (MainGprog M (Observer M) (Subject M))) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; reflexivity |]).
    congruence.
  * rewrite 2 make_tycontext_s_find_id.
    remember (find_id i (MainGprog M (Observer M) (Subject M))) as w.
    destruct w; [symmetry in Heqw | trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.

Require Import VST.floyd.VSU_addmain.

Definition SO_VSU: @LinkedProgVSU NullExtension.Espec LinkedVprog LinkedCompSpecs
      LinkedSYS nil linked_prog [mainspec] ApplibInitPred.
Proof.
 VSUAddMain_tac Applib_CanVSU.
   + apply disjoint_Vprog_linkedfuncts.
   + apply LinkedSYS_External. 
   + eapply semax_body_subsumption.
     * eapply semax_body_funspec_sub.
       - apply (@semax_body_cenv_sub MainCompSpecs _ CSSUB).
         2: apply (body_main M (Observer M) (Subject M)).
         repeat constructor; eauto.
       - apply main_sub. 
       - LNR_tac.
     * apply tycontext_sub_i99. apply tc_VG.
   + apply LinkedSYS_vacuous.
Qed.

Lemma prog_correct:
  exists G, 
 @semax_prog NullExtension.Espec LinkedCompSpecs linked_prog tt LinkedVprog G.
Proof.
(*boilerplate: same proof script as in VSUpile, connpool*)
  destruct SO_VSU as [G Comp MAIN]. exists G. 
  assert (DomG: map fst G = map fst (prog_funct linked_prog)).
  { destruct Comp. unfold Comp_G in *. rewrite CC_canonical.
    cbv; reflexivity. }
  Time prove_linked_semax_prog. (*7s*)
  all: rewrite augment_funspecs_eq by trivial.
  apply (@Canonical_semax_func _ _ _ _ _ _ _ _ Comp); cbv; reflexivity.
  destruct MAIN as [post [MainG MainExp]]. inv MainExp. rewrite MainG; eexists; reflexivity.
Qed.

(*Print Assumptions prog_correct.*)