Require Export VST.floyd.proofauto.
Require Export SubjectObserverModel. (*the semantic constructions*)

Require Import spec_stdlib. (*needed for mem_mgr*)
Require Import Subject. (*needed for function and type names *)

Global Open Scope funspec_scope.
Definition SubjectTP:type:=  Tstruct _subject noattr.
Definition ObserverTP:type:=  Tstruct _observer noattr.

Fixpoint Observers (OPred:ObserverRepTp -> val -> mpred) s (vals: list Z) (l:list val) :=
    match vals, l with
      nil, nil => emp
    | v::vs, o::os => (OPred (s,v) o * Observers OPred s vs os)%logic
    | _, _ => FF
    end.

Lemma Observers_app OPred s: forall vals1 l1 vals2 l2 (L: length vals1 = length l1),
      Observers OPred s (vals1 ++ vals2) (l1++l2) = 
      (Observers OPred s vals1 l1 * Observers OPred s vals2 l2)%logic.
  Proof.
    induction vals1; intros; destruct l1; simpl; inv L.
    + rewrite emp_sepcon; trivial.
    + rewrite IHvals1; [ rewrite sepcon_assoc |]; trivial.
  Qed.

Section SubjectASI.
  Variable M: MallocFreeAPD.
  Variable SP: SubjectAPD.
  Variable OP: ObserverAPD.

  Notation S_rep := (Subjectrep SP).
  Notation R_Pre := (RegisterPre SP).
  Notation U_Pre := (UpdatePre SP).
  Notation G_Pre := (GetPre SP).
  Notation G_Post := (GetPost SP).
  Notation O_rep := (Observerrep OP).
  Notation N_Pre:= (NotifyPre OP).

  Definition newSubject_spec :=
    DECLARE _newSubject
    WITH gv: globals
    PRE [ ] PROP() PARAMS () GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr SubjectTP ]
    EX p: val,
     PROP()
     LOCAL(temp ret_temp p)
     SEP(S_rep newSubject p; mem_mgr M gv).

  Definition get_smallspec :=
    DECLARE _get
    WITH s:val, S: SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(G_Pre S s)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr (snd S))))
      SEP(G_Post S s).

  Definition update_spec :=
    DECLARE _update
    WITH s:val, n: Z, l:list val
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(EX z, U_Pre (l,z) s; EX vals, Observers N_Pre s vals l)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (l,n) s; Observers O_rep s (list_repeat (length l) n) l).

  Definition register_spec :=
    DECLARE _registr
    WITH s:val, l:list val, n: Z, o:val, gv:globals
    PRE [ tptr SubjectTP, tptr ObserverTP  ]
     PROP ()
     PARAMS (s; o) GLOBALS (gv)
     SEP (R_Pre (l,n) s; mem_mgr M gv; EX v, N_Pre (s,v) o)
    POST[ tvoid ]
     PROP ()
     LOCAL ()
     SEP (S_rep (o::l, n) s; mem_mgr M gv; O_rep (s,n) o).

  Definition freeSubject_spec:=
    DECLARE _freeSubject 
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv)
     SEP(S_rep S s; mem_mgr M gv)
    POST[ tint ] EX r,
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr r)))
      SEP(if zeq r 0 then !!(fst S = @nil val)&&emp 
          else !!(r=1 /\ fst S <> @nil val) && S_rep S s; mem_mgr M gv).

  Definition detachfirst_spec:=
    DECLARE _detachfirst 
    WITH gv: globals, s:val, S:SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS (gv)
     SEP(S_rep S s; mem_mgr M gv)
    POST[ tptr ObserverTP ] EX r,
      PROP()
      LOCAL(temp ret_temp r)
      SEP(if Val.eq r nullval 
          then !!(fst S = @nil val) && S_rep S s 
          else EX O, !!(fst S=r::O) && S_rep (O, snd S) s; mem_mgr M gv).

  Definition SubjectASI:funspecs := [ detachfirst_spec; freeSubject_spec; newSubject_spec; register_spec; update_spec; get_smallspec].

  Definition get_largespec :=
    DECLARE _get
    WITH s:val, S: SubjectRepTp
    PRE [ tptr SubjectTP ]
     PROP()
     PARAMS (s) GLOBALS ()
     SEP(S_rep S s)
    POST[ tint ]
      PROP()
      LOCAL(temp ret_temp (Vint (Int.repr (snd S))))
      SEP(S_rep S s).

Lemma Get_sub: funspec_sub (snd get_smallspec) (snd get_largespec).
Proof.
  do_funspec_sub. destruct w as [s [l n]]; clear H.
  Exists (s, (l, n)).
  Exists (GetPost SP (l, n) s -* Subjectrep SP (l, n) s).
  simpl; entailer!.
  { intros. rewrite <- H0. entailer!.
    apply wand_frame_elim''. }
  rewrite sepcon_comm. apply (Subject_GetPrePost SP (l, n)).
Qed.

Definition update_Obs_spec :=
    DECLARE _update
    WITH s:val, n: Z, l:list val
    PRE [ tptr SubjectTP, tint ]
     PROP()
     PARAMS (s; Vint (Int.repr n)) GLOBALS ()
     SEP(EX z, U_Pre (l,z) s; EX vals, Observers O_rep s vals l)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(S_rep (l,n) s; Observers O_rep s (list_repeat (length l) n) l).

Lemma Upd_sub: funspec_sub (snd update_spec) (snd update_Obs_spec).
Proof.
  do_funspec_sub. destruct w as [[s n] l]; clear H.
  Exists ((s, n), l) emp. simpl. entailer!. apply exp_derives; intros vs.
  generalize dependent l; clear. induction vs; simpl; intros.
  + destruct l; trivial.
  + destruct l; trivial. sep_apply Observer_Notify; cancel.
Qed.
End SubjectASI.
