Require Export VST.floyd.proofauto.

Inductive tree :=
  Tnil: tree
| Tnode: val -> tree -> tree -> tree.

(*No need to expose this - it's also commented as being private in the Verifast development
Inductive context :=
  Root: context
| lcontext: val -> context -> tree -> context
| rcontext: val -> tree -> context -> context.
*)
Fixpoint contains (value:tree) (n:val) :bool :=
  match value with
   Tnil => false
  | Tnode r lhs rhs => (Val.eq n r) || (contains lhs n) || (contains rhs n)
  end.

Record CompositeAPD := {
  Tree: tree -> val -> mpred; (*We swap the order of arguments as Coq's Program fixpoint otherwise will lead to a much more complex definition*)
  (*Keep private Context: context -> val -> Z -> mpred; (*Same here - Jacobs uses val -> context Z -> mpred*)*)
  isTree: val -> tree-> mpred;
  rotate: forall r value n, !!(contains value n = true) && isTree r value|-- isTree n value
}.

Fixpoint replace (ts:tree) (n:val) (plug:tree): tree:=
  match ts with
    Tnil => Tnil
  | Tnode r lhs rhs => if Val.eq r n then plug else
                       (if contains lhs n then Tnode r (replace lhs n plug) rhs else Tnode r lhs (replace rhs n plug))
  end.

Fixpoint size(t:tree) :Z :=
  match t with
    Tnil => 0
  | Tnode node lhs rhs => 1 + size(lhs) + size(rhs)
  end.

Fixpoint disjoint (lhs rhs: tree): bool:=
  match lhs with
    Tnil => true
  | Tnode r leftT rightT => (negb(contains rhs r)) && disjoint leftT rhs && disjoint rightT rhs
  end.

Fixpoint uniqueNodes(t:tree):bool :=
  match t with
    Tnil => true
  | Tnode r leftT rightT => (negb(contains leftT r)) && (negb (contains rightT r))
                          && disjoint leftT rightT && uniqueNodes leftT && uniqueNodes rightT 
   end.

Fixpoint valueOf (value:tree) (n:val):tree :=
  match value with
    Tnil => Tnil
  | Tnode r lhs rhs => 
      if Val.eq n r then Tnode r lhs rhs else (if contains lhs n then valueOf lhs n else valueOf rhs n)
  end.