Require Import VST.floyd.VSU.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import Composite. (*needed for defining ConnCompSpecs and Vprog *)
Require Import Predicates.
Require Import ASI_Composite.
(*https://github.com/verifast/verifast/blob/master/examples/Composite.c#L60*)
Instance CompSpecs : compspecs. make_compspecs prog. Defined.

Inductive context :=
  Root: context
| lcontext: val -> context -> tree -> context
| rcontext: val -> tree -> context -> context.

Fixpoint TreeHeight (t:tree):nat :=
  match t with Tnil => O | Tnode p l r => 1 + max (TreeHeight l) (TreeHeight r) end.

Fixpoint ContextHeight (c:context):nat :=
  match c with Root => O | lcontext x cc t => 1 + (ContextHeight cc) | rcontext x t cc => 1 + ContextHeight cc end.

Section MyComposite_APD.
Variable M: MallocFreeAPD.

Fixpoint MyTree (value: tree)(node: val) : mpred :=
 match value with
  | Tnil => FF
  | Tnode node2 lhs rhs => EX l r, !!(isptr node /\ node=node2 /\ size value <= Int.max_signed) &&
       field_at Ews nodeTP [StructField _left] l node *
       field_at Ews nodeTP [StructField _right] r node *
       (if Val.eq l nullval then !!(lhs=Tnil) && emp else (MyTree lhs l * field_at Ews nodeTP [StructField _parent] node l)) *
       (if Val.eq r nullval then !!(rhs=Tnil) && emp else (MyTree rhs r * field_at Ews nodeTP [StructField _parent] node r)) *
       field_at Ews nodeTP [StructField _count]  (Vint (Int.repr (size value))) node *
       malloc_token M Ews nodeTP node 
  end.
Lemma MyTree_Fold : forall value node, MyTree value node =
 match value with
  | Tnil => FF
  | Tnode node2 lhs rhs => EX l r,
      !!(isptr node /\ node=node2 /\ size value <= Int.max_signed) &&
       field_at Ews nodeTP [StructField _left] l node *
       field_at Ews nodeTP [StructField _right] r node *
       (if Val.eq l nullval then !!(lhs=Tnil) && emp else (MyTree lhs l * field_at Ews nodeTP [StructField _parent] node l)) *
       (if Val.eq r nullval then !!(rhs=Tnil) && emp else (MyTree rhs r * field_at Ews nodeTP [StructField _parent] node r)) *
       field_at Ews nodeTP [StructField _count]  (Vint (Int.repr (size value))) node *
       malloc_token M Ews nodeTP node 
  end.
Proof. induction value; intros; simpl; trivial. Qed.

Lemma MyTree_isptr value node: MyTree value node |-- !!(isptr node).
Proof. destruct value; simpl. entailer!. Intros l r; entailer!. Qed.

  Definition MyIsTree (n: val) (value: tree) :=
    EX root, !!(root <> nullval /\ contains value n=true) && 
             MyTree value root * 
             field_at Ews nodeTP [StructField _parent] nullval root.

  Fixpoint MyContext (value:context)(node:val)  (holeCount: Z) (*{measure (ContextHeight value)}*): mpred :=
   match value with
     Root => field_at Ews nodeTP [StructField _parent] nullval node
   | lcontext n cont t => EX r c, !!(n<>nullval /\ node<>nullval /\ c= holeCount + 1 + size(t)
                                     /\ 0 <= c <= Int.max_signed) &&
             field_at Ews nodeTP [StructField _left] node n *
             field_at Ews nodeTP [StructField _right] r n *
             field_at Ews nodeTP [StructField _count] (Vint (Int.repr c)) n *
             (if Val.eq r nullval then !!(t=Tnil) && emp else (MyTree t r * field_at Ews nodeTP [StructField _parent] n r)) *
             MyContext cont n c * field_at Ews nodeTP [StructField _parent] n node * malloc_token M Ews nodeTP n
   | rcontext n t cont => EX l c, !!(n<>nullval /\ node <> nullval /\ c= holeCount + 1 + size(t)
                                     /\ 0 <= c <= Int.max_signed) &&
                          field_at Ews nodeTP [StructField _right] node n  *
                          field_at Ews nodeTP [StructField _left] l n *
                          field_at Ews nodeTP [StructField _count] (Vint (Int.repr c)) n *
                          (if Val.eq l nullval then !!(t=Tnil) && emp else (MyTree t l * field_at Ews nodeTP [StructField _parent] n l)) * 
                          MyContext cont n c * field_at Ews nodeTP [StructField _parent] n node * malloc_token M Ews nodeTP n
  end.

Lemma MyContext_Fold: forall value node holeCount,
   MyContext value node holeCount =
   match value with
     Root => field_at Ews nodeTP [StructField _parent] nullval node
   | lcontext n cont t => EX r c, !!(n<>nullval /\ node<>nullval /\ c= holeCount + 1 + size(t)
                                     /\ 0 <= c <= Int.max_signed) &&
             field_at Ews nodeTP [StructField _left] node n *
             field_at Ews nodeTP [StructField _right] r n *
             field_at Ews nodeTP [StructField _count] (Vint (Int.repr c)) n *
             (if Val.eq r nullval then !!(t=Tnil) && emp else (MyTree t r * field_at Ews nodeTP [StructField _parent] n r)) *
             MyContext cont n c * field_at Ews nodeTP [StructField _parent] n node * malloc_token M Ews nodeTP n
   | rcontext n t cont => EX l c, !!(n<>nullval /\ node <> nullval /\ c= holeCount + 1 + size(t)
                                     /\ 0 <= c <= Int.max_signed) &&
                          field_at Ews nodeTP [StructField _right] node n  *
                          field_at Ews nodeTP [StructField _left] l n *
                          field_at Ews nodeTP [StructField _count] (Vint (Int.repr c)) n *
                          (if Val.eq l nullval then !!(t=Tnil) && emp else (MyTree t l * field_at Ews nodeTP [StructField _parent] n l)) * 
                          MyContext cont n c * field_at Ews nodeTP [StructField _parent] n node * malloc_token M Ews nodeTP n
  end.
Proof. induction value; intros; simpl; trivial. Qed.

  Lemma Myrotate: forall r value n, !!(contains value n = true) && MyIsTree r value|-- MyIsTree n value.
  Proof. intros. unfold MyIsTree. Intros root. Exists root. entailer!. Qed.

  Definition MyCompositeAPD : CompositeAPD :=
   Build_CompositeAPD MyTree (*MyContext*) MyIsTree Myrotate.

Lemma nonidentityEws: sepalg.nonidentity Ews.
  Proof. apply readable_nonidentity. apply writable_readable. apply writable_Ews. Qed.

  Lemma size_nonneg: forall (t : tree), 0<=size t.
   Proof. induction t; simpl; lia. Qed.

  Lemma treeDistinct' node l: forall t root,
    MyTree t root * field_at Ews nodeTP [StructField _left] l node |-- !!(contains t node = false).
  Proof. simpl.
  induction t. { intros; entailer!. } rename v into r; rename t1 into lhs; rename t2 into rhs.
  simpl; intros; Intros ll rr. subst. 
  destruct (Val.eq node r); simpl.
  + subst; simpl. sep_apply field_at_conflict. apply nonidentityEws. entailer!.
  + destruct (Val.eq ll nullval); simpl.
    - subst. simpl. 
      destruct (Val.eq rr nullval); simpl.
      * Intros; subst. simpl; entailer!.
      * sep_apply (IHt2 rr); clear IHt2; Intros. rewrite H. entailer!.
    - sep_apply (IHt1 ll); clear IHt1; Intros. rewrite H; simpl.
      destruct (Val.eq rr nullval); simpl.
      * Intros; subst. simpl; entailer!.
      * sep_apply (IHt2 rr); clear IHt2; Intros. entailer!.
  Qed.

  Lemma treeDistinct t root node l:
  MyTree t root * field_at Ews nodeTP [StructField _left] l node |-- 
  !!(contains t node = false) &&
     MyTree t root * field_at Ews nodeTP [StructField _left] l node.
  Proof. entailer. apply treeDistinct'. Qed. 

  Lemma treesDisjoint': forall t1 t2 root1 root2,
    MyTree t1 root1 * MyTree t2 root2 |-- !!(disjoint t1 t2 = true).
  Proof. induction t1. { simpl; intros; entailer!. }
     rename v into r; rename t1_1 into lhs; rename t1_2 into rhs.
     intros. simpl. Intros ll rr; subst.
     sep_apply (treeDistinct' r ll t2); Intros. rewrite H; clear H; simpl.
     destruct (Val.eq ll nullval); Intros; subst.
     + simpl in *. clear IHt1_1.
       destruct (Val.eq rr nullval); Intros; subst.
       - simpl; entailer!.
       - sep_apply (IHt1_2 t2 rr); clear IHt1_2; Intros. entailer!.
     + destruct (Val.eq rr nullval); Intros; subst.
       - simpl in *. clear IHt1_2; rewrite andb_true_r.
         sep_apply IHt1_1; clear IHt1_1; Intros. entailer!.
       - sep_apply (IHt1_1 t2); clear IHt1_1; Intros. rewrite H; clear H; simpl.
         sep_apply (IHt1_2 t2); clear IHt1_2; Intros. entailer!.
  Qed. 
  Lemma treesDisjoint t1 t2 root1 root2:
    MyTree t1 root1 * MyTree t2 root2 |-- 
    !!(disjoint t1 t2 = true) && MyTree t1 root1 * MyTree t2 root2.
  Proof. entailer. apply treesDisjoint'. Qed.

  Lemma disjoint_Tnil: forall t, disjoint t Tnil = true.
  Proof. induction t; simpl; trivial. rewrite IHt1, IHt2; trivial. Qed.
 (* not neededLemma disjoint_symm: forall t s, disjoint t s = disjoint s t.
  Proof. induction t; simpl; trivial; intros.
   + rewrite disjoint_Tnil; trivial.
   + rewrite ! IHt1. rewrite ! IHt2. destruct s; simpl. trivial.
     destruct (contains t1 v0); simpl.  rewrite ! disjoint_Tnil; trivial.
     rewrite   rewrite IHt1, IHt2; trivial. Qed*)

  Lemma treeUniqueNodes': forall t root,
    MyTree t root |-- !!(uniqueNodes t = true).
  Proof. induction t; intros. simpl; entailer!.
    simpl; Intros ll rr; subst.
    destruct (Val.eq ll nullval); Intros; subst.
    + clear IHt1. simpl. rewrite 2 andb_true_r.
      destruct (Val.eq rr nullval); Intros; subst. entailer!.
      sep_apply (treeDistinct' v nullval t2); Intros. rewrite H. simpl.
      sep_apply IHt2; Intros. entailer!.
    + sep_apply IHt1; clear IHt1; Intros. rewrite H, andb_true_r; clear H.
      destruct (Val.eq rr nullval); Intros; subst.
      - simpl. clear IHt2. rewrite ! andb_true_r. 
        sep_apply (treeDistinct' v ll); Intros. rewrite H, disjoint_Tnil; simpl.
        entailer!.
      - sep_apply IHt2; clear IHt2; Intros. rewrite H, andb_true_r. 
        sep_apply (treesDisjoint' t1 t2); Intros. (*rewrite H0, andb_true_r. *)
        sep_apply (treeDistinct' v ll t1); Intros. rewrite H1; simpl.
        sep_apply (treeDistinct' v ll t2); Intros. rewrite H2; simpl; entailer!.
        rewrite andb_true_r. apply negb_true_iff; trivial.
  Qed.
  Lemma treeUniqueNodes t root:
    MyTree t root |-- !!(uniqueNodes t = true) && MyTree t root.
  Proof. entailer. apply treeUniqueNodes'. Qed. 

 Fixpoint inContext (con:context) (n:val):bool :=
   match con with
     Root => false
   | lcontext r lcon rhs => Val.eq r n || inContext lcon n || contains rhs n
   | rcontext r lhs rcon => Val.eq r n || contains lhs n || inContext rcon n
   end.

  Lemma contextDistinct': forall con count c n x,
    MyContext con c count * field_at Ews nodeTP [StructField _left] x n |-- 
    !!(inContext con n = false). 
  Proof. induction con; simpl; intros. entailer!; simpl in IHcon.
   + Intros rr d. destruct (Val.eq v n); subst; simpl.
     - sep_apply field_at_conflict. apply nonidentityEws. entailer!.
     - destruct (Val.eq rr nullval); Intros; subst; simpl in *.
       * sep_apply (IHcon (count + 1 + 0) v n); Intros; rewrite H1; entailer!.
       * sep_apply (IHcon (count + 1 + size t) v n); Intros; rewrite H1; simpl. 
         sep_apply (treeDistinct' n x t); Intros. entailer!. (*rewrite H2 / H3; entailer!*)
   + Intros ll d. destruct (Val.eq v n); subst; simpl.
     - sep_apply (field_at_conflict Ews nodeTP [StructField _left]). apply nonidentityEws. entailer!. 
     - destruct (Val.eq ll nullval); Intros; subst; simpl in *.
       * sep_apply (IHcon (count + 1 + 0) v n); Intros. entailer!.
       * sep_apply (IHcon (count + 1 + size t) v n); Intros. rewrite H1, orb_false_r.
         sep_apply (treeDistinct' n x t); Intros. entailer!. (* rewrite H2; entailer!.*)
  Qed.
  Lemma contextDistinct con count c n x:
    MyContext con c count * field_at Ews nodeTP [StructField _left] x n |-- 
    !!(inContext con n = false) && MyContext con c count * field_at Ews nodeTP [StructField _left] x n.
  Proof. entailer. apply contextDistinct'. Qed.

  Fixpoint disjointContext (con: context) (t:tree):bool :=
    match t with
      Tnil => true
    | Tnode v ltree rtree => (negb (inContext con v)) && disjointContext con ltree && disjointContext con rtree
    end.

  Lemma treeContextDisjoint': forall t con cc n c,
     MyTree t n * MyContext con c cc |-- !!(disjointContext con t = true).
  Proof.
    induction t; simpl; intros. entailer!.
    Intros ll rr. subst. rename t1 into ltree; rename t2 into rtree.
    sep_apply contextDistinct; Intros. rename H into IC. rewrite IC; simpl.
    destruct (Val.eq ll nullval); Intros; subst; simpl.
    + destruct (Val.eq rr nullval); Intros; subst; simpl. entailer!.
      sep_apply IHt2; entailer!.
    + sep_apply IHt1; Intros. rename H into DCleft; rewrite DCleft; simpl.
      destruct (Val.eq rr nullval); Intros; subst; simpl. entailer!.
      sep_apply IHt2; entailer!.
  Qed.
  Lemma treeContextDisjoint t con cc n c:
     MyTree t n * MyContext con c cc |-- 
     !!(disjointContext con t = true) && MyTree t n * MyContext con c cc.
  Proof. entailer. apply treeContextDisjoint'. Qed.

  Fixpoint contextUniqueNodes (con:context):bool :=
    match con with
      Root => true
    | lcontext r lcon rhs => (negb (inContext lcon r)) && (negb (contains rhs r)) && 
                             disjointContext lcon rhs && contextUniqueNodes lcon && uniqueNodes rhs
    | rcontext r lhs rcon => (negb (contains lhs r)) && (negb (inContext rcon r)) &&
                             disjointContext rcon lhs && uniqueNodes lhs && contextUniqueNodes rcon
    end.

  Lemma ctxtUniqueNodes': forall con n count,
     MyContext con n count |-- !!(contextUniqueNodes con = true).
  Proof.
    induction con; intros. simpl; entailer!.
  + rename v into r; rename con into lcon; rename t into rhs; simpl.
    Intros rr cc.
    sep_apply contextDistinct; Intros. rewrite H3; simpl.
    sep_apply IHcon; clear IHcon; Intros. rewrite H4; simpl.
    destruct (Val.eq rr nullval); Intros; subst; simpl.
    - entailer!.
    - sep_apply treeDistinct; Intros. (*rewrite H1; simpl.*)
      sep_apply treeContextDisjoint; Intros. (*rewrite H5; simpl.*)
      sep_apply treeUniqueNodes; Intros. (* rewrite H6; simpl;*) entailer!. rewrite H1, H5, H6; trivial.
  + rename v into r; rename con into rcon; rename t into lhs; simpl.
    Intros ll cc.
    sep_apply contextDistinct; Intros. rewrite H3; simpl.
    sep_apply IHcon; clear IHcon; Intros. rewrite H4; simpl.
    destruct (Val.eq ll nullval); Intros; subst; simpl.
    - entailer!.
    - sep_apply treeDistinct; Intros. (*rewrite H1; simpl.*)
      sep_apply treeContextDisjoint; Intros. (* rewrite H4; simpl.*)
      sep_apply treeUniqueNodes; Intros. (*rewrite H5; simpl;*) entailer!. rewrite H1, H5, H6; trivial.
  Qed.
  (*name of lemma is different than in Verifast, to void clash with the Fixpoint def above*)
  Lemma ctxtUniqueNodes con n count:
     MyContext con n count |-- !!(contextUniqueNodes con = true) && MyContext con n count.
  Proof. entailer. apply ctxtUniqueNodes'. Qed.

  Fixpoint getroot (con:context) (prev:val): val :=
  match con with
    Root => prev
  | lcontext r lcon rhs => getroot lcon r
  | rcontext r lhs rcon => getroot rcon r
  end.

  Fixpoint upsideDownMinus (value:tree) (n:val) (con:context): context :=
  match value with
    Tnil => con
  | Tnode r lhs rhs => if Val.eq n r then con
                       else if contains lhs n
                            then upsideDownMinus lhs n (lcontext r con rhs) 
                            else upsideDownMinus rhs n (rcontext r lhs con)
  end.

  Lemma tree2context: forall value n root oroot cvalue, 
   contains value n = true -> getroot cvalue root =oroot ->
   MyTree value root * MyContext cvalue root (size(value))
   |-- !!(getroot (upsideDownMinus value n cvalue) n = oroot) &&
       (MyContext (upsideDownMinus value n cvalue) n (size(valueOf value n)) *
        MyTree (valueOf value n) n).
  Proof. induction value. simpl; intros; entailer!.
    rename v into root. rename value1 into lhs. rename value2 into rhs.
    intros. simpl in H, H0. destruct (Val.eq n root).
    + subst; simpl. destruct (Val.eq root root); [ subst | congruence]. clear e; simpl in H; clear H.
      Intros ll rr. subst root0. simpl. Exists ll rr. entailer!.
    + simpl in *. Intros ll rr. subst root0.
      destruct (Val.eq n root); [ congruence |]. clear n1.
      remember (contains lhs n) as dl; symmetry in Heqdl; destruct dl; simpl; simpl in H.
      - clear H IHvalue2.
        specialize (IHvalue1 n ll oroot (lcontext root cvalue rhs) Heqdl H0).
        destruct (Val.eq ll nullval); Intros; subst; [ inv Heqdl |].
        eapply derives_trans; [ clear IHvalue1 | apply IHvalue1].
        simpl. Exists rr (1 + size lhs + size rhs); entailer!.
        specialize (size_nonneg lhs); specialize (size_nonneg rhs); rep_lia.
      - clear Heqdl IHvalue1. 
        specialize (IHvalue2 n rr oroot (rcontext root lhs cvalue) H H0).
        destruct (Val.eq rr nullval); Intros; subst; [ inv H |].
        eapply derives_trans; [ clear IHvalue2 | apply IHvalue2].
        simpl. Exists ll (1 + size lhs + size rhs).
        specialize (size_nonneg lhs); specialize (size_nonneg rhs); entailer!.
  Qed.
(*
Fixpoint uniqueNodes(t:tree):bool :=
  match t with
    Tnil => true
  | Tnode r leftT rightT => (negb(contains leftT r)) && (negb (contains rightT r))
                          && disjoint leftT rightT && uniqueNodes leftT && uniqueNodes rightT 
   end.
Goal uniqueNodes = Predicates.uniqueNodes. reflexivity.*)
End MyComposite_APD.

Section Composite_VSU.
  Variable M: MallocFreeAPD.

  Definition internalGetNbOfNodes_spec :=
    DECLARE _internalGetNbOfNodes
    WITH n:val, value:_
    PRE [ tptr nodeTP ] 
    PROP() PARAMS (n) GLOBALS ()
      SEP(MyTree M value n)
    POST[ tint ]
     PROP()
     RETURN (Vint (Int.repr (size value)))
     SEP(MyTree M value n).

  Definition internalCreate_spec :=
    DECLARE _internalCreate
    WITH parent:val, gv: globals
    PRE [ tptr nodeTP ] 
    PROP() PARAMS (parent) GLOBALS (gv) SEP(mem_mgr M gv)
    POST[ tptr nodeTP ]
    EX res: val,
     PROP(isptr res)
     RETURN (res)
     SEP(MyTree M (Tnode res Tnil Tnil) res; mem_mgr M gv; field_at Ews nodeTP [StructField _parent] parent res).

  Definition internalAddLeft_spec := 
    DECLARE _internalAddLeft
    WITH node:val, value:_, gv:globals
    PRE [ tptr nodeTP ] 
    PROP(isptr node) PARAMS (node) GLOBALS (gv)
      SEP(MyContext M value node 1; field_at Ews nodeTP [StructField _left] nullval node;
          field_at Ews nodeTP [StructField _right] nullval node; mem_mgr M gv;
          malloc_token M Ews nodeTP node; field_at Ews nodeTP [StructField _count] (Vint(Int.repr 1)) node)
    POST[ tptr nodeTP ]
    EX res: val,
     PROP()
     RETURN (res)
     SEP(MyContext M value node 2; field_at Ews nodeTP [StructField _left] res node; 
         field_at Ews nodeTP [StructField _right] nullval node; mem_mgr M gv;
         malloc_token M Ews nodeTP node; field_at Ews nodeTP [StructField _count] (Vint(Int.repr 2)) node;
         MyTree M (Tnode res Tnil Tnil) res; field_at Ews nodeTP [StructField _parent] node res).

  Definition fix_spec :=
    DECLARE _fix
    WITH node:val, c:Z, value:context
    PRE [ tptr nodeTP ] 
      PROP(isptr node; (*Int.min_signed*)0 <= c <= Int.max_signed)
      PARAMS (node) GLOBALS () 
      SEP(field_at Ews nodeTP [StructField _count] (Vint (Int.repr c)) node;
          MyContext M value node c)
    POST[ tvoid ]
     PROP(0 <= c+1 <= Int.max_signed)
     RETURN ()
     SEP(MyContext M value node (c+1); field_at Ews nodeTP [StructField _count] (Vint (Int.repr (c+1))) node).

  Definition Vprog : varspecs. mk_varspecs prog. Defined.

  Definition Comp_ASI: funspecs := CompositeASI M (MyCompositeAPD M).

  Definition Comp_imported_specs: funspecs := exit_spec :: SM_ASI M.

  Definition Comp_internal_specs: funspecs := [internalGetNbOfNodes_spec; fix_spec; internalCreate_spec; internalAddLeft_spec] ++ Comp_ASI. (*extend by other functions later*)

  Definition CompGprog:funspecs := Comp_imported_specs ++ Comp_internal_specs.

  Lemma internalGetNbOfNodes_verif:
    semax_body Vprog CompGprog f_internalGetNbOfNodes internalGetNbOfNodes_spec.
  Proof.
    start_function.
    simpl. rewrite MyTree_Fold.
    destruct value as [ | x ltree rtree]. apply semax_pre with FF. entailer!. apply semax_ff.
    Intros l r. subst x. forward. forward.
    rewrite (MyTree_Fold M (Tnode n ltree rtree)). Exists l r. entailer!.
  Qed.

  Fixpoint contextSize (con:context):=
    match con with
      Root => 0
    | lcontext r lcon rhs => 1 + contextSize lcon
    | rcontext r lhs rcon => 1 + contextSize rcon
    end.

  Lemma upsideDownMinusSize: forall (t:tree) n con, 
        contextSize con <= contextSize (upsideDownMinus t n con).
  Proof. induction t; simpl; intros. lia.
    destruct (Val.eq n v); subst. lia.
    destruct (contains t1 n).
    specialize (IHt1 n (lcontext v con t2)). simpl in IHt1; lia.
    specialize (IHt2 n (rcontext v t1 con)). simpl in IHt2; lia.
  Qed.

  Lemma upsideDownMinusIsRoot r (lhs rhs: tree) node (con:context):
        (con =upsideDownMinus (Tnode r lhs rhs) node con) -> node = r.
  Proof. simpl; intros.
    destruct (Val.eq node r); subst; trivial.
    destruct (contains lhs node); subst.
    specialize (upsideDownMinusSize lhs node (lcontext r con rhs));  rewrite <- H. clear; simpl; lia.
    specialize (upsideDownMinusSize rhs node (rcontext r lhs con));  rewrite <- H. clear; simpl; lia.
  Qed.

  Lemma valueOfNotNil: forall (t:tree) n, contains t n = true -> valueOf t n <> Tnil.
  Proof. induction t; simpl; intros. congruence. 
   specialize (IHt1 n). specialize (IHt2 n).
    destruct (Val.eq n v); subst; simpl in *. congruence.
    destruct (contains t1 n); simpl in *; auto.
  Qed.

  Lemma valueOfRoot: forall t n r lhs rhs,
        valueOf t n = Tnode r lhs rhs ->  n = r.
  Proof. induction t; simpl; intros. congruence.
   destruct (Val.eq n v); subst; simpl in *.
   - inv H; trivial.
   - destruct (contains t1 n); eauto.
  Qed.

  Lemma containsValueOf: forall t n v n2,
        uniqueNodes t =true -> valueOf t n =v -> contains v n2 =true ->
        contains t n2=true.
  Proof. induction t; simpl; intros; subst. inv H1.
    destruct (Val.eq n v); subst; simpl in *; trivial.
    repeat rewrite andb_true_iff in H. destruct H as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H. apply binop_lemmas3.negb_true in H0.
    destruct (contains t1 n).
    + rewrite (IHt1 n (valueOf t1 n)), orb_true_r; trivial.
    + rewrite (IHt2 n (valueOf t2 n)), orb_true_r; trivial.
  Qed.

  Lemma disjointContains: forall t1 t2 n, contains t2 n =true -> disjoint t1 t2 =true ->
        contains t1 n = false.
  Proof. induction t1; simpl; intros; trivial.
    repeat rewrite andb_true_iff in H0. destruct H0 as [[? ?] ?]. 
    apply binop_lemmas3.negb_true in H0.
    destruct (Val.eq n v ); subst; simpl. congruence. 
    rewrite (IHt1_1 t2 n); trivial. 
    rewrite (IHt1_2 t2 n); trivial.
  Qed. 

  Lemma valueOfValueOf: forall t r v n, 
        uniqueNodes t = true -> contains t r = true -> 
        valueOf t r = v -> contains v n = true -> valueOf t n = valueOf v n.
  Proof. induction t; intros. simpl in H0; congruence.
    simpl in *.
    destruct (Val.eq r v); subst; simpl in *; trivial.
    repeat rewrite andb_true_iff in H. destruct H as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H. apply binop_lemmas3.negb_true in H1.
    remember (contains t1 r) as c1; symmetry in Heqc1; destruct c1; simpl in *.
    + specialize (containsValueOf t1 r (*v*)_ n U1 (eq_refl _) H2); intros CAV1.
      rewrite CAV1.
      destruct (Val.eq n v); subst; simpl in *. congruence. apply (IHt1 r); trivial.
    + specialize (containsValueOf t2 r (*v*)_ n U2 (eq_refl _) H2); intros CAV2.
      specialize (disjointContains t1 t2 n CAV2 H3); intros. rewrite H4 in *.
      destruct (Val.eq n v ); subst; simpl. congruence. apply (IHt2 r); trivial.
  Qed.

  Lemma uniqueNodesValueOf: forall t n,
        uniqueNodes t = true -> contains t n = true ->
        uniqueNodes (valueOf t n) = true.
  Proof. induction t; simpl; intros. congruence.
    repeat rewrite andb_true_iff in H. destruct H as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H. apply binop_lemmas3.negb_true in H1.
    destruct (Val.eq n v); subst; simpl in *; trivial.
    + rewrite H, H1, H2, U1, U2; trivial.
    + specialize (IHt1 n U1). specialize (IHt2 n U2).
      destruct (contains t1 n); simpl in *.
      - apply IHt1; trivial.
      - apply IHt2; trivial.
  Qed.

  Lemma containsReplace: forall (t:tree) (node:val) (plug:tree),
        contains t node =true -> contains plug node = true ->
        contains (replace t node plug) node =true.
  Proof. induction t; simpl; intros; trivial.
    destruct (Val.eq v node); subst; simpl in *; trivial.
    destruct (Val.eq node v); subst; simpl in *. congruence. 
    specialize (IHt1 node plug). specialize (IHt2 node plug).
    destruct (contains t1 node); simpl in *.
    + rewrite IHt1, orb_true_r; trivial.
    + rewrite IHt2, orb_true_r; trivial.
  Qed.

  Lemma replaceItself: forall (value:tree) (node:val),
        uniqueNodes value =true ->
        replace value node (valueOf value node) = value.
  Proof. induction value; simpl; intros. trivial.
    simpl in H. repeat rewrite andb_true_iff in H. destruct H as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H. apply binop_lemmas3.negb_true in H0.
    destruct (Val.eq v node); subst; simpl in *.
    + rewrite if_true; trivial.
    + destruct (Val.eq node v). congruence.
      remember (contains value1 node); destruct b.
      rewrite IHvalue1; trivial. 
      rewrite IHvalue2; trivial.
  Qed.

  Lemma XX1: forall t1 v t2, Tnode v t1 t2 <> t1. 
  Proof. induction t1. congruence. intros. intros N. inv N. apply IHt1_1 in H1; trivial. Qed.
  Lemma XX2: forall t1 v t2, Tnode v t1 t2 <> t2. 
  Proof. induction t2. congruence. intros N. inv N. congruence. Qed.

  Lemma upsideDownMinusParentLeft: forall parent oldval child lhs rhs rhs2 con,
         contains oldval parent = true -> valueOf oldval parent = Tnode parent (Tnode child lhs rhs) rhs2 ->
         parent <> child -> uniqueNodes oldval = true ->
   upsideDownMinus oldval child con = lcontext parent (upsideDownMinus oldval parent con) rhs2.
  Proof. intros parent oldval. induction oldval; simpl; intros. congruence.
    repeat rewrite andb_true_iff in H2. destruct H2 as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H2. apply binop_lemmas3.negb_true in H3.
    destruct (Val.eq parent v); subst; simpl in *.
    + inv H0. simpl in *.
      destruct (Val.eq child v); subst.  congruence. 
      destruct (Val.eq child child); simpl in *; trivial. congruence.
    + destruct (contains oldval1 parent).
      - clear IHoldval2.
        rewrite <- (IHoldval1 child lhs rhs rhs2 (lcontext v con oldval2) (eq_refl _) H0 H1 U1); clear IHoldval1.
        specialize (containsValueOf oldval1 parent (Tnode parent (Tnode child lhs rhs) rhs2) child U1 H0); intros CVO.
        destruct (Val.eq child v); subst; simpl; trivial.
        * rewrite CVO in H2. congruence. 
          simpl. destruct (Val.eq v v); [ simpl; rewrite orb_true_r; trivial | congruence].
        * simpl. rewrite CVO; trivial. simpl.
          destruct (Val.eq child child). simpl; rewrite orb_true_r; trivial. congruence.
     - simpl in H. clear IHoldval1.
       rewrite <- (IHoldval2 child lhs rhs rhs2 (rcontext v oldval1 con) H H0 H1 U2); clear IHoldval2.
       specialize (containsValueOf oldval2 parent (Tnode parent (Tnode child lhs rhs) rhs2) child U2 H0); intros CVO. 
       specialize (disjointContains oldval1 oldval2 child); intros DC.
       rewrite DC; trivial. 2:{ apply  CVO. simpl. destruct (Val.eq child child ); simpl. rewrite orb_true_r; trivial. congruence. }
       destruct (Val.eq child v); trivial. subst. simpl in *.
       destruct (Val.eq v v); simpl in CVO. rewrite orb_true_r in CVO. rewrite CVO in H3; trivial; congruence. congruence.
  Qed.
       
  Lemma upsideDownMinusParentRight parent: forall oldval child lhs rhs lhs2 con,
        contains oldval parent = true -> valueOf oldval parent = Tnode parent lhs2 (Tnode child lhs rhs) ->
        parent <> child -> uniqueNodes oldval =true ->
  upsideDownMinus oldval child con = rcontext parent lhs2 (upsideDownMinus oldval parent con).
  Proof.  intros oldval.
    induction oldval; simpl; intros. congruence.
    repeat rewrite andb_true_iff in H2. destruct H2 as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H2. apply binop_lemmas3.negb_true in H3.
    destruct (Val.eq parent v); subst; simpl in *.
    + inv H0.
      destruct (Val.eq child v); subst. congruence. clear H.
      rewrite (disjointContains lhs2 (Tnode child lhs rhs) child); simpl; trivial.
      - destruct (Val.eq child child); simpl; trivial; congruence.
      - destruct (Val.eq child child); simpl; trivial; congruence.
    + destruct (contains oldval1 parent); simpl in H.
      - rewrite <- (IHoldval1 child lhs rhs lhs2 (lcontext v con oldval2)); trivial.
        clear IHoldval1 IHoldval2. 
        specialize (containsValueOf oldval1 parent (Tnode parent lhs2 (Tnode child lhs rhs)) child); intros COV.
        rewrite COV; trivial.
        * destruct (Val.eq child v); subst; trivial.
          rewrite COV in H2; trivial. congruence.
          simpl. destruct (Val.eq v v); [simpl | congruence]. rewrite orb_true_r; trivial.
        * simpl. destruct (Val.eq child child); [simpl | congruence]. rewrite orb_true_r; trivial.
      - rewrite <- (IHoldval2 child lhs rhs lhs2 (rcontext v oldval1 con)); trivial.
        clear IHoldval1 IHoldval2. 
        specialize (containsValueOf oldval2 parent (Tnode parent lhs2 (Tnode child lhs rhs)) child); intros COV.
        rewrite (disjointContains oldval1 oldval2 child); trivial.
        * destruct (Val.eq child v); trivial. subst.
          rewrite COV in H3; trivial. congruence.
          simpl. destruct (Val.eq v v); [rewrite orb_true_r; trivial | congruence].
        * apply COV; trivial. simpl. 
          destruct (Val.eq child child); [rewrite orb_true_r; trivial | congruence].
  Qed.

  Lemma replace_nocontains node: forall u v, contains u node = false -> replace u node v = u.
  Proof. induction u; simpl; intros; trivial.
    repeat rewrite orb_false_iff in H. destruct H as [[? ?] ?].
    destruct (Val.eq v node); subst; simpl in *. destruct ( Val.eq node node). inv H.  congruence.
    rewrite H0, H1 in *. rewrite IHu2; trivial.
  Qed.

  Lemma XX3: forall v n x y z, valueOf v n = Tnode x y z -> contains v n = true.
  Proof. induction v; simpl; intros. inv H.
   destruct ( Val.eq n v1); subst; simpl in *; trivial.
   destruct (contains v2 n); simpl in *;trivial. eauto.
  Qed.

  Lemma replaceParentLeft: forall oldval nodeParent node nlhs nrhs rhs v,
        contains oldval nodeParent = true ->
        valueOf oldval nodeParent = Tnode nodeParent (Tnode node nlhs nrhs) rhs ->
        uniqueNodes oldval = true ->
  replace oldval nodeParent (Tnode nodeParent v rhs) = replace oldval node v.
  Proof. induction oldval; simpl; intros. congruence.
    repeat rewrite andb_true_iff in H1. destruct H1 as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H1. apply binop_lemmas3.negb_true in H2.
    destruct (Val.eq nodeParent v); subst.
    + destruct (Val.eq v v); [simpl in * | congruence]. inv H0. simpl in *.
      destruct (Val.eq node node). 2: congruence. simpl.
      destruct (Val.eq v node); trivial. simpl in *. congruence.
    + simpl in *. 
      destruct (Val.eq v nodeParent). congruence.
      destruct (contains oldval1 nodeParent); simpl in *.
      - clear IHoldval2. 
        specialize (containsValueOf oldval1 nodeParent (Tnode nodeParent (Tnode node nlhs nrhs) rhs) node U1 H0); intros.
        simpl in H4. destruct (Val.eq node node). 2: congruence. rewrite orb_true_r in H4. simpl in H4.
        rewrite H4; trivial.
        destruct (Val.eq v node); subst. rewrite H4 in H1 by trivial; congruence.
        erewrite IHoldval1; trivial. 2: eassumption. eapply XX3. apply H0.
      - clear IHoldval1.
        specialize (containsValueOf oldval2 nodeParent (Tnode nodeParent (Tnode node nlhs nrhs) rhs) node U2 H0); intros.
        simpl in H4. destruct (Val.eq node node). 2: congruence. rewrite orb_true_r in H4. simpl in H4.
        rewrite (disjointContains oldval1 oldval2 node); auto.
        destruct (Val.eq v node); subst. rewrite H4 in H2 by trivial; congruence.
        erewrite IHoldval2; trivial. apply H0.
  Qed.

  Lemma replaceParentRight: forall oldval nodeParent node nlhs nrhs lhs v,
        contains oldval nodeParent = true ->
        valueOf oldval nodeParent = Tnode nodeParent lhs (Tnode node nlhs nrhs) ->
        uniqueNodes oldval = true ->
  replace oldval nodeParent (Tnode nodeParent lhs v) = replace oldval node v.
  Proof. induction oldval. simpl; intros. congruence. intros. 
    specialize (uniqueNodesValueOf (Tnode v oldval1 oldval2) nodeParent H1 H); intros UNVO.
    specialize (disjointContains lhs (Tnode node nlhs nrhs) node); intros DC.
    simpl in *. repeat rewrite andb_true_iff in H1. destruct H1 as [[[[? ?] ?] U1] U2].
    apply binop_lemmas3.negb_true in H1. apply binop_lemmas3.negb_true in H2.
    destruct (Val.eq nodeParent v); subst.
    + destruct (Val.eq v v); [simpl in * | congruence]. inv H0. clear e H.
      rewrite DC; trivial. 2: destruct (Val.eq node node); trivial; congruence.
      rewrite H1 in *;  simpl in H2.
      destruct (Val.eq v node); subst; simpl in H2. discriminate.
      simpl. destruct (Val.eq node node); [trivial | congruence].
    + specialize (IHoldval1 nodeParent node nlhs nrhs lhs).
      specialize (IHoldval2 nodeParent node nlhs nrhs lhs).
      destruct (contains oldval1 nodeParent); simpl in H.
      - clear H IHoldval2. 
        specialize (containsValueOf oldval1 nodeParent (Tnode nodeParent lhs (Tnode node nlhs nrhs)) node U1 H0); simpl; intros CVO.
        destruct (Val.eq node node); [ simpl in DC, CVO | congruence].
        rewrite (IHoldval1 v0 (eq_refl _) H0 U1); clear IHoldval1.
        rewrite orb_true_r in CVO. rewrite CVO; trivial.
        destruct (Val.eq v nodeParent); subst. congruence.
        destruct (Val.eq v node); subst; simpl in *; trivial.
        rewrite CVO in H1; [ congruence | trivial].
      - clear IHoldval1.
        specialize (containsValueOf oldval2 nodeParent (Tnode nodeParent lhs (Tnode node nlhs nrhs)) node U2 H0); simpl; intros CVO.
        specialize (disjointContains oldval1 oldval2 node); intros DC2.
        rewrite (IHoldval2 v0 H H0 U2); clear IHoldval2.
        destruct (Val.eq node node); [ simpl in DC, CVO | congruence].
        rewrite orb_true_r in CVO. specialize (CVO (eq_refl _)).
        specialize (DC2 CVO H3). rewrite DC2.
        destruct (Val.eq v nodeParent); subst. congruence.
        destruct (Val.eq v node); subst; simpl in *; trivial.
        congruence.
  Qed.

  Lemma upsideDownMinusLContext t0: forall tr t n nlhs nrhs con r lcon rhs,
    uniqueNodes t0 = true ->
    contains t0 tr = true ->
    upsideDownMinus t0 n Root = upsideDownMinus t n con ->
    upsideDownMinus t n con = lcontext r lcon rhs ->
    contains t n =true ->
    t = valueOf t0 tr ->
    valueOf t0 n = Tnode n nlhs nrhs ->
    tr <> n ->
   (contains t0 r = true /\
    valueOf t0 r = Tnode r (Tnode n nlhs nrhs) rhs).
  Proof. intros tr t. generalize dependent tr.
    induction t; simpl; intros. congruence. 
    rename v into tr0. rename t1 into tlhs. rename t2 into trhs.
    symmetry in H4; specialize (valueOfRoot _ _ _ _ _ H4); intros VOR. subst tr0.
    destruct (Val.eq n tr); subst; [ congruence | simpl in *; clear n0].
    remember (contains tlhs n) as cont1; symmetry in Heqcont1; destruct cont1.
    + clear H3. clear IHt2.
      destruct (tlhs) as [ | rtlhs tlhslhs tlhsrhs]; [ simpl in Heqcont1; discriminate | simpl in Heqcont1].
      destruct (Val.eq n rtlhs); subst.
      - simpl in H2, Heqcont1. destruct (Val.eq rtlhs rtlhs); [ | congruence]. clear e Heqcont1.
        specialize (valueOfValueOf t0 tr _ rtlhs H H0 H4). rewrite <- H4. simpl; intros X.
        inv H2. split; trivial. rewrite H4. f_equal.
        rewrite H4 in X. simpl in X.
        destruct (Val.eq rtlhs r); [ congruence |]. simpl in X.
        destruct (Val.eq rtlhs rtlhs); [ | congruence ]. simpl in X.
        rewrite X in H5; trivial.
      - assert (XX: contains (Tnode tr (Tnode rtlhs tlhslhs tlhsrhs) trhs) rtlhs = true).
        { simpl. destruct (Val.eq rtlhs rtlhs ). rewrite orb_true_r; trivial. congruence. }
        specialize (containsValueOf t0 tr (Tnode tr (Tnode rtlhs tlhslhs tlhsrhs) trhs) rtlhs H H4 XX); intros CVO.
        specialize (valueOfValueOf t0 tr (Tnode tr (Tnode rtlhs tlhslhs tlhsrhs) trhs) rtlhs H H0 H4 XX); intros VOVO.
        specialize (uniqueNodesValueOf t0 tr H H0); intros UNVO.
        specialize (IHt1 rtlhs n nlhs nrhs (lcontext tr con trhs) r lcon rhs H CVO H1 H2); clear H1 H2.
        simpl in Heqcont1.
        apply IHt1; trivial; clear IHt1.
        * simpl. destruct (Val.eq n rtlhs); [  congruence | simpl; rewrite Heqcont1; trivial].
        * rewrite VOVO. simpl in *. rewrite H4 in UNVO; simpl in UNVO.
          destruct (Val.eq rtlhs tr); subst; simpl.
          ++ destruct (Val.eq tr tr); simpl in *; congruence.
          ++ destruct (Val.eq rtlhs rtlhs); [ subst; simpl; trivial |  congruence].
        * congruence.
    + simpl in H3. clear IHt1. remember (Tnode tr tlhs trhs) as t.
      destruct (trhs) as [ | rtrhs trhslhs trhsrhs]; [ simpl in Heqcont1; discriminate | simpl in Heqcont1].
      clear trhs. remember (Tnode rtrhs trhslhs trhsrhs) as trhs.
      destruct (Val.eq n rtrhs).
      - clear IHt2. subst; simpl in *. destruct (Val.eq rtrhs rtrhs); [ simpl in * | congruence]. congruence.
      - assert (XX: contains t rtrhs = true). 
        { subst t trhs; simpl. destruct (Val.eq rtrhs rtrhs ). rewrite orb_true_r; trivial. congruence. }
        specialize (containsValueOf t0 tr t rtrhs H H4 XX); intros VCVO.
        specialize (valueOfValueOf t0 tr t rtrhs H H0 H4 XX); intros VOVO.
        specialize (uniqueNodesValueOf t0 tr H H0); intros UNVO.
        specialize (disjointContains tlhs trhs rtrhs); intros DC.
        specialize (IHt2 rtrhs n nlhs nrhs (rcontext tr tlhs con) r lcon rhs H VCVO H1 H2 H3).
        subst t trhs; simpl in *.
        rewrite H4 in UNVO. simpl in UNVO.
        repeat rewrite andb_true_iff in UNVO. destruct UNVO as [[[[U1 U2] U3] U4] U5].
        (*apply binop_lemmas3.negb_true in U1. apply binop_lemmas3.negb_true in U2. simpl in XX. *)
        simpl in VOVO. subst; simpl in *. 
        destruct (Val.eq n rtrhs). congruence. simpl in H3.
        destruct (Val.eq rtrhs rtrhs); [simpl in * | congruence].
        specialize (DC (eq_refl _) U3). rewrite DC in *.  rewrite orb_true_r in XX. clear XX e.
        destruct (Val.eq rtrhs tr); subst; simpl in *.
        * destruct (Val.eq tr tr); simpl in *; congruence.
        * rewrite <- VOVO in IHt2. apply IHt2; trivial. congruence.
  Qed.

  Lemma upsideDownMinusRContext t0: forall tr t n nlhs nrhs con r rcon lhs,
    uniqueNodes t0 = true ->
    contains t0 tr = true ->
    upsideDownMinus t0 n Root = upsideDownMinus t n con ->
    upsideDownMinus t n con = rcontext r lhs rcon ->
    contains t n =true ->
    t = valueOf t0 tr ->
    valueOf t0 n = Tnode n nlhs nrhs ->
    tr <> n ->
    contains t0 r = true /\
    valueOf t0 r = Tnode r lhs (Tnode n nlhs nrhs).
  Proof. intros tr t; generalize dependent tr.
  induction t. simpl; intros. congruence.
  rename t1 into tlhs. rename t2 into trhs. rename v into tr0.
  intros. exploit (valueOfRoot t0 tr tr0 tlhs trhs). rewrite H4; trivial. intros; subst tr0.
  remember (Tnode tr tlhs trhs)  as  t.
  remember (contains tlhs n); symmetry in Heqb; destruct b.
  + destruct tlhs as [ | rtlhs tlhslhs tlhsrhs]; simpl in *. congruence.
    remember (Tnode rtlhs tlhslhs tlhsrhs) as tlhs.
    destruct (Val.eq n rtlhs); simpl in Heqb.
    - subst n. symmetry in H4. exploit (valueOfValueOf t0 tr t rtlhs H H0); trivial. intros.
      subst. simpl in *. destruct ( Val.eq rtlhs tr); simpl in *. subst; congruence.
      destruct (Val.eq rtlhs rtlhs); simpl in *; subst; congruence.
    - (*738*) subst t. exploit (containsValueOf t0 tr (Tnode tr tlhs trhs) rtlhs H).
      { rewrite <- H4; trivial. }
      { subst tlhs; simpl. destruct (Val.eq rtlhs rtlhs); [ rewrite orb_true_r; trivial | congruence]. }
      intros CVO.
      specialize (uniqueNodesValueOf t0 tr H H0); intros UN.
      exploit (valueOfValueOf t0 tr (Tnode tr tlhs trhs) rtlhs ); trivial.
      { rewrite <- H4; trivial. }
      { subst tlhs; clear; simpl. destruct (Val.eq rtlhs rtlhs). rewrite orb_true_r; trivial. congruence. }
      intros VOVO.
      clear IHt2.
      specialize (IHt1 rtlhs n nlhs nrhs (lcontext tr con trhs) r rcon lhs H CVO).
      subst tlhs. simpl in *. 
      destruct (Val.eq n rtlhs); [ subst; congruence |]. simpl in *.
      destruct (Val.eq n tr); [ subst; congruence |]. simpl in *.
      rewrite <- H4 in UN. simpl in UN.
      destruct (Val.eq tr rtlhs); subst; simpl in *. congruence.
      destruct (Val.eq rtlhs tr); subst; simpl in *. congruence.
      destruct (Val.eq rtlhs rtlhs); subst; simpl in *. 2: congruence. clear e.
      rewrite Heqb in *. clear H3. specialize (IHt1 H1 H2).
      apply IHt1; trivial. symmetry; trivial. congruence.
  + clear IHt1. destruct trhs as [ | rtrhs trhslhs trhsrhs]; simpl in *.
    - subst. clear IHt2. simpl in *. rewrite Heqb, orb_false_r in *. 
      destruct (Val.eq n tr);  subst; simpl in *; congruence.
    - destruct (Val.eq rtrhs n); subst.
      * clear IHt2. 
        specialize (valueOfValueOf t0 r (Tnode r tlhs (Tnode n trhslhs trhsrhs)) n H); intros VOVO.
        specialize (uniqueNodesValueOf t0 tr H H0); intros UNVO1.
        specialize (uniqueNodesValueOf t0 r H); intros UNVO2.
        specialize (disjointContains tlhs (Tnode n trhslhs trhsrhs) n); intros DC.
        rewrite <- H4 in UNVO1. simpl in *.
        destruct (Val.eq n n); [simpl in * | congruence]. clear e H3. rewrite orb_true_r, Heqb in *. 
        destruct (Val.eq n tr); subst; simpl in *.
        ++ destruct (Val.eq tr tr); [simpl in * | congruence]. clear e. 
           rewrite andb_false_r in UNVO1. congruence.
        ++ inv H2. split; trivial. rewrite H0 in *.
           rewrite <- H4 in UNVO2, VOVO. specialize (UNVO2 (eq_refl _)).
           destruct (Val.eq n r). congruence. rewrite VOVO in H5; trivial; clear VOVO. inv H5.
           symmetry; trivial. 
      * symmetry in H4.
        assert (XX: contains (Tnode tr tlhs (Tnode rtrhs trhslhs trhsrhs)) rtrhs = true).
        { simpl.  destruct (Val.eq rtrhs rtrhs); simpl. rewrite orb_true_r; trivial. congruence. }
        specialize (containsValueOf t0 tr (Tnode tr tlhs (Tnode rtrhs trhslhs trhsrhs)) rtrhs H H4 XX); intros CVO.
        specialize (valueOfValueOf t0 tr (Tnode tr tlhs (Tnode rtrhs trhslhs trhsrhs)) rtrhs H H0 H4 XX); intros VOVO.
        specialize (uniqueNodesValueOf t0 tr H H0); intros UNVO.
        assert (YY: contains (Tnode rtrhs trhslhs trhsrhs) rtrhs = true).
        { simpl. destruct (Val.eq rtrhs rtrhs); trivial. congruence. }
        specialize (disjointContains tlhs (Tnode rtrhs trhslhs trhsrhs) rtrhs YY); intros DC.
        specialize (IHt2 rtrhs n nlhs nrhs (rcontext tr tlhs con) r rcon lhs H CVO).
        destruct (Val.eq n tr). congruence. simpl in *. 
        destruct (Val.eq n rtrhs); simpl in *. congruence.
        destruct (Val.eq rtrhs rtrhs). simpl in *. 2: congruence. clear e XX YY.
        destruct (Val.eq n tr). congruence. rewrite Heqb in *; simpl in *.
        specialize (IHt2 H1 H2 H3).
        rewrite H4 in UNVO; simpl in UNVO.
        destruct (Val.eq rtrhs tr); subst; simpl in *.
        ++ destruct (Val.eq tr tr); simpl in *. rewrite andb_false_r in UNVO. simpl in UNVO; congruence. congruence.
        ++ destruct (Val.eq tr rtrhs); subst; simpl in *.
           -- rewrite andb_false_r in UNVO; simpl in UNVO. congruence.
           -- destruct (disjoint tlhs (Tnode rtrhs trhslhs trhsrhs) ); simpl in *.
              2:{ rewrite andb_false_r in UNVO; simpl in UNVO. congruence. }
              rewrite DC in *; trivial. symmetry in VOVO; apply IHt2; trivial.
    Qed.

  Lemma context2tree theroot : forall node (contextval:context) (oldval: tree) v
        (GR: getroot contextval node = theroot)
        (CV: contextval = upsideDownMinus oldval node Root)
        (CONT: contains oldval node = true)
        (UNQ: uniqueNodes oldval = true),
     MyTree M v node * MyContext M contextval node (size v)
  |-- MyTree M (replace oldval node v) theroot * field_at Ews nodeTP [StructField _parent] nullval theroot.
  Proof. intros ? ?. generalize dependent node. induction contextval; intros.
  + simpl in *. destruct oldval. 
   - simpl in *; congruence.
   - specialize (upsideDownMinusIsRoot v0 oldval1 oldval2 node Root CV); intros.
     subst v0 theroot; simpl. rewrite if_true; trivial.
  + rename v into r. rename contextval into lcon. rename t into rhs. rename v0 into v.
    rewrite MyContext_Fold. Intros rr cc; Intros. (*reveals r=nodeParent*)
    specialize (valueOfNotNil _ _ CONT); intros NN.
    remember (valueOf oldval node) as t; destruct t; [ congruence |].
    
    rename v0 into node0. rename t1 into nodeLhs. rename t2 into nodeRhs. clear NN.
    symmetry in Heqt. specialize (valueOfRoot _ _ _ _ _ Heqt); intros; subst node0 (*-node*).
    eapply derives_trans with (MyTree M (Tnode r v rhs) r * MyContext M lcon r (size v + 1 + size rhs))%logic.
      { rewrite (MyTree_Fold M (Tnode r v rhs)).
        Exists node rr. entailer!. simpl; try lia.
        rewrite if_false by congruence. cancel. apply derives_refl'. f_equal. f_equal. f_equal. simpl. lia. }
          
    specialize (IHcontextval r oldval (Tnode r v rhs) ).
    assert (X: size (Tnode r v rhs) = size v + 1 + size rhs) by ( simpl; lia). rewrite <- X.
    simpl in GR; rewrite GR, UNQ in IHcontextval.
    specialize (uniqueNodesValueOf oldval r UNQ); intros UNVO.
    specialize (upsideDownMinusParentLeft r oldval node nodeLhs nodeRhs rhs Root); rewrite UNQ; intros DMPL.
    specialize (replaceParentLeft oldval r node nodeLhs nodeRhs rhs); rewrite UNQ; intros rPL.
    destruct oldval as [ | oldvalroot oldvallhs oldvalrhs].
    1: simpl in Heqt; congruence. 
    remember (Tnode oldvalroot oldvallhs oldvalrhs)  as oldval. 
    exploit (upsideDownMinusLContext oldval oldvalroot oldval node nodeLhs nodeRhs Root r lcon rhs UNQ); trivial.
    { subst oldval; simpl. destruct (Val.eq oldvalroot oldvalroot); trivial; congruence. }
    { rewrite CV; trivial. }
    { subst oldval; simpl. destruct (Val.eq oldvalroot oldvalroot); trivial; congruence. }
    { intros N; subst. clear rPL DMPL UNVO IHcontextval. simpl in *. 
      destruct (Val.eq node node); [simpl in * | congruence]. discriminate. }
    intros [COR VOR].
    destruct (Val.eq r node).
    { subst r. clear IHcontextval UNVO rPL DMPL. subst; simpl in *.
      rewrite VOR in Heqt; exfalso; clear - Heqt. 
      assert (CONTRA: size (Tnode node nodeLhs nodeRhs) = size (Tnode node (Tnode node nodeLhs nodeRhs) rhs)). rewrite Heqt; trivial.
      forget (Tnode node nodeLhs nodeRhs) as tre. clear - CONTRA. specialize (size_nonneg rhs). simpl in CONTRA. lia. }
    specialize (UNVO COR). specialize (DMPL COR VOR n (eq_refl _)).
    rewrite <- (rPL v COR VOR (eq_refl _)).
    apply IHcontextval; trivial. clear - DMPL CV. rewrite DMPL in CV. inv CV; trivial.
  + rename v into r. rename contextval into rcon. rename t into lhs. rename v0 into v.
    rewrite MyContext_Fold. Intros ll cc; Intros. (*reveals r=nodeParent*)
    specialize (valueOfNotNil _ _ CONT); intros NN.
    remember (valueOf oldval node) as t; destruct t; [ congruence |].
    
    rename v0 into node0. rename t1 into nodeLhs. rename t2 into nodeRhs. clear NN.
    symmetry in Heqt. specialize (valueOfRoot _ _ _ _ _ Heqt); intros; subst node0 (*-node*).
    eapply derives_trans with (MyTree M (Tnode r lhs v) r * MyContext M rcon r (size v + 1 + size lhs))%logic.
      { rewrite (MyTree_Fold M (Tnode r lhs v)).
        Exists ll node. entailer!. simpl; try lia.
        rewrite if_false by congruence. cancel. apply derives_refl'. f_equal. f_equal. f_equal. simpl. lia. }
          
    specialize (IHcontextval r oldval (Tnode r lhs v) ).
    assert (X: size (Tnode r lhs v) = size v + 1 + size lhs) by ( simpl; lia). rewrite <- X.
    simpl in GR; rewrite GR, UNQ in IHcontextval.
    specialize (uniqueNodesValueOf oldval r UNQ); intros UNVO.
    specialize (upsideDownMinusParentRight r oldval node nodeLhs nodeRhs lhs Root); rewrite UNQ; intros DMPL.
    specialize (replaceParentRight oldval r node nodeLhs nodeRhs lhs); rewrite UNQ; intros rPL.
    destruct oldval as [ | oldvalroot oldvallhs oldvalrhs].
    1: simpl in Heqt; congruence. 
    remember (Tnode oldvalroot oldvallhs oldvalrhs)  as oldval. 
    exploit (upsideDownMinusRContext oldval oldvalroot oldval node nodeLhs nodeRhs Root r rcon lhs UNQ); trivial.
    { subst oldval; simpl. destruct (Val.eq oldvalroot oldvalroot); trivial; congruence. }
    { rewrite CV; trivial. }
    { subst oldval; simpl. destruct (Val.eq oldvalroot oldvalroot); trivial; congruence. }
    { intros N; subst. clear rPL DMPL UNVO IHcontextval. simpl in *. 
      destruct (Val.eq node node); [simpl in * | congruence]. discriminate. }
    intros [COR VOR].
    destruct (Val.eq r node).
    { subst r. clear IHcontextval UNVO rPL DMPL. subst; simpl in *.
      rewrite VOR in Heqt; exfalso; clear - Heqt. 
      assert (CONTRA: size (Tnode node nodeLhs nodeRhs) = size (Tnode node lhs (Tnode node nodeLhs nodeRhs))). rewrite Heqt; trivial.
      forget (Tnode node nodeLhs nodeRhs) as tre. clear - CONTRA. specialize (size_nonneg lhs). simpl in CONTRA. lia. }
    specialize (UNVO COR). specialize (DMPL COR VOR n (eq_refl _)).
    rewrite <- (rPL v COR VOR (eq_refl _)).
    apply IHcontextval; trivial. clear - DMPL CV. rewrite DMPL in CV. inv CV; trivial.
  Qed.

  Lemma getNbOfNodes_verif:
    semax_body Vprog CompGprog f_getNbOfNodes (getNbOfNodes_spec (MyCompositeAPD M)).
  Proof.
    start_function.
    simpl.
    unfold MyIsTree; Intros root.
    sep_apply treeUniqueNodes; Intros.
    specialize (tree2context M value n root root Root H0 (eq_refl _)); intros X; simpl in X.
    sep_apply X; clear X; Intros. 
    forward_call (n,(valueOf value n)).
    forward.
    sep_apply context2tree; trivial.
    simpl. unfold MyIsTree. Exists (getroot (upsideDownMinus value n Root) n).
    entailer!. rewrite replaceItself; trivial.
  Qed. 

 Lemma fix_verif:
    semax_body Vprog CompGprog f_fix fix_spec.
  Proof.
    start_function.
    forward.
    forward_if (PROP (c <> 2147483647 )
      LOCAL (temp _tmp (Vint (Int.repr c)); temp _node node)
      SEP (field_at Ews nodeTP [StructField _count] (Vint (Int.repr c)) node; 
           MyContext M value node c)).
    { forward_call 1. entailer!. }
    { forward. entailer!. }
    Intros. rewrite MyContext_Fold.
    destruct value.
    + (*value=Root*)
      forward. forward. forward_if.
      - forward. rewrite MyContext_Fold. entailer!.
      - inv H2.
    + (*value = lcontext..*) Intros r c0. specialize (size_nonneg t); intros SZ.
      forward.
      forward.
      assert_PROP (isptr v) as Ptrv by entailer!.
      forward_if. congruence.
      forward_call (v, c0, value). (*{ split; [ trivial |]. split; try rep_lia. admit. }*)
      entailer. rewrite (MyContext_Fold M (lcontext v value t)).
      Exists r (c + 1 + size t + 1). entailer!.
   + (*value = rcontext..*) Intros l cc. 

      forward. forward.
      assert_PROP (isptr v) as Ptrv by entailer!.
      forward_if. congruence.
      forward_call (v, cc, value). (* { split; [ trivial |]. split; try rep_lia. admit. }*)
      entailer. rewrite (MyContext_Fold M (rcontext v t value)).
      Exists l (c + 1 + size t + 1). entailer!.
  Qed.

  Lemma internalCreate_verif:
    semax_body Vprog CompGprog f_internalCreate internalCreate_spec.
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (nodeTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros n. rewrite data_at__isptr; Intros.
    forward_if (PROP ( )
      LOCAL (temp _n n; gvars gv; temp _parent parent)
      SEP (mem_mgr M gv; malloc_token M Ews nodeTP n; data_at_ Ews nodeTP n)).
    { subst n; inv Pn. } 
    { forward. entailer!. }
    forward. forward. forward. forward. forward.
    unfold_data_at (data_at Ews nodeTP _ _).
    Exists n; entailer!. simpl. entailer!. Exists nullval nullval.
   rewrite if_true; trivial. entailer!. (*emp vs TT cancel.*)
  Qed.

  Lemma internalAddLeft_verif:
    semax_body Vprog CompGprog f_internalAddLeft internalAddLeft_spec.
  Proof.
    start_function.
    forward_call. Intros child. forward.
    assert_PROP (isptr node) as nodePtr by entailer!.
    forward_call (node, 1, value). { split; trivial.  rep_lia. }
    forward.
    Exists child. entailer!. apply derives_refl.
  Qed.

  Lemma create_verif:
    semax_body Vprog CompGprog f_create (create_spec M (MyCompositeAPD M)).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (nodeTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros n.
    assert_PROP (isptr n) as nPtr by entailer!.
    forward_if (PROP ( )
       LOCAL (temp _n n; gvars gv)
       SEP (mem_mgr M gv; malloc_token M Ews nodeTP n; data_at_ Ews nodeTP n)).
    { subst n; inv nPtr. }
    { forward. entailer!. }
    forward. forward. forward. forward. forward.
    Exists n; entailer!. simpl. unfold MyIsTree.
    unfold_data_at (data_at Ews nodeTP _ _). simpl.
    destruct (Val.eq n n); [  | elim n0; trivial].
    Exists n nullval nullval; entailer!. 
    destruct (Val.eq nullval nullval); [ trivial | elim n0; trivial]. (*cancel.*) (*emp vs TT *)
  Qed.

  Definition addLeft_specstrong :=
    DECLARE _addLeft
    WITH node:val, v:_, gv:globals
    PRE [ tptr nodeTP ] 
    PROP(valueOf v node = Tnode node Tnil Tnil) PARAMS (node) GLOBALS (gv)
      SEP(MyIsTree M node v; mem_mgr M gv)
    POST[ tptr nodeTP ]
    EX res: val,
     PROP()
     RETURN (res)
     SEP(MyIsTree M node(*res works*) (replace v node (Tnode node (Tnode res Tnil Tnil) Tnil)); mem_mgr M gv).

  Lemma addLeft_verifstrong:
    semax_body Vprog CompGprog f_addLeft (addLeft_specstrong).
  Proof.
    start_function.
    simpl. unfold MyIsTree; Intros root.
    sep_apply treeUniqueNodes; Intros. rename H2 into UN.  
    (*assert_PROP (exists x ltree rtree, v = Tnode x ltree rtree).
    {  rewrite MyTree_Fold. entailer.
       destruct v as [ | x ltree rtree]. entailer!. apply prop_right. do 3 eexists; reflexivity. (* apply semax_pre with FF. entailer!. apply semax_ff.*) }*)
    replace_SEP 1 (MyContext M Root root (size v))%logic by entailer.
    sep_apply (tree2context M v node root root Root); Intros.
    sep_apply MyTree_isptr; Intros.
    remember (valueOf v node) as q; destruct q. congruence. inv H.
    replace_SEP 0 (field_at Ews nodeTP [StructField _left] nullval node *
          field_at Ews nodeTP [StructField _right] nullval node *
          malloc_token M Ews nodeTP node *
          field_at Ews nodeTP [StructField _count] (Vint(Int.repr 1)) node)%logic.
    { entailer!. simpl. Intros ll rr.
      destruct (Val.eq rr nullval); Intros; subst; simpl.
      + destruct (Val.eq ll nullval); Intros; subst. cancel. entailer!.
      + entailer!. }
    Intros. simpl. forward_call (node, (upsideDownMinus v node Root), gv).
    Intros newChild. 
    assert_PROP (isptr newChild) as newChldPtr by (sep_apply MyTree_isptr; entailer!).

    forward. 

    Exists newChild. entailer!.
    eapply derives_trans with
    (MyContext M (upsideDownMinus v node Root) node 2 * 
        MyTree M (Tnode node (Tnode newChild Tnil Tnil) Tnil) node)%logic.
    { cancel. rewrite (MyTree_Fold M (Tnode node (Tnode newChild Tnil Tnil) Tnil)).
      Exists newChild nullval. entailer!. simpl; try rep_lia.
      rewrite if_false. rewrite if_true by trivial. entailer!. destruct newChild; inv newChldPtr. discriminate. }

    replace 2 with (1 + size (Tnode newChild Tnil Tnil)) by trivial. 
    forget (Tnode newChild Tnil Tnil) as childVAL.
    specialize ( context2tree (getroot(upsideDownMinus v node Root) node) node (upsideDownMinus v node Root) v (Tnode node childVAL Tnil) (eq_refl _) (eq_refl _) H1 UN). intros.
    replace (size (Tnode node childVAL Tnil))with  (1 + size childVAL)  in H11. 2: simpl; lia.
    sep_apply H11; clear H11. unfold MyIsTree.
    Exists (getroot (upsideDownMinus v node Root) node).
    entailer!. apply containsReplace. trivial. simpl.
    destruct (Val.eq node node); trivial. congruence.
  Qed.
 
  Lemma addLeft_verif:
    semax_body Vprog CompGprog f_addLeft (addLeft_spec M (MyCompositeAPD M)).
  Proof. eapply semax_body_funspec_sub. apply addLeft_verifstrong. 2: apply compute_list_norepet_e; cbv; trivial.
    do_funspec_sub. clear H. Exists w emp. destruct w as [[node v] gv]. entailer!.
   intros. Exists ( eval_id ret_temp x). entailer.
   unfold MyIsTree. Intros r. sep_apply treeUniqueNodes'. entailer!.
  Qed. 

  Definition CompositeComponent: @Component NullExtension.Espec Vprog _ 
      nil Comp_imported_specs prog Comp_ASI emp Comp_internal_specs.
  Proof. 
    mkComponent.
    + solve_SF_internal create_verif.
    + solve_SF_internal internalCreate_verif.
    + solve_SF_internal fix_verif.
    + solve_SF_internal internalAddLeft_verif.
    + solve_SF_internal internalGetNbOfNodes_verif.
    + solve_SF_internal addLeft_verif. 
    + solve_SF_internal getNbOfNodes_verif. 
  Qed.

  Definition CompositeVSU: @VSU NullExtension.Espec Vprog _ 
      nil Comp_imported_specs prog Comp_ASI emp.
  Proof. eexists; apply CompositeComponent. Qed.

End Composite_VSU.
