Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import VST.veric.initial_world.

Require Import VST.floyd.linking.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Connection.
Require Import spec_Connectionpool.
Require Import verif_Connection.
Require Import verif_Connectionpool.
Require Import verif_surelyMalloc.

Section Application_VSU.
Variable M: MallocFreeAPD.

Definition DM := D M.
Definition CM := C M.
Definition PM := P M.

(****** 1. A program unit containing Connection.c and Connectionpool.c ****)

Definition cp_prog': Clight.program := 
  match link_progs Connection.prog Connectionpool.prog with
    Errors.OK p => p
  | _ => Connection.prog (*arbitrary*)
  end.

Definition cp_prog := ltac:
  (let x := eval hnf in cp_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance cp_cs : compspecs. make_compspecs cp_prog. Defined.

Definition cp_Vprog:varspecs. mk_varspecs cp_prog. Defined.

Definition cp_Imports:funspecs := 
[free_spec' M Connection._free; surely_malloc_numericspec M].

Definition cp_Exports:funspecs := G_merge (P_ASI M DM CM) (C_ASI M).

Definition CP_VSU:
@VSU NullExtension.Espec
   cp_Vprog cp_cs nil cp_Imports cp_prog cp_Exports (NextConn Z0).
Proof.
  VSULink_tac (ConnectionpoolVSU M DM CM) (ConnectionVSU M).
  extensionality gv; simpl; rewrite emp_sepcon; trivial.
Qed.

(****** 2. Link surely_malloc to yield application ****)

Definition application_prog': Clight.program := 
  match link_progs surelyMalloc.prog cp_prog with
    Errors.OK p => p
  | _ => surelyMalloc.prog (*arbitrary*)
  end.

Definition application_prog := ltac:
  (let x := eval hnf in application_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance AppCs : compspecs. make_compspecs application_prog. Defined.

Definition AppVprog:varspecs. mk_varspecs application_prog. Defined.

Definition AppImports:funspecs := 
[free_spec' M Connection._free; malloc_spec' M surelyMalloc._malloc; exit_spec].

Definition AppExports:funspecs := G_merge cp_Exports (SM_ASI M).

Definition Application_VSU:
@VSU NullExtension.Espec
   AppVprog AppCs nil AppImports application_prog AppExports (NextConn Z0).
Proof. 
  VSULink_tac CP_VSU (SMVSU M). 
  extensionality gv; simpl; rewrite sepcon_emp; trivial.
Qed.

End Application_VSU.