Require Import VST.floyd.VSU.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import Connection. (*needed for defining ConnCompSpecs and Vprog *)
Require Import spec_Connection.

Instance ConnCompSpecs : compspecs. make_compspecs prog. Defined.

Section Connection_VSU.
  Variable M: MallocFreeAPD.

  Definition DB (i:DBIndex) (d:val): mpred :=
   (malloc_token M Ews DatabaseTP d * 
   data_at Ews DatabaseTP (Vint (Int.repr i)) d)%logic.

  Lemma DB_ptrnull i d: DB i d |-- !!(is_pointer_or_null d) && DB i d.
  Proof. unfold DB; entailer!. Qed.

  Definition D: DatabaseAPD := @Build_DatabaseAPD _ DB_ptrnull.

  Definition ConnModuleState := Z.
  Notation ConnIndex := ConnModuleState.

  Notation Connection := (val * ConnIndex)%type.

  Definition Conn (C:Connection) (s: val): mpred :=
  match C with
    ((*dbI*)d,cI) =>  !!(Int.min_signed <= cI < Int.max_signed) &&
             (malloc_token M Ews ConnectionTP s *
             @data_at ConnCompSpecs Ews ConnectionTP
                  (d, Vint (Int.repr cI)) s)%logic
  end.

  Definition NextConn (cI:ConnIndex) (gv: globals): mpred :=
   !!(Int.min_signed <= cI <= Int.max_signed /\ headptr (gv _nextID)) &&
   data_at Ews tint (Vint (Int.repr cI)) (gv _nextID).

  Lemma Conn_ptrnull C c: Conn C c |-- !!(is_pointer_or_null c) && Conn C c.
  Proof. destruct C as [v i]; simpl; entailer!. Qed.
 
  Lemma Conn_isptr C c: Conn C c |-- !!(isptr c).
  Proof. destruct C as [v i]; simpl; entailer!. Qed.

  Lemma Conn_validpointer: forall C c, Conn C c |-- valid_pointer c.
  Proof. destruct C as [v i]; simpl; entailer!. Qed.

  Lemma Conn_range C c: Conn C c |-- !!(Int.min_signed <= (snd C) < Int.max_signed).
  Proof. destruct C as [v i]; simpl; Intros. entailer!. Qed.
  
  Lemma NextConn_range i gv: NextConn i gv |-- !!(Int.min_signed <= i <= Int.max_signed).
  Proof. unfold NextConn. entailer!. Qed.

  Definition ConnVprog : varspecs. mk_varspecs prog. Defined.

  Lemma ConnInit_old: forall i t, In (i,t) ConnVprog -> exists (v:@reptype ConnCompSpecs t) (P:globals -> mpred),
            forall gv, !!(headptr (gv i)) && @data_at ConnCompSpecs Ews t v (gv i) |-- P gv.
  Proof. intros. destruct H; [inv H | contradiction].
    exists (Vint (Int.repr 0)). exists (EX cI, NextConn cI). intros. simpl. Exists Z0.
    unfold ConnIndex, NextConn; entailer!.
  Qed.

  Definition C := Build_ConnectionAPD ConnModuleState Conn NextConn Conn_isptr Conn_validpointer.

  Definition C_ASI: funspecs := ConnectionASI M D C.

  Definition conn_imported_specs: funspecs := SM_ASI M.

  Definition conn_internal_specs: funspecs := C_ASI.

  Definition ConnGprog:funspecs := conn_imported_specs ++ conn_internal_specs.

  Lemma consConn_verif:
    semax_body ConnVprog ConnGprog f_consConn (consConn_spec M D C).
  Proof.
    start_function.
    unfold C; simpl. unfold NextConn; Intros. forward. 
    assert(Int.max_signed = 2147483647) by reflexivity. 
    forward_if.
    + forward. rewrite H0 in *.
      assert (cI = 2147483647) by lia. subst cI.
      Exists nullval. rewrite if_true by trivial. entailer!.
      unfold D; simpl.
      unfold NextConn; entailer!.
    + forward_call surelymalloc_spec_sub (ConnectionTP, gv).
      { simpl. split3; trivial. rep_lia. }
      Intros c.
      assert_PROP (isptr c) as isPtrC by entailer!.
      forward. simpl; unfold NextConn.
      do 5 forward. Exists c.
      rewrite if_false by (intros X; subst; elim isPtrC).
      unfold C; simpl; entailer!.
      destruct (zle Int.max_signed cI). lia.
      Exists (cI+1). unfold NextConn; entailer!.
  Qed.

  Lemma newDB_verif:
    semax_body ConnVprog ConnGprog f_newDB (newDB_spec M D).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (DatabaseTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros c. forward. forward.
    Exists c. entailer!. apply derives_refl.
  Qed.
(*
  Lemma MyInitData gv (H: headptr (gv _nextID)):
        initialize.gv_globvar2pred  gv (_nextID, v_nextID) gv
        = data_at Ews tuint (Vint (Int.repr 0)) (gv _nextID).
  Proof. unfold initialize.gv_globvar2pred. simpl.
         unfold initialize.gv_lift2, initialize.gv_lift0; simpl.
         rewrite predicates_sl.sepcon_emp.
         apply mapsto_data_at''; trivial.
  Qed.*)

  Lemma create_NextConn0 gv: InitGPred (Vardefs Connection.prog) gv |-- NextConn 0 gv.
  Proof. unfold Vardefs. simpl. unfold InitGPred. simpl; Intros. rewrite sepcon_emp.
    unfold NextConn, globvar2pred; simpl; entailer!. 
    rewrite <- data_at_tuint_tint; erewrite <- mapsto_data_at''.
    apply derives_refl. cbv; trivial. trivial. apply JMeq_refl.
  Qed.

  Lemma create_NextConn gv: InitGPred (Vardefs Connection.prog) gv |-- EX i, NextConn i gv.
  Proof. sep_apply create_NextConn0. Exists 0; trivial. Qed.

  Definition ConnectionComponent: @Component NullExtension.Espec ConnVprog _ 
      nil conn_imported_specs prog C_ASI (NextConn Z0) conn_internal_specs.
  Proof. 
    mkComponent.
    + solve_SF_internal consConn_verif.
    + solve_SF_internal newDB_verif.
    + intros. apply create_NextConn0. 
      (*unfold NextConn, InitGPred; simpl; Intros. rewrite <- data_at_tuint_tint. 
      rewrite MyInitData; trivial. entailer!.*)
  Qed.

  Definition ConnectionVSU: @VSU NullExtension.Espec ConnVprog _ 
      nil conn_imported_specs prog C_ASI (NextConn Z0).
  Proof. eexists; apply ConnectionComponent. Qed.

  Definition ConnectionCanVSU := VSU_to_CanonicalVSU ConnectionVSU.

End Connection_VSU.
