Require Import VST.floyd.VSU.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Connection. (*Connectionpool.c calls a function exposed in Connection.h, namely consConn,
 so it's we obviously need access to the spec of this function*)

Require Import Connectionpool. (*needed for defining PoolCompSpecs and Vprog *)
Require Import spec_Connectionpool.

Instance PoolCompSpecs : compspecs. make_compspecs prog. Defined.

Notation NodeTP := (Tstruct _node noattr).

Section Connectionpool_VSU.
  Variable M: MallocFreeAPD.
  Variable D: DatabaseAPD.
  Variable C: ConnectionAPD.

  Fixpoint listrep db (l: list val) m: mpred :=
  match l with
    nil => !!(m=nullval) && emp
  | o::os => EX n cI, !!(isptr o) &&
                   @data_at PoolCompSpecs Ews NodeTP (o,n) m *
                   @malloc_token PoolCompSpecs M Ews NodeTP m *
                   Conn C (db,cI) o *
                   listrep db os n
  end.

  Lemma listrep_ptrnull d l s: listrep d l s = !!(is_pointer_or_null s) && listrep d l s.
  Proof.
    destruct l; simpl; apply pred_ext; try solve [entailer!].
    Intros n cI. Exists n cI. entailer!. 
    apply andp_left2; trivial. 
  Qed.

  Lemma listrep_valid_pointer d l s: listrep d l s |-- valid_pointer s. 
  Proof.
    destruct l; simpl. Intros; subst. entailer.
    Intros n cI. do 2 apply sepcon_valid_pointer1.
     apply sepcon_valid_pointer2. apply malloc_token_valid_pointer. 
  Qed.

  Definition CPool (d:val)(p:val): mpred :=
     EX l m, @data_at PoolCompSpecs Ews PoolTP (m,d) p *
               @malloc_token PoolCompSpecs M Ews PoolTP p *
               listrep d l m.

  Lemma CPool_ptrnull i p: CPool i p |-- !!(is_pointer_or_null p).
  Proof. unfold CPool. Intros l m. (*Exists l m.*) entailer!. Qed.

  Definition P := Build_PoolAPD _ CPool_ptrnull.
  
  Definition P_ASI: funspecs := PoolASI M D C P.

  Definition PoolVprog : varspecs. mk_varspecs prog. Defined.

  Definition pool_imported_specs := [free_spec' M _free] ++ SM_ASI M ++ [consConn_spec M D C].

  Definition pool_internal_specs: funspecs := P_ASI.

  Definition PoolGprog:funspecs := pool_imported_specs ++ pool_internal_specs.

  Lemma consPool_verif:
    semax_body PoolVprog PoolGprog f_consPool (consPool_spec M (*D*) P).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (PoolTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros p.
    forward. simpl. do 2 forward.
    Exists p. unfold P; simpl. entailer!. unfold CPool.
    Exists (@nil val) nullval. simpl. entailer!.
  Qed.

Lemma valid_pointer_is_null_or_ptrr p: valid_pointer p |-- !!( is_pointer_or_null p).
Proof. apply valid_pointer_is_pointer_or_null. Qed.

  Lemma getConn_verif:
    semax_body PoolVprog PoolGprog f_getConn (getConn_spec M D C P).
  Proof.
    start_function. (*rename H into Hd.*) unfold P. simpl.
    unfold CPool; Intros l n. rewrite listrep_ptrnull; Intros.
    forward.
    forward_if.
    + sep_apply listrep_valid_pointer; entailer!.
    + subst n. forward. sep_apply DB_ptrnull; entailer!.
      forward_call.
      Intros c.
      if_tac.
      - forward. Exists nullval. simpl. rewrite if_true by trivial. unfold P; entailer!.
        unfold CPool. Exists l nullval. cancel.
      - forward. Exists c. rewrite if_false by trivial. entailer!.
        Exists cI' cI. cancel. unfold P, CPool; simpl.
        Exists l nullval. cancel.
    + destruct l; simpl. Intros; subst; inv H.
      Intros n' cI'. rewrite listrep_ptrnull; Intros.
      do 2 forward.
      unfold malloc_token; Intros.
      forward_call (sizeof NodeTP, n, gv).
      { rewrite if_false by (intros N; subst; inv H).
        sep_apply (data_at_memory_block Ews NodeTP); cancel. }
      do 2 forward.
      Exists v; entailer!. 
      rewrite if_false by (intros N; subst; inv Pv).
      Exists cI cI'. entailer!.
      simpl. unfold CPool, malloc_token.
      Exists l n'. entailer!.
  Qed.

  Lemma freeConn_verif:
    semax_body PoolVprog PoolGprog f_freeConn (freeConn_spec M (*D*) C P).
  Proof.
    start_function.
    forward_call surelymalloc_spec_sub (NodeTP, gv).
    { simpl. split3; trivial. rep_lia. }
    Intros n. simpl. unfold CPool; Intros l m.
    rewrite listrep_ptrnull; Intros.
    do 5 forward. simpl. cancel.
    unfold CPool. Exists (c::l) n. cancel. simpl.
    Exists m cI. entailer. sep_apply Conn_isptr. entailer!.
  Qed.

  Definition ConnectionpoolComponent: @Component NullExtension.Espec PoolVprog _ 
      nil pool_imported_specs prog P_ASI emp pool_internal_specs.
  Proof.
    mkComponent (*pool_internal_specs*).
  + solve_SF_internal consPool_verif.
  + solve_SF_internal getConn_verif.
  + solve_SF_internal freeConn_verif.
  Qed.

  Definition ConnectionpoolVSU: @VSU NullExtension.Espec PoolVprog _ 
      nil pool_imported_specs prog P_ASI emp.
  Proof. eexists; apply ConnectionpoolComponent. Qed.

  Definition CPoolCanVSU := VSU_to_CanonicalVSU ConnectionpoolVSU.

End Connectionpool_VSU.
