(*Constructs a @CanonicalVSU consisting of the mmodules Connection, Connectionpool,
  surelymalloc, and stdlib. Explanation of linking with main, and how to obtain
  a semax_prog result are in verif_linked_prog.v*)
Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import VST.veric.initial_world.

Require Import VST.floyd.linking.
Require Import spec_stdlib.
Require Import spec_surelyMalloc.

Require Import spec_Connection.
Require Import spec_Connectionpool.
Require Import verif_Connection.
Require Import verif_Connectionpool.
Require Import verif_surelyMalloc.

Require Import verif_stdlib.

Section CORE_VSU.
Variable M: MallocFreeAPD.

Definition DM := D M.
Definition CM := C M.
Definition PM := P M. 

Definition pool_conn_prog': Clight.program := 
  match link_progs Connectionpool.prog Connection.prog with
    Errors.OK p => p
  | _ => Connectionpool.prog (*arbitrary*)
  end.
Definition pool_conn_prog := ltac:
  (let x := eval hnf in pool_conn_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).
Instance pool_conn_cs : compspecs. make_compspecs pool_conn_prog. Defined.

Definition pool_conn_Vprog:varspecs. mk_varspecs pool_conn_prog. Defined.

Definition pool_conn_Imports:funspecs := 
[free_spec' M Connectionpool._free; surely_malloc_numericspec M].

Definition pool_conn_Exports:funspecs := 
G_merge (P_ASI M DM CM) (C_ASI M).

Definition PoolConn_VSU:
@VSU NullExtension.Espec
   pool_conn_Vprog pool_conn_cs nil pool_conn_Imports pool_conn_prog pool_conn_Exports (NextConn Z0).
Proof.
  VSULink_tac (ConnectionpoolVSU M DM CM) (ConnectionVSU M).
  extensionality gv; simpl; rewrite emp_sepcon; trivial.
Qed.

Definition sm_poolconn_prog': Clight.program := 
  match link_progs surelyMalloc.prog pool_conn_prog with
    Errors.OK p => p
  | _ => surelyMalloc.prog (*arbitrary*)
  end.
Definition sm_poolconn_prog := ltac:
  (let x := eval hnf in sm_poolconn_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance sm_poolconn_cs : compspecs. make_compspecs sm_poolconn_prog. Defined.

Definition sm_poolconn_Vprog:varspecs. mk_varspecs sm_poolconn_prog. Defined.

Definition sm_poolconn_Imports:funspecs :=
[free_spec' M Connectionpool._free; malloc_spec' M Connection._malloc; exit_spec].

Definition sm_poolconn_Exports:funspecs := G_merge pool_conn_Exports (SM_ASI M).

Definition SM_PoolConn_VSU:
@VSU NullExtension.Espec
   sm_poolconn_Vprog sm_poolconn_cs nil sm_poolconn_Imports sm_poolconn_prog sm_poolconn_Exports (NextConn Z0).
Proof.
  VSULink_tac PoolConn_VSU (SMVSU M).
  + extensionality gv; simpl; rewrite sepcon_emp; trivial.
Qed.

Definition core_prog': Clight.program := 
  match link_progs stdlib.prog sm_poolconn_prog with
    Errors.OK p => p
  | _ => stdlib.prog (*arbitrary*)
  end.
Definition coreprog := ltac:
  (let x := eval hnf in core_prog' in
   let x := eval simpl in x in 
   let x := eval compute in x in 
       exact x).

Instance coreCS : compspecs. make_compspecs coreprog. Defined.

Definition coreVprog:varspecs. mk_varspecs coreprog. Defined.

Definition coreImports:funspecs := Malloc_ASI M malloc._malloc malloc._free.

Definition coreExports:funspecs := G_merge [exit_spec] sm_poolconn_Exports.
Definition coreBuiltins:funspecs := [exit_spec].

Definition Core_VSU: @VSU NullExtension.Espec coreVprog 
           coreCS coreBuiltins coreImports coreprog coreExports (NextConn Z0).
Proof.
  VSULink_tac (Stdlib_VSU M) SM_PoolConn_VSU.
  extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

Definition Core_CanVSU: @CanonicalVSU NullExtension.Espec coreVprog 
           coreCS coreBuiltins coreImports coreprog coreExports (NextConn Z0).
Proof. apply (VSU_to_CanonicalVSU Core_VSU). Qed.

End CORE_VSU.
