Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*For MallocFreeAPD.*)
Require Import spec_Connection.
Require Import spec_Connectionpool.
Require Import main.

(*We verify the body of main abstractly in M, C and P.
  We just provide the semax_body statement, no VSU.
  The spec of main is not VST standard spec, but a variant
  that can be instantiated to contain the appropriate initialization
  predidates, given suitable implementations (ie VSUs) of the code modules
  and the library.*)
Section MAIN_Verif.
  Variable M: MallocFreeAPD.
  Variable C: ConnectionAPD.
  Variable D: DatabaseAPD.
  Variable P: PoolAPD.

  (*Specification of "module-instantiated main" - we permit an arbitrary G here too.*)
  Definition main_inst_spec G:=
  DECLARE _main
  WITH gv: globals
  PRE [ ]
    PROP ()
    PARAMS () GLOBALS (gv)
    SEP (fold_right (fun g p => g gv * p) emp G; has_ext tt)
  POST[ tint ]
    PROP()
    LOCAL(temp ret_temp (Vint (Int.repr 0)))
    SEP((*mem_mgr M gv; *)TT;has_ext tt).

  Definition MainImports := [ newDB_spec M D; consPool_spec M P; 
   getConn_spec M D C P; freeConn_spec M C P; spec_stdlib.exit_spec].

  Definition MainInitPred: list (globals -> mpred) := [mem_mgr M; EX cI, NextConn C cI].
  Definition maininstspec := main_inst_spec MainInitPred.

  (*Definition main_internal_specs: funspecs := [mainspec (*M*)].*)

  Definition MainVprog : varspecs. mk_varspecs main.prog. Defined.
  Definition MainGprog:funspecs := MainImports(* ++ main_internal_specs*).

  Instance MainCompSpecs : compspecs. make_compspecs main.prog. Defined.
  Lemma body_main: semax_body MainVprog MainGprog f_main maininstspec.
  Proof.
  start_function. simpl; Intros.
  (*sep_apply (make_mem_mgr M gv).*)

  (*sep_apply (ConnInit_charGV C);*) Intros CMS1.
  forward_call; Intros db.
  forward_call; Intros p.
  forward_call (p, db, 1, CMS1, gv); Intros c1.
  if_tac; subst.
  { (*c1=nullval*)
    forward_if False.
    - forward_call 1. entailer!.
    - elim H; trivial. }
  Intros CMS2 CONN1. clear CMS1.
  forward_if True. contradiction. forward. entailer!.
  forward_call; Intros c2.
  if_tac; subst.
  { (*c2=nullval*)
    forward_call (p, c1, db, 1, CONN1, gv).
    forward_call; Intros c3.
    if_tac; subst.
    - forward. cancel.
    - Intros CMS3 CONN2. forward. cancel. }
  Intros CMS3 CONN2.
  forward_call (p, c1, db, 1, CONN1, gv).
  forward_call; Intros c3.
  if_tac; subst.
  - forward. cancel.
  - Intros COS8 CMS9. forward. cancel.
  Qed.

End MAIN_Verif.
(*
  (*Definition main_internal_specs: funspecs := [mainspec M].*)
  (*We instantiate main_inst_spec with M, and the (abstract) initial module predicate for the (abstract) C*) 
  (*Definition main_internal_specs: funspecs := [main_inst_spec M [EX cI, NextConn C cI]].*)
  Definition main_internal_specs: funspecs := [main_spec (*M*) linked_prog].
  Definition MainGprog:funspecs := MainImports ++ main_internal_specs.

  Definition MyInitPred: list (globals -> mpred) := [mem_mgr M; EX cI, NextConn C cI].
  Lemma body_main: semax_body LinkedVprog MainGprog f_main (main_inst_spec (*M*) MyInitPred).
  Proof.
  start_function. simpl.
  (*sep_apply (make_mem_mgr M gv).*)

  (*sep_apply (ConnInit_charGV C);*) Intros CMS1.
  forward_call; Intros db.
  forward_call; Intros p.
  forward_call (p, db, 1, CMS1, gv); Intros c1.
  if_tac; subst.
  { (*c1=nullval*)
    forward_if False.
    - forward_call 1. entailer!.
    - inv H. }
  Intros CMS2 CONN1. clear CMS1.
  forward_if True. contradiction. forward. entailer!.
  forward_call; Intros c2.
  if_tac; subst.
  { (*c2=nullval*)
    forward_call (p, c1, db, 1, CONN1, gv).
    forward_call; Intros c3.
    if_tac; subst.
    - forward. cancel.
    - Intros CMS3 CONN2. forward. cancel. }
  Intros CMS3 CONN2.
  forward_call (p, c1, db, 1, CONN1, gv).
  forward_call; Intros c3.
  if_tac; subst.
  - forward. cancel.
  - Intros COS8 CMS9. forward. cancel.
Qed.
(*
  (*Redundant*)
  Definition MainComponent: @Component NullExtension.Espec LinkedVprog LinkedCompSpecs (*MainCompSpecs*)
        nil MainImports main.prog [(*mainspec M*)main_inst_spec M (*[ConnInit C]*)[EX cI, NextConn C cI]] (NextConn C Z0) main_internal_specs.
  Proof. 
    mkComponent. clear; solve_SF_internal body_main.
  Qed.*)

End MAIN.
Section MAIN_Verif.
  Variable M: MallocFreeAPD.
  Variable OP: ObserverAPD.
  Variable SP: SubjectAPD.
  Definition S_ASI:= SubjectASI M SP OP.
  Definition O_ASI:= ObserverASI M OP SP.

  (*Specification of "module-instantiated main" - we permit an arbitrary G here too.*)
  Definition main_inst_spec G:=
  DECLARE _main
  WITH gv: globals
  PRE [ ]
    PROP ()
    PARAMS () GLOBALS (gv)
    SEP (fold_right (fun g p => g gv * p) emp G; has_ext tt)
  POST[ tint ]
    PROP()
    LOCAL(temp ret_temp (Vint (Int.repr 6)))
    SEP((*mem_mgr M gv; *)TT;has_ext tt).

  Definition MainImports := [ newObserver_spec M OP SP; val_spec OP; 
   freeObserver_spec M OP; newSubject_spec M SP; update_spec SP OP;
   detachfirst_spec M SP; freeSubject_spec M SP].

  Definition MainInitPred: list (globals -> mpred) := nil.
  Definition maininstspec := (main_inst_spec [mem_mgr M]).

  (*Definition main_internal_specs: funspecs := [mainspec (*M*)].*)

  Definition MainVprog : varspecs. mk_varspecs main.prog. Defined.
  Definition MainGprog:funspecs := MainImports(* ++ main_internal_specs*).

  Instance MainCompSpecs : compspecs. make_compspecs main.prog. Defined.
  Lemma body_main: semax_body MainVprog MainGprog f_main maininstspec.
  Proof.
  start_function. simpl; Intros. 
  (*sep_apply (spec_stdlib.make_mem_mgr M gv). *)
  forward_call. Intros s. simpl. (*nothing happens as SD is abstract*)
  sep_apply Subject_Register. 
  forward_call (gv,s, newSubject). Intros o1.
  simpl. (*only the definitions of newSubject and Subjectrep (both model functions!) are simplified*) 
  sep_apply Subject_Update.
  sep_apply Observer_Notify.
  forward_call (s, 2, o1::fst newSubject). 
  { Exists (snd newSubject). simpl. cancel.
    Exists [0].
    simpl. unfold newObserver; simpl. cancel. }

  simpl. (*again just the definition of newSubject is simplified*)
  sep_apply Subject_Register. 
  forward_call (gv,s, ([o1], 2)). Intros o2. simpl.

  sep_apply Subject_Update.
  sep_apply Observer_Notify.
  sep_apply Observer_Notify.
  forward_call (s, 3, [o2;o1]). 
  { Exists 2. cancel.
    Exists [2;2]. 
    simpl. unfold newObserver; simpl. cancel. }

  simpl. Intros.
  sep_apply Observer_Val.
  sep_apply Observer_Val.
  forward_call (o1,s,3).
  forward_call (o2,s,3).
  forward.
  forward_call. Intros p1. simpl. if_tac; Intros z. inv H0; deadvars.
  forward_call. Intros p2. simpl. if_tac; Intros z. inv H1; deadvars.
  forward_call; Intros v. simpl. if_tac; Intros; [ subst; simpl | congruence].
  subst; simpl. 
  forward_call (gv, p1, (s,3)).
  forward_call (gv, p2, (s,3)).
  forward.
  cancel.
  Qed.

End MAIN_Verif.*)