Require Import VST.floyd.proofauto.
Require Import VST.floyd.VSU.
Require Import ASI_malloc.
Require Import main_R.

Require Import ASI_mmap0.

(*We still verify the body of main abstractly in M*)
Section MAIN.
  Variable M: MallocFree_R_APD.

  Definition MainImports := ASI_mmap0.mmap0_spec _mmap0 :: Malloc_R_ASI M  _pre_fill _try_pre_fill _malloc _free. 

  Definition main_inst_spec G:=
   DECLARE _main
   WITH gv: globals
   PRE [ ]
     PROP ()
     PARAMS () GLOBALS (gv)
     SEP (fold_right (fun g p => g gv * p) emp G; has_ext tt)
   POST[ tint ]
     PROP()
     LOCAL(temp ret_temp (Vint (Int.repr 0)))
     SEP(TT;has_ext tt).

  Definition maininstspec := (main_inst_spec [mem_mgr_R M malloc_lemmas.emptyResvec]).

  Instance CompSpecs : compspecs. make_compspecs main_R.prog. Defined.

  Definition Vprog : varspecs. mk_varspecs main_R.prog. Defined.

  Definition Gprog:funspecs := MainImports.

  Lemma body_main: semax_body Vprog Gprog f_main maininstspec.
  Proof.
  start_function. simpl.
  change 8 with malloc_lemmas.BINS.

  forward_call malloc_lemmas.BIGBLOCK.  (*! *m = mmap0(BIGBLOCK ...) !*)  
  rep_lia.
  Intros m.
  forward_if.
  { if_tac; entailer!. }
  + (* m = null *)
    forward. (* return *)
    cancel.
  + (* m <> null *)
    if_tac. contradiction.
    forward_call(50,m,gv,malloc_lemmas.emptyResvec). (* pre_fill(50,m) *)
    split.  rep_lia. auto.
    set (rvec:=malloc_lemmas.add_resvec malloc_lemmas.emptyResvec (malloc_lemmas.size2binZ 50) (malloc_lemmas.chunks_from_block (malloc_lemmas.size2binZ 50))).
    forward_call(50,gv,rvec). (* t1 = malloc(50) --- using malloc_spec_R *)
    rep_lia.
    Intros p.
    change (malloc_lemmas.guaranteed rvec 50) with true.
    cbv iota.
    assert_PROP (p <> nullval) by entailer!.
    forward_call(50,p,gv,(malloc_lemmas.add_resvec rvec (malloc_lemmas.size2binZ 50) (-1))). (* free(p) *)
    rewrite if_false by auto.
    entailer!.
    rewrite if_false by auto.
    forward. simple_if_tac; cancel.
  Qed.
End MAIN.

Require Import VST.veric.initial_world.
Require Import VSU_malloc_definitions.

Require Import MM_prog.
Require Import VSU_MM. (*for mm_prog*)
Definition linked_prog : Clight.program :=
 ltac: (linking.link_progs_list [ mm_prog; main_R.prog]).

Definition LinkedSYS_pre:funspecs :=
   filter (fun x => in_dec ident_eq (fst x) (ExtIDs linked_prog)) (augment_funspecs linked_prog VSU_mmap0.SYS).
  (* Holds but dead code in this example*)
  Lemma coreSYS_in_LinkedSYS: forall i phi, find_id i VSU_mmap0.SYS = Some phi -> find_id i LinkedSYS_pre = Some phi.
  Proof. intros. specialize (find_id_In_map_fst _ _ _ H); intros.
    simpl in H0. repeat (destruct H0 as [HO | H0]; [ subst i; inv H; reflexivity |]).
    contradiction.
  Qed.
(*Eval compute in (map fst LinkedSYS_pre). contains 52:mmap and 53: munmap but not 61:mmap0*)

Definition LinkedSYS:funspecs := ltac:
    (let x := eval hnf in LinkedSYS_pre in
     let x := eval simpl in x in 
(*     let x := eval compute in x in *)
       exact x). (*Takes 30s to compute...*)

Lemma LinkedSYS_External: forall i : ident, In i (map fst LinkedSYS) ->
         exists (ef : external_function) (ts : typelist) (t : type) (cc : calling_convention),
           find_id i (prog_defs linked_prog) = Some (Gfun (External ef ts t cc)).
  Proof. intros.
    cbv in H. 
    repeat (destruct H as [H | H];
      [ subst; try solve [do 4 eexists; split; reflexivity ]
      | ]).
    contradiction.
  Qed.

Lemma LinkedSYS_vacuous i phi: find_id i LinkedSYS = Some phi -> find_id i VSU_mmap0.SYS = None ->
        exists ef argsig retsig cc, 
           phi = vacuous_funspec (External ef argsig retsig cc) /\ 
           find_id i (prog_funct mm_prog) = Some (External ef argsig retsig cc) /\
           ef_sig ef = {| sig_args := typlist_of_typelist argsig;
                          sig_res := rettype_of_type retsig;
                          sig_cc := cc_of_fundef (External ef argsig retsig cc) |}.
  Proof. 
    intros. unfold find_id in H. simpl in H. 
    (*Time*) repeat (if_tac in H;
    [ inv H; try inv H0; try solve [do 4 eexists; split3; reflexivity]
    | clear H1 ]); discriminate. (*13s -- was 3s*)
  Qed.

Definition main_spec:=
 DECLARE _main
 WITH gv: globals
 PRE [ ] main_pre linked_prog tt gv
 POST[ tint ]
    PROP()
    LOCAL(temp ret_temp (Vint (Int.repr 0)))
    SEP(TT;has_ext tt).

(*Finally, we choose a concrete APD*)
Definition M: MallocFree_R_APD := R_APD.

Definition LinkedVprog : varspecs. mk_varspecs linked_prog. Defined.

Lemma tc_VG: tycontext_subVG Vprog (Gprog M) LinkedVprog (main_spec :: mm_R_Exports).
Proof. split.
  * intros i. red. rewrite 2 semax_prog.make_context_g_char, 2 make_tycontext_s_find_id by LNR_tac.
    remember (find_id i (Gprog M)) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; reflexivity |]).
    congruence.
  * intros i; red. rewrite 2 make_tycontext_s_find_id.
    remember (find_id i (Gprog M)) as w.
    destruct w; [symmetry in Heqw | trivial]. simpl in Heqw.
    repeat (if_tac in Heqw;
            [ subst i; inv Heqw; eexists; split; [ reflexivity | apply funspec_sub_si_refl]
            | ]).
    congruence.
Qed.

Lemma disjoint_Vprog_linkedfuncts: 
      list_disjoint (map fst LinkedVprog) (map fst (prog_funct linked_prog)).
Proof.
  intros x y X Y ?; subst x; cbv in X; apply assoclists.find_id_None_iff in Y; [ trivial | clear H Y];
  repeat (destruct X as [X | X]; [ subst y; cbv; reflexivity |]); contradiction.
Qed.

Definition Imports:funspecs:=nil.

Lemma main_sub: funspec_sub (snd (main_inst_spec [mem_mgr_R M malloc_lemmas.emptyResvec]))
                            (snd main_spec).
Proof. do_funspec_sub. unfold main_pre; simpl; Intros; subst. clear. 
  Exists (initialize.genviron2globals g) emp. entailer!.
  + intros; entailer!.
  + specialize (globals_ok_genviron2globals g); intros G. 
    eapply derives_trans.
    2: apply (MkInitPred_of_CanonicalVSU MM_R_CanVSU); trivial.
    unfold InitGPred; simpl. entailer. unfold globvar2pred; simpl.
    sep_apply mapsto_zeros_isptr; entailer!.
Qed.

Require Import VST.floyd.VSU_addmain.

Instance LinkedCompSpecs : compspecs. make_compspecs linked_prog. Defined.

Definition Main_VSU: @LinkedProgVSU NullExtension.Espec LinkedVprog LinkedCompSpecs
      LinkedSYS Imports linked_prog [main_spec] (mem_mgr_R M malloc_lemmas.emptyResvec).
Proof.
 VSUAddMain_tac MM_R_CanVSU.
   + apply disjoint_Vprog_linkedfuncts.
   + apply LinkedSYS_External.
   + eapply semax_body_subsumption.
       * eapply semax_body_funspec_sub. 
         - apply (body_main M).
         - apply main_sub.
         - LNR_tac.
       * apply tycontext_sub_i99. apply tc_VG.
   + apply LinkedSYS_vacuous.
Qed.

Lemma prog_correct:
  exists G, 
 @semax_prog NullExtension.Espec LinkedCompSpecs linked_prog tt LinkedVprog G.
Proof.
(*boilerplate: same proof script as in VSUpile, connpool*)
  destruct Main_VSU as [G Comp MAIN]. exists G. 
  assert (DomG: map fst G = map fst (prog_funct linked_prog)).
  { destruct Comp. unfold Comp_G in *. rewrite CC_canonical.
    cbv; reflexivity. }
  Time prove_linked_semax_prog.
  all: rewrite augment_funspecs_eq by trivial.
  apply (@Canonical_semax_func _ _ _ _ _ _ _ _ Comp); cbv; reflexivity.
  destruct MAIN as [post [MainG MainExp]]. inv MainExp. rewrite MainG; eexists; reflexivity.
Qed.