Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import PositiveSubtypingTheory.

Require Import SimpleObjspec.
Require Import SimplePointModel.

Ltac do_compute_expr_warning::=idtac.

Global Open Scope funspec_scope.
Section Spec.
Variable point bpoint cpoint:type.
Variable T: Type. (*Fix/be parametric in a semantic representation, ie order and (Coq)types of record fields*)
Variable P:PointM T. (*For the given representation T, we assume model-level implementations of the exposed methods*)

  (*Parameter PR is the representation predicate that relates a concrete C struct to its model-level counterpart (an element of T*)
  Definition get_spec (PR: Pred T) :=
    WITH x:T, p: val
    PRE [ tptr point ]
     PROP()
     PARAMS (p) GLOBALS ()
     SEP(PR (x,p))
    POST[ tint ]
      PROP()
      RETURN (Vint (Int.repr (get _ P x)))
      SEP(PR (x,p)).

  Definition set_spec (PR: Pred T) :=
    WITH x:T, p: val, j:Z
    PRE [ tptr point, tint   ]
     PROP(Int.min_signed <= j <= Int.max_signed)
     PARAMS (p; Vint (Int.repr j)) GLOBALS ()
     SEP(PR (x,p))
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(PR (set _ P x j, p)).

Variable BP:BPointM T.
  Definition bump_spec  (PR: Pred T) :=
    WITH x:T, p:val
    PRE [ tptr bpoint ]
     PROP(bumpable _ BP x)
     PARAMS (p) GLOBALS ()
     SEP(PR (x,p))
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(PR (bump _ BP x, p)).

Variable CP:CPointM T.

  Definition getC_spec (PR: Pred T) :=
    WITH x:T, p: val
    PRE [ tptr cpoint ]
     PROP()
     PARAMS (p) GLOBALS ()
     SEP(PR (x,p))
    POST[ tint ]
      PROP()
      RETURN (Vint (Int.repr (color_encoding _ CP (getC _ CP x))))
      SEP(PR (x,p)).
End Spec.

Definition PointSpecs point T (P: PointM T):= [get_spec point T P; set_spec point T P].
Definition BPointSpecs point bpoint T (BP: BPointM T):= [get_spec point T BP; set_spec point T BP; bump_spec bpoint T BP].
Definition CPointSpecs point bpoint cpoint T (CP: CPointM T):= 
       [get_spec point T CP; set_spec point T CP; bump_spec bpoint T CP; getC_spec cpoint T CP].

Section PointPreds.
Variable point bpoint cpoint:type.
Variable methods bmethods cmethods:type.
Variable mtable g s b gC:ident.
Variable cs: compspecs.
Variable M:MallocFreeAPD.
Variable T: Type.

Definition Point (PM: PointM T) : Pred T:= 
  obj_mpred T cs M mtable point methods [g;s] (PointSpecs point _ PM). 

Definition makePoint_spec (pc:PointConstr T): funspec :=
 match pc with {| ptM := PM ; ptConstr := Init_Point |} =>
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr point ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; Point PM (Init_Point i,p))
 end.

Definition BPoint (BPM: BPointM T): Pred T:=
  obj_mpred T cs M mtable bpoint bmethods
           [ g; s; b] (BPointSpecs point bpoint _ BPM).

Definition makeBPoint_spec (bc:BPointConstr T) :=
  match bc with {| bptM := BM; bptConstr := Init_Point |} =>
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr bpoint ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; BPoint BM (Init_Point i,p))
 end.

Definition CPoint (CPM: CPointM T): Pred T:= 
  obj_mpred T cs M mtable cpoint cmethods 
            [ g; s; b; gC] (CPointSpecs point bpoint cpoint _ CPM).

Definition makeCPoint_spec (cc:CPointConstr T) :=
  match cc with {| cptM := CM; cptConstr:= Init_CPoint |} =>
 WITH gv: globals, i:Z, c:Color
 PRE [ tint, tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) 
    PARAMS (Vint(Int.repr i); Vint (Int.repr (color_encoding _ CM c))) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr cpoint ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; CPoint CM (Init_CPoint i c,p))
 end.

Lemma Point_fold_unfold PM:
  Point PM = fun x => EX cso, EX P:Pred T, 
                        P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt T point mtable m x *
                                     MTable T M methods [g; s] m (PointSpecs point _ PM) P).
Proof. unfold Point. apply obj_mpred_fold_unfold. Qed.

Lemma Point_isptr PM t v: Point PM (t,v) |-- !!(isptr v).
Proof. unfold Point. sep_apply obj_mpred_isptr; entailer!. Qed.
Lemma Point_isptr' PM tv: Point PM tv |-- !!(isptr (snd tv)).
Proof. unfold Point. sep_apply obj_mpred_isptr'; entailer!. Qed.

Lemma BPoint_fold_unfold BPM :
  BPoint BPM = fun x => EX cso, EX P : Pred T,
                        P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt T bpoint mtable m x *
                                     MTable T M bmethods [ g; s; b] m (BPointSpecs point bpoint _ BPM) P).
Proof. unfold BPoint. apply obj_mpred_fold_unfold. Qed.

Lemma BPoint_isptr BM t v: BPoint BM (t,v) |-- !!(isptr v).
Proof. unfold BPoint. sep_apply obj_mpred_isptr; entailer!. Qed. 
Lemma BPoint_isptr' BM tv: BPoint BM tv |-- !!(isptr (snd tv)).
Proof. unfold BPoint. sep_apply obj_mpred_isptr'; entailer!. Qed.

Lemma CPoint_fold_unfold CPM :
  CPoint CPM = fun x => EX cso, EX P : Pred T,
                        P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt T cpoint mtable m x *
                                     MTable T M cmethods [g; s; b; gC] m (CPointSpecs point bpoint cpoint _ CPM) P).
Proof. unfold CPoint. apply obj_mpred_fold_unfold. Qed.

Lemma CPoint_isptr CM t v: CPoint CM (t,v) |-- !!(isptr v).
Proof. unfold CPoint. sep_apply obj_mpred_isptr; entailer!. Qed. 
Lemma CPoint_isptr' CM tv: CPoint CM tv |-- !!(isptr (snd tv)).
Proof. unfold CPoint. sep_apply obj_mpred_isptr'; entailer!. Qed. 

Hint Resolve Point_isptr : saturate_local.
Hint Resolve Point_isptr' : saturate_local.
Hint Resolve BPoint_isptr : saturate_local.
Hint Resolve BPoint_isptr' : saturate_local.
Hint Resolve CPoint_isptr : saturate_local.
Hint Resolve CPoint_isptr' : saturate_local.

End PointPreds.
Definition PointASI (point bpoint cpoint methods bmethods cmethods:type) (mtable g s b gC: ident) T R pc bc cc
           cs M (makept makebpt makecpt:ident):=
           [(makept, makePoint_spec point methods mtable g s cs M T pc);
            (makebpt, makeBPoint_spec point bpoint bmethods mtable g s b cs M T bc);
            (makecpt, makeCPoint_spec point bpoint cpoint cmethods mtable g s b gC cs M R cc)].

Require Import SimpleClient.

Notation point := (Tstruct _point noattr).
Notation methods := (Tstruct _methods noattr).

Notation bpoint := (Tstruct _bpoint noattr).
Notation bmethods := (Tstruct _bmethods noattr).

Notation cpoint := (Tstruct _cpoint noattr).
Notation cmethods := (Tstruct _cmethods noattr).

Definition SimplePointASI := PointASI point bpoint cpoint methods bmethods cmethods 
           _mtable _get _set _bump _getC _ _ ptConstr1 bptConstr1 cptConstr1.

Section Funspec_Sub_set.
Variable T:Type.
Variable M:MallocFreeAPD.
Variable P: PointM T.
Variable BP: BPointM T.
Variable cs:compspecs.
Variable mtable gg s b:ident.

Goal funspec_sub (set_spec point T P (Point point methods mtable gg s cs M T P))
                 (set_spec point T BP (BPoint point bpoint bmethods mtable gg s b cs M T BP)).
Proof. do_funspec_sub. destruct w as [[t p] j]; simpl; Intros. simpl in H; subst. rewrite ! sepcon_emp.
 Exists ((t,p),j). simpl.
  rewrite BPoint_fold_unfold ,Point_fold_unfold. Intros cso Q m. Exists emp cso Q m.
remember (MTable T M bmethods [gg; s; b] m (BPointSpecs point bpoint T BP)) as TABLE1.
 Abort.
  
Goal funspec_sub (set_spec point T BP (BPoint point bpoint bmethods mtable gg s b cs M T BP))
                 (set_spec point T P (Point point methods mtable gg s cs M T P)).
Proof. do_funspec_sub. destruct w as [[x p] j]; simpl; Intros. simpl in H; subst. rewrite ! sepcon_emp.
  rewrite Point_fold_unfold. Intros PR m; simpl. Exists ((x,p),j). simpl.
Abort.
End Funspec_Sub_set.
