Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import spec_surelyMalloc.
  
Require Import SimplePointModel.
Require Import SimplePoints_ASI.
Require Import VST.floyd.VSU.
Require Import PointList.
Require Import DeepPoints.
Require Import ASI_DeepPointList.

Lemma Points_app T CT DP: forall L1 L2, 
      Points T CT DP (L1 ++ L2) = (Points T CT DP L1 *  Points T CT DP L2)%logic.
Proof. induction L1; simpl; intros. rewrite emp_sepcon; trivial.
destruct a. rewrite IHL1, sepcon_assoc; trivial.
Qed.

Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition Vprog : varspecs. mk_varspecs prog. Defined.
Notation lst := (Tstruct _list noattr).

Fixpoint LIST M (nodes:list val) (p:val): mpred :=
  match nodes with 
    nil => !!(p=nullval) && emp
  | (h::t) => !!(isptr h) && EX n, data_at Ews lst (h,n) p * malloc_token M Ews lst p * LIST M t n
  end.

Lemma LIST_validpointer M nodes l: LIST M nodes l |-- valid_pointer l.
Proof. induction nodes; simpl; entailer. Qed.
Hint Resolve LIST_validpointer : valid_pointer.

Lemma LIST_ptrnull M nodes l: LIST M nodes l |-- !! is_pointer_or_null l.
Proof. induction nodes; simpl; intros. entailer!. Intros n. entailer!. Qed.
Hint Resolve LIST_ptrnull : saturate_local.

Definition LIST_NIL M: emp |-- LIST M nil nullval.
Proof. simpl; entailer!. Qed.

(*Require Import ASI_HLWrapper. For WrapperAPD --  not needed any longer*)
Require Import ASI_DeepWrapper.

Section VSU_PointList.

Variable M:MallocFreeAPD.

Variable T CT : Type. 
Notation DeepPoint := (DeepPoint T CT).
Notation Deep_point := (Deep_point T CT).
Notation Deep_bpoint := (Deep_bpoint T CT).
Notation Deep_cpoint := (Deep_cpoint T CT).

Variable DP: SimpleObjspec.Pred DeepPoint.

Definition PL_Imports := SM_ASI M ++
           firstn 3 (DeepMethodsASI point bpoint cpoint T CT DP _GET _SET _BUMP _GETC).
(*GETC not used by this client*)

Definition PL_Internal_specs:= ASI_PointList lst point M (LIST M) T CT DP _insert _map_set _map_bump _sum.

Definition PL_Gprog := PL_Imports ++ PL_Internal_specs.

Lemma verif_insert: semax_body Vprog PL_Gprog f_insert (_insert, insert_spec lst point M (LIST M)).
Proof. start_function.
  forward_call surelymalloc_spec_sub (lst, gv).
    { simpl. split3; trivial. rep_lia. }
  Intros n.
  forward. forward. forward. Exists n. entailer. simpl. Exists l. entailer!.
Qed.

Fixpoint LISTaux (nodes:list val) (p q:val): mpred :=
  match nodes with 
    nil => !!(q=p /\ is_pointer_or_null p) && emp
  | (h::t) => !!(isptr h) && EX n, data_at Ews lst (h,n) p * malloc_token M Ews lst p * LISTaux t n q
  end.

Lemma LIST_LISTaux: forall nodes p, LIST M nodes p = LISTaux nodes p nullval.
Proof.
induction nodes; simpl; intros; apply pred_ext; intros; entailer.
Exists n; rewrite IHnodes; entailer!.
Exists n; rewrite IHnodes; entailer!.
Qed.

Lemma LISTaux_ptrnull nodes l p: LISTaux nodes l p |-- !! is_pointer_or_null l.
Proof. induction nodes; simpl; intros. entailer!. Intros n. entailer!. Qed.
Hint Resolve LISTaux_ptrnull : saturate_local.

Lemma LISTaux_app: forall nodes1 nodes2 l q, LISTaux (nodes1 ++ nodes2) l q =
                   EX p, LISTaux nodes1 l p * LISTaux nodes2 p q.
Proof.
  induction nodes1; simpl; intros; apply pred_ext; entailer.
+ Exists l. entailer!.
+ rewrite IHnodes1; clear IHnodes1. Intros p. Exists p n. entailer!.
+ Exists n. rewrite IHnodes1; clear IHnodes1. Exists p. entailer!.
Qed.

Lemma LISTaux_validpointer: forall nodes l, isptr l -> LISTaux nodes l nullval |-- valid_pointer l.
Proof.
  induction nodes; simpl; intros; entailer. Qed.
Hint Resolve LISTaux_validpointer : valid_pointer.

Lemma LIST_nullval nodes: LIST M nodes nullval |-- !!(nodes=[]) && emp.
Proof. destruct nodes; simpl; entailer. rewrite data_at_isptr; Intros. inv Pnullval. Qed.

Lemma verif_mapset: semax_body Vprog PL_Gprog f_map_set (_map_set, mapset_spec lst (LIST M) T CT DP).
Proof.
  unfold mapset_spec. start_function.
  forward. 
  forward_while (EX n, EX PL1, EX PL2, 
     PROP (PL=PL1 ++ PL2 )
     LOCAL (temp _n n; temp _l l; temp _i (Vint (Int.repr i)))
     SEP (LISTaux (map snd PL1) l n; LIST M (map snd PL2) n;
          Points T CT DP (map_set T CT PL1 i); Points T CT DP PL2)).
  { Exists l (@nil (DeepPoint * val)) PL. entailer!. simpl. entailer!. }
  { entailer!. }
  { destruct PL2.
    + simpl; Intros. subst n; inv HRE.
    + destruct p as [pt elem]. simpl. Intros next. forward.
      assert_PROP (is_pointer_or_null next) by entailer!.
      freeze [0;2;3;4;6] FR1.
      forward_call (pt, elem, i).
(*      { simpl.  cancel. }*)
      forward. entailer!. Exists (next, PL1 ++ [(pt, elem)], PL2). simpl. rewrite <- app_assoc.
      entailer!. unfold map_set. rewrite ! map_app, Points_app. simpl. cancel.
      thaw FR1. rewrite LISTaux_app. Exists n. cancel. simpl. Exists next.
      entailer!. apply derives_refl.  }
  forward. entailer!. destruct PL2.
    + simpl; Intros. rewrite ! app_nil_r, LIST_LISTaux. cancel. 
    + sep_apply (LIST_nullval (map snd (p :: PL2))); Intros.
Qed.

Lemma verif_mapbump: semax_body Vprog PL_Gprog f_map_bump (_map_bump, mapbump_spec lst (LIST M) T CT DP).
Proof.
  unfold mapbump_spec. start_function.
  forward.
  forward_while (EX n, EX PL1, EX PL2, 
     PROP (PL=PL1 ++ PL2 )
     LOCAL (temp _n n; temp _l l)
     SEP (LISTaux (map snd PL1) l n; LIST M (map snd PL2) n;
          Points T CT DP (map_bump T CT (*BM CM*) PL1); Points T CT DP PL2)).
  { Exists l (@nil (DeepPoint * val)) PL. entailer!. simpl. entailer!. }
  { entailer!. }
  { destruct PL2.
    + simpl; Intros. subst n; inv HRE.
    + destruct p as [pt elem]. simpl. Intros next. forward.
      assert_PROP (is_pointer_or_null next) by entailer!.
      freeze [0;2;3;4;6] FR1.
      forward_call (pt, elem).
      (*{ simpl. cancel. }*)
      { subst PL. clear - H. rewrite map_app in H. simpl in H.
        apply Forall_app in H; destruct H as [_ X]. inv X; trivial. }
      forward. entailer!.
        Exists (next, PL1 ++ [(pt, elem)], PL2). simpl. rewrite <- app_assoc.
        entailer!. unfold map_bump. rewrite ! map_app, Points_app. simpl. cancel.
        thaw FR1. rewrite LISTaux_app. Exists n. cancel. simpl. Exists next.
        entailer!. apply derives_refl. }
  { forward. entailer!. destruct PL2.
    + simpl; Intros. rewrite ! app_nil_r, LIST_LISTaux. cancel. 
    + sep_apply (LIST_nullval (map snd (p :: PL2))); Intros. }
Qed.

Definition sumR l:= fold_right (fun d z => Z.add z (dp_get T CT d)) 0 l.

Lemma sumR_app: 
  forall a b, sumR (a++b) =  sumR a + sumR b.
Proof.
  intros. unfold sumR; induction a; simpl; lia.
Qed.

Lemma sumR_sum: forall l, sumR l = sum T CT l.
Proof. induction l; simpl; intros.
+ unfold sum; simpl; trivial.
+ rewrite IHl; clear IHl. unfold sum; simpl.
  forget (dp_get _ _ a) as b. revert b; clear.
  generalize 0 as a.
  induction l; simpl; intros; trivial.
  rewrite IHl; clear IHl. f_equal. lia.
Qed.

Lemma sum_cons: forall L a, sum T CT (a :: L) = dp_get _ _ a + sum T CT L.
Proof. induction L; simpl; intros. unfold sum; simpl. lia.
rewrite IHL; clear IHL. rewrite <- ! sumR_sum. unfold sumR. simpl. lia.
Qed. 

Lemma sum_app: 
  forall a b, sum T CT (a++b) =  sum T CT a + sum T CT b.
Proof.
  intros. rewrite <- ! sumR_sum. apply sumR_app.
Qed. 

Lemma sum_nil: sum T CT nil= 0.
Proof. reflexivity. Qed. 

Lemma summable_cons: forall L a, summable T CT (a :: L) -> summable T CT L.
Proof. intros. destruct H. inv H. split; trivial. rewrite sum_cons in H0. lia. Qed.

Lemma sum_nonneg: forall l, summable T CT l -> 0 <= sum T CT l.
Proof.
  induction l; simpl; intros.
+ unfold sum; simpl; lia.
+ specialize (summable_cons _ _ H); intros. specialize (IHl H0).
  inv H. rewrite sum_cons in *. inv H1. lia.
Qed.

Lemma summable_app: forall L1 L2, summable T CT (L1++L2) -> summable T CT L1 /\  summable T CT L2.
Proof.
  induction L1; simpl; intros. 
+ split; trivial. split. constructor. unfold sum; simpl; rep_lia.
+ destruct (IHL1 L2); clear IHL1. apply summable_cons in H; trivial. split; trivial.
  destruct H. inv H. apply Forall_app in H6; destruct H6. split. constructor; trivial.
  specialize (sum_app (a::L1) L2); simpl; intros X. rewrite X in H2. clear X.
  apply sum_nonneg in H1. rep_lia.
Qed.

Lemma no_overflow (PL1 PL2: list (DeepPoint * val)) pt elem:
      summable T CT (map fst (PL1 ++ (pt, elem) :: PL2)) ->
      Int.min_signed <=
      Int.signed (Int.repr (sum T CT (map fst PL1))) + Int.signed (Int.repr (dp_get _ _ pt)) <=
      Int.max_signed.
Proof. intros. rewrite map_app in H. simpl in H.
          destruct (summable_app _ _ H). rewrite 2 Int.signed_repr.
          2: { destruct H1. inv H1. simpl in *. rep_lia. }
          2: { specialize (sum_nonneg _ H0); destruct H0. intros. rep_lia. } 
          specialize (summable_app (map fst PL1 ++ [pt]) (map fst PL2)). 
          rewrite <- app_assoc; simpl. intros X. destruct (X H).
          specialize (sum_nonneg _ H2); destruct H2. intros.
          rewrite sum_app, sum_cons, sum_nil in *. simpl in *; rep_lia.
Qed.

Lemma verif_sum: semax_body Vprog PL_Gprog f_sum (_sum, sum_spec lst (LIST M) T CT DP).
Proof.
  unfold sum_spec. start_function.
  forward. forward.
  forward_while (EX n, EX PL1, EX PL2, 
     PROP (PL=PL1 ++ PL2 )
     LOCAL (temp _s (Vint (Int.repr (sum T CT (map fst PL1)))); temp _n n; temp _l l)
     SEP (LISTaux (map snd PL1) l n; LIST M (map snd PL2) n;
          Points T CT DP PL1; Points T CT DP PL2)).
  { Exists l (@nil (DeepPoint * val)) PL. entailer!. simpl. entailer!. }
  { entailer!. }
  { destruct PL2.
    + simpl; Intros. subst n; inv HRE.
    + destruct p as [pt elem]. simpl. Intros next. forward.
      assert_PROP (is_pointer_or_null next) by entailer!.
      freeze [0;2;3;4;6] FR1.
      subst. apply no_overflow in H.
      forward_call (pt, elem).
      (*{ simpl; cancel. }*)
      forward. forward. entailer!.
        Exists (next, PL1 ++ [(pt, elem)], PL2). simpl. rewrite <- app_assoc.
        entailer!. 
        { f_equal. f_equal. unfold sum. rewrite map_app. rewrite fold_left_app. simpl. lia. }
        rewrite ! map_app, Points_app. simpl. cancel.
        thaw FR1. rewrite LISTaux_app. Exists n. cancel. simpl. Exists next.
        entailer!. }
  { forward. destruct PL2.
    + simpl; Intros. rewrite ! app_nil_r, LIST_LISTaux. entailer!.
    + sep_apply (LIST_nullval (map snd (p :: PL2))); Intros. }
Qed.

  Definition DeepPointList_Component: @Component NullExtension.Espec Vprog _ nil PL_Imports prog 
                 PL_Internal_specs emp PL_Internal_specs.
  Proof. mkComponent_cautious. 
     - solve_SF_internal verif_insert. 
     - solve_SF_internal verif_mapset.
     - solve_SF_internal verif_mapbump.
     - solve_SF_internal verif_sum.
  Qed.

  Definition DeepPointList_VSU: @VSU NullExtension.Espec Vprog _ nil PL_Imports prog PL_Internal_specs emp.
  Proof. eexists; apply DeepPointList_Component. Qed.

End VSU_PointList.