Require Import VST.floyd.proofauto.

Require Import ASI_malloc. (*needed for M*)
Require Import VST.floyd.VSU.
Require Import VST.veric.initial_world.

Require Import link_progs.

Require Import SimpleObjspec.
Require Import SimplePointModel.
Require Import SimplePoints_ASI.
Require Import SimplePoints_VSU1.
Require Import SimplePoints_VSU2.

Lemma obj_mpred_change_compspecs' T cs1 cs2 (M:MallocFreeAPD) mt base (*cso1 cso2*) MT idents specs
      (HFld: forall m x, @FldAt T cs1 base mt m x = @FldAt T cs2 base mt m x)
      (*(HMalloc: forall v, @malloc_token cs1 M Ews o v = @malloc_token cs2 M Ews o v)*)
      (HMTable: forall m P, @MTable T cs1 M MT idents m specs P = @MTable T cs2 M MT idents m specs P):
      @obj_mpred T cs1 M mt base (*cso1*) MT idents specs  = @obj_mpred T cs2 M mt base (*cso2*) MT idents specs.
Proof.
  rewrite 2 obj_mpred_fold_unfold. extensionality x. apply exp_congr; intros cso. 
  apply exp_congr; intros P.
  (*rewrite HMalloc. *)f_equal. apply exp_congr; intros m. rewrite HFld, HMTable. trivial.
Qed.

Lemma funspecs_sqsub_app l1 l2 m1 m2 (LM1: map fst l1 = map fst m1) (LM2: map fst l2= map fst m2)
      (LNR: list_norepet (map fst (l1++l2))):
      (funspecs_sqsub l1 m1 /\ funspecs_sqsub l2 m2) = funspecs_sqsub (l1++l2) (m1++m2).
Proof. unfold funspecs_sqsub. rewrite map_app in LNR. destruct (list_norepet_append_inv _ _ _ LNR) as [_[ _ D]]; clear LNR.
apply prop_ext; split; intros.
+ rewrite assoclists.find_id_app_char in *. destruct H. specialize (H i). specialize (H1 i).
  destruct (find_id i m1).
  - inv H0. destruct (H _ (eq_refl _)) as [? [X ?]]. rewrite X. eexists; split; auto.
  - destruct (H1 _ H0) as [? [X ?]]. 
    destruct (assoclists.find_id_None_iff i l1). rewrite H4, X. eexists; split; auto.
    apply list_disjoint_sym in D. 
    apply find_id_In_map_fst in H0; rewrite <- LM2 in H0.
    apply (list_disjoint_notin i D H0).
+ split; intros.
  - specialize (H i). rewrite assoclists.find_id_app_char, H0 in H.
    destruct (H _ (eq_refl _)) as [? [? ?]]; clear H.
    rewrite assoclists.find_id_app_char in H1.
    destruct (find_id i l1). eexists; split; eauto.
    apply find_id_In_map_fst in H0. apply find_id_In_map_fst in H1. rewrite <- LM1 in H0.
    elim (D i i H0 H1); trivial.
 - specialize (H i). rewrite assoclists.find_id_app_char, H0 in H.
   apply find_id_In_map_fst in H0. rewrite <- LM2 in H0.
   apply list_disjoint_sym in D. 
   specialize (list_disjoint_notin i D H0); intros.
   rewrite assoclists.find_id_app_char in H.
   destruct (assoclists.find_id_None_iff i l1). rewrite (H3 H1) in H.
   destruct (assoclists.find_id_None_iff i m1).
   rewrite LM1 in H1. rewrite (H5 H1) in H. apply H; trivial.
Qed.

Lemma funspecs_sqsub_refl l: funspecs_sqsub l l.
Proof. red; intros. exists phi'; split; trivial. apply funspec_sub_refl. Qed.

Lemma funspecs_sqsub_nil: funspecs_sqsub nil nil. 
Proof. red; intros. inv H. Qed.

Lemma funspecs_sqsub_cons i phi l phi' l': 
      funspec_sub phi phi' -> funspecs_sqsub l l' ->
      funspecs_sqsub ((i,phi)::l) ((i,phi')::l').
Proof. intros; simpl. red; intros; simpl. simpl in H1.
  if_tac in H1.
+ inv H1. exists phi; split; trivial.
+ apply H0; trivial.
 Qed.

Lemma funspecs_sqsub_D i phi l phi' l': 
      funspecs_sqsub ((i,phi)::l) ((i,phi')::l') ->
      list_norepet (map fst ((i,phi)::l)) -> map fst l = map fst l' ->
      funspec_sub phi phi' /\ funspecs_sqsub l l'.
Proof. intros; simpl; split.
+ specialize (H i); simpl in H. rewrite 2 if_true in H; trivial.
  destruct (H _ (eq_refl _)) as [? [X ?]]. inv X. trivial.
+ red; intros. specialize (H i0); simpl in H.
  assert (Y: i0<>i). 
  { intros N; subst. simpl in H0. inv H0.
    apply find_id_In_map_fst in H2. rewrite H1 in H5; contradiction. }
  rewrite 2 if_false in H; auto.
Qed.

Section VSU1_VSU2.
Variable M:MallocFreeAPD.
(*MF_ASI has different order (TODO: change VSULink_tac tactic so that
  imports need not be list equality but can be extensional equality*)
Goal (I1_Imports M) = (I2_Imports M). reflexivity. Qed.

Goal map fst (I1_Imports M) = [ SimplePoint_I1._surely_malloc]. reflexivity. Qed. (*So both import sets just cotain surelymalloc*)
Goal map fst (I1_ASI M) = [ SimplePoint_I1._makePoint_I1; SimplePoint_I1._makeBPoint_I1; SimplePoint_I1._makeCPoint_I1]. reflexivity. Qed.
Goal map fst (I2_ASI M) = [ SimplePoint_I2._makePoint_I2; SimplePoint_I2._makeBPoint_I2; SimplePoint_I2._makeCPoint_I2]. reflexivity. Qed.
Definition p12_Exports:funspecs := I1_ASI M ++ I2_ASI M.

Definition p12_VSU: @VSU NullExtension.Espec p12_Vprog p12_cs nil (I1_Imports M) prog_I1_I2 p12_Exports emp.
Proof. (*VSULink_tac (I1_VSU M) (I2_VSU M).*)
(*Print VSULink_tac.*) eapply VSU_ext.
+ eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ (I1_VSU M) (I2_VSU M)).
  - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - (* SC_tac. *)unfold SimplePoint_I1.prog, IntIDs; rewrite prog_defs_Clight_mkprogram. 
        SC_tac.
      - (* SC_tac.*)unfold SimplePoint_I2.prog, IntIDs; rewrite prog_defs_Clight_mkprogram.
        SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac. (*Goal list_norepet (map fst p12_Vprog)*)
      - domV_tac. (*Goal involves p12_Vprog*)
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity). 
      -  (*Proof should be like this: first [ find_id_subset_tac | idtac ].*)
         (*Print find_id_subset_tac. *)unfold Vardefs, SimplePoint_I1.prog; rewrite prog_defs_Clight_mkprogram; intros ? ? H.
         simpl in H; inv H.
      - (*Proof should be like this: first [ find_id_subset_tac | idtac ].*)
         unfold Vardefs, SimplePoint_I2.prog; rewrite prog_defs_Clight_mkprogram; intros ? ? H.
         simpl in H; inv H.
+ extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

End VSU1_VSU2.

Require Import ASI_SimpleClient.
Require Import VSU_SimpleClient.
Require Import SimpleClient.

Section WithClient.
Variable M: MallocFreeAPD.

Lemma align_compatible_point_A1 {b i} (AL: @align_compatible SimplePoints_VSU1.CompSpecs point (Vptr b i)):
      @align_compatible CompSpecs point (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma align_compatible_point_B1 {b i} (AL: @align_compatible CompSpecs point (Vptr b i)):
      @align_compatible SimplePoints_VSU1.CompSpecs point (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Instance CompSpecs_Preserve1: change_composite_env SimplePoints_VSU1.CompSpecs CompSpecs.
  make_cs_preserve SimplePoints_VSU1.CompSpecs CompSpecs.
Defined.

Lemma Point_FldAt1 T: forall (m : val) (x : T * val),
      @FldAt T SimplePoints_VSU1.CompSpecs point _mtable m x =
      @FldAt T CompSpecs point _mtable m x.
Proof. unfold FldAt. destruct x as [rep p]; simpl in *. clear. 
  apply pred_ext; Intros m'; simpl in *; apply JMeq_eq in H; subst m'.
  + specialize (@field_at_change_composite SimplePoints_VSU1.CompSpecs CompSpecs CompSpecs_Preserve1 Ews point [StructField _mtable] m m). simpl; intros.
    rewrite H; trivial. Exists m; entailer!. 
  + specialize (@field_at_change_composite SimplePoints_VSU1.CompSpecs CompSpecs CompSpecs_Preserve1 Ews point [StructField _mtable] m m). simpl; intros.
    rewrite <- H; trivial. Exists m; entailer!.
Qed.

Lemma Point_cs1_eq: Point point methods _mtable _get _set SimplePoints_VSU1.CompSpecs M PointRep m1 =
                    Point point methods _mtable _get _set CompSpecs M PointRep m1.
Proof. unfold Point. rewrite (obj_mpred_change_compspecs' PointRep SimplePoints_VSU1.CompSpecs CompSpecs).
    + reflexivity.
    + apply Point_FldAt1. 
    + (*MTable*) clear; intros. eapply (MTable_change_compspecs _ _ _ CompSpecs_Preserve1); trivial.
      intros. repeat (destruct H; [ subst i; reflexivity |]). simpl in H; contradiction.
Qed.
 
Lemma makePointSpec_eq1:
      makePoint_spec point methods _mtable _get _set SimplePoints_VSU1.CompSpecs M (*DynRep1*) PointRep (*m1 Init_PointRep*)ptConstr1 =
      makePoint_spec point methods _mtable _get _set CompSpecs M (*DynRep1*) PointRep (*m1 Init_PointRep*)ptConstr1.
Proof.
  apply semax_prog.funspec_eq. trivial. extensionality ts. extensionality w.
  destruct w as [gv i]. apply exp_congr. intros p. rewrite Point_cs1_eq; trivial.
Qed.

Lemma BPoint_FldAt1 T: forall (m : val) (x : T * val),
      @FldAt T SimplePoints_VSU1.CompSpecs bpoint _mtable m x =
      @FldAt T CompSpecs bpoint _mtable m x.
Proof. unfold FldAt. destruct x as [rep p]; simpl in *. clear. 
  apply pred_ext; Intros m'; simpl in *; apply JMeq_eq in H; subst m'.
  + specialize (@field_at_change_composite SimplePoints_VSU1.CompSpecs CompSpecs CompSpecs_Preserve1 Ews bpoint [StructField _mtable] m m). simpl; intros.
    rewrite H; trivial. Exists m; entailer!. 
  + specialize (@field_at_change_composite SimplePoints_VSU1.CompSpecs CompSpecs CompSpecs_Preserve1 Ews bpoint [StructField _mtable] m m). simpl; intros.
    rewrite <- H; trivial. Exists m; entailer!.
Qed.

Lemma BPoint_cs1_eq: BPoint point bpoint bmethods _mtable _get _set _bump SimplePoints_VSU1.CompSpecs M PointRep bm1 =
                     BPoint point bpoint bmethods _mtable _get _set _bump CompSpecs M PointRep bm1.
Proof. unfold BPoint. rewrite (obj_mpred_change_compspecs' PointRep SimplePoints_VSU1.CompSpecs CompSpecs).
    + reflexivity.
    + apply BPoint_FldAt1. 
    + (*MTable*) clear; intros. eapply (MTable_change_compspecs _ _ _ CompSpecs_Preserve1); trivial.
      intros. repeat (destruct H; [ subst i; reflexivity |]). simpl in H; contradiction.
Qed.
 
Lemma makeBPointSpec_eq1:
      makeBPoint_spec point bpoint bmethods _mtable _get _set _bump SimplePoints_VSU1.CompSpecs M (*DynRep1*) PointRep (*bm1 Init_PointRep*)bptConstr1 =
      makeBPoint_spec point bpoint bmethods _mtable _get _set _bump CompSpecs M (*DynRep1*) PointRep (*bm1 Init_PointRep*)bptConstr1.
Proof.
  apply semax_prog.funspec_eq. trivial. extensionality ts. extensionality w.
  destruct w as [gv i]. apply exp_congr. intros p. rewrite BPoint_cs1_eq; trivial.
Qed.

Lemma CPoint_FldAt1 T: forall (m : val) (x : T * val),
      @FldAt T SimplePoints_VSU1.CompSpecs cpoint _mtable m x =
      @FldAt T CompSpecs cpoint _mtable m x.
Proof. unfold FldAt. destruct x as [rep p]; simpl in *. clear. 
  apply pred_ext; Intros m'; simpl in *; apply JMeq_eq in H; subst m'.
  + specialize (@field_at_change_composite SimplePoints_VSU1.CompSpecs CompSpecs CompSpecs_Preserve1 Ews cpoint [StructField _mtable] m m). simpl; intros.
    rewrite H; trivial. Exists m; entailer!. 
  + specialize (@field_at_change_composite SimplePoints_VSU1.CompSpecs CompSpecs CompSpecs_Preserve1 Ews cpoint [StructField _mtable] m m). simpl; intros.
    rewrite <- H; trivial. Exists m; entailer!.
Qed.

Lemma CPoint_cs1_eq: CPoint point bpoint cpoint cmethods _mtable _get _set _bump _getC SimplePoints_VSU1.CompSpecs M CPointRep cm1 =
                     CPoint point bpoint cpoint cmethods _mtable _get _set _bump _getC CompSpecs M CPointRep cm1.
Proof. unfold CPoint. rewrite (obj_mpred_change_compspecs' CPointRep SimplePoints_VSU1.CompSpecs CompSpecs).
    + reflexivity.
    + apply CPoint_FldAt1. 
    + (*MTable*) clear; intros. eapply (MTable_change_compspecs _ _ _ CompSpecs_Preserve1); trivial.
      intros. repeat (destruct H; [ subst i; reflexivity |]). simpl in H; contradiction.
Qed.
 
Lemma makeCPointSpec_eq1:
      makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC SimplePoints_VSU1.CompSpecs M (*DynRep1*) CPointRep (*cm1 Init_CPointRep*) cptConstr1=
      makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC CompSpecs M (*DynRep1*) CPointRep (*cm1 Init_CPointRep*)cptConstr1.
Proof.
  apply semax_prog.funspec_eq. trivial. extensionality ts. extensionality w.
  destruct w as [[gv i] c]. apply exp_congr. intros p. rewrite CPoint_cs1_eq; trivial.
Qed. 

(*Same for VSU2*)
Lemma align_compatible_point_A2 {b i} (AL: @align_compatible SimplePoints_VSU2.CompSpecs point (Vptr b i)):
      @align_compatible CompSpecs point (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma align_compatible_point_B2 {b i} (AL: @align_compatible CompSpecs point (Vptr b i)):
      @align_compatible SimplePoints_VSU2.CompSpecs point (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Instance CompSpecs_Preserve2: change_composite_env SimplePoints_VSU2.CompSpecs CompSpecs.
  make_cs_preserve SimplePoints_VSU2.CompSpecs CompSpecs.
Defined.

Lemma Point_FldAt2 T: forall (m : val) (x : T * val),
      @FldAt T SimplePoints_VSU2.CompSpecs point _mtable m x =
      @FldAt T CompSpecs point _mtable m x.
Proof. unfold FldAt. destruct x as [rep p]; simpl in *. clear. 
  apply pred_ext; Intros m'; simpl in *; apply JMeq_eq in H; subst m'.
  + specialize (@field_at_change_composite SimplePoints_VSU2.CompSpecs CompSpecs CompSpecs_Preserve2 Ews point [StructField _mtable] m m). simpl; intros.
    rewrite H; trivial. Exists m; entailer!. 
  + specialize (@field_at_change_composite SimplePoints_VSU2.CompSpecs CompSpecs CompSpecs_Preserve2 Ews point [StructField _mtable] m m). simpl; intros.
    rewrite <- H; trivial. Exists m; entailer!.
Qed.

Lemma Point_cs2_eq: Point point methods _mtable _get _set SimplePoints_VSU2.CompSpecs M PointRep m1 =
                    Point point methods _mtable _get _set CompSpecs M PointRep m1.
Proof. unfold Point. rewrite (obj_mpred_change_compspecs' PointRep SimplePoints_VSU2.CompSpecs CompSpecs).
    + reflexivity.
    + apply Point_FldAt2. 
    + (*MTable*) clear; intros. eapply (MTable_change_compspecs _ _ _ CompSpecs_Preserve2); trivial.
      intros. repeat (destruct H; [ subst i; reflexivity |]). simpl in H; contradiction.
Qed. 

Lemma makePointSpec_eq2:
      makePoint_spec point methods _mtable _get _set SimplePoints_VSU2.CompSpecs M (*DynRep1*) PointRep (*m1 Init_PointRep*)ptConstr1 =
      makePoint_spec point methods _mtable _get _set CompSpecs M (*DynRep1*) PointRep (*m1 Init_PointRep*)ptConstr1.
Proof.
  apply semax_prog.funspec_eq. trivial. extensionality ts. extensionality w.
  destruct w as [gv i]. apply exp_congr. intros p. rewrite Point_cs2_eq; trivial.
Qed.

Lemma BPoint_FldAt2 T: forall (m : val) (x : T * val),
      @FldAt T SimplePoints_VSU2.CompSpecs bpoint _mtable m x =
      @FldAt T CompSpecs bpoint _mtable m x.
Proof. unfold FldAt. destruct x as [rep p]; simpl in *. clear. 
  apply pred_ext; Intros m'; simpl in *; apply JMeq_eq in H; subst m'.
  + specialize (@field_at_change_composite SimplePoints_VSU2.CompSpecs CompSpecs CompSpecs_Preserve2 Ews bpoint [StructField _mtable] m m). simpl; intros.
    rewrite H; trivial. Exists m; entailer!. 
  + specialize (@field_at_change_composite SimplePoints_VSU2.CompSpecs CompSpecs CompSpecs_Preserve2 Ews bpoint [StructField _mtable] m m). simpl; intros.
    rewrite <- H; trivial. Exists m; entailer!.
Qed.

Lemma BPoint_cs2_eq: BPoint point bpoint bmethods _mtable _get _set _bump SimplePoints_VSU2.CompSpecs M PointRep bm1 =
                     BPoint point bpoint bmethods _mtable _get _set _bump CompSpecs M PointRep bm1.
Proof. unfold BPoint. rewrite (obj_mpred_change_compspecs' PointRep SimplePoints_VSU2.CompSpecs CompSpecs).
    + reflexivity.
    + apply BPoint_FldAt2. 
    + (*MTable*) clear; intros. eapply (MTable_change_compspecs _ _ _ CompSpecs_Preserve2); trivial.
      intros. repeat (destruct H; [ subst i; reflexivity |]). simpl in H; contradiction.
Qed.
 
Lemma makeBPointSpec_eq2:
      makeBPoint_spec point bpoint bmethods _mtable _get _set _bump SimplePoints_VSU2.CompSpecs M (*DynRep1*) PointRep (*bm1 Init_PointRep*)bptConstr1 =
      makeBPoint_spec point bpoint bmethods _mtable _get _set _bump CompSpecs M (*DynRep1*) PointRep (*bm1 Init_PointRep*)bptConstr1.
Proof.
  apply semax_prog.funspec_eq. trivial. extensionality ts. extensionality w.
  destruct w as [gv i]. apply exp_congr. intros p. rewrite BPoint_cs2_eq; trivial.
Qed.

Lemma CPoint_FldAt2 T: forall (m : val) (x : T * val),
      @FldAt T SimplePoints_VSU2.CompSpecs cpoint _mtable m x =
      @FldAt T CompSpecs cpoint _mtable m x.
Proof. unfold FldAt. destruct x as [rep p]; simpl in *. clear. 
  apply pred_ext; Intros m'; simpl in *; apply JMeq_eq in H; subst m'.
  + specialize (@field_at_change_composite SimplePoints_VSU2.CompSpecs CompSpecs CompSpecs_Preserve2 Ews cpoint [StructField _mtable] m m). simpl; intros.
    rewrite H; trivial. Exists m; entailer!. 
  + specialize (@field_at_change_composite SimplePoints_VSU2.CompSpecs CompSpecs CompSpecs_Preserve2 Ews cpoint [StructField _mtable] m m). simpl; intros.
    rewrite <- H; trivial. Exists m; entailer!.
Qed.

Lemma CPoint_cs2_eq: CPoint point bpoint cpoint cmethods _mtable _get _set _bump _getC SimplePoints_VSU2.CompSpecs M CPointRep cm1 =
                     CPoint point bpoint cpoint cmethods _mtable _get _set _bump _getC CompSpecs M CPointRep cm1.
Proof. unfold CPoint. rewrite (obj_mpred_change_compspecs' CPointRep SimplePoints_VSU2.CompSpecs CompSpecs).
    + reflexivity.
    + apply CPoint_FldAt2. 
    + (*MTable*) clear; intros. eapply (MTable_change_compspecs _ _ _ CompSpecs_Preserve2); trivial.
      intros. repeat (destruct H; [ subst i; reflexivity |]). simpl in H; contradiction.
Qed.
 
Lemma makeCPointSpec_eq2:
      makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC SimplePoints_VSU2.CompSpecs M (*DynRep1*) CPointRep (*cm1 Init_CPointRep*)cptConstr1 =
      makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC CompSpecs M (*DynRep1*) CPointRep (*m1 Init_CPointRep*)cptConstr1.
Proof.
  apply semax_prog.funspec_eq. trivial. extensionality ts. extensionality w.
  destruct w as [[gv i] c]. apply exp_congr. intros p. rewrite CPoint_cs2_eq; trivial.
Qed.

(* Goal map fst (SimpleClient_Imports M DynRep1 DynRep2) = [_makePoint_I1; _makeBPoint_I1; _makePoint_I2; _makeBPoint_I2; _makeCPoint_I2].
 reflexivity. Qed. (*So can use (I1_Imports M) as withClient Imports*) *)
(*Definition withClient_Exports: funspecs := ASI_SimpleClient.ASI_SimpleClient M SimpleClient._test1 SimpleClient._test2 SimpleClient._test12.*)
Definition withClient_Exports: funspecs := p12_Exports M ++ ASI_SimpleClient M _test1 _test2 _test12.

Definition withClient_VSU: @VSU NullExtension.Espec with_client_Vprog with_client_cs nil (I1_Imports M) with_client withClient_Exports emp.
Proof. (*
eapply VSU_Exports_sub.
Print Ltac VSULink_tac.*) (*VSULink_tac (I1_VSU M) (I2_VSU M).*)
* eapply VSU_ext. 
  + apply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ (p12_VSU M) (VSU_SimpleClient.SimpleClient_VSU M (*DynRep1 DynRep2*))).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub. (*cspecs_sub p12_cs with_client_cs*)
      - prove_cspecs_sub. (*cspecs_sub CompSpecs with_client_cs*)
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - (*Print SC_tac.*)
        simpl; intros ? ? X H.
        rewrite Point_cs1_eq, BPoint_cs1_eq, CPoint_cs1_eq.
        rewrite Point_cs2_eq, BPoint_cs2_eq, CPoint_cs2_eq.
(*        rewrite makePointSpec_eq,. makeBPointSpec_eq1, makeCPointSpec_eq1, makePointSpec_eq2, makeBPointSpec_eq2, makeCPointSpec_eq2.*)
        repeat (destruct H; [ subst; simpl; simpl in X; try discriminate | ]); inv X;
        try (eexists; split; [ reflexivity | apply funspec_sub_refl ]).
        contradiction.
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.
End WithClient.

Require Import VSU_LLWrapper.
(*
Section With_LLWrapper.
Variable M:MallocFreeAPD.
Definition with_LLWrapper_Exports: funspecs := G_merge (withClient_Exports M) (WS M PointRep CPointRep).
Definition with_LLWrapper_VSU: @VSU NullExtension.Espec with_wrapper_Vprog with_wrapper_cs nil (I1_Imports M) with_wrapper with_LLWrapper_Exports emp.
Proof. 
(*eapply VSU_Exports_sub. (*VSULink_tac (I1_VSU M) (I2_VSU M).*)
* (*Print VSULink_tac.*) *)
eapply VSU_ext.
  + eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ (withClient_VSU M) (LLWrapper_VSU M PointRep CPointRep)).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - SC_tac.
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

End With_LLWrapper.*)

Require Import VSU_HLWrapper. (*for MyWrapper.*)
(*
Section With_HLWrapper.
Variable M:MallocFreeAPD.
Definition with_HLWrapper_Exports: funspecs := G_merge (withClient_Exports M) (HLWrapperMethods M CompSpecs PointRep CPointRep).
Definition with_HLWrapper_VSU: @VSU NullExtension.Espec with_wrapper_Vprog with_wrapper_cs nil (I1_Imports M) with_wrapper with_HLWrapper_Exports emp.
Proof. 
(*eapply VSU_Exports_sub. (*VSULink_tac (I1_VSU M) (I2_VSU M).*)
* (*Print VSULink_tac.*) *)
eapply VSU_ext.
  + eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ (withClient_VSU M) (HLWrapper_VSU M PointRep CPointRep)).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - SC_tac.
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

End With_HLWrapper.*)

Require Import VSU_DeepWrapper.

Section With_DeepWrapper.

Variable M:MallocFreeAPD.
Definition with_DeepWrapper_Exports: funspecs := withClient_Exports M ++ MyDeepWrapperMethods M PointRep CPointRep.
Goal exists phi, find_id _makePoint_I1 with_DeepWrapper_Exports = Some phi. simpl. Abort. (*makePoint_spec *)

Eval compute in (map fst with_DeepWrapper_Exports). (*makePoint_I1, _makeBPoint_I1, _makeCPoint_I1,
   _makePoint_I2, _makeBPoint_I2, _makeCPoint_I2
   146:test1 147:test2 155:test12, 126:GET , 127:SET, 128: BUMP, 129:GETC*)
Definition WithDeepWrapper_deepConstructors_aux := ASI_DeepWrapper.DeepConstructorsASI point bpoint cpoint _ _ (DP (MyWrapperAPD M _) _ _ ) M
               (* m1 bm1 cm1 Init_PointRep Init_CPointRep*)ptConstr1 bptConstr1 cptConstr1.
Definition WithDeepWrapper_deepConstructors := WithDeepWrapper_deepConstructors_aux _makePoint_I1 _makeBPoint_I1 _makeCPoint_I1 ++
                                               WithDeepWrapper_deepConstructors_aux _makePoint_I2 _makeBPoint_I2 _makeCPoint_I2.
Definition with_DeepWrapper_DeepExports := WithDeepWrapper_deepConstructors ++ skipn 6 with_DeepWrapper_Exports.

Goal (map fst with_DeepWrapper_DeepExports) = (map fst with_DeepWrapper_Exports). simpl; reflexivity. Qed.
Goal exists phi, find_id _makePoint_I1 with_DeepWrapper_DeepExports = Some phi. simpl. Abort. (*ASI_DeepWrapper.DP_makePOINT_spec*)

Definition with_DeepWrapper_VSU: @VSU NullExtension.Espec with_wrapper_Vprog with_wrapper_cs nil (I1_Imports M) with_wrapper with_DeepWrapper_Exports emp.
Proof. 
eapply VSU_ext.
  + eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ (withClient_VSU M) (DeepWrapper_VSU M PointRep CPointRep)).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - SC_tac.
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.
(*Eval compute in (map fst with_DeepWrapper_Exports).*) (*makePoint_I1, _makeBPoint_I1, _makeCPoint_I1,
   _makePoint_I2, _makeBPoint_I2, _makeCPoint_I2
   146:test1 147:test2 155:test12, 126:GET , 127:SET, 128: BUMP, 129:GETC*)
Goal compute_list_norepet (map fst (withClient_Exports M) ++ map fst (MyDeepWrapperMethods M PointRep CPointRep)) = true.
cbv; trivial. Qed.

Lemma my_funspec_sub_trans
     : forall f1 f2 f3 : funspec, funspec_sub f1 f2 -> funspec_sub f2 f3 -> funspec_sub f1 f3.
apply funspec_sub_trans.
Qed.

Definition with_DeepWrapper_DeepVSU: @VSU NullExtension.Espec with_wrapper_Vprog with_wrapper_cs nil (I1_Imports M) with_wrapper with_DeepWrapper_DeepExports emp.
Proof.
eapply VSU_Exports_sub. apply with_DeepWrapper_VSU.
LNR_tac. 
rewrite <- (firstn_skipn 6 with_DeepWrapper_Exports).
rewrite <- (firstn_skipn 6 with_DeepWrapper_DeepExports).
rewrite <-funspecs_sqsub_app.
- split.
  + simpl. (* unfold with_DeepWrapper_Exports, withClient_Exports, p12_Exports. simpl. *)
    specialize (ConstructorsSub (MyWrapperAPD M SimplePoints_VSU1.CompSpecs) _ _ (*m1 bm1 cm1*)ptConstr1 bptConstr1 cptConstr1 point bpoint cpoint M (*Init_PointRep Init_CPointRep*) _makePoint_I1 _makeBPoint_I1 _makeCPoint_I1); simpl; intros X.
    unfold ASI_DeepConstructors, ASI_DeepWrapper.DeepConstructorsASI  in X.
    apply funspecs_sqsub_D in X; [ destruct X as [? X] | LNR_tac | LNR_tac | cbv; reflexivity].
    apply funspecs_sqsub_D in X; [ destruct X as [? X] | LNR_tac | cbv; reflexivity].
    apply funspecs_sqsub_D in X; [ destruct X as [? X] | LNR_tac | cbv; reflexivity]. clear X.
    specialize (ConstructorsSub (MyWrapperAPD M SimplePoints_VSU2.CompSpecs) _ _ (*m1 bm1 cm1*)ptConstr1 bptConstr1 cptConstr1 point bpoint cpoint M (*Init_PointRep Init_CPointRep*) _makePoint_I2 _makeBPoint_I2 _makeCPoint_I2); simpl; intros X.
    unfold ASI_DeepConstructors, ASI_DeepWrapper.DeepConstructorsASI  in X.
    apply funspecs_sqsub_D in X; [ destruct X as [? X] | LNR_tac | LNR_tac | cbv; reflexivity].
    apply funspecs_sqsub_D in X; [ destruct X as [? X] | LNR_tac | cbv; reflexivity].
    apply funspecs_sqsub_D in X; [ destruct X as [? X] | LNR_tac | cbv; reflexivity].
    apply funspecs_sqsub_cons.
    { clear -H.  eapply my_funspec_sub_trans.
      eapply (makePt_sub M SimplePoints_VSU1.CompSpecs PointRep ptConstr1).
      eapply funspec_sub_trans. apply H.
      unfold ASI_DeepWrapper.DP_makePOINT_spec. simpl DP.
      unfold MyPOINT.
      clear. do_funspec_sub. Exists w emp. destruct w. entailer!. intros.
      Exists (eval_id ret_temp x). rewrite Point_cs1_eq. entailer!. apply derives_refl. }
    apply funspecs_sqsub_cons.
    { clear -H0. eapply funspec_sub_trans.
      apply (makeBPt_sub M SimplePoints_VSU1.CompSpecs PointRep bptConstr1).
      eapply funspec_sub_trans. apply H0.
      unfold ASI_DeepWrapper.DP_makeBPOINT_spec. simpl DP.
      unfold MyBPOINT.
      clear. do_funspec_sub. Exists w emp. destruct w. entailer!. intros.
      Exists (eval_id ret_temp x). rewrite BPoint_cs1_eq. entailer!. apply derives_refl. }
    apply funspecs_sqsub_cons.
    { clear -H1. eapply funspec_sub_trans.
      apply (makeCPt_sub M SimplePoints_VSU1.CompSpecs CPointRep cptConstr1).
      eapply funspec_sub_trans. apply H1.
      unfold ASI_DeepWrapper.DP_makeCPOINT_spec. simpl DP.
      unfold MyCPOINT.
      clear. do_funspec_sub. Exists w emp. destruct w as [[gv i] c]. entailer!. intros.
      Exists (eval_id ret_temp x). rewrite CPoint_cs1_eq. entailer!. apply derives_refl. }
    apply funspecs_sqsub_cons.
    { clear -H2. eapply funspec_sub_trans. 
      eapply (makePt_sub M SimplePoints_VSU2.CompSpecs PointRep ptConstr1).
      eapply funspec_sub_trans. apply H2.
      unfold ASI_DeepWrapper.DP_makePOINT_spec. simpl DP.
      unfold MyPOINT.
      clear. do_funspec_sub. Exists w emp. destruct w. entailer!. intros.
      Exists (eval_id ret_temp x). rewrite Point_cs2_eq. entailer!. apply derives_refl. }
    apply funspecs_sqsub_cons.
    { clear -H3. eapply funspec_sub_trans.
      apply (makeBPt_sub M SimplePoints_VSU2.CompSpecs PointRep bptConstr1).
      eapply funspec_sub_trans. apply H3.
      unfold ASI_DeepWrapper.DP_makeBPOINT_spec. simpl DP.
      unfold MyBPOINT.
      clear. do_funspec_sub. Exists w emp. destruct w. entailer!. intros.
      Exists (eval_id ret_temp x). rewrite BPoint_cs2_eq. entailer!. apply derives_refl. }
    apply funspecs_sqsub_cons.
    { clear -H4. eapply funspec_sub_trans. 
      apply (makeCPt_sub M SimplePoints_VSU2.CompSpecs CPointRep cptConstr1).
      eapply funspec_sub_trans. apply H4.
      unfold ASI_DeepWrapper.DP_makeCPOINT_spec. simpl DP.
      unfold MyCPOINT.
      clear. do_funspec_sub. Exists w emp. destruct w as [[gv i] c]. entailer!. intros.
      Exists (eval_id ret_temp x). rewrite CPoint_cs2_eq. entailer!. apply derives_refl. }
      trivial.
  + simpl. apply funspecs_sqsub_refl.
- cbv; reflexivity.
- cbv; reflexivity.
- LNR_tac.
Qed.
End With_DeepWrapper.
(*
Require Import WrapperClient.
Require Import ASI_WrapperClient.
Require Import VSU_DeepWrapperClient.

Section With_DeepWrapperClient.

Variable M:MallocFreeAPD.
Definition with_DeepWrapperClient_Exports: funspecs := with_DeepWrapper_Exports M ++ (ASI_WrapperClient M _test3 _test4).

Definition with_DeepWrapperClient_VSU: @VSU NullExtension.Espec with_wrapperclient_Vprog with_wrapperclient_cs nil (I1_Imports M) 
           with_wrapperclient with_DeepWrapperClient_Exports emp.
Proof. 
eapply VSU_ext. Locate LIST. Check (DeepWrapperClient_VSU M (DP (MyWrapperAPD M _) _ _ ) (LIST  M)). (LIST_NIL M)).
  + eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ (with_DeepWrapper_DeepVSU M) (DeepWrapper_VSU M PointRep CPointRep)).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - SC_tac.
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.*)
Require Import PointList.
Require Import ASI_DeepPointList.
Require Import VSU_DeepPointList.

Section With_DeepPointList.
Variable M:MallocFreeAPD. 

Definition with_DeepPointListExports:funspecs :=
       with_DeepWrapper_DeepExports M ++ PL_Internal_specs M (*(MyWrapperAPD M VSU_DeepWrapper.CompSpecs)*) PointRep CPointRep 
                                              (DP (MyWrapperAPD M VSU_DeepWrapper.CompSpecs) _ _).

Definition with_DeepPointList_VSU: @VSU NullExtension.Espec with_pointlist_Vprog with_pointlist_cs nil 
           (I1_Imports M) with_pointlist with_DeepPointListExports emp.
Proof. 
eapply VSU_ext.
  + eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ 
    (with_DeepWrapper_DeepVSU M) 
    (DeepPointList_VSU M (*(MyWrapperAPD M VSU_DeepWrapper.CompSpecs)*) PointRep CPointRep (DP (MyWrapperAPD M VSU_DeepWrapper.CompSpecs) _ _))).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - SC_tac. (*eexists; split. reflexivity.
         specialize (MethodsSub (MyWrapperAPD M VSU_LLWrapper.CompSpecs)) . Search ASI_DeepWrapper.DP_GET_spec. admit. admit. admit.*)
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.
End With_DeepPointList.

Require Import WrapperClient.
Require Import ASI_WrapperClient.
Require Import VSU_DeepWrapperClient.

Section With_DeepWrapperClient.
Variable M:MallocFreeAPD.
Definition with_DeepWrapperClient_Exports: funspecs := with_DeepPointListExports M ++ ASI_WrapperClient M _test3 _test4.
Definition with_DeepWrapperClient_VSU: @VSU NullExtension.Espec with_wrapperclient_Vprog with_wrapperclient_cs nil 
           (I1_Imports M) with_wrapperclient with_DeepWrapperClient_Exports emp.
Proof.
eapply VSU_ext.
  + eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ 
       (with_DeepPointList_VSU M) 
       (DeepWrapperClient_VSU M (DP (MyWrapperAPD M VSU_DeepWrapper.CompSpecs) PointRep CPointRep) (LIST  M) (LIST_NIL M))).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - SC_tac. (*Great*)
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

End With_DeepWrapperClient.

Require Import surelyMalloc. 
Require Import spec_surelyMalloc.
Require Import verif_surelyMalloc.

Section Application_VSU.

Variable M:MallocFreeAPD.
Definition Application_Exports: funspecs := with_DeepWrapperClient_Exports M ++ SM_ASI M.
Definition Application_VSU: @VSU NullExtension.Espec application_Vprog application_cs nil 
           [malloc_spec' M _malloc; spec_stdlib.exit_spec] application_prog Application_Exports emp.
Proof.
eapply VSU_ext.
  + eapply (VSULink _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ (with_DeepWrapperClient_VSU M) (SMVSU M)).
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - LNR_tac.
      - LNR_tac.
      - prove_cspecs_sub.
      - prove_cspecs_sub.
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
      - FDM_tac.
      - FunctionsPreserved_tac.
      - list_disjoint_tac.
      - list_disjoint_tac.
      - ExternsHyp_tac.
      - SC_tac.
      - SC_tac. 
      - HImports_tac.
      - ImportsDef_tac.
      - ExportsDef_tac.
      - LNR_tac.
      - LNR_tac.
      - domV_tac.
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - try (compute; reflexivity).
      - first [ find_id_subset_tac | idtac ].
      - first [ find_id_subset_tac | idtac ].
  + extensionality gv; simpl. rewrite emp_sepcon; trivial.
Qed.

End Application_VSU.
(*
Print Assumptions Application_VSU.
Parameter M:MallocFreeAPD.
Eval compute in (map fst (Application_Exports M)). (*constrcutors, insert, tests, .. -- probably just the test suffice*)
*)