Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import PositiveSubtypingTheory.

Require Import SimpleObjspec.
Require Import SimplePointModel.

Ltac do_compute_expr_warning::=idtac.

Global Open Scope funspec_scope.
Section Spec.
Variable point bpoint cpoint:type.
Variable T: Type. (*Fix/be parametric in a semantic representation, ie order and (Coq)types of record fields*)
Variable P:PointM T. (*For the given representation T, we assume model-level implementations of the exposed methods*)

  (*Parameter PR is the representation predicate that relates a concrete C struct to its model-level counterpart (an element of T*)
  Definition get_spec (PR: Pred T) :=
    WITH x:T, p: val
    PRE [ tptr point ]
     PROP()
     PARAMS (p) GLOBALS ()
     SEP(PR (x,p))
    POST[ tint ]
      PROP()
      RETURN (Vint (Int.repr (get _ P x)))
      SEP(PR (x,p)).

  Definition set_spec (PR: Pred T) :=
    WITH x:T, p: val, j:Z
    PRE [ tptr point, tint   ]
     PROP(Int.min_signed <= j <= Int.max_signed)
     PARAMS (p; Vint (Int.repr j)) GLOBALS ()
     SEP(PR (x,p))
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(PR (set _ P x j, p)).

Variable BP:BPointM T.
  Definition bump_spec  (PR: Pred T) :=
    WITH x:T, p:val
    PRE [ tptr bpoint ]
     PROP(bumpable _ BP x)
     PARAMS (p) GLOBALS ()
     SEP(PR (x,p))
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(PR (bump _ BP x, p)).

Variable CP:CPointM T.

  Definition getC_spec (PR: Pred T) :=
    WITH x:T, p: val
    PRE [ tptr cpoint ]
     PROP()
     PARAMS (p) GLOBALS ()
     SEP(PR (x,p))
    POST[ tint ]
      PROP()
      RETURN (Vint (Int.repr (color_encoding _ CP (getC _ CP x))))
      SEP(PR (x,p)).
End Spec.

Definition PointSpecs point T (P: PointM T):= [get_spec point T P; set_spec point T P].
Definition BPointSpecs point bpoint T (BP: BPointM T):= [get_spec point T BP; set_spec point T BP; bump_spec bpoint T BP].
Definition CPointSpecs point bpoint cpoint T (CP: CPointM T):= 
       [get_spec point T CP; set_spec point T CP; bump_spec bpoint T CP; getC_spec cpoint T CP].

Section Specs_Fash.

Lemma get_spec_fash point T (PM:PointM T) P Q v: funcptr_fash T P Q (get_spec point T PM) v.
Proof.
  unfold funcptr_fash. unfold func_ptr'.
  apply subp_andp; [ clear | apply subp_refl]. 
  eapply derives_trans; [ | apply fash_func_ptr_ND].
  apply allp_right; intros [ t p].
  apply andp_right. 
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
(*    unfold convertPre.*)
    unfold PROPx, PARAMSx, GLOBALSx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    (*rewrite prop_true_andp by tauto.*)
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (t,p).
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left2. apply derives_refl. apply derives_refl.
      rewrite andp_comm. apply modus_ponens.
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
    unfold convertPre.
    unfold PROPx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (t,p). (*([], (t,p)).*)
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left1. apply derives_refl. apply derives_refl.
      rewrite andp_comm. apply modus_ponens.
Qed.

Lemma set_spec_fash point T (PM:PointM T) P Q v: funcptr_fash T P Q (set_spec point T PM) v.
Proof.
  unfold funcptr_fash. unfold func_ptr'.
  apply subp_andp; [ clear | apply subp_refl]. 
  eapply derives_trans; [ | apply fash_func_ptr_ND].
  apply allp_right; intros [[t p] j].
  apply andp_right. 
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
(*    unfold convertPre.*)
    unfold PROPx, PARAMSx, GLOBALSx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    (*rewrite prop_true_andp by tauto.*)
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (t,p).
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left2. apply derives_refl. apply derives_refl.
      rewrite andp_comm. entailer!. apply modus_ponens.
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
    unfold convertPre.
    unfold PROPx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (set T PM t j,p). (*([], (t,p)).*)
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left1. apply derives_refl. apply derives_refl.
      rewrite andp_comm. apply modus_ponens.
Qed.

Lemma bump_spec_fash bpoint T (PM:BPointM T) P Q v: funcptr_fash T P Q (bump_spec bpoint T PM) v.
Proof.
  unfold funcptr_fash. unfold func_ptr'.
  apply subp_andp; [ clear | apply subp_refl]. 
  eapply derives_trans; [ | apply fash_func_ptr_ND].
  apply allp_right; intros [t p].
  apply andp_right. 
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
(*    unfold convertPre.*)
    unfold PROPx, PARAMSx, GLOBALSx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    (*rewrite prop_true_andp by tauto.*)
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (t,p).
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left2. apply derives_refl. apply derives_refl.
      rewrite andp_comm. entailer!. apply modus_ponens.
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
    unfold convertPre.
    unfold PROPx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (bump T PM t,p).
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left1. apply derives_refl. apply derives_refl.
      rewrite andp_comm. apply modus_ponens.
Qed.

Lemma getC_spec_fash cpoint T (CM:CPointM T) P Q v: funcptr_fash T P Q (getC_spec cpoint T CM) v.
Proof.
  unfold funcptr_fash. unfold func_ptr'.
  apply subp_andp; [ clear | apply subp_refl]. 
  eapply derives_trans; [ | apply fash_func_ptr_ND].
  apply allp_right; intros [t p].
  apply andp_right. 
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
(*    unfold convertPre.*)
    unfold PROPx, PARAMSx, GLOBALSx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    (*rewrite prop_true_andp by tauto.*)
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (t,p).
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left2. apply derives_refl. apply derives_refl.
      rewrite andp_comm. entailer!. apply modus_ponens.
  + apply allp_right; intro rho.
    apply subp_i1.
    rewrite unfash_allp.
    set (zz := allp _).
    unfold convertPre.
    unfold PROPx, LOCALx, SEPx, local, lift1; simpl.
    unfold_lift.
    normalize.
    subst zz.
    eapply derives_trans.
    apply andp_derives; [ | apply derives_refl].
    apply allp_left with (t,p).
    - apply unfash_fash.
    - eapply derives_trans.
      apply andp_derives.
      apply andp_left1. apply derives_refl. apply derives_refl.
      rewrite andp_comm. apply modus_ponens.
Qed.

Lemma Point_specs_fash point T (PM:PointM T) phi: In phi (PointSpecs point T PM) ->
            forall P Q v, funcptr_fash T P Q phi v.
Proof. simpl; intros. destruct H.
{ (*get*) subst phi. apply get_spec_fash. }
destruct H; [ | contradiction].
{ (*set*) subst phi. apply set_spec_fash. }
Qed.

Lemma BPoint_specs_fash point bpoint T (BM:BPointM T) phi: In phi (BPointSpecs point bpoint T BM) ->
            forall P Q v, funcptr_fash T P Q phi v.
Proof. simpl; intros. destruct H.
{ (*get*) subst phi. apply get_spec_fash. }
destruct H.
{ (*set*) subst phi. apply set_spec_fash. }
destruct H; [ | contradiction].
{ (*bump*) subst phi. apply bump_spec_fash. }
Qed.

Lemma CPoint_specs_fash point bpoint cpoint T (CM:CPointM T) phi: In phi (CPointSpecs point bpoint cpoint T CM) ->
            forall P Q v, funcptr_fash T P Q phi v.
Proof. simpl; intros. destruct H.
{ (*get*) subst phi. apply get_spec_fash. }
destruct H.
{ (*set*) subst phi. apply set_spec_fash. }
destruct H.
{ (*bump*) subst phi. apply bump_spec_fash. }
destruct H; [ | contradiction].
{ (*getC*) subst phi. apply getC_spec_fash. }
Qed.

End Specs_Fash.

Section PointPreds.
Variable point bpoint cpoint:type.
Variable methods bmethods cmethods:type.
Variable mtable g s b gC:ident.
Variable cs: compspecs.
Variable M:MallocFreeAPD.
Variable T: Type.

Definition prePoint PM:= Self_Mpred T cs M mtable point methods [g;s] (PointSpecs point _ PM).
Definition Point (PM: PointM T) : Pred T:= EX P, prePoint PM P.

Definition makePoint_spec (pc:PointConstr T): funspec :=
 match pc with {| ptM := PM ; ptConstr := Init_Point |} =>
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr point ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; Point PM (Init_Point i,p))
 end.

Definition preBPoint BPM := Self_Mpred T cs M mtable bpoint bmethods
           [ g; s; b] (BPointSpecs point bpoint _ BPM).
Definition BPoint (BPM: BPointM T): Pred T:= EX P, preBPoint BPM P.

Definition makeBPoint_spec (bc:BPointConstr T) :=
  match bc with {| bptM := BM; bptConstr := Init_Point |} =>
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr bpoint ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; BPoint BM (Init_Point i,p))
 end.

Definition preCPoint CPM := Self_Mpred T cs M mtable cpoint cmethods 
            [ g; s; b; gC] (CPointSpecs point bpoint cpoint _ CPM).
Definition CPoint (CPM: CPointM T): Pred T:= EX P, preCPoint CPM P.

Definition makeCPoint_spec (cc:CPointConstr T) :=
  match cc with {| cptM := CM; cptConstr:= Init_CPoint |} =>
 WITH gv: globals, i:Z, c:Color
 PRE [ tint, tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) 
    PARAMS (Vint(Int.repr i); Vint (Int.repr (color_encoding _ CM c))) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr cpoint ]
    EX p: val, PROP () RETURN (p)
     SEP (mem_mgr M gv; CPoint (*_*) CM (Init_CPoint i c,p))
 end.

Lemma prePoint_fold_unfold PM P x:
  prePoint PM P x = EX cso, (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                    (EX m : val, !! isptr m && FldAt T point mtable m x *
                                 |> MTable T M methods [g; s] m (PointSpecs point _ PM) (prePoint PM P)))%logic.
Proof. unfold prePoint. apply SelfMpred_fold_unfold'. apply Point_specs_fash. Qed.
Lemma Point_fold_unfold PM:
  Point PM = fun x => EX cso P,
                        P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt T point mtable m x *
                                     |> MTable T M methods [g; s] m (PointSpecs point _ PM) (prePoint PM P)).
Proof. unfold Point. simpl. extensionality x. 
  apply pred_ext.
   + Intros P. rewrite prePoint_fold_unfold. Intros cso. Exists cso P. entailer!. 
   + Intros cso P. Exists P. rewrite prePoint_fold_unfold. Exists cso. entailer!. Qed.

Lemma prePoint_isptr PM P t v: prePoint PM P (t,v) |-- !!(isptr v).
Proof. rewrite prePoint_fold_unfold.  entailer.
  sep_apply (@malloc_token_local_facts (@fst compspecs type cso) M Ews (@snd compspecs type cso)).
  entailer!. apply H. Qed.
Lemma prePoint_isptr' PM P tv: prePoint PM P tv |-- !!(isptr (snd tv)).
Proof. destruct tv. sep_apply prePoint_isptr. entailer!. Qed.
Lemma Point_isptr PM t v: Point PM (t,v) |-- !!(isptr v).
Proof. rewrite Point_fold_unfold. Intros cso P. 
sep_apply (@malloc_token_local_facts (@fst compspecs type cso) M Ews (@snd compspecs type cso)).
  entailer!. apply H. Qed.
Lemma Point_isptr' PM tv: Point PM tv |-- !!(isptr (snd tv)).
Proof. destruct tv. sep_apply Point_isptr. entailer!. Qed.

Lemma preBPoint_fold_unfold BPM P x:
  preBPoint BPM P x = EX cso, (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                    (EX m : val, !! isptr m && FldAt T bpoint mtable m x *
                                 |> MTable T M bmethods [g; s; b] m (BPointSpecs point bpoint _ BPM) (preBPoint BPM P)))%logic.
Proof. unfold preBPoint. apply SelfMpred_fold_unfold'. apply BPoint_specs_fash. Qed.
Lemma BPoint_fold_unfold BPM:
  BPoint BPM = fun x => EX cso P,
                        P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt T bpoint mtable m x *
                                     |> MTable T M bmethods [g; s; b] m (BPointSpecs point bpoint _ BPM) (preBPoint BPM P)).
Proof. unfold BPoint. simpl. extensionality x. 
  apply pred_ext.
   + Intros P. rewrite preBPoint_fold_unfold. Intros cso. Exists cso P. entailer!. 
   + Intros cso P. Exists P. rewrite preBPoint_fold_unfold. Exists cso. entailer!. Qed.

Lemma preBPoint_isptr PM P t v: preBPoint PM P (t,v) |-- !!(isptr v).
Proof. rewrite preBPoint_fold_unfold. entailer.
  sep_apply (@malloc_token_local_facts (@fst compspecs type cso) M Ews (@snd compspecs type cso)).
  entailer!. apply H. Qed.
Lemma preBPoint_isptr' PM P tv: preBPoint PM P tv |-- !!(isptr (snd tv)).
Proof. destruct tv. sep_apply preBPoint_isptr. entailer!. Qed.
Lemma BPoint_isptr PM t v: BPoint PM (t,v) |-- !!(isptr v).
Proof. rewrite BPoint_fold_unfold. Intros cso P. 
sep_apply (@malloc_token_local_facts (@fst compspecs type cso) M Ews (@snd compspecs type cso)).
  entailer!. apply H. Qed.
Lemma BPoint_isptr' PM tv: BPoint PM tv |-- !!(isptr (snd tv)).
Proof. destruct tv. sep_apply BPoint_isptr. entailer!. Qed.

Lemma preCPoint_fold_unfold CPM P x:
  preCPoint CPM P x = EX cso,(P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                    (EX m : val, !! isptr m && FldAt T cpoint mtable m x *
                                 |> MTable T M cmethods [g; s; b; gC] m (CPointSpecs point bpoint cpoint _ CPM) (preCPoint CPM P)))%logic.
Proof. unfold preCPoint. apply SelfMpred_fold_unfold'. apply CPoint_specs_fash. Qed.
Lemma CPoint_fold_unfold CPM:
  CPoint CPM = fun x => EX cso P,
                        P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
                        (EX m : val, !! isptr m && FldAt T cpoint mtable m x *
                                     |> MTable T M cmethods [g; s; b; gC] m (CPointSpecs point bpoint cpoint _ CPM) (preCPoint CPM P)).
Proof. unfold CPoint. simpl. extensionality x.
  apply pred_ext.
   + Intros P. rewrite preCPoint_fold_unfold. Intros cso. Exists cso P. entailer!. 
   + Intros cso P. Exists P. rewrite preCPoint_fold_unfold. Exists cso. entailer!. Qed.

Lemma preCPoint_isptr PM P t v: preCPoint PM P (t,v) |-- !!(isptr v).
Proof. rewrite preCPoint_fold_unfold. entailer.
  sep_apply (@malloc_token_local_facts (@fst compspecs type cso) M Ews (@snd compspecs type cso)).
  entailer!. apply H. Qed.
Lemma preCPoint_isptr' PM P tv: preCPoint PM P tv |-- !!(isptr (snd tv)).
Proof. destruct tv. sep_apply preCPoint_isptr. entailer!. Qed.
Lemma CPoint_isptr PM t v: CPoint PM (t,v) |-- !!(isptr v).
Proof. rewrite CPoint_fold_unfold. Intros cso P. 
sep_apply (@malloc_token_local_facts (@fst compspecs type cso) M Ews (@snd compspecs type cso)).
  entailer!. apply H.  Qed.
Lemma CPoint_isptr' PM tv: CPoint PM tv |-- !!(isptr (snd tv)).
Proof. destruct tv. sep_apply CPoint_isptr. entailer!. Qed.

Hint Resolve prePoint_isptr : saturate_local.
Hint Resolve prePoint_isptr' : saturate_local.
Hint Resolve preBPoint_isptr : saturate_local.
Hint Resolve preBPoint_isptr' : saturate_local.
Hint Resolve preCPoint_isptr : saturate_local.
Hint Resolve preCPoint_isptr' : saturate_local.
Hint Resolve Point_isptr : saturate_local.
Hint Resolve Point_isptr' : saturate_local.
Hint Resolve BPoint_isptr : saturate_local.
Hint Resolve BPoint_isptr' : saturate_local.
Hint Resolve CPoint_isptr : saturate_local.
Hint Resolve CPoint_isptr' : saturate_local.

End PointPreds.

Definition PointASI (point bpoint cpoint methods bmethods cmethods:type) (mtable g s b gC: ident) (T R:Type) pc bc cc
           cs M (makept makebpt makecpt:ident):=
           [(makept, makePoint_spec point methods mtable g s cs M T pc);
            (makebpt, makeBPoint_spec point bpoint bmethods mtable g s b cs M T bc);
            (makecpt, makeCPoint_spec point bpoint cpoint cmethods mtable g s b gC cs M R cc)].

Require Import SimpleClient.

Notation point := (Tstruct _point noattr).
Notation methods := (Tstruct _methods noattr).

Notation bpoint := (Tstruct _bpoint noattr).
Notation bmethods := (Tstruct _bmethods noattr).

Notation cpoint := (Tstruct _cpoint noattr).
Notation cmethods := (Tstruct _cmethods noattr).

Definition SimplePointASI := PointASI point bpoint cpoint methods bmethods cmethods 
           _mtable _get _set _bump _getC _ _ ptConstr1 bptConstr1 cptConstr1.


Section Funspec_Sub_set.
Variable T:Type.
Variable M:MallocFreeAPD.
Variable P: PointM T.
Variable BP: BPointM T.
Variable cs:compspecs.
Variable  mtable gg s b:ident.

Goal funspec_sub (set_spec point T P (Point point methods mtable gg s cs M T P))
                 (set_spec point T BP (BPoint point bpoint bmethods mtable gg s b cs M T BP)).
Proof. do_funspec_sub. destruct w as [[t p] j]; simpl; Intros. simpl in H; subst. rewrite ! sepcon_emp.
 Exists ((t,p),j). simpl.
  rewrite BPoint_fold_unfold ,Point_fold_unfold. Intros cso Q m. Exists emp cso Q m.
remember (MTable T M bmethods [gg; s; b] m (BPointSpecs point bpoint T BP)) as TABLE1.
 Abort.
  
Goal funspec_sub (set_spec point T BP (BPoint point bpoint bmethods mtable gg s b cs M T BP))
                 (set_spec point T P (Point point methods mtable gg s cs M T P)).
Proof. do_funspec_sub. destruct w as [[x p] j]; simpl; Intros. simpl in H; subst. rewrite ! sepcon_emp.
  rewrite Point_fold_unfold. Intros PR m; simpl. Exists ((x,p),j). simpl.
Abort.
End Funspec_Sub_set.
