Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import PositiveSubtypingTheory.
Require Import SimpleObjspec.
Require Import SimplePointModel.
Require Import SimplePoints_ASI.

Require Import SimplePoint_I1.
Instance CompSpecs : compspecs. make_compspecs prog. Defined.

Notation point1 := (Tstruct _point_I1 noattr).
Notation bpoint1 := (Tstruct _bpoint_I1 noattr).
Notation cpoint1 := (Tstruct _cpoint_I1 noattr).

Lemma align_compatible_point1_point {b i} (AL: align_compatible point1 (Vptr b i)):
      align_compatible point (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma align_compatible_bpoint1_bpoint {b i} (AL: align_compatible bpoint1 (Vptr b i)):
      align_compatible bpoint (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma align_compatible_cpoint1_cpoint {b i} (AL: align_compatible cpoint1 (Vptr b i)):
      align_compatible cpoint (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma field_compatible_point1_point_mtable p: field_compatible point1 [StructField _mtable] p ->
              field_compatible point [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_point1_point; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Lemma field_compatible_bpoint1_bpoint_mtable p: field_compatible bpoint1 [StructField _mtable] p ->
              field_compatible bpoint [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_bpoint1_bpoint; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Lemma field_compatible_cpoint1_cpoint_mtable p: field_compatible cpoint1 [StructField _mtable] p ->
              field_compatible cpoint [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_cpoint1_cpoint; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Definition Pred1: Pred PointRep := fun xp =>
 field_at Ews point1 [StructField _value] (Vint (Int.repr (value (fst xp)))) (snd xp).

Notation prePred1 MF := (prePoint point methods _mtable _get _set CompSpecs MF PointRep m1 Pred1).

Lemma get_core Espec x p: @semax CompSpecs Espec (func_tycontext f_get_I1 nil nil nil)
  (PROP ( )  LOCAL (temp _p p)  SEP (Pred1 (x, p)))
  (fn_body f_get_I1) (frame_ret_assert
     (function_body_ret_assert tint
        (PROP ( )
         RETURN (Vint (Int.repr (get PointRep m1 x))) 
         SEP (Pred1 (x, p)))) (stackframe_of f_get_I1)).
Proof. abbreviate_semax. unfold Pred1; simpl. forward. forward. Qed.
(*
Lemma get_core_normal Espec x p AA BB CC: @semax CompSpecs Espec (func_tycontext f_get_I1 nil nil nil)
  (PROP ( )  LOCAL (temp _p p)  SEP (Pred1 (x, p)))
  (fn_body f_get_I1)
(normal_ret_assert (PROPx AA (LOCALx BB (SEPx CC)))).
Proof. abbreviate_semax. unfold Pred1; simpl. forward.
  eapply semax_post.
  5:{ forward. }
 Search main_post. entailer. Qed.*)
(*Lemma dataverif_get: semax_body nil nil f_get_I1 (_get_I1, get_spec point _ m1 Pred1).
Proof.
  start_function. unfold Pred1; simpl. forward. forward.
Qed.
Lemma dataverif_set: semax_body nil nil f_set_I1 (_set_I1, set_spec point _ m1 Pred1).
Proof.
  start_function. unfold Pred1; simpl. forward. forward.
Qed.*)

Lemma dataverif_get_Point M: semax_body nil nil f_get_I1 (_get_I1, get_spec point _ m1 (prePred1 M)).
Proof.
  start_function. unfold Pred1; simpl. rewrite prePoint_fold_unfold. Intros cso m; simpl. forward. forward.
  unfold Pred1; simpl. rewrite prePoint_fold_unfold. Exists cso m. simpl. entailer!.
Qed. 

Lemma dataverif_set_Point M: semax_body nil nil f_set_I1 (_set_I1, set_spec point _ m1 (prePred1 M)).
Proof.
  start_function. unfold Pred1; simpl. rewrite prePoint_fold_unfold. Intros cso m. forward. forward.
  unfold Pred1; simpl. rewrite prePoint_fold_unfold. Exists cso m. simpl. entailer!. apply derives_refl.
Qed.
 
Definition BPred1: Pred PointRep := fun xp =>
 field_at Ews bpoint1 [StructField _value] (Vint (Int.repr (value (fst xp)))) (snd xp).
Notation preBPred1 MF := (preBPoint point bpoint bmethods _mtable _get _set _bump CompSpecs MF PointRep bm1 BPred1).
(*
Lemma dataverif_get_BPoint M: semax_body nil nil f_get_I1 (_get_I1, get_spec point _ bm1 (preBPred1 M)).
Proof.
  start_function. unfold BPred1; simpl. rewrite preBPoint_fold_unfold. Intros m; simpl. forward. forward.
  unfold Pred1; simpl. rewrite prePoint_fold_unfold. Exists m. simpl. entailer!.
Qed.
Lemma dataverif_get_CPoint M: semax_body nil nil f_get_I1 (_get_I1, get_spec point _ m1 (prePred1 M)).
Proof.
  start_function. unfold Pred1; simpl. rewrite prePoint_fold_unfold. Intros m; simpl. forward. forward.
  unfold Pred1; simpl. rewrite prePoint_fold_unfold. Exists m. simpl. entailer!.
Qed.*)
(*
Lemma dataverif_bump: semax_body nil nil f_bump_I1 (_bump_I1, bump_spec bpoint _ bm1 BPred1). (*Pred1*)
Proof.
  start_function. unfold BPred1; simpl.
  forward. forward.
  { entailer!. red in H. simpl in H.
    do 2 rewrite Int.signed_repr by rep_lia. rep_lia. }
  forward. (*unfold Pred1; simpl.
  unfold field_at; simpl; entailer!.*)
Qed.*)

(*Note: The variation in directory positivePointsSelf_BumpAsGetSet shows
 that it's actually preferable to verify wrt Gprog:=(BPoint1_MethSpecs M) 
  or even Gprog=(at least the method part of) CommonGprog rtaher than nil*)
Lemma dataverif_bump_BPoint M: semax_body nil nil f_bump_I1 (_bump_I1, bump_spec bpoint _ bm1 (preBPred1 M)). (*Pred1*)
Proof.
  start_function. unfold BPred1; simpl. rewrite preBPoint_fold_unfold. Intros cso m.
  forward. forward.
  { entailer!. red in H. simpl in H.
    do 2 rewrite Int.signed_repr by rep_lia. rep_lia. }
  forward.
  unfold BPred1; simpl. rewrite preBPoint_fold_unfold. Exists cso m. entailer!. apply derives_refl.
Qed.

Definition CPred1: Pred CPointRep := fun xp =>
 (field_at Ews cpoint1 [StructField _value] (Vint (Int.repr (value (fst xp)))) (snd xp) *
  field_at Ews cpoint1 [StructField _color] (Vint (Int.repr (color_encoding _ cm1 (color (fst xp))))) (snd xp))%logic.
Notation preCPred1 MF := (preCPoint point bpoint cpoint cmethods _mtable _get _set _bump _getC CompSpecs MF CPointRep cm1 CPred1).

Lemma dataverif_getC MF: semax_body nil nil f_getC_I1 (_getC_I1, getC_spec cpoint _ cm1 (preCPred1 MF)).
Proof.
  start_function. unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Intros cso m.
  forward. forward. unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Exists cso m. entailer!. apply derives_refl.
Qed.

(*We might be tempted to verify get wrt the following spec:*)
Lemma objverif_get M: semax_body nil nil f_get_I1 
      (_get_I1, get_spec point _ m1 (Point point methods _mtable _get _set CompSpecs M _ m1)).
Proof.
  start_function. rewrite Point_fold_unfold. Intros cso P m. simpl.
Abort. (*But there's no way we can complete this proof - P is abstract!
  Point - being based on obj_mpred - is a view that hides the C representation/layout of
  a Point1Rep object - spec based on the obj_mpred predicate (eg Point) are 
  hence only of use to for clients.
  The mechanism by which inner (C-layout-aware) specs are related to outer (layout-unaware but
(  semantic)representation-aware) specifications is object creation.*)

Require Import spec_surelyMalloc.

  Definition PointMspecs:= [get_spec; set_spec].
  Definition PointMT T M P m := MTable T M methods [_get; _set] m
           [get_spec point T P; set_spec point T P].

  Definition BPointMT T M (P:BPointM T)  m := MTable T M bmethods [_get; _set; _bump] m
           [get_spec point T P; set_spec point T P; bump_spec bpoint T P].

  Definition CPointMT T M (P:CPointM T)  m := MTable T M cmethods [_get; _set; _bump; _getC] m
           [get_spec point T P; set_spec point T P; bump_spec bpoint T P; getC_spec cpoint T P].

Section ConstructorSpecs.
Variable MF: MallocFreeAPD. 

Lemma make_PointMT:
  forall T P PR g s m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  malloc_token MF Ews (Tstruct _methods noattr) m *
  data_at Ews (Tstruct _methods noattr) (g,s) m
  |-- PointMT T MF P m PR)%logic.
Proof.
  intros.
  unfold PointMT, MTable. simpl.
  Exists Ews; entailer!.
  unfold_data_at (data_at Ews (Tstruct _methods noattr) _ m). cancel.
  Exists s g g s. entailer!. 
Qed.

Lemma make_laterPointMT:
  forall T P PR g s m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  malloc_token MF Ews (Tstruct _methods noattr) m *
  data_at Ews (Tstruct _methods noattr) (g,s) m
  |-- |> PointMT T MF P m PR)%logic.
Proof. intros. sep_apply make_PointMT. apply now_later. Qed.


Lemma make_BPointMT:
  forall T (P:BPointM T) PR g s b m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  func_ptr' (bump_spec bpoint T P PR) b *
  malloc_token MF Ews (Tstruct _bmethods noattr) m *
  data_at Ews (Tstruct _bmethods noattr) (g,(s,b)) m
  |-- BPointMT T MF P m PR)%logic. 
Proof.
  intros.
  unfold BPointMT, MTable. simpl.
  Exists Ews; entailer!.
  unfold_data_at (data_at Ews (Tstruct _bmethods noattr) _ m). cancel.
  Exists b s g g b s. entailer!. 
Qed.

Lemma make_laterBPointMT:
  forall T (P:BPointM T) PR g s b m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  func_ptr' (bump_spec bpoint T P PR) b *
  malloc_token MF Ews (Tstruct _bmethods noattr) m *
  data_at Ews (Tstruct _bmethods noattr) (g,(s,b)) m
  |-- |> BPointMT T MF P m PR)%logic. 
Proof. intros. sep_apply make_BPointMT. apply now_later. Qed.

Lemma make_CPointMT:
  forall T (P:CPointM T) PR g s b gC m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  func_ptr' (bump_spec bpoint T P PR) b *
  func_ptr' (getC_spec cpoint T P PR) gC *
  malloc_token MF Ews (Tstruct _cmethods noattr) m *
  data_at Ews (Tstruct _cmethods noattr) (g,(s,(b,gC))) m
  |-- CPointMT T MF P m PR)%logic. 
Proof.
  intros.
  unfold CPointMT, MTable. simpl.
  Exists Ews; entailer!.
  unfold_data_at (data_at Ews (Tstruct _cmethods noattr) _ m). cancel.
  Exists gC b s g. Exists g gC b s. entailer!. 
Qed.

Lemma make_laterCPointMT:
  forall T (P:CPointM T) PR g s b gC m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  func_ptr' (bump_spec bpoint T P PR) b *
  func_ptr' (getC_spec cpoint T P PR) gC *
  malloc_token MF Ews (Tstruct _cmethods noattr) m *
  data_at Ews (Tstruct _cmethods noattr) (g,(s,(b,gC))) m
  |-- |> CPointMT T MF P m PR)%logic. 
Proof. intros. sep_apply make_CPointMT. apply now_later. Qed.

Definition makePoint_I1_spec:= (_makePoint_I1,
           makePoint_spec point methods _mtable _get _set CompSpecs MF _ (*m1 Init_PointRep*)ptConstr1).

(*The Grog contains get/setspec for m1 and Pred1, but makePoint's spec mentions Point!
Definition Point1_Gprog : funspecs :=   (*ltac:(with_library prog*)
    ([(_get_I1, get_spec point _ m1 Pred1);
      (_set_I1, set_spec point _ m1 Pred1);
    makePoint_I1_spec] ++ SM_ASI MF)*)

Definition Point1_Gprog : funspecs :=   (*ltac:(with_library prog*)
    ([(_get_I1, get_spec point _ m1 (prePred1 MF));
      (_set_I1, set_spec point _ m1 (prePred1 MF));
    makePoint_I1_spec] ++ SM_ASI MF).

Definition Vprog : varspecs. mk_varspecs prog. Defined.

Lemma body_makePoint_I1: semax_body Vprog Point1_Gprog f_makePoint_I1 makePoint_I1_spec.
Proof.
unfold makePoint_I1_spec, makePoint_spec, ptConstr1.
    (*redundant function_pointers.*)
start_function. 
forward_call surelymalloc_spec_sub (point1, gv).
    { simpl. split3; trivial. rep_lia. }
Intros p.
forward.
 
forward_call surelymalloc_spec_sub (methods, gv).
    { simpl. split3; trivial. rep_lia. }
Intros m.
(*needed for application of make_PointMT 5 lines below*) make_func_ptr _get_I1.  make_func_ptr _set_I1. 
forward. forward. forward. forward.
Exists p. entailer!.

sep_apply (make_laterPointMT PointRep m1 (prePred1 MF)).
rewrite Point_fold_unfold. simpl.
unfold PointMT; simpl. Exists (CompSpecs, point1) Pred1 m. cancel.
unfold Pred1, FldAt. Exists m.
unfold_data_at (data_at Ews point1 _ p). entailer!. 
clear - H6. apply field_compatible_point1_point_mtable in H6. unfold field_at; entailer.
Qed.

(*Similarly for makeBPoint_I1*)
Definition makeBPoint_I1_spec := (_makeBPoint_I1,
           makeBPoint_spec point bpoint bmethods _mtable _get _set _bump CompSpecs MF _ (*bm1 Init_PointRep*)bptConstr1).
(*
(*Note that we use  get_spec _ bm1 BPred1 rather than get_spec _m1 Pred1)*)
Definition BPoint1_Gprog : funspecs := 
    ([(_get_I1, get_spec point _ bm1 BPred1);
      (_set_I1, set_spec point _ bm1 BPred1);
      (_bump_I1, bump_spec bpoint _ bm1 BPred1);
    makeBPoint_I1_spec] ++ SM_ASI MF).*)
Definition BPoint1_Gprog : funspecs := 
    ([(_get_I1, get_spec point _ bm1 (preBPred1 MF));
      (_set_I1, set_spec point _ bm1 (preBPred1 MF));
      (_bump_I1, bump_spec bpoint _ bm1 (preBPred1 MF));
    makeBPoint_I1_spec] ++ SM_ASI MF).

Lemma body_makeBPoint_I1: semax_body Vprog BPoint1_Gprog f_makeBPoint_I1 makeBPoint_I1_spec.
Proof.
unfold makeBPoint_I1_spec, makeBPoint_spec, bptConstr1.
    (*redundant function_pointers.*)
start_function. 
forward_call surelymalloc_spec_sub (bpoint1, gv).
    { simpl. split3; trivial. rep_lia. }
Intros p.
forward.
 
forward_call surelymalloc_spec_sub (bmethods, gv).
    { simpl. split3; trivial. rep_lia. }
Intros m.
(*needed for application of make_PointMT 5 lines below*) make_func_ptr _get_I1. make_func_ptr _set_I1. make_func_ptr _bump_I1.
forward. forward. forward. forward. forward.
Exists p. entailer!.

sep_apply (make_laterBPointMT PointRep bm1 (preBPred1 MF) (gv _get_I1)  (gv _set_I1)  (gv _bump_I1) m).
rewrite BPoint_fold_unfold. simpl. Exists (CompSpecs, bpoint1) BPred1 m. entailer!.
unfold BPointMT. simpl. unfold BPointSpecs. cancel.
rewrite sepcon_comm. apply sepcon_derives; [ | apply derives_refl ].
unfold BPred1; simpl.
unfold_data_at (data_at Ews bpoint1 _ p). cancel.
unfold FldAt. Exists m; simpl. entailer!.
clear - H6. apply field_compatible_bpoint1_bpoint_mtable in H6. unfold field_at; entailer.
Qed.

Definition makeCPoint_I1_spec := (_makeCPoint_I1,
           makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC CompSpecs MF _ (* cm1 Init_CPointRep*)cptConstr1).

(*Note that we use  get_spec _ cm1 CPred1 rather than get_spec _bm1 Pred1 or ...*)
(*Definition CPoint1_Gprog : funspecs :=  (* ltac:(with_library prog*)
    ([(_get_I1, get_spec point _ cm1 CPred1);
      (_set_I1, set_spec point _ cm1 CPred1);
      (_bump_I1, bump_spec bpoint _ cm1 CPred1);
      (_getC_I1, getC_spec cpoint _ cm1 CPred1);
    makeCPoint_I1_spec] ++ SM_ASI MF).*)
Definition CPoint1_Gprog : funspecs :=  (* ltac:(with_library prog*)
    ([(_get_I1, get_spec point _ cm1 (preCPred1 MF));
      (_set_I1, set_spec point _ cm1 (preCPred1 MF));
      (_bump_I1, bump_spec bpoint _ cm1 (preCPred1 MF));
      (_getC_I1, getC_spec cpoint _ cm1 (preCPred1 MF));
    makeCPoint_I1_spec] ++ SM_ASI MF).

Lemma body_makeCPoint_I1: semax_body Vprog CPoint1_Gprog f_makeCPoint_I1 makeCPoint_I1_spec.
Proof.
unfold makeCPoint_I1_spec, makeCPoint_spec, cptConstr1.
    (*redundant function_pointers.*)
start_function. 
forward_call surelymalloc_spec_sub (cpoint1, gv).
    { simpl. split3; trivial. rep_lia. }
Intros p.
forward. forward.
 
forward_call surelymalloc_spec_sub (cmethods, gv).
    { simpl. split3; trivial. rep_lia. }
Intros m.
(*needed for application of make_PointMT 5 lines below*) make_func_ptr _get_I1. make_func_ptr _set_I1.
   make_func_ptr _bump_I1. make_func_ptr _getC_I1.
forward. forward. forward. forward. forward. forward.
Exists p. entailer!.

sep_apply (make_laterCPointMT CPointRep cm1 (preCPred1 MF)).
rewrite CPoint_fold_unfold. simpl.
unfold CPointMT; simpl. Exists (CompSpecs, cpoint1) CPred1 m. cancel.
unfold CPred1, FldAt. Exists m.
unfold_data_at (data_at Ews cpoint1 _ p). entailer!. 
clear - H6; apply field_compatible_cpoint1_cpoint_mtable in H6. unfold field_at; entailer.
Qed.

Section Subsumption.

Lemma align_compatible_bpoint1_point1 {b i} (AL:align_compatible bpoint1 (Vptr b i)):
      align_compatible point1 (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1. 
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    inv H.
Qed.

Lemma field_compatible_bpoint1_point1_value {p} (FC:field_compatible bpoint1 [StructField _value] p):
      field_compatible point1 [StructField _value] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_bpoint1_point1; trivial.
  - split; simpl; auto. right; left; trivial.
Qed.

Lemma gsub_point_bpoint pt: funspec_sub (get_spec pt _ m1 Pred1) (get_spec pt _ bm1 BPred1).
Proof. do_funspec_sub. destruct w as [P p]; simpl; Intros. clear H.
unfold BPred1, Pred1; simpl.
Exists (P, p) emp.
entailer!.
+ intros. entailer!. unfold field_at; simpl. entailer!.
+ simpl in *. unfold field_at; simpl. entailer!.
  clear - H. apply field_compatible_bpoint1_point1_value; trivial. 
Qed.
 
Lemma gsub_point_bpoint_pre pt M: funspec_sub (get_spec pt _ m1 (prePred1 M)) (get_spec pt _ bm1 (preBPred1 M)).
Proof. do_funspec_sub. destruct w as [P p]; simpl; Intros. clear H.
unfold BPred1, Pred1; simpl.
Exists (P, p) emp.
entailer!.
+ intros. entailer!. unfold field_at; simpl. entailer!.
  rewrite prePoint_fold_unfold, preBPoint_fold_unfold. Intros m. Exists m. entailer. cancel.
(*+ simpl in *. unfold field_at; simpl. entailer!.
  clear - H. apply field_compatible_bpoint1_point1_value; trivial. 
Qed.*)Abort.

Lemma align_compatible_cpoint1_bpoint1 {b i} (AL:align_compatible cpoint1 (Vptr b i)):
      align_compatible bpoint1 (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1. 
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    inv H.
Qed.

Lemma field_compatible_cpoint1_bpoint1_value {p} (FC:field_compatible cpoint1 [StructField _value] p):
      field_compatible bpoint1 [StructField _value] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_cpoint1_bpoint1; trivial.
  - split; simpl; auto. right; left; trivial.
Qed.

Lemma gsub_bpoint_cpoint pt : funspec_sub (get_spec pt _ bm1 BPred1) (get_spec pt _ cm1 CPred1).
Proof. do_funspec_sub. destruct w as [CP p]; simpl; Intros. clear H.
destruct CP as [PT COL].
unfold CPred1, BPred1; simpl.
Exists (PT, p) 
       (field_at Ews cpoint1 [StructField _color] (Vint (Int.repr (mycolor_encoding COL))) p).
entailer!.
+ intros. entailer!. unfold field_at; simpl. entailer!.
+ simpl in *. unfold field_at; simpl. entailer!.
  clear - H. apply field_compatible_cpoint1_bpoint1_value; trivial.
Qed.

Lemma gsub_point_cpoint pt: funspec_sub (get_spec pt _ m1 Pred1) (get_spec pt _ cm1 CPred1).
Proof. eapply funspec_sub_trans. apply gsub_point_bpoint. apply gsub_bpoint_cpoint. Qed.

Lemma ssub_point_bpoint pt: funspec_sub (set_spec pt _ m1 Pred1) (set_spec pt _ bm1 BPred1).
Proof. do_funspec_sub. destruct w as [[P p] i]; simpl; Intros. clear H. rename H0 into BNDS.
unfold BPred1, Pred1; simpl.
Exists ((P, p), i) emp.
entailer!.
+ intros. entailer!. unfold field_at; simpl. entailer!.
+ unfold field_at; simpl. entailer!.
  clear - H. apply field_compatible_bpoint1_point1_value; trivial.
Qed.

Lemma ssub_bpoint_cpoint pt: funspec_sub (set_spec pt _ bm1 BPred1) (set_spec pt _ cm1 CPred1).
Proof. do_funspec_sub. destruct w as [[[PT COL] p] c]; simpl; Intros. clear H. rename H0 into BND.
unfold CPred1, BPred1; simpl.
Exists ((PT, p), c) 
       (field_at Ews cpoint1 [StructField _color] (Vint (Int.repr (mycolor_encoding COL))) p).
entailer!.
+ intros. entailer!. unfold field_at; simpl. entailer!.
+ unfold field_at; simpl. entailer!.
  clear - H. apply field_compatible_cpoint1_bpoint1_value; trivial.
Qed. 

Lemma ssub_point_cpoint pt: funspec_sub (set_spec pt _ m1 Pred1) (set_spec pt _ cm1 CPred1).
Proof. eapply funspec_sub_trans. apply ssub_point_bpoint. apply ssub_bpoint_cpoint. Qed.

Lemma bsub_bpoint_cpoint pt: funspec_sub (bump_spec pt _ bm1 BPred1) (bump_spec pt _ cm1 CPred1).
Proof. do_funspec_sub. destruct w as [[PT COL] p]; simpl; Intros. clear H. rename H0 into BND.
unfold CPred1, BPred1; simpl.
Exists (PT, p) 
       (field_at Ews cpoint1 [StructField _color] (Vint (Int.repr (mycolor_encoding COL))) p).
entailer!.
+ intros. entailer!. unfold field_at; simpl. entailer!.
+ unfold field_at; simpl. entailer!.
  clear - H. apply field_compatible_cpoint1_bpoint1_value; trivial.
Qed.
 
End Subsumption.

Definition gspec pt := binary_intersection' 
                   (binary_intersection' (get_spec pt _ m1 (prePred1 MF)) (get_spec pt _ bm1 (preBPred1 MF)) (eq_refl _) (eq_refl _))
                   (get_spec pt _ cm1 (preCPred1 MF)) (eq_refl _) (eq_refl _).

Definition sspec pt:= binary_intersection' 
                   (binary_intersection' (set_spec pt _ m1 (prePred1 MF)) (set_spec pt _ bm1 (preBPred1 MF)) (eq_refl _) (eq_refl _))
                   (set_spec pt _ cm1 (preCPred1 MF)) (eq_refl _) (eq_refl _).

Definition bspec pt:= binary_intersection' (bump_spec pt _ bm1 (preBPred1 MF)) (bump_spec pt _ cm1 (preCPred1 MF)) (eq_refl _) (eq_refl _).

Definition commonGprog:= (*ltac:(with_library prog*)
    ([(_get_I1, gspec point);
      (_set_I1, sspec point);
      (_bump_I1, bspec bpoint);
      (_getC_I1, getC_spec cpoint _ cm1 (preCPred1 MF));
    makePoint_I1_spec;
    makeBPoint_I1_spec;
    makeCPoint_I1_spec] ++ (*free_spec' MF _free ::*) SM_ASI MF )(*)*).

Lemma body_makeCPoint_I1_common: semax_body Vprog commonGprog f_makeCPoint_I1 makeCPoint_I1_spec.
Proof. eapply VSU.semax_body_subsumespec. apply body_makeCPoint_I1.
  * intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id
        by (apply compute_list_norepet_e; reflexivity). (* by LNR_tac.*)
    remember (initial_world.find_id i CPoint1_Gprog) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; reflexivity |]).
    congruence.
  * intros i; red. 
    remember (initial_world.find_id i CPoint1_Gprog ) as w.
    destruct w; [symmetry in Heqw | solve [trivial]]. simpl in Heqw.
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub. }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub. }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub. }
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.

Lemma body_makeBPoint_I1_common: semax_body Vprog commonGprog f_makeBPoint_I1 makeBPoint_I1_spec.
Proof. eapply VSU.semax_body_subsumespec. apply body_makeBPoint_I1.
  * intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). (* by LNR_tac.*)
    remember (initial_world.find_id i BPoint1_Gprog) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; try reflexivity |]).
    congruence.
  * intros i; red. 
    remember (initial_world.find_id i BPoint1_Gprog ) as w.
    destruct w; [symmetry in Heqw | solve [trivial]]. simpl in Heqw.
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply seplog.funspec_sub_trans
      with (f2 := binary_intersection' (get_spec point _ m1 (prePred1 MF)) (get_spec point _ bm1 (preBPred1 MF)) (eq_refl _) (eq_refl _));
      apply (binary_intersection'_sub). }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply seplog.funspec_sub_trans
      with (f2 := binary_intersection' (set_spec point _ m1 (prePred1 MF)) (set_spec point _ bm1 (preBPred1 MF)) (eq_refl _) (eq_refl _));
      apply (binary_intersection'_sub). }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si. apply binary_intersection'_sub. }
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.

Lemma body_makePoint_I1_common: semax_body Vprog commonGprog f_makePoint_I1 makePoint_I1_spec.
Proof. eapply VSU.semax_body_subsumespec. apply body_makePoint_I1.
  * intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). (* by LNR_tac.*)
    remember (initial_world.find_id i Point1_Gprog) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; try reflexivity |]).
    congruence.
  * intros i; red. 
    remember (initial_world.find_id i Point1_Gprog ) as w.
    destruct w; [symmetry in Heqw | solve [trivial]]. simpl in Heqw.
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply seplog.funspec_sub_trans
      with (f2 := binary_intersection' (get_spec point _ m1 (prePred1 MF)) (get_spec point _ bm1 (preBPred1 MF)) (eq_refl _) (eq_refl _));
      apply (binary_intersection'_sub). }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply seplog.funspec_sub_trans
      with (f2 := binary_intersection' (set_spec point _ m1 (prePred1 MF)) (set_spec point _ bm1 (preBPred1 MF)) (eq_refl _) (eq_refl _));
      apply (binary_intersection'_sub). }
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.
End ConstructorSpecs.

Require Import VST.floyd.VSU.
Section MakeVSU.
Variable M: MallocFreeAPD.

  Definition I1_Imports := (*[free_spec' M _free] ++*) SM_ASI M.
  Definition I1_InternalSpecs: funspecs:=[(_get_I1, gspec M point);
      (_set_I1, sspec M point);
      (_bump_I1, bspec M bpoint);
      (_getC_I1, getC_spec cpoint _ cm1 (preCPred1 M));
    makePoint_I1_spec M;
    makeBPoint_I1_spec M;
    makeCPoint_I1_spec M].

  Definition I1_Gprog:funspecs := I1_Imports ++ I1_InternalSpecs.
  Definition I1_ASI := SimplePointASI CompSpecs M _makePoint_I1 _makeBPoint_I1 _makeCPoint_I1.

Lemma dataverif_get_BPoint: @semax_body [] [] CompSpecs f_get_I1 (_get_I1, get_spec point PointRep bm1 (preBPred1 M)).
Proof.
  start_function. unfold BPred1; simpl. rewrite preBPoint_fold_unfold. Intros cso m; simpl.
  assert_PROP (field_compatible bpoint1 [StructField _value] p) as FC by entailer!.
  replace_SEP  0 (field_at Ews point1 [StructField _value] (Vint (Int.repr (value x))) p).
  { apply field_compatible_bpoint1_point1_value in FC. entailer!.
    unfold field_at; entailer. }
  forward. forward.
  unfold BPred1; simpl. rewrite preBPoint_fold_unfold. Exists cso m. simpl. entailer!.
  unfold field_at; entailer.
Qed. 

Lemma dataverif_get_CPoint: semax_body [] [] f_get_I1 (_get_I1, get_spec point CPointRep cm1 (preCPred1 M)).
Proof.
  start_function. unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Intros cso m; simpl.
  assert_PROP (field_compatible cpoint1 [StructField _value] p) as FC by entailer!.
  replace_SEP 0 (field_at Ews point1 [StructField _value] (Vint (Int.repr (value x))) p).
  { apply field_compatible_cpoint1_bpoint1_value in FC. 
    apply field_compatible_bpoint1_point1_value in FC. unfold field_at; entailer. }
  forward. forward.
  unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Exists cso m. simpl. entailer!.
  unfold field_at; entailer.
Qed.

Lemma dataverif_get_common: semax_body Vprog (commonGprog M) f_get_I1 (_get_I1, gspec M point).
Proof. eapply VSU.semax_body_subsumespec.
  + unfold gspec. 
    apply semax_body_binaryintersection''.
    - apply semax_body_binaryintersection''. apply dataverif_get_Point. apply dataverif_get_BPoint.
    - apply dataverif_get_CPoint. (*
    apply binary_intersection'_sub3.
    - apply binary_intersection'_sub3. apply funspec_sub_refl.
      apply gsub_point_bpoint.
    - apply gsub_point_cpoint.*)
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.

Lemma dataverif_set_BPoint: @semax_body [] [] CompSpecs f_set_I1 (_set_I1, set_spec point PointRep bm1 (preBPred1 M)).
Proof.
  start_function. unfold BPred1; simpl. rewrite preBPoint_fold_unfold. Intros cso m; simpl.
  assert_PROP (field_compatible bpoint1 [StructField _value] p) as FC by entailer!.
  replace_SEP 0 (field_at Ews point1 [StructField _value] (Vint (Int.repr (value x))) p).
  { apply field_compatible_bpoint1_point1_value in FC.
    unfold field_at; entailer. }
  forward. forward.
  unfold BPred1; simpl. rewrite preBPoint_fold_unfold. Exists cso m. simpl. entailer!.
  unfold field_at; entailer.
Qed. 

Lemma dataverif_set_CPoint: semax_body [] [] f_set_I1 (_set_I1, set_spec point CPointRep cm1 (preCPred1 M)).
Proof.
  start_function. unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Intros cso m; simpl.
  assert_PROP (field_compatible cpoint1 [StructField _value] p) as FC by entailer!.
  replace_SEP 0 (field_at Ews point1 [StructField _value] (Vint (Int.repr (value x))) p).
  { apply field_compatible_cpoint1_bpoint1_value in FC. 
    apply field_compatible_bpoint1_point1_value in FC. unfold field_at; entailer. }
  forward. forward.
  unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Exists cso m. simpl. entailer!.
  unfold field_at; entailer.
Qed.

Lemma dataverif_set_common: semax_body Vprog (commonGprog M) f_set_I1 (_set_I1, sspec M point).
Proof. eapply VSU.semax_body_subsumespec.
  + (*eapply semax_body_funspec_sub.  apply dataverif_set. 2: LNR_tac.
    apply binary_intersection'_sub3.
    - apply binary_intersection'_sub3. apply funspec_sub_refl.
      apply ssub_point_bpoint.
    - apply ssub_point_cpoint.*)
    apply semax_body_binaryintersection''.
    - apply semax_body_binaryintersection''. apply dataverif_set_Point. apply dataverif_set_BPoint.
    - apply dataverif_set_CPoint.
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.

(*see again the variation in directory positivePointsSelf_BumpAsGetSet*)
Lemma dataverif_bump_CPoint: semax_body [] [] f_bump_I1 (_bump_I1, bump_spec bpoint CPointRep cm1 (preCPred1 M)).
Proof.
  start_function. unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Intros cso m; simpl.
  assert_PROP (field_compatible cpoint1 [StructField _value] p) as FC by entailer!.
  replace_SEP 0 (field_at Ews bpoint1 [StructField _value] (Vint (Int.repr (value x))) p).
  { apply field_compatible_cpoint1_bpoint1_value in FC. unfold field_at; entailer. }
  forward. forward.
  { entailer!. red in H. simpl in H.
    do 2 rewrite Int.signed_repr by rep_lia. rep_lia. }
  forward.
  unfold CPred1; simpl. rewrite preCPoint_fold_unfold. Exists cso m. entailer!.
  unfold field_at; entailer.
Qed.

(*Note: The variation in directory positivePointsSelf_BumpAsGetSet shows
 that it's actually preferable to do semax_body_binaryintersection'' first, and
  then two separate applications of eapply VSU.semax_body_subsumespec., 
  or verify bump wrt CommonGrpog in the first place*)
Lemma dataverif_bump_common: semax_body Vprog (commonGprog M) f_bump_I1 (_bump_I1, bspec M bpoint).
Proof. eapply VSU.semax_body_subsumespec.
  + (*eapply semax_body_funspec_sub. apply dataverif_bump. 2: LNR_tac.
    apply binary_intersection'_sub3.
    - apply funspec_sub_refl.
    - apply bsub_bpoint_cpoint.*)
    apply semax_body_binaryintersection''. apply dataverif_bump_BPoint. apply dataverif_bump_CPoint.
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.

Lemma dataverif_getC_common: semax_body Vprog (commonGprog M) f_getC_I1 (_getC_I1, getC_spec cpoint _ cm1 (preCPred1 M)).
Proof. eapply VSU.semax_body_subsumespec. apply dataverif_getC.
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.

  Definition I1_Component: @Component NullExtension.Espec Vprog _ 
      nil I1_Imports prog I1_ASI emp I1_InternalSpecs.
  Proof. mkComponent_cautious. 
     - solve_SF_internal dataverif_get_common.
     - solve_SF_internal dataverif_set_common.
     - solve_SF_internal dataverif_bump_common.
     - solve_SF_internal dataverif_getC_common.
     - solve_SF_internal body_makePoint_I1_common.
     - solve_SF_internal body_makeBPoint_I1_common. 
     - solve_SF_internal body_makeCPoint_I1_common.
Qed.

  Definition I1_VSU: @VSU NullExtension.Espec Vprog _ 
      nil I1_Imports prog I1_ASI emp.
  Proof. eexists; apply I1_Component. Qed.

End MakeVSU.