Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import SimplePointModel.
Require Import SimplePoints_ASI.
Require Import VST.floyd.VSU.
Require Import Wrapper.
Require Import ASI_LLWrapper.

Instance CompSpecs : compspecs. make_compspecs prog. Defined.

Section Alignment_And_FieldCompatible_Lemmas.

Lemma align_compatible_point_bpoint {b i} (AL:align_compatible point (Vptr b i)):
      align_compatible bpoint (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1.
    inv H.
Qed.
Lemma align_compatible_bpoint_cpoint {b i} (AL:align_compatible bpoint (Vptr b i)):
      align_compatible cpoint (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1.
    inv H.
Qed.
Lemma align_compatible_point_cpoint {b i} (AL:align_compatible point (Vptr b i)):
      align_compatible cpoint (Vptr b i).
Proof. apply align_compatible_bpoint_cpoint. apply align_compatible_point_bpoint; trivial. Qed.

Lemma FC_mtable_point_bpoint {m}: field_compatible point [StructField _mtable] m ->
      field_compatible bpoint [StructField _mtable] m. 
Proof. intros [? [? [? [AL ?]]]]. split3; trivial. destruct m; inv H.
    split3.
    + simpl. simpl in H1. rep_lia.
    + clear - AL. apply align_compatible_point_bpoint; trivial.
    + split; simpl; trivial. left; trivial.
Qed.
 
Lemma FC_mtable_point_cpoint {m}: field_compatible point [StructField _mtable] m ->
      field_compatible cpoint [StructField _mtable] m. 
Proof. intros [? [? [? [AL ?]]]]. split3; trivial. destruct m; inv H.
    split3.
    + simpl. simpl in H1. rep_lia.
    + clear - AL. apply align_compatible_point_cpoint; trivial.
    + split; simpl; trivial. left; trivial.
Qed.

Lemma align_compatible_bmethods_methods {b i} (AL:align_compatible bmethods (Vptr b i)):
      align_compatible methods (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1. 
    inv H.
Qed.
Lemma align_compatible_cmethods_bmethods {b i} (AL:align_compatible cmethods (Vptr b i)):
      align_compatible bmethods (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1.
    if_tac in H.
    { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
      eapply align_compatible_rec_by_value. reflexivity.
      inv H0. inv H3. inv H. apply H0. }
    clear H1.
    inv H.
Qed.
Lemma align_compatible_cmethods_methods {b i} (AL:align_compatible cmethods (Vptr b i)):
      align_compatible methods (Vptr b i).
Proof. apply align_compatible_bmethods_methods. apply align_compatible_cmethods_bmethods; trivial. Qed.

Lemma FC_get_cmethods_bmethods {m}: field_compatible cmethods [StructField _get] m ->
      field_compatible bmethods [StructField _get] m. 
Proof. intros [? [? [? [AL ?]]]]. split3; trivial. destruct m; inv H.
    split3.
    + simpl. simpl in H1. rep_lia.
    + clear - AL. apply align_compatible_cmethods_bmethods; trivial. 
    + split; simpl; trivial. left; trivial.
Qed. 
Lemma FC_get_bmethods_methods {m}: field_compatible bmethods [StructField _get] m ->
      field_compatible methods [StructField _get] m.
Proof. intros [? [? [? [AL ?]]]]. split3; trivial. destruct m; inv H.
    split3.
    + simpl. simpl in H1. rep_lia.
    + clear - AL. apply align_compatible_bmethods_methods; trivial. 
    + split; simpl; trivial. left; trivial.
Qed.
Lemma FC_get_cmethods_methods {m}: field_compatible cmethods [StructField _get] m ->
      field_compatible methods [StructField _get] m.
Proof. intros. apply FC_get_bmethods_methods. apply FC_get_cmethods_bmethods; trivial. Qed.
Lemma FC_set_cmethods_bmethods {m}: field_compatible cmethods [StructField _set] m ->
      field_compatible bmethods [StructField _set] m. 
Proof. intros [? [? [? [AL ?]]]]. split3; trivial. destruct m; inv H.
    split3.
    + simpl. simpl in H1. rep_lia.
    + clear - AL. apply align_compatible_cmethods_bmethods; trivial. 
    + split; simpl; trivial. right; left; trivial.
Qed. 
Lemma FC_set_bmethods_methods {m}: field_compatible bmethods [StructField _set] m ->
      field_compatible methods [StructField _set] m.
Proof. intros [? [? [? [AL ?]]]]. split3; trivial. destruct m; inv H.
    split3.
    + simpl. simpl in H1. rep_lia.
    + clear - AL. apply align_compatible_bmethods_methods; trivial. 
    + split; simpl; trivial. right; left; trivial.
Qed.
Lemma FC_set_cmethods_methods {m}: field_compatible cmethods [StructField _set] m ->
      field_compatible methods [StructField _set] m.
Proof. intros. apply FC_set_bmethods_methods. apply FC_set_cmethods_bmethods; trivial. Qed.
Lemma FC_bump_cmethods_bmethods {m}: field_compatible cmethods [StructField _bump] m ->
      field_compatible bmethods [StructField _bump] m. 
Proof. intros [? [? [? [AL ?]]]]. split3; trivial. destruct m; inv H.
    split3.
    + simpl. simpl in H1. rep_lia.
    + clear - AL. apply align_compatible_cmethods_bmethods; trivial.
    + split; simpl; trivial. right; right; left; trivial.
Qed.

Lemma align_compatible_bpoint_point {b i} (AL: align_compatible bpoint (Vptr b i)):
      align_compatible point (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.
Lemma field_compatible_bpoint_point_mtable p: field_compatible bpoint [StructField _mtable] p ->
              field_compatible point [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_bpoint_point; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Lemma align_compatible_cpoint_bpoint {b i} (AL: align_compatible cpoint (Vptr b i)):
      align_compatible bpoint (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.
Lemma field_compatible_cpoint_bpoint_mtable p: field_compatible cpoint [StructField _mtable] p ->
              field_compatible bpoint [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_cpoint_bpoint; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Lemma align_compatible_cpoint_point {b i} (AL: align_compatible cpoint (Vptr b i)):
      align_compatible point (Vptr b i).
Proof. apply align_compatible_bpoint_point. apply align_compatible_cpoint_bpoint; trivial. Qed.
Lemma field_compatible_cpoint_point_mtable p: field_compatible cpoint [StructField _mtable] p ->
              field_compatible point [StructField _mtable] p.
Proof. intros. apply field_compatible_bpoint_point_mtable. apply field_compatible_cpoint_bpoint_mtable; trivial. Qed.

End Alignment_And_FieldCompatible_Lemmas.

Require Import SimpleObjspec.
(*
Section MyPredicates.
Variable M: MallocFreeAPD.
(*
(*Check (Point_GET_LLspec point M _ _mtable _get _set T). point (MyPOINT M) T*)
Definition MyPOINT DR T (PM: PointM T): Pred T :=
fun x => Point point methods _mtable _get _set _ M T DR PM x.

Definition MyBPOINT DR T (BPM: BPointM T): Pred T :=
fun x => BPoint point bpoint bmethods _mtable _get _set _bump _ M T DR BPM x.

Definition MyCPOINT DR T (CPM: CPointM T): Pred T :=
fun x=> CPoint point bpoint cpoint cmethods _mtable _get _set _bump _getC _ M T DR CPM x.
*)
(*
Definition MyWrapperAPD := Build_WrapperAPD MyPOINT MyBPOINT MyCPOINT.

Lemma makePt_sub T PM D initP: 
      funspec_sub (makePoint_spec point methods _mtable _get _set _ M T D PM initP)
                  (makePOINT_spec point MyWrapperAPD M T PM initP).
Proof. do_funspec_sub. Exists w emp. destruct w as [gv i]; clear H; entailer!.
  intros. Exists (eval_id ret_temp x); entailer!. Exists D; cancel.
Qed.
 
Lemma makeBPt_sub T BM D initP: 
      funspec_sub (makeBPoint_spec point bpoint bmethods _mtable _get _set _bump _ M T D BM initP)
                  (makeBPOINT_spec bpoint MyWrapperAPD  M T BM initP).
Proof. do_funspec_sub. Exists w emp. destruct w as [gv i]; clear H; entailer!.
  intros. Exists (eval_id ret_temp x); entailer!. Exists D; cancel.
Qed.

Lemma makeCPt_sub T CM D initCP: 
      funspec_sub (makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC _ M T D CM initCP)
                  (makeCPOINT_spec cpoint MyWrapperAPD M T CM initCP).
Proof. do_funspec_sub. Exists w emp. destruct w as [[gv i] c]; clear H; entailer!.
  intros. Exists (eval_id ret_temp x); entailer!. Exists D; cancel.
Qed. 
*)
(*
Lemma makePt_subLL T PM D initP: 
      funspec_sub (makePoint_spec point methods _mtable _get _set _ M T D PM initP)
                  (makePOINT_spec point MyPOINT M T PM initP).
Proof. do_funspec_sub. Exists w emp. destruct w as [gv i]; clear H; entailer!.
  intros. Exists (eval_id ret_temp x); entailer!. Exists D; cancel.
Qed.
 
Lemma makeBPt_sub T BM D initP: 
      funspec_sub (makeBPoint_spec point bpoint bmethods _mtable _get _set _bump _ M T D BM initP)
                  (makeBPOINT_spec bpoint MyBPOINT M T BM initP).
Proof. do_funspec_sub. Exists w emp. destruct w as [gv i]; clear H; entailer!.
  intros. Exists (eval_id ret_temp x); entailer!. Exists D; cancel.
Qed.

Lemma makeCPt_sub T CM D initCP: 
      funspec_sub (makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC _ M T D CM initCP)
                  (makeCPOINT_spec cpoint MyCPOINT M T CM initCP).
Proof. do_funspec_sub. Exists w emp. destruct w as [[gv i] c]; clear H; entailer!.
  intros. Exists (eval_id ret_temp x); entailer!. Exists D; cancel.
Qed.
Definition MyConstructorAPD : 
           WrapperConstructorAPD point bpoint cpoint methods bmethods cmethods MyWrapperAPD _mtable _get _set _bump _getC _ M.
Proof. econstructor. apply makePt_sub. apply makeBPt_sub. apply makeCPt_sub. Qed.*)
End MyPredicates.
*)
(*
Section VSU_Wrapper_HL. 
Variable M: MallocFreeAPD.
Notation WR := (MyWrapperAPD M).

Definition Vprog : varspecs. mk_varspecs prog. Defined.

Lemma verif_Point_GET T:semax_body Vprog nil f_GET (_GET, Point_GET_spec point WR T).
Proof.
  start_function. simpl; unfold MyPOINT; Intros DR.
  rewrite Point_fold_unfold. Intros P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. forward.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'.
  freeze [1;2;3;6;7;8] FR2. forward.
  forward_call. forward.
  Exists DR. rewrite Point_fold_unfold. Exists P m; simpl. thaw FR2; entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s. entailer!.
Qed.

Lemma verif_CPoint_GET T:semax_body Vprog nil f_GET (_GET, CPoint_GET_spec point WR T).
Proof.
  start_function; simpl. unfold MyCPOINT; Intros DR.
  rewrite CPoint_fold_unfold. Intros P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'.
  apply JMeq_eq in H. subst m'.
  assert_PROP (field_compatible cpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { apply field_compatible_cpoint_point_mtable in FC_mtable. unfold field_at; entailer. }
  forward.

  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'. apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  freeze [1;2;3;6;7;8;9;10;11;12] FR1. 
  replace_SEP 3 (field_at sh methods [StructField _get] g m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_get_cmethods_methods; trivial. }
  forward.
  forward_call. forward.
  Exists DR. rewrite CPoint_fold_unfold. Exists P m; simpl. thaw FR1. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s. entailer!. Exists b b gC gC. entailer!.
  unfold field_at; simpl; entailer.
Qed.

Lemma verif_Point_SET T:semax_body Vprog nil f_SET (_SET, Point_SET_spec point WR T).
Proof.
  start_function; simpl. rename H into BND. unfold MyPOINT; Intros DR.
  rewrite Point_fold_unfold. Intros P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. forward.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'. forward.
  forward_call (r,p,i). forward.
  Exists DR. rewrite Point_fold_unfold. Exists P m; simpl. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s. entailer!.
Qed.

Lemma verif_CPoint_SET T:semax_body Vprog nil f_SET (_SET, CPoint_SET_spec point WR T).
Proof.
  start_function; simpl. rename H into BND. unfold MyCPOINT; Intros DR.
  rewrite CPoint_fold_unfold. Intros P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. 
  assert_PROP (field_compatible cpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { apply field_compatible_cpoint_point_mtable in FC_mtable. unfold field_at; entailer. }
  forward.

  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'. 
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  freeze [1;2;3;4;5;8;9;10;11;12] FR1.
  replace_SEP 3 (field_at sh methods [StructField _set] s m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_set_cmethods_methods; trivial. }
  forward.
  forward_call (r,p,i). forward.
  Exists DR. rewrite CPoint_fold_unfold. Exists P m; simpl. thaw FR1. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s. entailer!. Exists b b gC gC. entailer!.
  unfold field_at; simpl; entailer.
Qed. 

Lemma verif_BPoint_BUMP T:semax_body Vprog nil f_BUMP (_BUMP, BPoint_BUMP_spec bpoint WR T).
Proof.
  start_function; simpl. rename H into BUMPABLE. unfold MyBPOINT; Intros DR.
  rewrite BPoint_fold_unfold. Intros BP m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'.
  (*replace_SEP 2 (field_at Ews bpoint [StructField _mtable] m p).
  { unfold field_at; entailer.  }*)
  forward. forward. forward.
  forward_call (r,p). forward.
  Exists DR. rewrite BPoint_fold_unfold. Exists BP m; simpl. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s b b. entailer!. (*unfold field_at. simpl. entailer!. apply derives_refl.*)
Qed.

Lemma verif_CPoint_BUMP T:semax_body Vprog nil f_BUMP (_BUMP, CPoint_BUMP_spec bpoint WR T).
Proof.
  start_function; simpl. rename H into BUMPABLE. unfold MyCPOINT; Intros DR.
  rewrite CPoint_fold_unfold. Intros CP m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'.
  assert_PROP (field_compatible cpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews bpoint [StructField _mtable] m p).
  { apply field_compatible_cpoint_bpoint_mtable in FC_mtable. unfold field_at; entailer. }
  forward.

  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'. apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  (*replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { unfold field_at; entailer. simpl. apply field_compatible_bpoint_point_mtable  in H0. entailer!. }*)
  forward. 
  replace_SEP 9 (field_at sh bmethods [StructField _bump] b m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_bump_cmethods_bmethods; trivial. }
  forward.
  forward_call (r,p). forward.
  Exists DR. rewrite CPoint_fold_unfold. Exists CP m; simpl. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s b b gC gC. entailer!. unfold field_at. entailer.
Qed.

Lemma verif_GETC T:semax_body Vprog nil f_GETC (_GETC, GETC_spec cpoint WR T).
Proof.
  start_function; simpl. unfold MyCPOINT; Intros DR.
  rewrite CPoint_fold_unfold. Intros CP m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  (*replace_SEP 2 (field_at Ews cpoint [StructField _mtable] m p).
  { unfold field_at; entailer. }*)
  forward. forward. forward.
  forward_call (r,p). forward.
  Exists DR. rewrite CPoint_fold_unfold. Exists CP m; simpl. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s b b gC gC. entailer!.
Qed.

Variable T CT:Type.
  (*Definition Wrapper_Imports:funspecs := nil.*)

  (*Plays role of internal_specs, Gprog, and WrapperASI*)
  Definition WS: funspecs:= 
    ASI_WrapperMethods point bpoint cpoint WR T CT _GET _SET _BUMP _GETC.

  Lemma verif_GET_common: semax_body Vprog WS f_GET (_GET, GET_spec point WR T CT). 
  Proof. eapply semax_body_subsumespec_GprogNil; [ | apply compute_list_norepet_e; reflexivity].
    eapply semax_body_binaryintersection''. apply verif_Point_GET. apply verif_CPoint_GET.
  Qed.
  Lemma verif_SET_common: semax_body Vprog WS f_SET (_SET, SET_spec point WR T CT). 
  Proof. eapply semax_body_subsumespec_GprogNil; [ | apply compute_list_norepet_e; reflexivity].
    eapply semax_body_binaryintersection''. apply verif_Point_SET. apply verif_CPoint_SET.
  Qed.
  Lemma verif_BUMP_common: semax_body Vprog WS f_BUMP (_BUMP, BUMP_spec bpoint WR T CT). 
  Proof. eapply semax_body_subsumespec_GprogNil; [ | apply compute_list_norepet_e; reflexivity].
    eapply semax_body_binaryintersection''. apply verif_BPoint_BUMP. apply verif_CPoint_BUMP.
  Qed.
  Lemma verif_GETC_common: semax_body Vprog WS f_GETC (_GETC, GETC_spec cpoint WR CT). 
  Proof. eapply semax_body_subsumespec_GprogNil; [ apply verif_GETC | apply compute_list_norepet_e; reflexivity]. Qed.

  Definition WrapperMethods_Component: @Component NullExtension.Espec Vprog _ nil nil prog WS emp WS.
  Proof. mkComponent_cautious.
     - solve_SF_internal verif_GET_common. 
     - solve_SF_internal verif_SET_common.
     - solve_SF_internal verif_BUMP_common.
     - solve_SF_internal verif_GETC_common.
  Qed.

  Definition WrapperMethods_VSU: @VSU NullExtension.Espec Vprog _ nil nil prog WS emp.
  Proof. eexists; apply WrapperMethods_Component. Qed.

End VSU_Wrapper.
*)
Section VSU_LLWrapper.
Variable M: MallocFreeAPD.

Definition Vprog : varspecs. mk_varspecs prog. Defined.
(*Definition MyPOINT' T DR (PM: PointM T): Pred T :=
fun x => Point point methods _mtable _get _set _ M T DR PM x.
Check (Point_GET_LLspec point M _ _mtable _get _set). (*: Type -> funspec*)
Check (MyPOINT M).*)
Lemma verif_Point_GET T:semax_body Vprog nil f_GET (_GET, Point_GET_LLspec point M _ _mtable _get _set T). (* (MyPOINT M) T).*)
Proof. 
  unfold Point_GET_LLspec.
  start_function. (* unfold MyPOINT; Intros DR.*)
  rewrite Point_fold_unfold. Intros cso P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. forward.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'.
  freeze [1;2;3;6;7;8] FR2. forward.
  forward_call (r,p). { assert (Frame = nil); subst Frame; trivial. simpl.
                        rewrite prePoint_fold_unfold. Exists cso m. thaw FR2; entailer!.
                        cancel. unfold FldAt. Exists m. entailer!.
                        eapply derives_trans; [ | apply now_later].
                        unfold MTable. Exists sh; simpl. Exists g g s s. entailer!. }
  forward. unfold Point. simpl. Exists P. trivial.
(*  (*Exists DR.*) rewrite Point_fold_unfold. Exists P m; simpl. (*thaw FR2; *)entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s. entailer!.*)
Qed.

Lemma verif_BPoint_GET T:semax_body Vprog nil f_GET (_GET, BPoint_GET_LLspec point bpoint M _ _mtable _get _set _bump T).
Proof. 
  unfold BPoint_GET_LLspec.
  start_function. (* unfold MyPOINT; Intros DR.*)
  rewrite BPoint_fold_unfold. Intros cso P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'.
  assert_PROP (field_compatible bpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { apply field_compatible_bpoint_point_mtable in FC_mtable. unfold field_at; entailer. }
  forward.

  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b'. apply JMeq_eq in H0; subst g'. apply JMeq_eq in H1; subst s'. apply JMeq_eq in H2; subst b'.
  replace_SEP 5 (field_at sh methods [StructField _get] g m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_get_bmethods_methods; trivial. }
  forward.
  forward_call (r,p). { assert (Frame = nil); subst Frame; trivial. simpl.
                        rewrite preBPoint_fold_unfold. Exists cso m. entailer!.
                        cancel. unfold FldAt. Exists m. entailer!.
                        rewrite ! sepcon_assoc. apply sepcon_derives.
                        { unfold field_at; simpl; entailer. } 
                        eapply derives_trans; [ | apply now_later].
                        unfold MTable. Exists sh; simpl. Exists g g s s b b. entailer!.
                        unfold field_at; simpl; entailer. }
  forward. unfold BPoint; simpl. Exists P; trivial.
(*  (*Exists DR.*) rewrite BPoint_fold_unfold. Exists P m; simpl. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s b b. entailer!.
  unfold field_at; simpl; entailer.*)
Qed.

Lemma verif_CPoint_GET T:semax_body Vprog nil f_GET (_GET, CPoint_GET_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC T).
Proof.
  unfold CPoint_GET_LLspec.
  start_function. (*unfold MyCPOINT; Intros DR.*)
  rewrite CPoint_fold_unfold. Intros cso P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'.
  apply JMeq_eq in H. subst m'.
  assert_PROP (field_compatible cpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { apply field_compatible_cpoint_point_mtable in FC_mtable. unfold field_at; entailer. }
  forward.

  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'. apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  freeze [1;2;3;6;7;8;9;10;11;12] FR1. 
  replace_SEP 3 (field_at sh methods [StructField _get] g m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_get_cmethods_methods; trivial. }
  forward.
  forward_call (r,p). { assert (Frame = nil); subst Frame; trivial. simpl.
                        rewrite preCPoint_fold_unfold. Exists cso m. thaw FR1; entailer!.
                        cancel. unfold FldAt. Exists m. entailer!.
                        rewrite ! sepcon_assoc. apply sepcon_derives.
                        { unfold field_at; simpl; entailer. } 
                        eapply derives_trans; [ | apply now_later].
                        unfold MTable. Exists sh; simpl. Exists g g s s b b gC gC. entailer!.
                        unfold field_at; simpl; entailer. }
  forward. unfold CPoint; simpl. Exists P; trivial.
(*  (*Exists DR.*) rewrite CPoint_fold_unfold. Exists P m; simpl. thaw FR1. entailer!.
  unfold FldAt. Exists m; simpl. entailer!.
  unfold MTable. Exists sh. unfold funcptrs; simpl.
  Exists g g s s. entailer!. Exists b b gC gC. entailer!.
  unfold field_at; simpl; entailer.*)
Qed.

Lemma verif_Point_SET T:semax_body Vprog nil f_SET (_SET, Point_SET_LLspec point M _ _mtable _get _set T).
Proof.
  unfold Point_SET_LLspec.
  start_function. rename H into BND. (* unfold MyPOINT; Intros DR.*)
  rewrite Point_fold_unfold. Intros cso P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. forward.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'. forward.
  forward_call (r,p,i).
  { assert (Frame = nil); subst Frame; trivial. simpl.
    rewrite prePoint_fold_unfold. Exists cso m; unfold FldAt; Exists m; simpl; entailer!.
    eapply derives_trans; [ | apply now_later].
    unfold MTable. Exists sh; simpl. Exists g g s s. entailer!. }
  forward. unfold Point; simpl. Exists P; trivial.
Qed.

Lemma verif_BPoint_SET T:semax_body Vprog nil f_SET (_SET, BPoint_SET_LLspec point bpoint M _ _mtable _get _set _bump T).
Proof.
  unfold BPoint_SET_LLspec.
  start_function. rename H into BND. (* unfold MyPOINT; Intros DR.*)
  rewrite BPoint_fold_unfold. Intros cso P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'.
  assert_PROP (field_compatible bpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { apply field_compatible_bpoint_point_mtable in FC_mtable. unfold field_at; entailer. }
  forward. 
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b'. apply JMeq_eq in H0; subst g'. apply JMeq_eq in H1; subst s'. apply JMeq_eq in H2; subst b'.
  replace_SEP 7 (field_at sh methods [StructField _set] s m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_set_bmethods_methods; trivial. }
  forward.
  forward_call (r,p,i).
  { assert (Frame = nil); subst Frame; trivial. simpl.
    rewrite preBPoint_fold_unfold. Exists cso m. unfold FldAt. Exists m. entailer!. cancel.
    rewrite ! sepcon_assoc. apply sepcon_derives.
    { unfold field_at; simpl; entailer. } 
    eapply derives_trans; [ | apply now_later].
    unfold MTable. Exists sh; simpl. Exists g g s s b b. entailer!.
    unfold field_at; simpl; entailer. }
  entailer!. unfold BPoint; simpl. Exists P; trivial.
Qed.

Lemma verif_CPoint_SET T:semax_body Vprog nil f_SET (_SET, CPoint_SET_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC T).
Proof.
  unfold CPoint_SET_LLspec.
  start_function. rename H into BND. (*unfold MyCPOINT; Intros DR.*)
  rewrite CPoint_fold_unfold. Intros cso P m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. 
  assert_PROP (field_compatible cpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { apply field_compatible_cpoint_point_mtable in FC_mtable. unfold field_at; entailer. }
  forward.

  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'. 
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  freeze [1;2;3;4;5;8;9;10;11;12] FR1.
  replace_SEP 3 (field_at sh methods [StructField _set] s m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_set_cmethods_methods; trivial. }
  forward.
  forward_call (r,p,i).
  { assert (Frame = nil); subst Frame; trivial. simpl.
    rewrite preCPoint_fold_unfold. Exists cso m. unfold FldAt. Exists m. thaw FR1; entailer!. cancel.
    rewrite ! sepcon_assoc. apply sepcon_derives.
    { unfold field_at; simpl; entailer. } 
    eapply derives_trans; [ | apply now_later].
    unfold MTable. Exists sh; simpl. Exists g g s s b b gC gC. entailer!.
    unfold field_at; simpl; entailer. }
  entailer!. unfold CPoint; simpl. Exists P; trivial.
Qed. 

Lemma verif_BPoint_BUMP T:semax_body Vprog nil f_BUMP (_BUMP, BPoint_BUMP_LLspec point bpoint M _ _mtable _get _set _bump T).
Proof.
  unfold BPoint_BUMP_LLspec.
  start_function. rename H into BUMPABLE. (*unfold MyBPOINT; Intros DR.*)
  rewrite BPoint_fold_unfold. Intros cso BP m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. 
  (*moved a forward here, so that later is consumed*)forward.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'.
  (*replace_SEP 2 (field_at Ews bpoint [StructField _mtable] m p).
  { unfold field_at; entailer.  }*)
  forward. forward. (*remoave a forward here forward.*)
  forward_call (r,p).
  { assert (Frame = nil); subst Frame; trivial. simpl.
    rewrite preBPoint_fold_unfold. Exists cso m. unfold FldAt. Exists m. entailer!. cancel.
    eapply derives_trans; [ | apply now_later].
    unfold MTable. Exists sh; simpl. Exists g g s s b b. entailer!. }
  entailer!. unfold BPoint; simpl. Exists BP; trivial.
Qed.

Lemma verif_CPoint_BUMP T:semax_body Vprog nil f_BUMP (_BUMP, CPoint_BUMP_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC T).
Proof.
  unfold CPoint_BUMP_LLspec.
  start_function. rename H into BUMPABLE. (*unfold MyCPOINT; Intros DR.*)
  rewrite CPoint_fold_unfold. Intros cso CP m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'.
  assert_PROP (field_compatible cpoint [StructField _mtable] p) as FC_mtable by entailer!. (*We need to remember this fact*)
  replace_SEP 2 (field_at Ews bpoint [StructField _mtable] m p).
  { apply field_compatible_cpoint_bpoint_mtable in FC_mtable. unfold field_at; entailer. }
  forward.

  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'. apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  (*replace_SEP 2 (field_at Ews point [StructField _mtable] m p).
  { unfold field_at; entailer. simpl. apply field_compatible_bpoint_point_mtable  in H0. entailer!. }*)
  forward. 
  replace_SEP 9 (field_at sh bmethods [StructField _bump] b m).
  { entailer. unfold field_at; simpl; entailer!. apply FC_bump_cmethods_bmethods; trivial. }
  forward.
  forward_call (r,p).
  { assert (Frame = nil); subst Frame; trivial. simpl.
    rewrite preCPoint_fold_unfold. Exists cso m. unfold FldAt. Exists m. entailer!. cancel.
    rewrite ! sepcon_assoc. apply sepcon_derives.
    { unfold field_at; simpl; entailer. } 
    eapply derives_trans; [ | apply now_later].
    unfold MTable. Exists sh; simpl. Exists g g s s b b gC gC. entailer!.
    unfold field_at; simpl; entailer. }
  entailer!. unfold CPoint; simpl. Exists CP; trivial. 
Qed.

Lemma verif_GETC T:semax_body Vprog nil f_GETC (_GETC, GETC_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC T).
Proof.
  unfold GETC_LLspec.
  start_function. (*unfold MyCPOINT; Intros DR.*)
  rewrite CPoint_fold_unfold. Intros cso CP m; simpl.
  unfold FldAt; Intros m'; simpl. simpl in m'. apply JMeq_eq in H; subst m'. (*moved a forward here*)forward.
  unfold MTable; Intros sh. unfold funcptrs; simpl.
  Intros g g' s s' b b' gC gC'. apply JMeq_eq in H0; subst g'.  apply JMeq_eq in H1; subst s'.
  apply JMeq_eq in H2; subst b'. apply JMeq_eq in H3; subst gC'.
  (*replace_SEP 2 (field_at Ews cpoint [StructField _mtable] m p).
  { unfold field_at; entailer. }*)
  forward. forward. (*remove a forward here forward.*)
  forward_call (r,p).
  { assert (Frame = nil); subst Frame; trivial. simpl.
    rewrite preCPoint_fold_unfold. Exists cso m. unfold FldAt. Exists m. entailer!. cancel.
    eapply derives_trans; [ | apply now_later].
    unfold MTable. Exists sh; simpl. Exists g g s s b b gC gC. entailer!. }
  forward. unfold CPoint; simpl. Exists CP; trivial. 
Qed. 

Variable T CT:Type.

  (*Definition Wrapper_Imports:funspecs := nil.*)

  Definition WS: funspecs:= ASI_LLWrapper point bpoint cpoint M _ _mtable _get _set _bump _getC _GET _SET _BUMP _GETC T CT.
  (*Plays role of internal_specs, Gprog, and WrapperASI*)

  Lemma verif_GET_common: semax_body Vprog WS f_GET (_GET, GET_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC T CT). 
  Proof. eapply semax_body_subsumespec_GprogNil; [ | apply compute_list_norepet_e; reflexivity].
    eapply semax_body_binaryintersection''. apply verif_Point_GET. 
    eapply semax_body_binaryintersection''. apply verif_BPoint_GET. apply verif_CPoint_GET.
  Qed.
  Lemma verif_SET_common: semax_body Vprog WS f_SET (_SET, SET_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC T CT).
  Proof. eapply semax_body_subsumespec_GprogNil; [ | apply compute_list_norepet_e; reflexivity].
    eapply semax_body_binaryintersection''. apply verif_Point_SET.
    eapply semax_body_binaryintersection''. apply verif_BPoint_SET. apply verif_CPoint_SET.
  Qed.
  Lemma verif_BUMP_common: semax_body Vprog WS f_BUMP (_BUMP, BUMP_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC T CT).
  Proof. eapply semax_body_subsumespec_GprogNil; [ | apply compute_list_norepet_e; reflexivity].
    eapply semax_body_binaryintersection''. apply verif_BPoint_BUMP. apply verif_CPoint_BUMP.
  Qed.

  Lemma verif_GETC_common: semax_body Vprog WS f_GETC (_GETC, GETC_LLspec point bpoint cpoint M _ _mtable _get _set _bump _getC CT).
  Proof. eapply semax_body_subsumespec_GprogNil; [ apply verif_GETC | apply compute_list_norepet_e; reflexivity]. Qed.

  Definition LLWrapper_Component: @Component NullExtension.Espec Vprog _ nil nil prog WS emp WS.
  Proof. mkComponent_cautious.
     - solve_SF_internal verif_GET_common. 
     - solve_SF_internal verif_SET_common.
     - solve_SF_internal verif_BUMP_common.
     - solve_SF_internal verif_GETC_common.
  Qed.

  Definition LLWrapper_VSU: @VSU NullExtension.Espec Vprog _ nil nil prog WS emp.
  Proof. eexists; apply LLWrapper_Component. Qed.

End VSU_LLWrapper.
