Require Import VST.floyd.proofauto.
Require Import ASI_malloc.
Require Import SimpleObjspec.
Require Import SimplePointModel.
Require Import DeepPoints.

Require Import ASI_HLWrapper.

Section DeepSpec. (* ;-) *)
Variable point bpoint cpoint:type.
Variable T CT:Type.
Notation DeepPoint := (DeepPoint T CT).
Notation Deep_point := (Deep_point T CT).
Notation Deep_bpoint := (Deep_bpoint T CT).
Notation Deep_cpoint := (Deep_cpoint T CT).

Notation dp_get := (dp_get T CT).
Notation dp_set := (dp_set T CT).

Notation dp_bumpable := (dp_bumpable T CT).
Notation dp_bump := (dp_bump T CT).
Notation dp_getc := (dp_getc T CT).

Variable DP: Pred DeepPoint.
 
Definition DP_GET_spec:=
  WITH d:DeepPoint, p:val
  PRE [ tptr point]
  PROP () PARAMS (p) GLOBALS () SEP (DP (d,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (dp_get d))) SEP (DP (d,p)).

Definition DP_SET_spec :=
  WITH d:DeepPoint, p:val, i: Z
  PRE [ tptr point, tint]
  PROP (Int.min_signed <= i <= Int.max_signed) PARAMS (p; Vint(Int.repr i)) GLOBALS ()
  SEP (DP (d,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (DP (dp_set d i,p)).

(*dp_bumpable enforces that d is in fact a bpoint or a cpoint*)
Definition DP_BUMP_spec:=
  WITH d:DeepPoint, p:val
  PRE [ tptr bpoint]
  PROP (dp_bumpable d) PARAMS (p) GLOBALS () SEP (DP (d,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (DP (dp_bump d,p)).
(*
Definition DP_GETC_spec :=
  WITH c:CT, cm:CPointM CT, p:val
  PRE [ tptr cpoint]
  PROP () PARAMS (p) GLOBALS ()
  SEP (DP (Deep_cpoint c cm,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (dp_getc (Deep_cpoint c cm))))
          SEP (DP (Deep_cpoint c cm,p)).*)
Definition DP_GETC_spec:=
  WITH d:DeepPoint, p:val
  PRE [ tptr cpoint]
  PROP (dp_colored _ _ d) PARAMS (p) GLOBALS ()
  SEP (DP (d,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (dp_getc d)))
          SEP (DP (d,p)).

Variable M: MallocFreeAPD.

Definition DP_makePOINT_spec pc :=
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr point ]
    EX p: val, PROP (isptr p) RETURN (p)
     SEP (mem_mgr M gv; DP (Deep_point (ptConstr _ pc i) (ptM _ pc),p)).

Definition DP_makeBPOINT_spec bc:=
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr bpoint ]
    EX p: val, PROP (isptr p) RETURN (p)
     SEP (mem_mgr M gv; DP (Deep_bpoint (bptConstr _ bc i) (bptM _ bc),p)).

Definition DP_makeCPOINT_spec cc:=
 WITH gv: globals, i:Z, c:Color
 PRE [ tint, tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) 
    PARAMS (Vint(Int.repr i); Vint (Int.repr (color_encoding _ (cptM _ cc) c))) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr cpoint ]
    EX p: val, PROP (isptr p) RETURN (p)
     SEP (mem_mgr M gv; DP (Deep_cpoint (cptConstr _ cc i c) (cptM _ cc),p)).

Definition DeepMethodsASI g s b gc:funspecs := 
  [(g,DP_GET_spec); (s,DP_SET_spec); (b, DP_BUMP_spec); (gc, DP_GETC_spec)].
(*
Definition DeepConstructorsASI pm bm cm InitPoint Init_CPoint mp mb mc:funspecs :=
   [(mp, DP_makePOINT_spec pm InitPoint); (mb, DP_makeBPOINT_spec bm InitPoint); (mc,DP_makeCPOINT_spec cm Init_CPoint)].
*)
Definition DeepConstructorsASI pc bc cc mp mb mc:funspecs :=
   [(mp, DP_makePOINT_spec pc); (mb, DP_makeBPOINT_spec bc); (mc,DP_makeCPOINT_spec cc)].

End DeepSpec.