Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import SimpleObjspec.
Require Import SimplePointModel.
Require Import SimplePoints_ASI. (*for Point etc*)

Section LowLevelWrapperSpecs.
Variable point bpoint cpoint:type.
Variable M:MallocFreeAPD.
Variable cs:compspecs.
Variable mt g s b gC:ident.
Definition Point_GET_LLspec T:=
  WITH PM: PointM T, r:T, p:val
  PRE [ tptr point]
  PROP () PARAMS (p) GLOBALS () SEP (Point point methods mt g s _ M T PM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (get T PM r))) SEP (Point point methods mt g s _ M T PM (r,p)).

Definition BPoint_GET_LLspec T:=
  WITH BM: BPointM T, r:T, p:val
  PRE [ tptr point]
  PROP () PARAMS (p) GLOBALS () SEP (BPoint point bpoint bmethods mt g s b _ M T BM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (get T BM r))) SEP (BPoint point bpoint bmethods mt g s b _ M T BM (r,p)).

Definition CPoint_GET_LLspec T:=
  WITH CM: CPointM T, r:T, p:val
  PRE [ tptr point]
  PROP () PARAMS (p) GLOBALS () SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (get T CM r))) SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (r,p)).

Definition Point_SET_LLspec T :=
  WITH PM: PointM T, r:T, p:val, i: Z
  PRE [ tptr point, tint]
  PROP (Int.min_signed <= i <= Int.max_signed) PARAMS (p; Vint(Int.repr i)) GLOBALS ()
  SEP (Point point methods mt g s _ M T PM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (Point point methods mt g s _ M T PM (set _ PM r i,p)).

Definition BPoint_SET_LLspec T:=
  WITH BM: BPointM T, r:T, p:val, i: Z
  PRE [ tptr point, tint]
  PROP (Int.min_signed <= i <= Int.max_signed) PARAMS (p; Vint(Int.repr i)) GLOBALS ()
  SEP (BPoint point bpoint bmethods mt g s b _ M T BM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (BPoint point bpoint bmethods mt g s b _ M T BM (set _ BM r i,p)).

Definition CPoint_SET_LLspec T:=
  WITH CM: CPointM T, r:T, p:val, i: Z
  PRE [ tptr point, tint]
  PROP (Int.min_signed <= i <= Int.max_signed) PARAMS (p; Vint(Int.repr i)) GLOBALS ()
  SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (set _ CM r i,p)).

Definition BPoint_BUMP_LLspec T:=
  WITH BM: BPointM T, r:T, p:val
  PRE [ tptr bpoint]
  PROP (bumpable T BM r) PARAMS (p) GLOBALS ()
  SEP (BPoint point bpoint bmethods mt g s b _ M T BM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (BPoint point bpoint bmethods mt g s b _ M T BM (bump _ BM r,p)).

Definition CPoint_BUMP_LLspec T:=
  WITH CM: CPointM T, r:T, p:val
  PRE [ tptr bpoint]
  PROP (bumpable T CM r) PARAMS (p) GLOBALS ()
  SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (bump _ CM r,p)).

Definition GETC_LLspec T :=
  WITH CM: CPointM T, r:T, p:val
  PRE [ tptr cpoint]
  PROP () PARAMS (p) GLOBALS ()
  SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (color_encoding _ CM (getC _ CM r)))) 
          SEP (CPoint point bpoint cpoint cmethods mt g s b gC _ M T CM (r,p)).

Definition GET_LLspec T CT:= binary_intersection' (Point_GET_LLspec T) 
                              (binary_intersection' (BPoint_GET_LLspec T)(CPoint_GET_LLspec CT) eq_refl eq_refl)
                              eq_refl eq_refl.
Lemma GETsubPoint_LL T CT: funspec_sub (GET_LLspec T CT) (Point_GET_LLspec T).
Proof. apply binary_intersection'_sub. Qed.
Lemma GETsubBPoint_LL T CT: funspec_sub (GET_LLspec T CT) (BPoint_GET_LLspec T).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub. apply binary_intersection'_sub. Qed.
Lemma GETsubCPoint_LL T CT: funspec_sub (GET_LLspec T CT) (CPoint_GET_LLspec CT).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub. apply binary_intersection'_sub. Qed.

Definition SET_LLspec T CT:= binary_intersection' (Point_SET_LLspec T)
                              (binary_intersection' (BPoint_SET_LLspec T) (CPoint_SET_LLspec CT) eq_refl eq_refl)
                             eq_refl eq_refl.
Lemma SETsubPoint_LL T CT: funspec_sub (SET_LLspec T CT) (Point_SET_LLspec T).
Proof. apply binary_intersection'_sub. Qed.
Lemma SETsubBPoint_LL T CT: funspec_sub (SET_LLspec T CT) (BPoint_SET_LLspec T).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub. apply binary_intersection'_sub. Qed.
Lemma SETsubCPoint_LL T CT: funspec_sub (SET_LLspec T CT) (CPoint_SET_LLspec CT).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub.  apply binary_intersection'_sub. Qed.

Definition BUMP_LLspec BT CT:= binary_intersection' (BPoint_BUMP_LLspec BT) (CPoint_BUMP_LLspec CT) eq_refl eq_refl.

Lemma BUMPsubBPoint_LL BT CT: funspec_sub (BUMP_LLspec BT CT) (BPoint_BUMP_LLspec BT).
Proof. apply binary_intersection'_sub. Qed.
Lemma BUMPsubCPoint_LL BT CT: funspec_sub (BUMP_LLspec BT CT) (CPoint_BUMP_LLspec CT).
Proof. apply binary_intersection'_sub. Qed.

Definition ASI_LLWrapper GET SET BUMP GETC T CT:funspecs := [ (GET,GET_LLspec T CT); (SET,SET_LLspec T CT); (BUMP,BUMP_LLspec T CT); (GETC,GETC_LLspec CT)].

End LowLevelWrapperSpecs.
