Require Import VST.floyd.proofauto.
Require Import PositiveSubtypingTheory.
Require Import SimplePointModel.

Definition Class (M: Type -> Type) (R:Type): Type :=
  forall S, PosSubT S R -> M S.

Definition simplePointClass : Class PointM PointRep := fun S SR =>
{| get := fun s : S => value (project _ _ SR  s);
    set := fun (s : S) (i:Z) => recomb _ _ SR  s {| value := i |} |}.

Goal m1 = simplePointClass PointRep (PosSubT_Refl _).
Proof. reflexivity. Qed.

Goal PointM_of_BPointM _ bm1 = simplePointClass PointRep (PosSubT_Refl _).
Proof. reflexivity. Qed.

Goal PointM_of_BPointM _ (BPointM_of_CPointM _ cm1) = simplePointClass CPointRep SubT_CpointRep_PointRep.
Proof. reflexivity. Qed.

Definition simpleBPointClass : Class BPointM PointRep := fun S SR =>
let super := simplePointClass S SR in
{| PointM_of_BPointM := super;
   bump := fun s : S =>
           recomb S PointRep SR s {| value := value (project S PointRep SR s) + 1|};
   bumpable := fun s : S =>
           Int.min_signed <= value (project S PointRep SR s) < Int.max_signed |}.

Goal bm1 = simpleBPointClass PointRep (PosSubT_Refl _).
Proof. reflexivity. Qed.

Goal BPointM_of_CPointM _ cm1 = simpleBPointClass CPointRep SubT_CpointRep_PointRep.
Proof. reflexivity. Qed.

Definition simpleCPointClass : Class CPointM CPointRep := fun S SCR =>
let SR := (PosSubT_Trans _ _ _ SCR SubT_CpointRep_PointRep) in
let super := simpleBPointClass S SR in
{| BPointM_of_CPointM := super;
   getC := fun s : S => color (project _ _ SCR s);
   color_encoding := mycolor_encoding |}.

Goal m1 = simplePointClass PointRep (PosSubT_Refl _).
Proof. reflexivity. Qed.

Goal cm1 = simpleCPointClass CPointRep (PosSubT_Refl _).
Proof. reflexivity. Qed.

Definition resetPointClass : Class PointM PointRep := fun S SR =>
{| get := fun s : S => value (project _ _ SR  s);
    set := fun (s : S) (i:Z) => recomb _ _ SR  s {| value := 0 |} |}.

Check (CPointM CPointRep). (*Type*)
Check (simplePointClass CPointRep SubT_CpointRep_PointRep). (*: PointM CPointRep*)

Goal (project _ _ (SubT_CpointM_PointM _ (PosSubT_Refl _)) cm1) =
      (simplePointClass CPointRep SubT_CpointRep_PointRep).
Proof. (* forget (simplePointClass CPointRep SubT_CpointRep_PointRep) as U.*) reflexivity. Qed.

Goal (project _ _ (SubT_CpointM_PointM _ (PosSubT_Refl _)) cm1) =
      (resetPointClass CPointRep SubT_CpointRep_PointRep).
Proof. simpl. unfold resetPointClass. simpl. f_equal. Abort.

Definition blueCPointClassStatic : Class CPointM CPointRep := fun S SCR =>
let SR := PosSubT_Trans S CPointRep PointRep SCR SubT_CpointRep_PointRep in
let superP := {| get := fun s => value (project S PointRep SR s);
                 set := fun s j => recomb S CPointRep SCR s
                                   {| pointRep := {| value := j |}; color := blue |} |} in
let superB := {| PointM_of_BPointM := superP;
                 bump := fun s => recomb S PointRep SR s {| value := value (project S PointRep SR s) + 1 |};
                 bumpable := fun s => Int.min_signed <= value (project S PointRep SR s) < Int.max_signed |} in
{| BPointM_of_CPointM := superB;
   getC := fun s : S => color (project S CPointRep SCR s);
   color_encoding := mycolor_encoding |}.

Definition cmBlueStatic: CPointM CPointRep := blueCPointClassStatic CPointRep (PosSubT_Refl _).
(*
Definition blueCPointClassVirtual : Class CPointM CPointRep := fun S SCR =>
let SR := PosSubT_Trans S CPointRep PointRep SCR SubT_CpointRep_PointRep in
{|
BPointM_of_CPointM := {|
                      PointM_of_BPointM := {|
                                           get := fun s : S => value (project S PointRep SR s);
                                           set := fun (s : S) (j : Z) =>
                                                  recomb S CPointRep SCR s
                                                    {|
                                                    pointRep := (*project S PointRep SR s;*){| value := j |};
                                                    color := blue |} |};
                      bump := let p := simplePointClass S SR in
                              match p with
                              | {| get := get; set := set |} => fun s : S => set s (get s + 1)
                              end;
                      bumpable := fun s : S =>
                                  Int.min_signed <= value (project S PointRep SR s) < Int.max_signed |};
getC := fun s : S => color (project S CPointRep SCR s);
color_encoding := mycolor_encoding |}.*)
Definition blueCPointClassVirtual : Class CPointM CPointRep := fun S SCR =>
let SR := PosSubT_Trans S CPointRep PointRep SCR SubT_CpointRep_PointRep in
let superP := {| get := fun s => value (project S PointRep SR s);
                 set := fun s j => recomb S CPointRep SCR s
                                    {| pointRep := {| value := j |}; color := blue |} |} in
let superB := {| PointM_of_BPointM := superP;
                 bump := fun s => set _ superP s (get _ superP s + 1);
                 bumpable := fun s => Int.min_signed <= value (project S PointRep SR s) < Int.max_signed |} in
{| BPointM_of_CPointM := superB;
   getC := fun s : S => color (project S CPointRep SCR s);
   color_encoding := mycolor_encoding |}.

Definition cmBlueVirtual: CPointM CPointRep := blueCPointClassVirtual CPointRep (PosSubT_Refl _).


Section SuperClass.
Variable BaseM: Type -> Type.
Variable R:Type.
Variable Base: Class BaseM R.

Variable SubM: Type -> Type.
Variable S:Type.
Variable SR: PosSubT S R.

Definition superClass (C: Class SubM S): Class BaseM S := fun U US => Base U (PosSubT_Trans _ _ _ US SR).
End SuperClass.

Section LSP.
Variable BaseM: Type -> Type.
Variable R:Type.
Variable Base: Class BaseM R.

Variable SubM: Type -> Type.
Variable S:Type.
Variable SR: PosSubT S R.

Variable Sub: PosSubT (SubM S) (BaseM S).

Definition LSP (m :SubM S):= Base S SR = project _ _ Sub m.
End LSP.
Print LSP.

Lemma LSP_simple_cm1: LSP PointM PointRep simplePointClass CPointM CPointRep SubT_CpointRep_PointRep
        (SubT_CpointM_PointM _ (PosSubT_Refl _)) cm1.
Proof. reflexivity. Qed.

Lemma LSP_reset_cm1: 
      LSP PointM PointRep resetPointClass CPointM CPointRep SubT_CpointRep_PointRep
          (SubT_CpointM_PointM _ (PosSubT_Refl _)) cm1.
Proof. unfold resetPointClass, LSP; simpl. f_equal. Abort.

Section LSP_base.
Variable M: Type -> Type.
Variable R:Type.
Variable Base: Class M R.

Variable S:Type.
Variable SR: PosSubT S R.

Lemma LSP_base: LSP M R Base M S SR (PosSubT_Refl (M S)) (Base S SR).
Proof. unfold LSP, p_refl. trivial. Qed.
End LSP_base.

Section LSP_refl.
Variable M: Type -> Type.
Variable R:Type.
Variable Base: Class M R.

Lemma LSP_refl: LSP M R Base M R (PosSubT_Refl _) (PosSubT_Refl (M R)) (Base R (PosSubT_Refl _) ).
Proof. apply LSP_base. Qed.
End LSP_refl.

Section LSPstrong.
Variable BaseM: Type -> Type.
Variable R:Type.
Variable Base: Class BaseM R.

Variable SubM: Type -> Type.
Variable S:Type.
Variable SR: PosSubT S R.

Variable Sub: forall T, PosSubT T S -> PosSubT (SubM T) (BaseM T).
Variable T:Type.
Variable TS: PosSubT T S.

Definition LSPstrong (m :SubM T):= Base T (PosSubT_Trans _ _ _ TS SR) = project _ _ (Sub _ TS) m.
End LSPstrong.

Lemma LSP_LSTStrong BaseM R Base SubM S SR Sub m:
   LSPstrong BaseM R Base SubM S SR Sub S (PosSubT_Refl _) m ->
   LSP BaseM R Base SubM S SR (Sub S (PosSubT_Refl _)) m.
Proof. intros.
unfold LSPstrong in H. unfold LSP. rewrite <- H. simpl.
f_equal.
unfold PosSubT_Trans. destruct SR. f_equal; apply proof_irr.
Qed.

Lemma LSP_LSTStrong_inv BaseM R Base SubM S SR Sub m:
   LSP BaseM R Base SubM S SR (Sub S (PosSubT_Refl _)) m ->
   LSPstrong BaseM R Base SubM S SR Sub S (PosSubT_Refl _) m. 
Proof. intros.
unfold LSP in H. unfold LSPstrong. rewrite <- H.
f_equal.
unfold PosSubT_Trans. destruct SR. f_equal; apply proof_irr.
Qed.
(*
Section LSPstrong_trans.
Variable BaseM: Type -> Type.
Variable R:Type.
Variable Base: Class BaseM R.

Variable MidM: Type -> Type.
Variable S:Type.
Variable SR: PosSubT S R.

Variable Sub1: forall T, PosSubT T S -> PosSubT (MidM T) (BaseM T).
Variable T:Type.
Variable TS: PosSubT T S.
Variable mt: MidM T.

Variable LSP1: LSPstrong _ R Base _ _ SR Sub1 T TS mt.

Variable MidClass: Class MidM S.
Variable SubM: Type -> Type.

Variable Sub2: forall U, PosSubT U T -> PosSubT (SubM U) (MidM U).
Variable U:Type.
Variable UT: PosSubT U T.
Variable mu: SubM U.

Variable LSP2: LSPstrong _ S MidClass _ _ TS Sub2 U UT mu.

Lemma AUX: forall Q, PosSubT Q T -> PosSubT (SubM Q) (BaseM Q).
Proof. intros. clear LSP1 LSP2. 
  eapply PosSubT_Trans. apply (Sub2 Q X). apply Sub1.
  eapply PosSubT_Trans; eassumption.
Defined.

Lemma LSTStrong_trans: LSPstrong _ _ Base SubM T (PosSubT_Trans _ _ _ TS SR) AUX U UT mu.
Proof. unfold LSPstrong in *.
  unfold PosSubT_Trans, p_trans, r_trans in *. simpl in *.
  unfold PosSubT_Trans, p_trans, r_trans in *. simpl in *.
  extensionality z.
  rewrite <- LSP2; clear LSP2. simpl. unfold Class in *.
  remember (Sub1 U
     {|
     project := fun s : U => project T S TS (project U T UT s);
     recomb := fun (s : U) (t : S) => recomb U T UT s (recomb T S TS (project U T UT s) t);
     Ax1 := Ax1_trans U T S UT TS;
     Ax2 := Ax2_trans U T S UT TS;
     Ax3 := Ax3_trans U T S UT TS |}) as B. destruct B; simpl in *.
unfold Ax1_trans.
  remember (MidClass U
     {|
     project := fun s : U => PositiveSubtypingTheory.project T S TS (PositiveSubtypingTheory.project U T UT s);
     recomb := fun (s : U) (t : S) => PositiveSubtypingTheory.recomb U T UT s (PositiveSubtypingTheory.recomb T S TS (PositiveSubtypingTheory.project U T UT s) t);
     Ax1 := Ax1_trans U T S UT TS;
     Ax2 := Ax2_trans U T S UT TS;
     Ax3 := Ax3_trans U T S UT TS |}) as A. destruct A.
  unfold PosSubT_Trans, p_trans, r_trans in *; simpl in *.
  rewrite f_equal.
  unfold Sub1.
  remember  (AUX U UT)  as z. destruct z. simpl in *.
  remember (Sub2 U UT) as q. destruct q; simpl in *. 
  destruct   simpl in *. destruct LSP2. destruct LSP1.
  f_equal. inv LSP2. simpl.
 SR).

Check (LSPstrong _ _ Base SubM T).
Check PosSubT_Trans.

Definition LSP (m :SubM S):= Base S SR = project _ _ Sub m.
End LSP.

Variable BaseM: Type -> Type.
Variable R:Type.
Variable Base: Class BaseM R.

Variable SubM: Type -> Type.
Variable S:Type.
Variable SR: PosSubT S R.

Variable Sub: forall T, PosSubT T S -> PosSubT (SubM T) (BaseM T).
Variable T:Type.
Variable TS: PosSubT T S.

Definition LSPstrong (m :SubM T):= Base T (PosSubT_Trans _ _ _ TS SR) = project _ _ (Sub _ TS) m.
End LSPstrong.
Lemma LSP_refl (m :M S): LSP M R Base Base S SR = project _ _ (PosSubT_Refl (M S)) m.


Search PosSubT CPointRep.  
Lemma LSPcm1: LSP (simplePointClass CPointRep SubT_CpointRep_PointRep).

PosSubT 

Check (Class PointM CPointRep).

Goal PosSubT (CPointM CPointRep) (Class PointM CPointRep).
econstructor.
red.

 (simplePointClass _ SubT_CpointRep_PointRep).*)