Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import SimpleObjspec.
Require Import SimplePointModel.
Require Import SimplePoints_ASI.
Require Import VST.floyd.VSU.
Require Import Wrapper.
Require Import ASI_HLWrapper.
Require Import DeepPoints.
Require Import ASI_DeepWrapper.

Instance CompSpecs : compspecs. make_compspecs prog. Defined.

Section DeepWrapperAPD.
Variable W: WrapperAPD.

Notation POINT:= (Wr_POINT W).
Notation BPOINT:= (Wr_BPOINT W).
Notation CPOINT:= (Wr_CPOINT W).

Variable T CT:Type.

Definition DP: Pred (DeepPoint T CT) := fun dp =>
  match dp with
  | (Deep_point t pm,p) => POINT T pm (t, p)
  | (Deep_bpoint t bm,p) => BPOINT T bm (t, p)
  | (Deep_cpoint c cm,p) => CPOINT CT cm (c, p)
  end.

  Definition makeGetInstance ts2 (d: DeepPoint T CT) (v:val): functors.MixVariantFunctor._functor
               (functors.MixVariantFunctorGenerator.fsig
                  (fun i : bool =>
                   rmaps.dependent_type_functor_rec ts2
                     (if i
                      then rmaps.ConstType (PointM T * T * val)
                      else
                       rmaps.SigType bool
                         (fun b : bool =>
                          if b
                          then rmaps.ConstType (BPointM T * T * val)
                          else rmaps.ConstType (CPointM CT * CT * val))))) mpred.
  Proof. destruct d.
  + simpl. exists true; simpl.  apply ((*PM*)p, t, v).
  + simpl. exists false. exists true; simpl. apply ((*BM*)b, t, v).
  + simpl. exists false; exists false; simpl. apply ((*CM*)c0, c, v).
  Defined.
  Definition makeSetInstance ts2 (d: DeepPoint T CT) (i:Z)(v:val)
             : functors.MixVariantFunctor._functor
               (functors.MixVariantFunctorGenerator.fsig
                  (fun i0 : bool =>
                   rmaps.dependent_type_functor_rec ts2
                     (if i0
                      then rmaps.ConstType (PointM T * T * val * Z)
                      else
                       rmaps.SigType bool
                         (fun b : bool =>
                          if b
                          then rmaps.ConstType (BPointM T * T * val * Z)
                          else rmaps.ConstType (CPointM CT * CT * val * Z)))))
               mpred.
  Proof. destruct d.
  + simpl. exists true; simpl. apply ((*PM*)p, t, v, i).
  + simpl. exists false. exists true; simpl. apply ((*BM*)b, t, v, i).
  + simpl. exists false; exists false; simpl. apply ((*CM*)c0, c, v, i).
  Defined.
  Definition makeBumpInstance ts2 (d: DeepPoint T CT) (v:val)
             (B:dp_bumpable T CT (*BM CM*) d):
            functors.MixVariantFunctor._functor
               (functors.MixVariantFunctorGenerator.fsig
                  (fun i : bool =>
                   rmaps.dependent_type_functor_rec ts2
                     (if i
                      then rmaps.ConstType (BPointM T * T * val)
                      else rmaps.ConstType (CPointM CT * CT * val)))) mpred.
  Proof. destruct d.
  + simpl in B. contradiction.
  + simpl. exists true; simpl. apply ((*BM*)b, t, v).
  + simpl. exists false; simpl. apply ((*CM*)c0, c, v).
  Defined.
(*redundant - can just apply manually inside proof
  Definition makeGetcInstance (*(d: Deep_cpoint T CT d, p)*) (c:CT) (v:val):
             functors.MixVariantFunctor._functor
               (functors.MixVariantFunctorGenerator.fconst (CPointM CT * CT * val))
               mpred.
  simpl. apply (CM, c, v). Defined.*)

(*Variable point bpoint cpoint:type.*)
Variable gg ss bb gC: ident.
Variable MethodsDistinct: list_norepet [gg; ss; bb; gC].

  Definition HLWrapperMethods: funspecs:=
             ASI_HLWrapperMethods point bpoint cpoint W T CT gg ss bb gC.

  Definition DeepWrapperMethods: funspecs:=
             DeepMethodsASI point bpoint cpoint _ _ (*PM BM CM*) DP gg ss bb gC.

  Lemma MethodsSub: VSU.funspecs_sqsub HLWrapperMethods DeepWrapperMethods.
  Proof. red; intros.  inv MethodsDistinct. inv H3. inv H5.
     specialize (initial_world.find_id_In_map_fst _ _ _ H); intros.
     destruct H0 as [? | [? | [? | [? | ?]]]]; subst; try solve [contradiction].
  + simpl in H; simpl.
    destruct (Memory.EqDec_ident gg gg); [clear e; inv H | congruence].
    eexists; split. reflexivity. do_funspec_sub_nonND.
    destruct w as [d p]. Exists ts (makeGetInstance ts d p) emp.
    unfold binarySUM, makeGetInstance; simpl.
    destruct d; simpl; entailer!.
  + simpl in H; simpl.
    destruct (Memory.EqDec_ident ss gg); [ subst; elim H2; left; trivial | clear n].
    destruct (Memory.EqDec_ident ss ss); [ clear e; inv H | congruence]. 
    eexists; split. reflexivity. do_funspec_sub_nonND.
    destruct w as [[d p] i]. Exists ts (makeSetInstance ts d i p) emp.
    unfold binarySUM, makeSetInstance; simpl.
    destruct d; simpl; entailer!.
  + simpl in H; simpl.
    destruct (Memory.EqDec_ident bb gg); [ subst; elim H2; right; left; trivial | clear n].
    destruct (Memory.EqDec_ident bb ss); [ subst; elim H4; left; trivial | clear n].
    destruct (Memory.EqDec_ident bb bb); [ clear e; inv H | congruence]. 
    eexists; split. reflexivity. do_funspec_sub_nonND.
    destruct w as [d p].
    assert_PROP (dp_bumpable T CT (*BM CM*) d) as BMP  by (unfold PROPx; simpl; Intros; apply prop_right; trivial).
    Exists ts (makeBumpInstance ts d p BMP) emp.
    unfold binarySUM, makeBumpInstance; simpl.
    destruct d; simpl; entailer!.
  + simpl in H; simpl.
    destruct (Memory.EqDec_ident gC gg); [ subst; elim H2; right; right; left; trivial | clear n].
    destruct (Memory.EqDec_ident gC ss); [ subst; elim H4; right; left; trivial | clear n].
    destruct (Memory.EqDec_ident gC bb); [ subst; elim H3; left; trivial | clear n].
    destruct (Memory.EqDec_ident gC gC); [ clear e; inv H | congruence]. 
    eexists; split. reflexivity. do_funspec_sub_nonND.
    destruct w as [d p]. unfold PROPx; simpl. Intros. destruct d; simpl in H0; try contradiction.
      Exists ts (c0, c, p) emp. simpl; entailer!.
    (*    destruct w as [d p]. Exists ts (CM, d, p) emp. simpl; entailer!.
    destruct w as [[c cm] p]. Exists ts (cm, c, p) emp. simpl; entailer!.*)
  Qed.

  Variable PC: PointConstr T.
  Variable BC: BPointConstr T.
  Variable CC: CPointConstr CT. (*
Variable PM: PointM T.
Variable BM: BPointM T.
Variable CM: CPointM CT.*)

  Variable point bpoint cpoint:type.
  Variable M: MallocFreeAPD. (*
  Variable Init_Point:Z -> T.
  Variable Init_CPoint: Z -> Color -> CT.*)
  Variable mp mb mc:ident.
  Variable ConstructorsDistinct: list_norepet [mp; mb; mc].

  Definition ASI_HLConstructors:= ASI_HLWrapperConstructors point bpoint cpoint W M T CT (*PM BM CM Init_Point Init_CPoint*)PC BC CC mp mb mc.
  Definition ASI_DeepConstructors:= DeepConstructorsASI point bpoint cpoint T CT DP M (*PM BM CM Init_Point Init_CPoint*)PC BC CC mp mb mc.

  Lemma ConstructorsSub: VSU.funspecs_sqsub ASI_HLConstructors ASI_DeepConstructors.
  Proof. red; intros. inv ConstructorsDistinct. inv H3. clear H5.
     specialize (initial_world.find_id_In_map_fst _ _ _ H); intros.
     destruct H0 as [? | [? | [? | ?]]]; subst; try solve [contradiction].
  + simpl in H. simpl. destruct (Memory.EqDec_ident mp mp); [clear e | congruence].
    inv H. eexists; split. reflexivity.
    destruct PC. do_funspec_sub. Exists w emp; destruct w as [gv i]; Intros; entailer!.
    intros. Exists (eval_id ret_temp x). entailer!.
  + simpl in H. simpl. destruct (Memory.EqDec_ident mb mp); [subst; elim H2; left; trivial | clear n].
    destruct (Memory.EqDec_ident mb mb); [ clear e; inv H| congruence].
    eexists; split. reflexivity.
    destruct BC. do_funspec_sub. Exists w emp; destruct w as [gv i]; Intros; entailer!.
    intros. Exists (eval_id ret_temp x). entailer!.
  + simpl in H. simpl. destruct (Memory.EqDec_ident mc mp); [subst; elim H2; right; left; trivial | clear n].
    destruct (Memory.EqDec_ident mc mb); [subst; elim H4; left; trivial | clear n].
    destruct (Memory.EqDec_ident mc mc); [ clear e; inv H| congruence].
    eexists; split. reflexivity.
    destruct CC. do_funspec_sub. Exists w emp; destruct w as [[gv i] c]; Intros; entailer!.
    intros. Exists (eval_id ret_temp x). entailer!.
  Qed.
End DeepWrapperAPD.

Require Import VSU_HLWrapper.

Section VSU_DeepWrapper.
  Variable M: MallocFreeAPD.
  Variable T CT:Type.
  Variable PM: PointM T.
  Variable BM: BPointM T.
  Variable CM: CPointM CT.

  Definition Vprog : varspecs. mk_varspecs prog. Defined.

  Definition MyDeepWrapperMethods: funspecs:= DeepWrapperMethods (MyWrapperAPD M _) T CT (*PM BM CM point bpoint cpoint*) _GET _SET _BUMP _GETC.

  Definition DeepWrapper_VSU: @VSU NullExtension.Espec Vprog _ nil nil prog MyDeepWrapperMethods emp.
  Proof. eapply VSU_Exports_sub. apply (HLWrapper_VSU M T CT).
     apply compute_list_norepet_e; cbv; reflexivity.
     apply MethodsSub.
     apply compute_list_norepet_e; cbv; reflexivity.
  Qed.
End VSU_DeepWrapper.