Require Import VST.floyd.proofauto. 
Require Import ASI_malloc. (*needed for MallocFreeAPD*) 

Require Import SimplePointModel. (*For PointM etc*)
Require Import SimpleObjspec. (*For Pred*)
Require Import DeepPoints.

Require Import ASI_DeepWrapper. (*for DeepPoint ect*)

Section ListSpecs.
Variable lst point:type.

Variable M:MallocFreeAPD.

Variable LIST: list val -> val -> mpred.
Variable LIST_NIL: emp |-- LIST nil nullval.

Definition insert_spec :=
  WITH l:val, p:val, L:list val, gv:globals
  PRE [ tptr lst, tptr point ]
  PROP (isptr p) PARAMS (l; p) GLOBALS (gv) SEP (mem_mgr M gv; LIST L l)
  POST [ tptr lst ]
  EX l', PROP () RETURN (l') SEP (mem_mgr M gv; LIST (p::L) l').

Variable T CT:Type.
Notation DeepPoint := (DeepPoint T CT).
Notation Deep_point := (Deep_point T CT).
Notation Deep_bpoint := (Deep_bpoint T CT).
Notation Deep_cpoint := (Deep_cpoint T CT).

Notation dp_get := (dp_get T CT).
Notation dp_set := (dp_set T CT).

Notation dp_bumpable := (dp_bumpable T CT).
Notation dp_bump := (dp_bump T CT).
Notation dp_getc := (dp_getc T CT).

Variable DP: Pred DeepPoint.

Fixpoint Points (PL: list (DeepPoint * val)) :=
  match PL with
   nil => emp
  | dp::l => (DP dp * Points l)%logic
  end. 

Definition map_set (PL: list (DeepPoint  * val)) (i:Z): list (DeepPoint * val) :=
  map (fun x => (dp_set (fst x) i, snd x)) PL.

Definition mapset_spec :=
  WITH l:val, i:Z, PL: list (DeepPoint * val)
  PRE [ tptr lst, tint ]
  PROP (Int.min_signed <= i <= Int.max_signed) 
  PARAMS (l; Vint (Int.repr i)) GLOBALS () SEP (LIST (map snd PL) l; Points PL)
  POST [ tvoid ]
  PROP () RETURN () SEP (LIST (map snd PL) l; Points (map_set PL i)).

Definition map_bump (PL: list (DeepPoint * val)): list (DeepPoint * val) :=
  map (fun x => (dp_bump (fst x), snd x)) PL.

Definition mapbump_spec :=
  WITH l:val, PL: list (DeepPoint* val)
  PRE [ tptr lst ]
  PROP (Forall dp_bumpable (map fst PL))
  PARAMS (l) GLOBALS () SEP (LIST (map snd PL) l; Points PL)
  POST [ tvoid ]
  PROP () RETURN () SEP (LIST (map snd PL) l; Points (map_bump PL)).

Definition sum l:= fold_left (fun z d => z + dp_get d) l 0. 

Definition summable (l:list (DeepPoint)) :=
  Forall (fun d => 0 <= dp_get d <= Int.max_signed) l /\ sum l <= Int.max_signed.

Definition sum_spec :=
  WITH l:val, PL: list (DeepPoint * val)
  PRE [ tptr lst ]
  PROP (summable (map fst PL)) 
  PARAMS (l) GLOBALS () SEP (LIST (map snd PL) l; Points PL)
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (sum (map fst PL)))) SEP (LIST (map snd PL) l; Points PL).

Definition ASI_PointList i ms mb sm:funspecs:= [(i,insert_spec); (ms,mapset_spec); (mb,mapbump_spec); (sm,sum_spec)].

End ListSpecs.