Require Import VST.floyd.proofauto.
Require Import ASI_malloc.
Require Import SimpleObjspec.
Require Import SimplePointModel.
Require Import SimplePoints_ASI.

Record WrapperAPD := {
 Wr_POINT: forall T, PointM T -> Pred T;
 Wr_BPOINT: forall T, BPointM T -> Pred T;
 Wr_CPOINT: forall T, CPointM T -> Pred T
}.

Section WrapperSpecs.
Variable point bpoint cpoint:type.
Variable W: WrapperAPD.

Notation POINT:= (Wr_POINT W).
Notation BPOINT:= (Wr_BPOINT W).
Notation CPOINT:= (Wr_CPOINT W).

Definition Point_GET_HLspec T:=
  WITH PM: PointM T, r:T, p:val
  PRE [ tptr point]
  PROP () PARAMS (p) GLOBALS () SEP (POINT T PM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (get T PM r))) SEP (POINT T PM (r,p)).

Definition BPoint_GET_HLspec T:=
  WITH BM: BPointM T, r:T, p:val
  PRE [ tptr point]
  PROP () PARAMS (p) GLOBALS () SEP (BPOINT T BM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (get T BM r))) SEP (BPOINT T BM (r,p)).

Definition CPoint_GET_HLspec T:=
  WITH CM: CPointM T, r:T, p:val
  PRE [ tptr point]
  PROP () PARAMS (p) GLOBALS () SEP (CPOINT T CM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (get T CM r))) SEP (CPOINT T CM (r,p)).

Definition Point_SET_HLspec T :=
  WITH PM: PointM T, r:T, p:val, i: Z
  PRE [ tptr point, tint]
  PROP (Int.min_signed <= i <= Int.max_signed) PARAMS (p; Vint(Int.repr i)) GLOBALS ()
  SEP (POINT T PM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (POINT T PM (set _ PM r i,p)).

Definition BPoint_SET_HLspec T :=
  WITH BM: BPointM T, r:T, p:val, i: Z
  PRE [ tptr point, tint]
  PROP (Int.min_signed <= i <= Int.max_signed) PARAMS (p; Vint(Int.repr i)) GLOBALS ()
  SEP (BPOINT T BM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (BPOINT T BM (set _ BM r i,p)).

Definition CPoint_SET_HLspec T:=
  WITH CM: CPointM T, r:T, p:val, i: Z
  PRE [ tptr point, tint]
  PROP (Int.min_signed <= i <= Int.max_signed) PARAMS (p; Vint(Int.repr i)) GLOBALS ()
  SEP (CPOINT T CM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (CPOINT T CM (set _ CM r i,p)).

Definition BPoint_BUMP_HLspec T:=
  WITH BM: BPointM T, r:T, p:val
  PRE [ tptr bpoint]
  PROP (bumpable T BM r) PARAMS (p) GLOBALS ()
  SEP (BPOINT T BM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (BPOINT T BM (bump _ BM r,p)).

Definition CPoint_BUMP_HLspec T:=
  WITH CM: CPointM T, r:T, p:val
  PRE [ tptr bpoint]
  PROP (bumpable T CM r) PARAMS (p) GLOBALS ()
  SEP (CPOINT T CM (r,p))
  POST [ tvoid ]
  PROP () RETURN () SEP (CPOINT T CM (bump _ CM r,p)).

Definition GETC_HLspec T :=
  WITH CM: CPointM T, r:T, p:val
  PRE [ tptr cpoint]
  PROP () PARAMS (p) GLOBALS ()
  SEP (CPOINT T CM (r,p))
  POST [ tint ]
  PROP () RETURN (Vint (Int.repr (color_encoding _ CM (getC _ CM r)))) 
          SEP (CPOINT T CM (r,p)).

Definition GET_HLspec T CT:= binary_intersection' (Point_GET_HLspec T)
                            (binary_intersection' (BPoint_GET_HLspec T) (CPoint_GET_HLspec CT) eq_refl eq_refl)
                            eq_refl eq_refl.
Lemma GETsubPoint T CT: funspec_sub (GET_HLspec T CT) (Point_GET_HLspec T).
Proof. apply binary_intersection'_sub. Qed.
Lemma GETsubBPoint T CT: funspec_sub (GET_HLspec T CT) (BPoint_GET_HLspec T).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub. apply binary_intersection'_sub. Qed.
Lemma GETsubCPoint T CT: funspec_sub (GET_HLspec T CT) (CPoint_GET_HLspec CT).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub. apply binary_intersection'_sub. Qed.

Definition SET_HLspec T CT:= binary_intersection' (Point_SET_HLspec T)
                             (binary_intersection' (BPoint_SET_HLspec T) (CPoint_SET_HLspec CT) eq_refl eq_refl)
                             eq_refl eq_refl.
Lemma SETsubPoint T CT: funspec_sub (SET_HLspec T CT) (Point_SET_HLspec T).
Proof. apply binary_intersection'_sub. Qed.
Lemma SETsubBPoint T CT: funspec_sub (SET_HLspec T CT) (BPoint_SET_HLspec T).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub. apply binary_intersection'_sub. Qed.
Lemma SETsubCPoint T CT: funspec_sub (SET_HLspec T CT) (CPoint_SET_HLspec CT).
Proof. eapply funspec_sub_trans. apply binary_intersection'_sub. apply binary_intersection'_sub. Qed.

Definition BUMP_HLspec BT CT:= binary_intersection' (BPoint_BUMP_HLspec BT) (CPoint_BUMP_HLspec CT) eq_refl eq_refl.

Lemma BUMPsubBPoint BT CT: funspec_sub (BUMP_HLspec BT CT) (BPoint_BUMP_HLspec BT).
Proof. apply binary_intersection'_sub. Qed.
Lemma BUMPsubCPoint BT CT: funspec_sub (BUMP_HLspec BT CT) (CPoint_BUMP_HLspec CT).
Proof. apply binary_intersection'_sub. Qed.

Definition ASI_HLWrapperMethods T CT g s b gC:funspecs := [ (g,GET_HLspec T CT); (s,SET_HLspec T CT); (b,BUMP_HLspec T CT); (gC,GETC_HLspec CT)].
End WrapperSpecs.

Section ConstructorSpecs.
Variable point bpoint cpoint:type.
Variable W: WrapperAPD.

Notation POINT:= (Wr_POINT W).
Notation BPOINT:= (Wr_BPOINT W).
Notation CPOINT:= (Wr_CPOINT W).

Definition makePOINT_spec M T(* (PM: PointM T) (Init_Point:Z -> T)*)(pc:PointConstr T):=
 match pc with {| ptM := PM ; ptConstr := Init_Point |} =>
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr point ]
    EX p: val, PROP (isptr p) RETURN (p)
     SEP (mem_mgr M gv; POINT T PM (Init_Point i,p))
 end.

Definition makeBPOINT_spec M T (* (BPM: BPointM T) (Init_Point:Z -> T)*)(bc:BPointConstr T):=
  match bc with {| bptM := BM; bptConstr := Init_Point |} =>
 WITH gv: globals, i:Z
 PRE [ tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) PARAMS(Vint(Int.repr i)) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr bpoint ]
    EX p: val, PROP (isptr p) RETURN (p)
     SEP (mem_mgr M gv; BPOINT T BM (Init_Point i,p))
 end.

Definition makeCPOINT_spec M T(* (CPM: CPointM T) (Init_CPoint: Z -> Color -> T)*)(cc:CPointConstr T):=
  match cc with {| cptM := CM; cptConstr:= Init_CPoint |} =>
 WITH gv: globals, i:Z, c:Color
 PRE [ tint, tint ]
    PROP (Int.min_signed <= i <= Int.max_signed) 
    PARAMS (Vint(Int.repr i); Vint (Int.repr (color_encoding T CM c))) GLOBALS (gv) 
    SEP (mem_mgr M gv)
 POST [ tptr cpoint ]
    EX p: val, PROP (isptr p) RETURN (p)
     SEP (mem_mgr M gv; CPOINT T CM (Init_CPoint i c,p))
 end.
(*
Definition ASI_HLWrapperConstructors M T CT (PM: PointM T) (BPM: BPointM T) (CPM: CPointM CT) Init_Point Init_CPoint m mb mc:funspecs := 
           [ (m,makePOINT_spec M T PM Init_Point); (mb,makeBPOINT_spec M T BPM Init_Point); (mc,makeCPOINT_spec M CT CPM Init_CPoint)].
*)
Definition ASI_HLWrapperConstructors M T R pc bc cc m mb mc:funspecs := 
           [ (m,makePOINT_spec M T pc); (mb,makeBPOINT_spec M T bc); (mc,makeCPOINT_spec M R cc)].
End ConstructorSpecs.

(*
Record JustifiedWrapperAPD point bpoint cpoint methods bmethods cmethods mt g s b gC cs M:= {
 JustifiedWrapper :> WrapperAPD;
(* JW_DynRep: dynRep;*)
 Pt_sub: forall T PM initP, funspec_sub (makePoint_spec point methods mt g s cs M (*JW_DynRep*) T PM initP)
                  (makePOINT_spec point JustifiedWrapper M T PM initP);
 BPt_sub: forall T BM initP, funspec_sub (makeBPoint_spec point bpoint bmethods mt g s b _ M (*JW_DynRep*) T BM initP)
                  (makeBPOINT_spec bpoint JustifiedWrapper M T BM initP);
 CPt_sub: forall CT CM initCP, funspec_sub (makeCPoint_spec point bpoint cpoint cmethods mt g s b gC _ M (*JW_DynRep*) CT CM initCP)
                    (makeCPOINT_spec cpoint JustifiedWrapper M CT CM initCP)
}.*)

Record JustifiedWrapperAPD point bpoint cpoint methods bmethods cmethods mt g s b gC cs M T R := {
 JustifiedWrapper :> WrapperAPD;
(* JW_DynRep: dynRep;*)
 Pt_sub: forall pc, funspec_sub (makePoint_spec point methods mt g s cs M (*JW_DynRep*) T pc)
                  (makePOINT_spec point JustifiedWrapper M _ pc);
 BPt_sub: forall bc, funspec_sub (makeBPoint_spec point bpoint bmethods mt g s b _ M (*JW_DynRep*)T bc)
                  (makeBPOINT_spec bpoint JustifiedWrapper M _ bc);
 CPt_sub: forall cc, funspec_sub (makeCPoint_spec point bpoint cpoint cmethods mt g s b gC _ M (*JW_DynRep*)R cc)
                    (makeCPOINT_spec cpoint JustifiedWrapper M _ cc)
}.