Require Import VST.floyd.proofauto.
Require Import PositiveSubtypingTheory.

(*Page 27*)
Parameter Object: forall (M: Type -> Type) (R:Type) (s:R) (m:M R), Type.
Parameter object: forall (M: Type -> Type) (R:Type) (s:R) (m:M R), Object M R s m.

(*And here's an example realization - an object is a pair containing the state
and the methods - rather unsurprising*)
Definition ObjectDef (M: Type -> Type) (R:Type) (s:R) (m:M R): Type := (R * M R).
Definition objectDef (M: Type -> Type) (R:Type) (s:R) (m:M R): ObjectDef M R s m := (s,m).

Record PointM X := {
  get : X -> Z;
  set : X -> Z -> X;
  bump : X -> X
}.

Record PointR1 := { x : Z }.
Definition m1: PointM PointR1 := {|
  get := fun s : PointR1 => x s;
  set := fun (_ : PointR1) (i : Z) => {| x := i |};
  bump := fun s : PointR1 => match s with
                            | {| x := x0 |} => {| x := x0 + 1 |}
                            end |}.


Record PointR2 := { y : Z * Z }.
Definition m2: PointM PointR2 := {|
  get := fun s : PointR2 => fst (y s);
  set := fun (s : PointR2) (i : Z) => {| y := (i, snd (y s)) |};
  bump := fun s : PointR2 => match s with
                            | {| y := (a,b) |} => {| y := (a+1,b) |}
                            end |}.

Lemma SubTP_R2R1: PosSubT PointR2 PointR1.
Proof.
  apply (Build_PosSubT PointR2 PointR1 (fun r2 => {| x:= fst (y r2) |})
      (fun r2 r1 => {| y := (x r1, snd (y r2)) |})); simpl; intros; trivial.
  destruct t; simpl; trivial. destruct s as [[? ?]]; simpl; trivial.
Defined.

Definition Class (M: Type -> Type) (R:Type): Type :=
  forall S, PosSubT S R -> M S.

Definition simplePointClass : Class PointM PointR1 := fun S SR =>
match SR with
| {| project := proj; recomb := rec |} =>
    {|
    get := fun s : S => x (proj s);
    set := fun (s : S) (i : Z) => rec s {| x := i |};
    bump := fun s : S => rec s {| x := 1 + x (proj s) |} |}
end.

Section NewSectionNotInHBPaper_isSUB.
Section isSUB.
Variable M: Type -> Type.
Variable R:Type. 
Variable C : Class M R.

Definition isSub S SR (m:M S):= C S SR = m.
End isSUB.

Goal isSub PointM PointR1 simplePointClass PointR2 SubTP_R2R1 m2.
Proof. red. unfold m2, simplePointClass. simpl. f_equal.
extensionality r2. destruct r2 as [[? ?]]. simpl. f_equal. f_equal. lia. 
Qed.

Definition m2wrong: PointM PointR2 := {|
  get := fun s : PointR2 => fst (y s);
  set := fun (s : PointR2) (i : Z) => {| y := (42, snd (y s)) |};
  bump := fun s : PointR2 => match s with
                            | {| y := (a,b) |} => {| y := (a+1,b) |}
                            end |}.
Goal isSub PointM PointR1 simplePointClass PointR2 SubTP_R2R1 m2wrong.
Proof. red. unfold m2wrong, simplePointClass. simpl. f_equal.
2:{ extensionality r2. destruct r2 as [[? ?]]. simpl. f_equal. f_equal. lia. }
extensionality r2. destruct r2 as [[? ?]]. simpl. extensionality i. f_equal. f_equal.
Abort.

Definition BaseClass (M: Type -> Type) (R:Type) (C: Class M R): M R := C R (PosSubT_Refl R).
Definition isGenericImplementation M R C (m:M R) := m = BaseClass M R C.

Goal forall M R C, isSub M R C _ (PosSubT_Refl R) (BaseClass M R C).
Proof. intros. reflexivity. Qed.
 
Goal forall M R C m (Hm: isGenericImplementation M R C m),
     isSub M R C _ (PosSubT_Refl R) m.
Proof. intros. unfold isGenericImplementation in Hm; subst. reflexivity. Qed.

Definition Class_Trans M R (C:Class M R) S (SR:PosSubT S R): Class M S :=
 fun U US => C U (PosSubT_Trans U S R US SR).

Section X.
Variable M: Type -> Type.
Variable R:Type. 
Variable C : Class M R.
Variable S:Type.
Variable mS: M S.
Variable SR:PosSubT S R.
Variable D : Class M S.
Variable U:Type.
Variable US:PosSubT U S.
Variable mU: M U.

(*This is trival*)
Lemma isSub_Trans: isSub M R C S SR mS -> isSub M S D U US mU ->
      isSub M S (Class_Trans M R C S SR) U US mU <->
      isSub M R C U (PosSubT_Trans U S R US SR) mU.
Proof. intros. red. red in H0. red in H. subst; split; trivial. Qed.
End X.
End NewSectionNotInHBPaper_isSUB.

Definition mypoint := object PointM PointR1 {| x := 0|}
                             (simplePointClass PointR1 (PosSubT_Refl _)).

Lemma Equation4: forall (S:Type) (Sub: PosSubT S PointR1) (s:S) (i:Z), 
      get _ (simplePointClass S Sub) (set _ (simplePointClass S Sub) s i) = i.
Proof. intros. destruct Sub as [proj rec AX1 AX2 AX3]. simpl.
  rewrite AX1. reflexivity.
Qed. (*That wasn't too hard ;-) *)

Parameter Color:Type.

(*We can't reuse record field names in Coq but since the interfaces are
  intended to be in sub-relationship anyway it's ok to use record extension*)
Record CPointM X := {
  myPointM :> PointM X; (*get : X -> Z;
  set : X -> Z -> X;
  bump : X -> X;*)
  getC: X -> Color;
}.

Record CPointR1 := { pointR1 :> PointR1; c : Color }.

(*As a consequence, here's a simple form of record subtyping*)
Lemma CPointR1_sub: PosSubT CPointR1 PointR1.
Proof.
eapply (Build_PosSubT CPointR1 PointR1 pointR1 (fun cp rp => {| pointR1 := rp; c := c cp|})); intros.
reflexivity. destruct s; simpl; trivial. simpl. reflexivity.
Defined. (*Using "Defined here" enables proof of Equation5_color below*)

(*We also have this -- use in below and implicitly CPointSClass*)
Lemma CPointR1_sub_Color: PosSubT CPointR1 Color.
Proof.
eapply (Build_PosSubT _ _ c (fun p d => {| pointR1 := p; c := d|})); intros.
reflexivity. destruct s; simpl; trivial. simpl; reflexivity.
Defined. (*Again use "Defined"*)

(*This does not yet  support subtyping in the types of the fields*)
(*Since filed names can't be reasoned about in Coq, the real definition can't be done using just
Coq's recored. We need to replicate the encoding*)

Definition EncRecord := PTree.t Type. (*partial map from positive numbers, ie identifiers, to Types*)

Definition SubTPEncRec (S T:EncRecord) :=
  forall i t, T ! i = Some t -> exists s, S ! i = Some s /\ PosSubTProp s t. 
(*Using PosSubT leads to universe inconsistency - maybe use propositional variant throughout?*)

(*In principle, we should probably try to derive propositional variants of
  the subtyping rules above, ie typing rules for PosSubTProp wrt product, Top, refl, trans, ALL,
  Then, SubTPEncRec may even be derivable?*)

(*For many cases, we can fall back on the type-level results using the following three conversion lemmas*) 
Lemma PostST_TypeProp {S T} (X:PosSubT S T): PosSubTProp S T.
Proof. destruct X as [pr rec]. exists pr, rec. split3; trivial. Qed.

(*Trivial, this is just the definition*)
Lemma PostST_Prop_Elim {S T} (X: PosSubTProp S T):
  exists (project: S -> T) (recomb: S -> T -> S), 
  (forall s t, project (recomb s t) = t) /\
  (forall s, recomb s (project s) = s) /\
  (forall s t1 t2, recomb (recomb s t1) t2 = recomb s t2).
Proof. apply X. Qed.

Lemma PostST_PropType {S T} (project: S -> T) (recomb: S -> T -> S)
  (HST:(forall s t, project (recomb s t) = t) /\
  (forall s, recomb s (project s) = s) /\
  (forall s t1 t2, recomb (recomb s t1) t2 = recomb s t2)): PosSubT S T.
Proof. econstructor; apply HST. Qed.

Definition EncPointR1 (x:ident):EncRecord := @PTree.set Type x Z (PTree.empty _).
Definition EncCPointR1 (x c:ident):EncRecord := PTree.set c Color (@PTree.set Type x Z (PTree.empty _)).

Lemma SubTPEncRec_CPointR1_PointR1 x c (XC: x <> c) : SubTPEncRec (EncCPointR1 x c) (EncPointR1 x).
Proof. red; intros. unfold EncPointR1 in H. unfold EncCPointR1. 
  rewrite PTree.gsspec in H. destruct (peq i x).
  + inv H. rewrite PTree.gso, PTree.gss by trivial. exists Z; split; trivial. 
    apply PostST_TypeProp. apply PosSubT_Refl.
  + rewrite PTree.gempty in H; congruence.
Qed.

Lemma PosSubT_Inherit S (SCR: PosSubT S CPointR1): PosSubT S PointR1.
Proof. apply (PosSubT_Trans _ _ _ SCR  CPointR1_sub). Defined.

(*A variant of the same result using Encoded records*)
Lemma PosSubT_InheritEnc S x c (XC:x<>c) (SCR: SubTPEncRec S (EncCPointR1 x c)): SubTPEncRec S (EncPointR1 x).
Proof. red; intros. red in SCR.
  destruct (SubTPEncRec_CPointR1_PointR1 _ _ XC _ _ H) as [s [S1 S2]].
  specialize (SCR _ _ S1) as [r [R1 R2]].
  exists r; split. trivial. clear - S2 R2.
  (*transitivity of PosSubTProp; destructing S1 and R2 before before applying PostST_TypeProp
    avoids the type vs prop problmes/ universe inconsisteny*) 
  destruct (PostST_Prop_Elim R2) as [projA [recA A]]; clear R2.
  apply PostST_PropType in A.
  destruct (PostST_Prop_Elim S2) as [projB [recB B]]; clear S2.
  apply PostST_PropType in B. clear - A B.
  apply PostST_TypeProp. apply (PosSubT_Trans _ s); trivial.
Qed.

Definition simpleCPointClass : Class CPointM CPointR1 := fun S SCR =>
let SR := PosSubT_Inherit S SCR in
let super := simplePointClass S SR in
{| myPointM := super; getC := fun s : S => c (project S CPointR1 SCR s) |}.

Lemma Equation5: forall (S:Type) (Sub: PosSubT S CPointR1) (s:S) (i:Z), 
      get _ (simpleCPointClass S Sub) (set _ (simpleCPointClass S Sub) s i) = i.
Proof. intros. unfold simpleCPointClass. apply Equation4. Qed. (*Indeed*)

(*on color, we have to do the proof manually since the superclass does not have this field*)
Lemma Equation5_color: forall (S:Type) (Sub: PosSubT S CPointR1) (s:S) (i:Z), 
      getC _ (simpleCPointClass S Sub) (set _ (simpleCPointClass S Sub) s i) = c (project _ _ Sub s).
Proof. intros. destruct Sub as [projS recS AX1 AX2 AX3].
 simpl. rewrite AX1. simpl; trivial.
Qed. 

Section BoundedFixpoints. (*ie Section6.1.2*)

Fixpoint fixN {T} n (fa:((T -> T) * T)):T := match n with O => snd fa | S k => fst fa (fixN k fa) end.

Lemma fixN_inner {T} n f a: f (@fixN T n (f, a)) = fixN n (f, f a).
Proof. induction n; simpl. trivial. rewrite IHn; trivial. Qed.

Definition fixN_sound {T} n (fa:((T -> T) * T)) := forall (x y:T), @fixN T n (fst fa, x) = @fixN T n (fst fa, y).

Lemma fixN_sound_AB1 {T} n f a b: @fixN_sound T n (f,a) -> @fixN_sound T n (f,b).
Proof. intros ? x y; simpl. apply H. Qed.

Lemma fixN_sound_AB2 {T} n f a b: @fixN_sound T n (f,a) -> fixN n (f,a) = fixN n (f,b).
Proof. intros ?. apply (H a b). Qed. 

Lemma fix_fixpoint {T} n f a: @fixN_sound T n (f,a) -> @fixN T n (f,a) = f (fixN n (f,a)).
Proof. intros. rewrite fixN_inner; apply H. Qed.

Lemma BoundedInduction {T} f (* a*): forall n (*Not needed: (Sound: @fixN_sound T n (f,a))*) (phi:T->Prop) 
      t (Base: phi t) (Step: forall u, phi u -> phi (f u)),
      phi (@fixN T n (f,t)).
Proof. intros. assert (forall m, (m<=n)%nat -> phi (fixN m (f,t))); [ | apply H; lia].
  induction m; simpl; intros. trivial. apply Step. apply IHm. lia.
Qed.

(*The fact that soundness of fixN is not needed for bounded induction seems weird.
  But maybe we'll want to apply induction somehwere and only exploit soundness of fixN tat the result?*)

End BoundedFixpoints.

Section Section6_3. (*Classes with Self*)

Definition SelfClass (M: Type -> Type) (R:Type): Type :=
  forall S, PosSubT S R -> M S -> M S.

Definition PointSClass: SelfClass PointM PointR1 := fun S SR self =>
match SR with
| {| project := project0; recomb := recomb0 |} =>
    {|
    get := fun s : S => x (project0 s);
    set := fun (s : S) (i : Z) => recomb0 s {| x := i |};
    bump := fun s : S => set S self s (1 + get S self s) |}
end.

(*Variant of simpleCPointClass wrt Self-aware notion of classes*)
Definition simpleCPointSClass : SelfClass CPointM CPointR1 := fun S SCR self =>
let SR := PosSubT_Inherit S SCR in
let super := simplePointClass S SR in
{| myPointM := super; getC := fun s : S => c (project S CPointR1 SCR s) |}.

(*Three properties of simpleCPointSClass: a) get-set from above*)
Lemma simpleEquation5_SClass S Sub self s i:
      get _ (simpleCPointSClass S Sub self) (set _ (simpleCPointSClass S Sub self) s i) = i.
Proof.
 unfold simpleCPointSClass. simpl. destruct Sub as [pr rec AX1 AX2 AX3]; simpl.
 rewrite AX1. reflexivity.
Qed.

(* b) setting x to i leaves the color unchanged *)
Lemma simpleEquation5_SClass_Color S Sub self s i:
      getC _ (simpleCPointSClass S Sub self) (set _ (simpleCPointSClass S Sub self) s i) = 
      c (project _ _ Sub s).
Proof.
  unfold simpleCPointSClass. simpl. destruct Sub as [pr rec AX1 AX2 AX3]; simpl.
  rewrite AX1; reflexivity.
Qed.

(* c) bumping the point also leaves the color unchanged *)
Lemma simpleEquation5_SClass_ColorBump S Sub self s:
      getC _ (simpleCPointSClass S Sub self) (bump _ (simpleCPointSClass S Sub self) s) = 
      c (project _ _ Sub s).
Proof.
  unfold simpleCPointSClass. simpl. destruct Sub as [pr rec AX1 AX2 AX3]; simpl.
  rewrite AX1; reflexivity.
Qed.

(*The self-aware class from the paper that overrided "set" so that it also
  changes the color to blue.*)
Parameter blue:Color.

(*Overriding "set" so that it also changes the color to blue*)
Definition cpointSClass : SelfClass CPointM CPointR1 := fun S SCR  self =>
let super := PointSClass S (PosSubT_Inherit _ SCR) self
in Build_CPointM S
     (Build_PointM S
       (get _ super)
       (fun s i => let Q := PosSubT_Trans S CPointR1 Color SCR CPointR1_sub_Color in
                   recomb S Color Q (set S super s i) blue)
       (bump _ super))
    (fun s => c (project S CPointR1 SCR s)).

(*Again, three simple properties of overrCPointSClass: a) get-set from above*)
Lemma overrEquation5_SClass S Sub self s i:
      get _ (cpointSClass S Sub self) (set _ (cpointSClass S Sub self) s i) = i.
Proof.
 unfold cpointSClass. simpl. destruct Sub as [pr rec AX1 AX2 AX3]; simpl.
 rewrite 2 AX1. reflexivity.
Qed.

(*b) setting x to i has indeed the side effect of setting the color to blue*)
Lemma overrEquation5_SClass_Color S Sub self s i:
      getC _ (cpointSClass S Sub self) (set _ (cpointSClass S Sub self) s i) = blue.
Proof.
  unfold cpointSClass. simpl. destruct Sub as [pr rec AX1 AX2 AX3]; simpl.
 rewrite 2 AX1. reflexivity.
Qed.

(* c) bumping the point also sets the color to blue - see below*)

Section subclassConstruction.
Variable M M': Type -> Type.
Variable R R': Type.
Variable SubRR': PosSubT R' R.
Variable SubMM': forall S, PosSubT S R' -> PosSubT (M' S) (M S).
Variable superClass : SelfClass M R.
Variable build: forall S, PosSubT S R' -> M' S -> M S -> M' S.

Definition subClass: SelfClass M' R' := fun S SR self =>
let super := superClass _ (PosSubT_Trans _ _ _ SR SubRR') (project _ _ (SubMM' _ SR) self)
in (build S SR self super).
End subclassConstruction.

End Section6_3.

Section Section6_4.

Section SubclassReasoningClaim.
Variable M M': Type -> Type.
Variable R R': Type.
Variable SubRR': PosSubT R' R.
Variable SubMM': forall S, PosSubT S R' -> PosSubT (M' S) (M S).
Variable superClass : SelfClass M R.
Variable build: forall S, PosSubT S R' -> M' S -> M S -> M' S.

Variable phi: forall S (SR: PosSubT S R) (m: M S), Prop.
Variable phi': forall S (SR': PosSubT S R') (m: M' S), Prop.
Variable phi'_entails_phi: forall S (SR': PosSubT S R') (m: M' S),
         phi' S SR' m -> phi S (PosSubT_Trans _ _ _ SR' SubRR') (project _ _ (SubMM' _ SR') m). 

(*My guess is that "phi is invariant in the same sense" means this:*)
Variable Phi_is_invariant: forall S (SR: PosSubT S R) (self: M S),
      phi S SR self -> 
      phi S SR (superClass S SR self).

Variable Hyp: forall S (SR': PosSubT S R') (self: M' S) (super: M S),
         phi' S SR' self /\ phi S (PosSubT_Trans _ _ _ SR' SubRR') super ->
         phi' S SR' (build S SR' self super).

Lemma Phi'_is_invariant S (SR': PosSubT S R') (self: M' S) (H: phi' S SR' self):
      phi' S SR' (subClass M M' R R' SubRR' SubMM' superClass build S SR' self).
Proof.  eauto. Qed. 
(*Explicit proof: eapply Hyp; clear Hyp. split. apply H.
  apply Phi_is_invariant; clear Phi_is_invariant.
  apply phi'_entails_phi. apply H.
*)
End SubclassReasoningClaim.

Section FirstExample.

(*Variable m: CPointM CPointR1. (*eg CPointSClass.*)*)
Definition CpointMeth m:= fixN 2 (cpointSClass CPointR1 (PosSubT_Refl _), m).

Lemma Equation6 s m: get _ (CpointMeth m) (bump _ (CpointMeth m) s) = 1 + get _ (CpointMeth m) s.
Proof. reflexivity. Qed.
(*
Definition phi (S:Type) (SubS: PosSubT S PointR1) (m:CPointM CPointR1):Prop:= 
(forall s, get _ m (bump _ m s) = 1 + get _ m s) /\
(forall s, get _ m s = x s) /\
(forall s i, project _ _ CPointR1_sub (set _ m s i) = {| x := i |}).*)

Definition phiSub (S:Type) (SubS: PosSubT S CPointR1) (m:CPointM S):Prop:= 
(forall s, get _ m (bump _ m s) = 1 + get _ m s) /\
(forall s, get _ m s = x (project _ _ SubS s)) /\
(forall s i, set _ m s i = recomb _ _ (PosSubT_Trans _ _ _ SubS CPointR1_sub) s {| x := i |}).


Definition phiSuper (S:Type) (SubS: PosSubT S PointR1) (m:PointM S):Prop:= 
(forall s, get _ m (bump _ m s) = 1 + get _ m s) /\
(forall s, get _ m s = x (project _ _ SubS s)) /\
(forall s i, set _ m s i = recomb _ _ SubS s {| x := i |}).

Lemma MySubMM' : forall S : Type, PosSubT S CPointR1 -> PosSubT (CPointM S) (PointM S).
Proof. intros. 
apply (Build_PosSubT (CPointM S) (PointM S) (myPointM _)
  (fun cm mm => {| myPointM := mm; getC := getC _ cm |})).
+ intros. destruct t; trivial.
+ intros. destruct s; trivial.
+ intros. simpl; trivial.
Defined.

Definition MyBuild : forall S : Type, PosSubT S CPointR1 -> CPointM S -> PointM S -> CPointM S.
Proof. intros S SR CM PM. apply (Build_CPointM S PM (fun (s:S) => getC _ CM s)). Defined.

Lemma Aux1: forall (S : Type) (SR' : PosSubT S CPointR1) (m : CPointM S),
phiSub S SR' m -> phiSuper S (PosSubT_Trans S CPointR1 PointR1 SR' CPointR1_sub) (project (CPointM S) (PointM S) (MySubMM' S SR') m).
Proof. intros. apply H. Qed.

Lemma Aux2: forall (S : Type) (SR : PosSubT S PointR1) (self : PointM S), phiSuper S SR self -> phiSuper S SR (PointSClass S SR self).
Proof. intros. destruct SR as [proj rec AX1 AX2 AX3]. simpl.
  destruct H as [H1 [H2 H3]]. simpl in *. split3; simpl; intros; trivial.
  rewrite H3, AX1, H2; simpl; trivial.
Qed.

Lemma Aux3: forall (S : Type) (SR' : PosSubT S CPointR1) (self : CPointM S) (super : PointM S),
phiSub S SR' self /\ phiSuper S (PosSubT_Trans S CPointR1 PointR1 SR' CPointR1_sub) super -> phiSub S SR' (MyBuild S SR' self super).
Proof. intros. destruct SR' as [proj rec AX1 AX2 AX3].
  destruct H. destruct H as [H1 [H2 H3]]. simpl in *.
  destruct H0 as [K1 [K2 K3]]. simpl in *.
  split3; simpl; intros; trivial.
Qed.

Lemma Equation6_otherProof s m: 
  get _ (CpointMeth m) (bump _ (CpointMeth m) s) = 1 + get _ (CpointMeth m) s.
Proof.
apply (Phi'_is_invariant PointM CPointM PointR1 CPointR1 CPointR1_sub MySubMM' PointSClass MyBuild
  phiSuper phiSub Aux1 Aux2 Aux3 CPointR1).
split3; simpl; intros; trivial.
unfold p_refl; simpl. admit.
unfold r_refl. f_equal. unfold p_refl; simpl. admit.
Abort.
(*
destruct s0; simpl. CPointR1. in *.
 destruct m; simpl. destruct myPointM0. simpl. 
  destruct s0. simpl. unfold get0. unfold set, myPointM. CPointR1; simpl. unfold set.
 reflexivity.
 unfold phiSub in X.  destruct X as [X1 _]. apply X1. Super phiSub).



Lemma FirstClaim: forall (S:Type) (SubS: PosSubT S PointR1) (self: PointM S),
      phi S SubS self -> True.
   


Definition phi (S:Type) (SubS: PosSubT S CPointR1) (m:CPointM CPointR1):Prop.
Check (forall s, get _ m (bump _ m s) = 1 + get _ m s).
Check (forall s, get _ m s = x s). 
Check (forall s i, set _ m s i = {| x := i |}). 

(project _ _ SubS s)) = 1 + get _ m (project _ _ SubS s)).
apply (forall (s:S), get _ m s = 3). (bump _ m (project _ _ SubS s)) = 1 + get _ m s).
pointR1
Lemma Equation6_logic s: get _ CpointMeth (bump _ CpointMeth s) = 1 + get _ CpointMeth s.
Proof. reflexivity. Qed.
 simpl.
Check (cpointSClass CPointR1 (PosSubT_Refl _)).
Check (@fixN (CPointM CPointR1) 2 (cpointSClass CPointR1 (PosSubT_Refl _), m)).
*)
End FirstExample.

Section LateBindingBlue.

Lemma Equation8 (s : CPointR1) (m : CPointM CPointR1): 
      getC _ (CpointMeth m) (bump _ (CpointMeth m) s) = blue.
Proof. reflexivity. Qed.

(*Maybe there's no need to formalize a proof using bounded induction?*)
End LateBindingBlue.
End Section6_4.
(* Search CPointM.
 *)
Definition SubBeh1 S T (Sub:PosSubT S T) (fS: S -> S) (fT: T -> T):=
  forall s, fT (project _ _ Sub s) = project _ _ Sub (fS s).

Definition SubBeh2 S T (Sub:PosSubT S T) (fS: S -> S) (fT: T -> T):=
  forall s, recomb _ _ Sub s (fT (project _ _ Sub s)) = fS s.

Lemma SB2_implies_SB1 S T Sub fS fT: SubBeh2 S T Sub fS fT -> SubBeh1 S T Sub fS fT.
Proof. intros H s. rewrite <- (H s), (Ax1 _ _ Sub); trivial. Qed.

Program Definition CrossBool S:  PosSubT (S * bool) S :=
Build_PosSubT _ _ (fun s => fst s) (fun s t => (t, snd s)) _ _ _.

Lemma CrossSubBeh1 S: SubBeh1 _ _ (CrossBool S) (fun x => (fst x, true)) id. 
Proof. red; intros; simpl; trivial. Qed.

Lemma CrossSubBeh2 S: SubBeh2 _ _ (CrossBool S) (fun x => (fst x, true)) id. 
Proof. red; intros; simpl. f_equal. Abort.

(*
Definition SB1_ID (M:Type -> Type) S T (Sub:PosSubT S T) SubM (fS: S -> M S) (fT: T -> M T):=
  forall s, fT (project _ _ Sub s) = project _ _ SubM (fS s).
Check SB1.*)

Definition PointMSim (X Y:Type) (mX: PointM X) (mY:PointM Y) (Sim: X -> Y -> Prop) :=
   (forall x y, Sim x y -> get _ mX x = get _ mY y) /\ 
   (forall x y z, Sim x y -> Sim (set _ mX x z) (set _ mY y z)) /\
   (forall x y, Sim x y -> Sim (bump _ mX x) (bump _ mY y)).

Definition Method (Args Res X:Type):Type := (X * Args) -> (X * Res).

Definition GetM := Method unit Z.
(*
Definition Method' (Args Res X:Type):Type := (X * Args) -> (option X * Res).
Definition GetFunctor' := Method' unit Z. (ore faithful to PointM? Not quite right*)

Definition SetM := Method Z unit.
Definition BumpM := Method unit unit.

(*
Definition MethodSim {A R} (M: forall X, Method A R X) X Y (Sim:X -> Y -> Prop) := 
  forall a x y, Sim x y -> 
    match M X (x,a), M Y (y, a) with
      (x', rx), (y', ry) => Sim x' y' /\ rx=ry
   end.*)
Definition MethodSim {A R X Y} (Mx: Method A R X) (My: Method A R Y) (Sim:X -> Y -> Prop) := 
  forall a x y, Sim x y -> 
    match Mx (x,a), My (y, a) with
      (x', rx), (y', ry) => Sim x' y' /\ rx=ry
   end.

Definition MyGetZ:GetM Z := fun x => (fst x, fst x).
Definition MyGetZColor:GetM (Z * Color) := fun x => (fst x, fst (fst x)).

Lemma GetSim : MethodSim MyGetZ MyGetZColor (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

Definition MySetZ:SetM Z := fun x => (snd x, tt).
Definition MySetZColor:SetM (Z * Color) := fun x => ((snd x, snd (fst x)), tt).

Lemma SetSim: MethodSim MySetZ MySetZColor (fun z zc => z = fst zc).
Proof. intros i z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

Definition MyBumpZ:BumpM Z:= fun x => (fst x+1, snd x).
Definition MyBumpZColor:BumpM (Z * Color) := fun x => ((fst (fst x)+1, snd(fst x)), snd x).

Lemma BumpSim : MethodSim MyBumpZ MyBumpZColor (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

(*Bump sets color to blue*)
Definition MyBumpZColor':BumpM (Z * Color):= fun x => ((fst (fst x)+1, blue), snd x).

(*That's ok, since the sim relation does not look at the color*)
Lemma BumpSim' : MethodSim MyBumpZ MyBumpZColor' (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

(*Do the same on a product functor:

Definition MultiSim' (l: list (sigT (fun A => sigT (fun R => forall X, (X * A)  -> (X * R)))))
  (X Y:Type) (Sim:X -> Y -> Prop) 
  (e: sigT (fun A => sigT (fun R => forall X, (X * A)  -> (X * R)))) (*(He: In e l)*):Prop.
Proof. destruct e as [A [R HH]].
apply (forall x y, Sim x y -> 
   forall a, match HH _ (x,a), HH _ (y, a) with (x',rx), (y',ry) => Sim x' y' /\ rx=ry 
 end).
Defined.

Definition MultiSim l X Y (Sim:X -> Y -> Prop):Prop := fold_right (fun e p => p /\ MultiSim' l _ _ Sim e) True l.

Definition ApplyTable (l: list (sigT (fun A => sigT (fun R => forall X, (X * A)  -> (X * R)))))
            (X:Type): list (sigT (fun A => sigT (fun R => (X * A)  -> (X * R)))).
Proof.
induction l. apply nil.
clear l. destruct a as [a [r H]].
apply cons. + exists a. exists r. apply (H X). + apply IHl.
Defined.

Inductive Table:list (Type * Type) -> Type :=
  T_nil: Table nil
| T_cons: forall A R T, Method A R -> Table ((A,R)::T).

Definition Table (l: list (Type * Type)):Type -> .


Definition MyZPoint: list (sigT (fun A => sigT (fun R => forall X, (X * A)  -> (X * R)))).
Proof.
apply cons.
+ exists unit. exists Z. intros X [x a]. apply (x,x).   apply (fun x => (fst x, fst x)).
Definition MyGetZColor:GetM (Z * Color) := fun x => (fst x, fst (fst x)).

Lemma GetSim : MethodSim MyGetZ MyGetZColor (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

Definition MySetZ:SetM Z := fun x => (snd x, tt).
Definition MySetZColor:SetM (Z * Color) := fun x => ((snd x, snd (fst x)), tt).

Lemma SetSim: MethodSim MySetZ MySetZColor (fun z zc => z = fst zc).
Proof. intros i z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

Definition MyBumpZ:BumpM Z:= fun x => (fst x+1, snd x).
Definition MyBumpZColor:BumpM (Z * Color) := fun x => ((fst (fst x)+1, snd(fst x)), snd x).

Lemma BumpSim : MethodSim MyBumpZ MyBumpZColor (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.*/
Print MultiSim.
 destruct e as [A [R HH]].
apply (forall x y, Sim x y -> 
   forall a, match HH _ (x,a), HH _ (y, a) with (x',rx), (y',ry) => Sim x' y' /\ rx=ry 
 end).
Defined.
Definition PointMethods:list (sigT (fun A => sigT (fun R => forall X, (X * A)  -> (X * R)))).
eapply cons.
+ exists unit. exists Z. intros. apply (Method unit Z). GetM.
Print MultiSim.
 
Definition MultiSim' (l: list (sigT (fun A => sigT (fun R => forall X, (X * A)  -> (X * R)))))
  (X Y:Type) (Sim:X -> Y -> Prop):Prop.
Proof.
Check fold_right. (fun e p => ) True l).
 destruct e as [A [R HH]].
apply (forall x y, Sim x y -> 
   forall a, match HH _ (x,a), HH _ (y, a) with (x',rx), (y',ry) => Sim x' y' /\ rx=ry 
 end).
Defined.
Print MultiSim.forall a x y, Sim x y -> 
    match Mx (x,a), My (y, a) with
      (x', rx), (y', ry) => Sim x' y' /\ rx=ry
   end end.  

 destruct He. simpl in He.
destruct e as [Mx My].
apply (forall x y, Sim x y -> 
   forall a, match Mx (x,a), My (y, a) with (x',rx), (y',ry) => Sim x' y' end). forall a x y, Sim x y -> 
    match Mx (x,a), My (y, a) with
      (x', rx), (y', ry) => Sim x' y' /\ rx=ry
   end end.  


Definition MultiSim (l: list (((Type * Type)-> (Type * Type)) * ((Type * Type) -> (Type * Type))))
  (X Y:Type) (Sim:X -> Y -> Prop) := 
  forall e, In e l -> match e with (Mx, My) =>
                       forall a x y, Sim x y -> 
    match Mx (x,a), My (y, a) with
      (x', rx), (y', ry) => Sim x' y' /\ rx=ry
   end end.  

(*
Record Method := { Args: Type; Res: Type; f: forall (X:Type), (X * Args) -> (X * Res) }.

Definition GetM := Method unit Z.
(*
Definition Method' (Args Res X:Type):Type := (X * Args) -> (option X * Res).
Definition GetFunctor' := Method' unit Z. (ore faithful to PointM? Not quite right*)

Definition SetM := Method Z unit.
Definition BumpM := Method unit unit.

(*
Definition MethodSim {A R} (M: forall X, Method A R X) X Y (Sim:X -> Y -> Prop) := 
  forall a x y, Sim x y -> 
    match M X (x,a), M Y (y, a) with
      (x', rx), (y', ry) => Sim x' y' /\ rx=ry
   end.*)
Definition MethodSim {A R X Y} (Mx: Method A R X) (My: Method A R Y) (Sim:X -> Y -> Prop) := 
  forall a x y, Sim x y -> 
    match Mx (x,a), My (y, a) with
      (x', rx), (y', ry) => Sim x' y' /\ rx=ry
   end.

Definition MyGetZ:GetM Z := fun x => (fst x, fst x).
Definition MyGetZColor:GetM (Z * Color) := fun x => (fst x, fst (fst x)).

Lemma GetSim : MethodSim MyGetZ MyGetZColor (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

Definition MySetZ:SetM Z := fun x => (snd x, tt).
Definition MySetZColor:SetM (Z * Color) := fun x => ((snd x, snd (fst x)), tt).

Lemma SetSim: MethodSim MySetZ MySetZColor (fun z zc => z = fst zc).
Proof. intros i z [z' c] Hz. simpl in *. subst; split; trivial. Qed.

Definition MyBumpZ:BumpM Z:= fun x => (fst x+1, snd x).
Definition MyBumpZColor:BumpM (Z * Color) := fun x => ((fst (fst x)+1, snd(fst x)), snd x).

Lemma BumpSim : MethodSim MyBumpZ MyBumpZColor (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.*/

/*Bump sets color to blue*/
Definition MyBumpZColor':BumpM (Z * Color):= fun x => ((fst (fst x)+1, blue), snd x).

/*That's ok, since the sim relation does not look at the color*/
Lemma BumpSim' : MethodSim MyBumpZ MyBumpZColor' (fun z zc => z = fst zc).
Proof. intros a z [z' c] Hz. simpl in *. subst; split; trivial. Qed.
*)
(*Do the same on a product functor:*)
Definition MRes := PTree.t Type.
Definition MMeth (X:Type):Type := PTree.t. (X * Args) -> (X * Res).

(*Definition MyPoint(X:Type):Type := (GetM X * SetM X * BumpM X).
Definition MyPointZ: MyPoint Z := (MyGetZ, MySetZ, MyBumpZ).
Definition MyPointZColor: MyPoint (Z * Color) := (MyGetZColor, MySetZColor, MyBumpZColor).*)
Record TableEntry := { ARG: Type; RES:Type; ENTRY: Type -> Type; Hentry: forall X, ENTRY X = Method ARG RES X }.
Definition Table := PTree.t TableEntry.

Definition MkMethod (A R:Type) (E:Type -> Type) (H: forall X, E X = Method A R X) X (m:E X): Method A R X.
rewrite <- H. apply m. Defined.
Definition MkMethod' (TE: TableEntry) X (m:ENTRY TE X): Method (ARG TE) (RES TE) X.
(*destruct TE as [A R E H]; simpl in *. rewrite H in m; apply m. Defined.*)
rewrite (Hentry TE) in m; apply m. Defined. Print MkMethod'.

Definition PointTable (g s b:ident):Table.
Proof.
   apply (PTree.set g). eapply (@Build_TableEntry unit Z GetM). reflexivity.
   apply (PTree.set s). eapply (@Build_TableEntry Z unit SetM). reflexivity.
   apply (PTree.set b). eapply (@Build_TableEntry unit unit BumpM). reflexivity.
   apply PTree.empty.
Defined.
Print PointTable.

Parameter TE: TableEntry.
Definition MTableSim (MT: Table) X Y (Sim:X -> Y -> Prop): Prop.
(*  forall i TE, MT ! i = Some TE ->*) 
eapply (@MethodSim (ARG TE) (RES TE) X Y). 3: apply Sim. eapply (MkMethod' TE X). MkMethod likely wrong. (MkMethod' TE Y)

(*Definition PointTable g s b:PTree.t (Type -> Type):= PTree.set g GetM (PTree.set s SetM (PTree.set b BumpM (PTree.empty _))).*)
(*Definition MyPoint' g s b (X:Type):PTree.t Type:= PTree.set g (GetM X) (PTree.set s (SetM X) (PTree.set b (BumpM X) (PTree.empty _))).
Check GetM.*)
Definition MTableSim (MTable: PTree.t (Type -> Type)) X Y (Sim:X -> Y -> Prop) :=
  forall i entry, MTable ! i = Some entry -> True.

Check (MethodSim MyPointZ MyPointZColor).
Lemma MyPointSim: MethodSim MyPointZ MyPointZColor (fun z zc => z = fst zc).



Lemma GetSim_ZZ_Z : MethodSim 

Definition GetM_Sim:= MethodSim 

Definition MethodProduct {A1 R1 A2 R2 X:Type} (M1 : Method A1 R1 X) (M2: Method A2 R2 X): Method (A1 * A2) (R1 * R2) X.
intros arg. unfold Method. simpl.
  BuildGetFunctor unit Z
Section SubBeh1_implies_PointMSim.
Variable X Y:Type.
Variable Sub:PosSubT X Y.
Variable mX: PointM X.
Variable mY:PointM Y.

Lemma SubBeh1_implies_PointMSim: PointMSim _ _ mX mY (fun x y => y = project _ _ Sub x).
Proof. split3; intros x y; intros;  subst.
+ destruct Sub. simpl.

Check (SubBeh1 _ _ Sub). Check PointM X. Check get.
Check (PointMSim _ _ (Sub:PosSubT X Y) ).
check 
Definition SubBeh1_PointMSim (X Y:Type) (Sub:PosSubT X Y)(mX: PointM X) (mY:PointM Y):
   (SB:SubBeh1 _ _ Sub)  
   PointMSim _ _ (Sub:PosSubT S T) (fS: S -> S) (fT: T -> T):=
  forall s, fT (project _ _ Sub s) = project _ _ Sub (fS s).

Definition GeneralSim (M:Type -> Type) (X Y:Type) (mX:M X) (mY:M Y) (Sim: X -> Y -> Prop) :=
  forall x y, Sim x y -> Sim (mX x) (mY y).

 T (Sub:PosSubT S T) (fS: S -> S) (fT: T -> T):=
  forall s, fT (project _ _ Sub s) = project _ _ Sub (fS s).
Proof. 
Next Obligation. Ax1_prod Ax2_prod Ax3_prod.
Proof.
Next Obligation. simpl.
Program Definition PosSubT_Degenerate (S:Type): PosSubT S S :=
Build_PosSubT _ _ (fun s => s) (fun s t => t) _ _ _. 

Lemma DegBeh1 S: SubBeh1 _ _ (PosSubT_Degenerate S) id id.
Proof. red; intros. simpl. trivial. Qed.
Lemma DegBeh2 S: SubBeh2 _ _ (PosSubT_Degenerate S) id id.
Proof. red; intros. simpl. trivial. Qed.

Lemma SB1_implies_SB2 S T Sub fS fT: SubBeh1 S T Sub fS fT -> SubBeh2 S T Sub fS fT.
Proof. intros H s. destruct Sub as [pi rho AX1 AX2 AX3]. red in H. simpl in *.
rewrite H.
 red in H. rewrite <- (H s), (Ax1 _ _ Sub); trivial. Qed.


Definition SubBeh1 S T (Sub:PosSubT S T) (fS: S -> S) (fT: T -> T):=
  forall s, fT (project _ _ Sub s) = project _ _ Sub (fS s).
Proof. 
Next Obligation. Ax1_prod Ax2_prod Ax3_prod.
*)