Require Export VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for MallocFreeAPD*)

Lemma eqb_list_sym {A} f: forall l1 l2, @eqb_list A f l1 l2 = @eqb_list A (fun x y => f y x) l2 l1.
Proof. induction l1; simpl; intros; destruct l2; simpl; trivial. f_equal; auto. Qed.

Lemma eqb_ident_sym i j: eqb_ident i j = eqb_ident j i.
Proof. apply Pos.eqb_sym. Qed.
Lemma eqb_member_sym: (fun x y : ident * type => eqb_member y x) = eqb_member.
Proof.
  extensionality x. extensionality y. unfold eqb_member.
  rewrite eqb_ident_sym, expr_lemmas4.eqb_type_sym; trivial.
Qed.

Lemma eqb_su_sym a b: eqb_su a b = eqb_su b a.
Proof. destruct a; destruct b; trivial. Qed. 
Lemma eqb_attr_sym a b: eqb_attr a b = eqb_attr b a.
Proof. destruct a; destruct b; simpl; f_equal.
  apply Raux.eqb_sym. unfold eqb_option.
  destruct attr_alignas; destruct attr_alignas0; trivial. apply N.eqb_sym.
Qed.

Lemma test_aux_sym cs1 cs2 b i: test_aux cs1 cs2 b i = test_aux cs2 cs1 b i. 
Proof. unfold test_aux. f_equal.
  destruct ((@cenv_cs cs1) ! i); destruct ((@cenv_cs cs2) ! i); trivial.
  rewrite eqb_list_sym, eqb_su_sym, eqb_member_sym, eqb_attr_sym; trivial.
Qed.

Lemma cs_preserve_type_sym cs1 cs2: forall t CCE, cs_preserve_type cs1 cs2 CCE t = cs_preserve_type cs2 cs1 CCE t. 
Proof. induction t; simpl; trivial; intros; destruct (CCE ! i); trivial; apply test_aux_sym. Qed.

Lemma malloc_token_change_compspecs cs1 cs2 (CCE : change_composite_env cs1 cs2) t
      (CST : cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE) t = true) m (M : MallocTokenAPD):
@malloc_token cs1 M Ews t m = @malloc_token cs2 M Ews t m.
Proof. unfold malloc_token. f_equal.
+ apply pred_ext; apply prop_derives; intros.
  2:{ eapply field_compatible_change_composite. apply CST. apply H. }
  clear - CST H.
  destruct H as [? [? [? [? ?]]]]. split; trivial. repeat split.
  - rewrite (@complete_legal_cosu_type_change_composite _ _ CCE _ CST) in H0; trivial.
  - unfold size_compatible in *.  destruct m; trivial.
    rewrite (@sizeof_change_composite _ _ CCE _ CST) in H1; trivial.
  - unfold align_compatible. destruct m; trivial. simpl in H2.
    apply (@align_compatible_rec_change_composite _ _ CCE _ CST); trivial.
+ rewrite (@sizeof_change_composite _ _ CCE _ CST); trivial.
Qed.

Section ObjectDefs.
Variable T: Type. (*model-level info associated with an object, eg list Z*)

Definition Pred:= (T*val) -> mpred.
(*
Record HOpred:= {
  RP: Pred;
  RP_contr: HOcontractive (fun (_:T*val -> mpred) (x : T*val) => RP x)
}.
*)
Fixpoint funcptrs {cs} (sh:share) (t:type) (meths:list ident) (m:val) (specs: list (Pred -> funspec))(P:Pred): mpred :=
  match specs, meths with 
    nil, nil => emp
  | phi::specs', meth::meths' => ((EX v v', !!(JMeq v v') &&  func_ptr' (phi P) v' * @field_at cs sh t [StructField meth] v m)
                                      * funcptrs sh t meths' m specs' P)%logic
  | _,_ => FF
  end.

Definition MTable {cs} M (t:type) (meths:list ident) (m:val)(specs: list (Pred -> funspec)) P: mpred :=
  (malloc_token M Ews t m * 
  EX sh: share, 
  !! readable_share sh && @funcptrs cs sh t meths m specs P)%logic.

Lemma MTable_is_pointer_or_null {cs} M t meths specs P m (L: Zlength meths <> 0): @MTable cs M t meths m specs  P |-- !!(is_pointer_or_null m).
  Proof. destruct meths. elim L. reflexivity. unfold MTable; Intros sh. entailer!.
(*    destruct specs. apply FF_left. 
    Intros v v'. entailer!. *)
  Qed.
  Hint Resolve MTable_is_pointer_or_null : saturate_local.

Definition FldAt{cs:compspecs} (o:type) (mt:ident) (m:val) (x:T * val):= EX m', !! JMeq m m' && field_at Ews o [StructField mt] m' (snd x).

Lemma funcptrs_change_compspecs cs1 cs2 (CCE : change_composite_env cs1 cs2) t
      (CST: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE) t = true) sh u P: forall specs idents
      (Hidents: forall i, In i idents -> @reptype cs1 (@nested_field_type cs1 t [StructField i]) = @reptype cs2 (@nested_field_type cs2 t [StructField i])),
      @funcptrs cs1 sh t idents u specs P = @funcptrs cs2 sh t idents u specs P.
Proof. 
  induction specs; intros; simpl funcptrs. destruct idents; simpl; trivial.
  destruct idents. simpl; trivial.
  unfold funcptrs.
  f_equal. 2:{ apply IHspecs. intros. apply Hidents. right; trivial. }
  assert (R: @reptype cs1 (@nested_field_type cs1 t [StructField i]) =
             @reptype cs2 (@nested_field_type cs2 t [StructField i])). apply Hidents; left; trivial.
  clear - R CST.
  apply pred_ext.
  + Intros v1 v. destruct (@JMeq_sigT _ _ v1 R) as [v2 V2]. 
    Exists v2 v. entailer!.
    - eapply JMeq_trans. apply JMeq_sym. apply V2. apply H. 
    - apply change_compspecs_field_at_cancel; trivial.
  + Intros v2 v. symmetry in R. destruct (@JMeq_sigT _ _ v2 R) as [v1 V1]. 
    Exists v1 v. apply JMeq_sym in V1. entailer!; clear H1.
    - eapply JMeq_trans. apply V1. apply H. 
    - erewrite (@field_at_change_composite _ _ CCE). apply derives_refl. apply V1. apply CST.
Qed.

Lemma MTable_change_compspecs cs1 cs2 (CCE : change_composite_env cs1 cs2) t
      (CST: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE) t = true) m M P: forall specs idents
      (Hidents: forall i, In i idents -> @reptype cs1 (@nested_field_type cs1 t [StructField i]) = @reptype cs2 (@nested_field_type cs2 t [StructField i])),
      @MTable cs1 M t idents m specs  P = @MTable cs2 M t idents m specs P.
Proof.
  intros. unfold MTable. f_equal.
  + clear - CST. eapply malloc_token_change_compspecs. apply CST. 
  + apply exp_congr. intros sh. rewrite (funcptrs_change_compspecs _ _ CCE _ CST); trivial.
Qed.

Notation StructOf n := (Tstruct n noattr).
(*
(*mt: the field name "mtable"; 
  MT: the struct of the method table eg Tstruct methods noattr
  o: the struct of the object eg Tstruct point noattr*)
Definition ObjectFunctor {cs:compspecs} M (mt:ident) (o MT:type) (meths:list ident) specs (P X:Pred) (x:T*val): mpred :=
  (P x * malloc_token M Ews o (snd x) *
  EX m, !!(isptr m) && FldAt o mt m x *
        MTable M MT meths m specs X)%logic.

Definition ObjectFunctorSub {cs:compspecs} M (mt:ident) (base o MT:type) (meths:list ident) specs (P X:Pred) (x:T*val): mpred :=
  (P x * malloc_token M Ews o (snd x) *
  EX m, !!(isptr m) && FldAt base mt m x *
        MTable M MT meths m specs X)%logic.*)
End ObjectDefs.

Section ObjMpred.
Variable T:Type.
Variable cs:compspecs.
Variable M :MallocFreeAPD.
Variable mt:ident.
Variable base:type.
(*Variable o: type.*)
Variable MT:type.

Variable meths:list ident.
Variable specs : list (Pred T -> funspec).

Definition obj_mpred':= fun x => EX cso, EX P:Pred T, (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
  EX m, !!(isptr m) && FldAt T base mt m x *
        MTable T M MT meths m specs P)%logic.

Definition obj_mpred:= fun x => EX cso, EX P:Pred T, (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
  EX m, FldAt T base mt m x *
        MTable T M MT meths m specs P)%logic.

Lemma obj_mpred_obj_mpred': obj_mpred = obj_mpred'. extensionality x. unfold obj_mpred, obj_mpred'.
  f_equal. extensionality cso. f_equal. extensionality P. f_equal. f_equal. extensionality m.
  apply pred_ext; entailer. unfold MTable. simpl; entailer!.
Qed.

Lemma obj_mpred_fold_unfold: 
obj_mpred = fun x => EX cso, EX P:Pred T, (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
  EX m, !!(isptr m) && FldAt T base mt m x *
        MTable T M MT meths m specs P)%logic.
Proof. rewrite obj_mpred_obj_mpred'. reflexivity. Qed.
Lemma obj_mpred_fold_unfold' x: 
obj_mpred x = EX cso, EX P:Pred T, (P x * @malloc_token (fst cso) M Ews (snd cso) (snd x) *
  EX m, !!(isptr m) && FldAt T base mt m x *
        MTable T M MT meths m specs P)%logic.
Proof. rewrite obj_mpred_obj_mpred'. reflexivity. Qed.

Lemma obj_mpred_isptr t p: obj_mpred (t,p) |-- !!(isptr p).
Proof. rewrite obj_mpred_fold_unfold; Intros cso P m; simpl. entailer!. apply H. Qed.

Lemma obj_mpred_isptr' tp: obj_mpred tp |-- !!(isptr (snd tp)).
Proof. rewrite obj_mpred_fold_unfold; Intros csi P m; entailer!. apply H. Qed.

End ObjMpred.

Lemma obj_mpred_change_compspecs cs1 cs2 (CCE12 : change_composite_env cs1 cs2) T M mt base MT 
      (CST_MT: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) MT = true) 
      (CST_Base: cs_preserve_type cs1 cs2 (@coeq cs1 cs2 CCE12) base = true) specs idents
      (HMT: forall i, In i idents ->
            @reptype cs1 (@nested_field_type cs1 MT [StructField i]) = @reptype cs2 (@nested_field_type cs2 MT [StructField i]))
      (Hbase: @reptype cs1 (@nested_field_type cs1 base [StructField mt]) = @reptype cs2 (@nested_field_type cs2 base [StructField mt])):
      @obj_mpred T cs1 M mt base MT idents specs  = @obj_mpred T cs2 M mt base MT idents specs.
Proof.
  intros. rewrite 2 obj_mpred_fold_unfold. extensionality x. apply exp_congr; intros cso. apply exp_congr; intros P; simpl.
  f_equal.
  apply exp_congr; intros m. rewrite (MTable_change_compspecs _ _ _ CCE12 _ CST_MT _ _ _ _ _ HMT).
  f_equal. f_equal. unfold FldAt. apply pred_ext.
  + Intros m1. specialize (@JMeq_sigT _ _ m1 Hbase) as [m2 M2]. 
    Exists m2. entailer!; clear H1.
    - eapply JMeq_trans. apply H. apply M2.
    - apply change_compspecs_field_at_cancel; trivial.
  + Intros m2. symmetry in Hbase. destruct (@JMeq_sigT _ _ m2 Hbase) as [m1 M1]. 
    Exists m1. entailer!; clear H1.
    - eapply JMeq_trans. apply H. apply M1.
    - erewrite (@field_at_change_composite _ _ CCE12). apply derives_refl. apply JMeq_sym; apply M1. apply CST_Base.
Qed. 