Require Import VST.floyd.proofauto.
Require Import PositiveSubtypingTheory.

Record PointM X := {
  get : X -> Z;
  set : X -> Z -> X
}.

Record BPointM X := {
  PointM_of_BPointM :> PointM X;
  bump : X -> X;
  
  (*we include the "precondition" of bump in this record-maybe rethink*)
  bumpable : X -> Prop; (*Int.min_signed <= x < Int.max_signed*)
}.

Inductive Color:Type := blue | red | green.

Record CPointM X := {
  BPointM_of_CPointM :> BPointM X; 
  getC: X -> Color;

  (*we include the encoding of colors in this record -maybe rethink*)
  color_encoding: Color -> Z
}.

Record PointConstr (ptRep: Type) := {
(*  ptRep: Type;*)
  ptM: PointM ptRep;
  ptConstr: Z -> ptRep
}.

Record BPointConstr (bptRep: Type):= {
(*  bptRep: Type;*)
  bptM: BPointM bptRep;
  bptConstr: Z -> bptRep
}.

Record CPointConstr (cptRep:Type):= {
(*  cptRep: Type;*)
  cptM: CPointM cptRep;
  cptConstr: Z -> Color -> cptRep;
}.

(*** "Model": The expected (semantic) object definitions - shared between iimplementations I1/I2 and SimpleClient**)
(*This is "Rep1" in the theory files, but since we're not using Rep2 much we just call it Rep*)
Record PointRep := { value : Z }.
Definition Init_PointRep (i:Z): PointRep := {|value := i|}.

Definition m1: PointM PointRep := {|
  get := fun s : PointRep => value s;
  set := fun (_ : PointRep) (i : Z) => {| value := i |} |}.

Definition bm1: BPointM PointRep := {|
  PointM_of_BPointM := m1;
  bump := fun s => {| value := value s + 1 |};
  bumpable := fun s => Int.min_signed <= value s < Int.max_signed
|}.

Record CPointRep := { pointRep :> PointRep; color : Color }.

Definition mycolor_encoding (C:Color) :Z :=
  match C with blue => 1 | red => 2 | green => 3 end.

Definition Init_CPointRep (i: Z) (c:Color):= Build_CPointRep (Init_PointRep i) c.

Definition cm1: CPointM CPointRep :=
{| BPointM_of_CPointM := {|
                      PointM_of_BPointM := {|
                                           get := fun s : CPointRep => value s;
                                           set := fun (s : CPointRep) (i : Z) => {| pointRep := {| value := i |}; color := color s |} |};
                      bump := fun s => {| pointRep := {| value := (value s)  + 1 |}; color := color s |};
                      bumpable := fun s => Int.min_signed <= value s < Int.max_signed |};
   getC := color;
   color_encoding := mycolor_encoding |}.

(*As a consequence, here's a simple form of record subtyping*)
Lemma SubT_CpointRep_PointRep: PosSubT CPointRep PointRep.
Proof. 
apply (Build_PosSubT _ _ pointRep (fun c r => {| pointRep := r; color := color c |})); intros; simpl; trivial.
destruct s; simpl; trivial.
Defined. (*Using "Defined here" enables proof of Equation5_color below*)

Lemma SubT_BpointM_PointM: forall S : Type, PosSubT S PointRep -> PosSubT (BPointM S) (PointM S).
Proof. intros. 
apply (Build_PosSubT (BPointM S) (PointM S) (PointM_of_BPointM _)
   (fun bm mm => {| PointM_of_BPointM := mm; bump := bump _ bm; bumpable  := bumpable _ bm|})); intros; simpl; trivial.
destruct s; simpl; trivial.
Defined.

Lemma SubT_CpointM_BPointM: forall S : Type, PosSubT S CPointRep -> PosSubT (CPointM S) (BPointM S).
Proof. intros. 
apply (Build_PosSubT (CPointM S) (BPointM S) (BPointM_of_CPointM _)
   (fun cm bm => {| BPointM_of_CPointM := bm; getC := getC _ cm; color_encoding := color_encoding _ cm |})); intros; simpl; trivial.
destruct s; simpl; trivial.
Defined.

Lemma SubT_CpointM_PointM: forall S : Type, PosSubT S CPointRep -> PosSubT (CPointM S) (PointM S).
Proof. intros. eapply PosSubT_Trans. apply SubT_CpointM_BPointM; trivial.
eapply SubT_BpointM_PointM. eapply PosSubT_Trans. apply X. apply SubT_CpointRep_PointRep.
Defined.

Definition ptConstr1: PointConstr PointRep:= {| ptM := m1; ptConstr := Init_PointRep |}.
Definition bptConstr1: BPointConstr PointRep:= {| bptM := bm1; bptConstr := Init_PointRep |}.
Definition cptConstr1: CPointConstr CPointRep:= {| cptM := cm1; cptConstr := Init_CPointRep |}.
