Require Import VST.floyd.proofauto.
Require Import ASI_malloc. (*needed for mem_mgr*)
Require Import PositiveSubtypingTheory.
Require Import SimpleObjspec.
Require Import SimplePointModel.
Require Import SimplePoints_ASI.

Require Import SimplePoint_I2.
Instance CompSpecs : compspecs. make_compspecs prog. Defined.

Ltac do_compute_expr_warning::=idtac.
Notation point2 := (Tstruct _point_I2 noattr).
Notation bpoint2 := (Tstruct _bpoint_I2 noattr).
Notation cpoint2 := (Tstruct _cpoint_I2 noattr).

Lemma align_compatible_point2_point {b i} (AL: align_compatible point2 (Vptr b i)):
      align_compatible point (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma align_compatible_bpoint2_bpoint {b i} (AL: align_compatible bpoint2 (Vptr b i)):
      align_compatible bpoint (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma align_compatible_cpoint2_cpoint {b i} (AL: align_compatible cpoint2 (Vptr b i)):
      align_compatible cpoint (Vptr b i).
Proof. inv AL. inv H. inv H1. 
      eapply align_compatible_rec_Tstruct; [reflexivity |].
      simpl co_members in *; intros. specialize (H3 i0).
      simpl in H.
      if_tac in H.
      { inv H. specialize (H3 _ _ (eq_refl _) (eq_refl _)).
        eapply align_compatible_rec_by_value. reflexivity.
        inv H0. inv H3. inv H. apply H0. }
      clear H1. congruence.
Qed.

Lemma field_compatible_point2_point_mtable p: field_compatible point2 [StructField _mtable] p ->
              field_compatible point [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_point2_point; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Lemma field_compatible_bpoint2_bpoint_mtable p: field_compatible bpoint2 [StructField _mtable] p ->
              field_compatible bpoint [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_bpoint2_bpoint; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Lemma field_compatible_cpoint2_cpoint_mtable p: field_compatible cpoint2 [StructField _mtable] p ->
              field_compatible cpoint [StructField _mtable] p.
Proof. intros.
  destruct H as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_cpoint2_cpoint; trivial.
  - split; simpl; auto. left; trivial.
Qed.

Definition Pred2: Pred PointRep := fun xp =>EX minint z, 
   (match xp with (r, p) =>
   !!((minint=Vtrue /\ value r = Int.min_signed) \/
     (minint=Vfalse /\ Int.min_signed < value r <= Int.max_signed /\ z=Vint (Int.repr (- value r)))) &&
    (@field_at CompSpecs Ews point2 [StructField _intmin] minint p * 
     @field_at CompSpecs Ews point2 [StructField _negvalue] z p *
     spacer Ews 5 8 p)
   end)%logic.

Lemma dataverif_get: semax_body nil nil f_get_I2 (_get_I2, get_spec point _ m1 Pred2).
Proof.
  start_function. destruct x as [i]. unfold Pred2; simpl. Intros minint x. forward.
    { entailer!. destruct H as [[? ?] | [? ?]]; subst; red; simpl. right; trivial.
      left; trivial. }
    forward_if (PROP ( )
      LOCAL (temp _t'1 (Vint (Int.repr i)))
      SEP (field_at Ews point2 [StructField _intmin] minint p;
           field_at Ews point2 [StructField _negvalue] x p; 
           spacer Ews 5 8 p)).
    { destruct H as [[? ?] | [? ?]]; subst.
      2:{ red in H0; simpl in H0. inv H0. }
      red in H0; simpl in H0. clear H0.
      forward. (*rewrite I2Rep_fold_unfold'; unfold I2_data.
      Intros m; Exists Vtrue x m. entailer!; try cancel. *) entailer!. }
    { destruct H as [[? ?] | [? [? ?]]]; subst.
      1:{ red in H0; simpl in H0. inv H0. }
      red in H0; simpl in H0. clear H0.
      forward. forward.
      { entailer!. rewrite Int.signed_zero; rep_lia. } 
      (*rewrite I2Rep_fold_unfold'; unfold I2_data.
      Intros m; Exists Vfalse (Vint (Int.repr (- i))) m.*)
      entailer!; try cancel. rewrite Z.opp_involutive; trivial. }
    forward.
  { unfold Pred2. Exists minint x. entailer!. }
Qed.

Lemma signed_cases i: Int.min_signed <= i <= Int.max_signed ->
     (i=Int.min_signed) \/ 
     (Int.min_signed <= -i <= Int.max_signed).
Proof. destruct (zeq i Int.min_signed). left; trivial. right. rep_lia. Qed.

Lemma dataverif_set: semax_body nil nil f_set_I2 (_set_I2, set_spec point _ m1 Pred2).
Proof.
  start_function. unfold Pred2; simpl. destruct x as [i]; simpl. unfold Pred2; simpl. Intros minint x. 
    (*rewrite I2Rep_fold_unfold'. Intros.  simpl. unfold I2_data; simpl; Intros minint x.*)
    forward_if (PROP ( False) LOCAL () SEP ()).
    + forward. forward.
      (*rewrite I2Rep_fold_unfold'; unfold I2_data.
      Exists Vtrue x mtable. entailer!; try cancel.*)
      unfold Pred2. Exists (Vint Int.one) x. entailer!.
      specialize (Int.signed_repr _ H). rewrite H1, Int.signed_repr; intros; subst.
      left; split; trivial. rep_lia.
    + forward. forward.
      { entailer!. rewrite Int.signed_zero. simpl in H1.
        change (0-j) with (-j). apply signed_cases in H.
        destruct H; trivial. subst. elim H1. reflexivity. }
      forward. simpl in *. 
      (*rewrite I2Rep_fold_unfold'; unfold I2_data.
      Exists (Vint Int.zero) (Vint (Int.neg (Int.repr j))) mtable.
      entailer!; try cancel.*)unfold Pred2. Exists (Vint Int.zero)  (Vint (Int.neg (Int.repr j))); entailer!. 
      right. split3; trivial.
      clear - H H1. specialize (signed_cases _ H); intros [? | ?].
      - subst. elim H1; trivial.
      - simpl. rep_lia.
    + forward.
Qed.

Definition BPred2: Pred PointRep :=fun xp =>EX minint z, 
   (match xp with (r, p) =>
   !!((minint=Vtrue /\ value r = Int.min_signed) \/
     (minint=Vfalse /\ Int.min_signed < value r <= Int.max_signed /\ z=Vint (Int.repr (- value r)))) &&
    (@field_at CompSpecs Ews bpoint2 [StructField _intmin] minint p * 
     @field_at CompSpecs Ews bpoint2 [StructField _negvalue] z p *
     spacer Ews 5 8 p)
   end)%logic.

Lemma dataverif_bump: semax_body nil nil f_bump_I2 (_bump_I2, bump_spec bpoint _ bm1 BPred2). (*Pred1*)
Proof.
  start_function. destruct x as [i]; simpl. simpl in H.
    unfold BPred2; simpl; Intros minint x.
    forward.
     { entailer!. destruct H0 as [[? ?] | [? [? ?]]]; subst; red; simpl; [ right | left]; trivial. }
    forward_if (PROP ( False) LOCAL () SEP ()).
    + forward. forward. forward. unfold BPred2; simpl. red in H1. simpl in H1. inv H1.
      apply negb_true_iff in H8.
      destruct H0 as [[? ?] | [? [? ?]]]; inv H0.
      2:{ rewrite Int.eq_true in H8; inv H8. }
      (*rewrite I2Rep_fold_unfold'; unfold I2_data.*)
      Exists (Vint Int.zero) (Vint (Int.repr 2147483647)) (*mtable*). entailer!; try cancel.
      right; split3; trivial. lia.
    + forward.
      { entailer!. red in H1. simpl in H1. inv H1.
        apply negb_false_iff in H8. apply Int.same_if_eq in H8; subst minint.
        destruct H0 as [[? ?] | [? [? ?]]]; inv H0.
        trivial. }
      forward.
      { entailer!. red in H1. simpl in H1. inv H1. 
        apply negb_false_iff in H8. apply Int.same_if_eq in H8; subst minint.
        destruct H0 as [[? ?] | [? [? ?]]]; inv H0. inv H7.
        rewrite Int.signed_repr; rep_lia. }
      forward. simpl.
      (*rewrite I2Rep_fold_unfold'; unfold I2_data.*)unfold BPred2.
      Exists (Vint minint) (Vint (Int.sub x (Int.repr 1))) (*mtable*).
      entailer!; try cancel. red in H1. simpl in H1. inv H1. 
      apply negb_false_iff in H9. apply Int.same_if_eq in H9; subst minint.
      destruct H0 as [[? ?] | [? [? ?]]]; inv H0. inv H8. right. split3; trivial. simpl;lia.
      f_equal. rewrite Int.sub_signed, 2 Int.signed_repr by rep_lia. simpl; f_equal; lia.
    + forward.
Qed.

Definition CPred2: Pred CPointRep := fun xp =>EX minint z, 
   (match xp with (r, p) =>
   !!((minint=Vtrue /\ value r = Int.min_signed) \/
     (minint=Vfalse /\ Int.min_signed < value r <= Int.max_signed /\ z=Vint (Int.repr (- value r)))) &&
    (@field_at CompSpecs Ews cpoint2 [StructField _intmin] minint p * 
     @field_at CompSpecs Ews cpoint2 [StructField _negvalue] z p *
     @field_at CompSpecs Ews cpoint2 [StructField _color] (Vint (Int.repr (color_encoding _ cm1 (color (fst xp))))) (snd xp)) *
     spacer Ews 5 8 p
   end)%logic. 

Lemma dataverif_getC: semax_body nil nil f_getC_I2 (_getC_I2, getC_spec cpoint _ cm1 CPred2).
Proof.
  start_function. unfold CPred2; simpl. destruct x as [i]; simpl. Intros minint x. 
  forward. forward. unfold CPred2. Exists minint x. entailer!.
Qed.

(*We might be tempted to verify get wrt the following spec:*)
Lemma objverif_get M: semax_body nil nil f_get_I2
      (_get_I2, get_spec point _ m1 (Point point methods _mtable _get _set CompSpecs M (*DynRep2*) _ m1)).
Proof.
  start_function. rewrite Point_fold_unfold. Intros P m. simpl.
Abort. (*But there's no way we can complete this proof - P is abstract!
  Point - being based on obj_mpred - is a view that hides the C representation/layout of
  a Point1Rep object - spec based on the obj_mpred predicate (eg Point) are 
  hence only of use to for clients.
  The mechanism by which inner (C-layout-aware) specs are related to outer (layout-unaware but
(  semantic)representation-aware) specifications is object creation.*)

Section Compatible_lemmas.
Ltac align_compatible_tac H H3 H0 :=
 repeat (if_tac in H;
      [ inv H; specialize (H3 _ _ (eq_refl _) (eq_refl _));
      eapply align_compatible_rec_by_value; [ reflexivity |];
      inv H0; inv H3; inv H; apply H0 | ]); congruence.

Lemma align_compatible_bpoint_point {b i} (AL:align_compatible bpoint2 (Vptr b i)):
      align_compatible point2 (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H. align_compatible_tac H H3 H0.
Qed.

Lemma field_compatible_bpoint_point_intmin {p} (FC:field_compatible bpoint2 [StructField _intmin] p):
      field_compatible point2 [StructField _intmin] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_bpoint_point; trivial.
  - split; simpl; auto. right; left; trivial.
Qed.

Lemma field_compatible_bpoint_point_nvalue {p} (FC:field_compatible bpoint2 [StructField _negvalue] p):
      field_compatible point2 [StructField _negvalue] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_bpoint_point; trivial.
  - split; simpl; auto. right; right; left; trivial.
Qed.

Lemma field_compatible_bpoint_point_mtable {p} (FC:field_compatible bpoint2 [StructField _mtable] p):
      field_compatible point2 [StructField _mtable] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_bpoint_point; trivial.
  - split; simpl; auto. left; trivial.
Qed.

(*The next three lemmas (inverses of the above three) do not have counterparts in I1, but they're also not used below*)
Lemma align_compatible_point_bpointDEAD {b i} (AL:align_compatible point2 (Vptr b i)):
      align_compatible bpoint2 (Vptr b i).
Proof. intros. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H. align_compatible_tac H H3 H0.
Qed.

Lemma field_compatible_point_bpoint_intminDEAD {p} (FC:field_compatible point2 [StructField _intmin] p):
      field_compatible bpoint2 [StructField _intmin] p.
Proof. intros.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_point_bpointDEAD; trivial.
  - split; simpl; auto. right; left; trivial.
Qed.

Lemma field_compatible_point_bpoint_nvalueDEAD {p} (FC:field_compatible point2 [StructField _negvalue] p):
      field_compatible bpoint2 [StructField _negvalue] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_point_bpointDEAD; trivial.
  - split; simpl; auto. right; right; left; trivial.
Qed.

Lemma align_compatible_cpoint_bpoint {b i} (AL:align_compatible cpoint2 (Vptr b i)):
      align_compatible bpoint2 (Vptr b i).
Proof. inv AL. inv H. inv H1. 
    eapply align_compatible_rec_Tstruct; [reflexivity |].
    simpl co_members in *; intros. specialize (H3 i0).
    simpl in H. align_compatible_tac H H3 H0.
Qed.

Lemma field_compatible_cpoint_bpoint_intmin {p} (FC:field_compatible cpoint2 [StructField _intmin] p):
      field_compatible bpoint2 [StructField _intmin] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_cpoint_bpoint; trivial.
  - split; simpl; auto. right; left; trivial.
Qed.

Lemma field_compatible_cpoint_bpoint_nvalue {p} (FC:field_compatible cpoint2 [StructField _negvalue] p):
      field_compatible bpoint2 [StructField _negvalue] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_cpoint_bpoint; trivial.
  - split; simpl; auto. right; right; left; trivial.
Qed.

Lemma field_compatible_cpoint_bpoint_mtable {p} (FC:field_compatible cpoint2 [StructField _mtable] p):
      field_compatible bpoint2 [StructField _mtable] p.
Proof.
  destruct FC as [Pt [COMPL [SZ [AL LNF]]]].
  destruct p; inv Pt. split3; try reflexivity.
  split3.
  - red; simpl. red in SZ; simpl in SZ. lia.
  - clear - AL. apply align_compatible_cpoint_bpoint; trivial.
  - split; simpl; auto. left; trivial.
Qed.

End Compatible_lemmas.

Require Import spec_surelyMalloc.
Definition Vprog : varspecs. mk_varspecs prog. Defined.

  Definition PointMspecs:= [get_spec; set_spec].
  Definition PointMT T M P m := MTable T M methods [_get; _set] m
           [get_spec point T P; set_spec point T P].

  Definition BPointMT T M (P:BPointM T)  m := MTable T M bmethods [_get; _set; _bump] m
           [get_spec point T P; set_spec point T P; bump_spec bpoint T P].

  Definition CPointMT T M (P:CPointM T)  m := MTable T M cmethods [_get; _set; _bump; _getC] m
           [get_spec point T P; set_spec point T P; bump_spec bpoint T P; getC_spec cpoint T P].

Section ConstructorSpecs.
Variable MF: MallocFreeAPD. 

Lemma make_PointMT:
  forall T P PR g s m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  malloc_token MF Ews (Tstruct _methods noattr) m *
  data_at Ews (Tstruct _methods noattr) (g,s) m
  |-- PointMT T MF P m PR)%logic.
Proof.
  intros.
  unfold PointMT, MTable. simpl.
  Exists Ews; entailer!.
  unfold_data_at (data_at Ews (Tstruct _methods noattr) _ m). cancel.
  Exists s g g s. entailer!. 
Qed.

Lemma make_BPointMT:
  forall T (P:BPointM T) PR g s b m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  func_ptr' (bump_spec bpoint T P PR) b *
  malloc_token MF Ews (Tstruct _bmethods noattr) m *
  data_at Ews (Tstruct _bmethods noattr) (g,(s,b)) m
  |-- BPointMT T MF P m PR)%logic. 
Proof.
  intros.
  unfold BPointMT, MTable. simpl.
  Exists Ews; entailer!.
  unfold_data_at (data_at Ews (Tstruct _bmethods noattr) _ m). cancel.
  Exists b s g g b s. entailer!. 
Qed.

Lemma make_CPointMT:
  forall T (P:CPointM T) PR g s b gC m,
  (func_ptr' (get_spec point T P PR) g *
  func_ptr' (set_spec point T P PR) s *
  func_ptr' (bump_spec bpoint T P PR) b *
  func_ptr' (getC_spec cpoint T P PR) gC *
  malloc_token MF Ews (Tstruct _cmethods noattr) m *
  data_at Ews (Tstruct _cmethods noattr) (g,(s,(b,gC))) m
  |-- CPointMT T MF P m PR)%logic. 
Proof.
  intros.
  unfold CPointMT, MTable. simpl.
  Exists Ews; entailer!.
  unfold_data_at (data_at Ews (Tstruct _cmethods noattr) _ m). cancel.
  Exists gC b s g. Exists g gC b s. entailer!. 
Qed.

  Definition set_internalspec:=
    WITH p:val, i:Z, m:val, x:PointRep
    PRE [ tptr point, tint   ]
     PROP(Int.min_signed <= i <= Int.max_signed)
     PARAMS (p;Vint (Int.repr i)) GLOBALS ()
     SEP(data_at Ews point2 (m, snd (default_val point2)) p)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(field_at Ews point2 [StructField _mtable] m p; (* spacer Ews 5 8 p;*)
          Pred2 (set PointRep m1 x i, p)).

Lemma dataverif_setinternal: semax_body nil nil f_set_I2 (_set_I2, set_internalspec).
Proof.
  start_function. unfold Pred2; simpl.
    (*rewrite I2Rep_fold_unfold'. Intros.  simpl. unfold I2_data; simpl; Intros minint x.*)
    forward_if (PROP ( False) LOCAL () SEP ()).
    + forward. forward.
      (*rewrite I2Rep_fold_unfold'; unfold I2_data.
      Exists Vtrue x mtable. entailer!; try cancel.*)
      unfold Pred2. simpl. Exists (Vint Int.one) Vundef. entailer!.
      left. split. reflexivity.
      specialize (Int.signed_repr _ H). rewrite H0, Int.signed_repr; intros; subst. reflexivity.
      rep_lia.
      unfold_data_at (data_at Ews point2 _ p). cancel.
    + forward. forward.
      { entailer!. rewrite Int.signed_zero. simpl in H0.
        change (0-i) with (-i). apply signed_cases in H.
        destruct H; trivial. subst. elim H0. reflexivity. }
      forward. simpl in *. 
      (*rewrite I2Rep_fold_unfold'; unfold I2_data.
      Exists (Vint Int.zero) (Vint (Int.neg (Int.repr j))) mtable.
      entailer!; try cancel.*)unfold Pred2. Exists (Vint Int.zero)  (Vint (Int.neg (Int.repr i))); entailer!. 
      right. split3; trivial.
      clear - H H0. specialize (signed_cases _ H); intros [? | ?].
      - subst. elim H0; trivial.
      - simpl. rep_lia.
      - unfold_data_at (data_at Ews point2 _ p). cancel.
    + forward.
Qed.

  Definition set_internalspecB:=
    WITH p:val, i:Z, m:val, x:PointRep
    PRE [ tptr point, tint   ]
     PROP(Int.min_signed <= i <= Int.max_signed)
     PARAMS (p;Vint (Int.repr i)) GLOBALS ()
     SEP(data_at Ews bpoint2 (m, snd (default_val bpoint2)) p)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(field_at Ews bpoint2 [StructField _mtable] m p; (* spacer Ews 5 8 p;*)
          BPred2 (set PointRep bm1 x i, p)).

Lemma set_internal_internalB:
      funspec_sub set_internalspec set_internalspecB.
Proof. do_funspec_sub. Exists w emp. destruct w as [[[p i] m] rep]; Intros. clear H; simpl in H1; subst.
  entailer. unfold Pred2, BPred2; simpl.
  unfold_data_at (data_at Ews point2 _ _). 
  unfold_data_at (data_at Ews bpoint2 _ _). entailer!.
  + intros. Exists x0 z. entailer!.
    unfold field_at; entailer.
  + apply field_compatible_bpoint_point_mtable in H2.
    apply field_compatible_bpoint_point_nvalue in H6.
    apply field_compatible_bpoint_point_intmin in H4.
    unfold field_at; entailer.
Qed.

Lemma dataverif_setinternalB: semax_body nil nil f_set_I2 (_set_I2, set_internalspecB).
Proof.
  eapply semax_body_funspec_sub. apply dataverif_setinternal.
  apply set_internal_internalB. apply compute_list_norepet_e; cbv; trivial.
Qed.

 Definition set_internalspecC:=
    WITH p:val, i:Z, m:val, x:CPointRep
    PRE [ tptr point, tint  ]
     PROP(Int.min_signed <= i <= Int.max_signed)
     PARAMS (p;Vint (Int.repr i)) GLOBALS ()
     SEP(data_at Ews cpoint2 (m, (Vundef, (Vundef, Vint (Int.repr (color_encoding _ cm1 (color x)))))) p)
    POST[ tvoid ]
      PROP()
      LOCAL()
      SEP(field_at Ews cpoint2 [StructField _mtable] m p; (* spacer Ews 5 8 p;*)
          CPred2 ({|pointRep := set CPointRep cm1 x i; color:=color x|} , p)).

Lemma set_internalB_internalC:
      funspec_sub set_internalspecB set_internalspecC.
Proof. do_funspec_sub. destruct w as [[[p i] m] rep]; Intros. clear H; simpl in H1; subst.
  destruct rep as [rep1 col]; simpl. Exists (p,i,m,rep1).
  unfold_data_at (data_at Ews cpoint2 _ _).
  Exists (field_at Ews cpoint2 [StructField _color] (Vint (Int.repr (mycolor_encoding col))) p).
  entailer. unfold CPred2, BPred2; simpl.
  unfold_data_at (data_at Ews bpoint2 _ _). entailer!.
  + intros. Exists x0 z. entailer!.
    unfold field_at; entailer.
  + apply field_compatible_cpoint_bpoint_mtable in H.
    apply field_compatible_cpoint_bpoint_nvalue in H4.
    apply field_compatible_cpoint_bpoint_intmin in H2.
    unfold field_at; entailer.
Qed.

Lemma dataverif_setinternalC: semax_body nil nil f_set_I2 (_set_I2, set_internalspecC).
Proof.
  eapply semax_body_funspec_sub. apply dataverif_setinternalB.
  apply set_internalB_internalC. apply compute_list_norepet_e; cbv; trivial.
Qed.

Lemma set_internal_sub1:
       funspec_sub (binary_intersection' set_internalspec (set_spec point _ m1 Pred2) (eq_refl _) (eq_refl _))
                   set_internalspec.
Proof. eapply binary_intersection'_sub. Qed.

Lemma binary_set_internal_sub1:
       funspec_sub (binary_intersection' set_internalspec (set_spec point _ m1 Pred2) (eq_refl _) (eq_refl _))
                   (set_spec point _ m1 Pred2).
Proof. eapply binary_intersection'_sub. Qed.
Lemma binary_set_internal_func_ptr1 p:
  func_ptr' ((binary_intersection' set_internalspec
     (set_spec point PointRep m1 Pred2) eq_refl eq_refl)) p
|-- func_ptr' (set_spec point PointRep m1 Pred2) p.
Proof. apply func_ptr'_mono. apply binary_set_internal_sub1. Qed.

Definition makePoint_I2_spec:= (_makePoint_I2,
           makePoint_spec point methods _mtable _get _set CompSpecs MF _ (*m1 Init_PointRep*)ptConstr1).

(*The Grog contains get/setspec for m1 and Pred1, but makePoint's spec mentions Point!*)
Definition Point2_Gprog : funspecs := 
    ([(_get_I2, get_spec point _ m1 Pred2);
      (_set_I2, binary_intersection' set_internalspec (set_spec point _ m1 Pred2) (eq_refl _) (eq_refl _));
    makePoint_I2_spec] ++ SM_ASI MF).

Lemma body_makePoint_I2: semax_body Vprog Point2_Gprog f_makePoint_I2 makePoint_I2_spec.
Proof.
unfold makePoint_I2_spec, makePoint_spec, ptConstr1.
    (*redundant function_pointers.*)
start_function. 
forward_call surelymalloc_spec_sub (methods, gv).
    { simpl. split3; trivial. rep_lia. }
Intros m.
forward. forward.
 
forward_call surelymalloc_spec_sub (point2, gv).
    { simpl. split3; trivial. rep_lia. }
Intros p.
(*needed for application of make_PointMT 5 lines below make_func_ptr _get_I2.  make_func_ptr _set_I2.*)
forward.
forward_call set_internal_sub1 (p,i,m,{| value := 3 |}).
(*needed for application of make_PointMT 2 lines below*) 
make_func_ptr _get_I2.  make_func_ptr _set_I2. sep_apply (binary_set_internal_func_ptr1).
forward.
Exists p. entailer!.
sep_apply (make_PointMT PointRep m1 Pred2 (gv _get_I2)  (gv _set_I2) m).
rewrite Point_fold_unfold. simpl. Exists (CompSpecs, point2) Pred2 m. entailer!.
unfold FldAt. Exists m; simpl. entailer!.
clear - H0. apply field_compatible_point2_point_mtable in H0. unfold field_at; entailer.
simpl. rewrite sepcon_comm, sepcon_assoc.
apply sepcon_derives; [ | rewrite sepcon_comm]; apply derives_refl.
Qed.

(*Similarly for makeBPoint_I2*)
Definition makeBPoint_I2_spec := (_makeBPoint_I2, 
           makeBPoint_spec point bpoint bmethods _mtable _get _set _bump CompSpecs MF _ (*bm1 Init_PointRep*)bptConstr1).

Lemma set_internal_subB:
       funspec_sub (binary_intersection' set_internalspec (set_spec point _ bm1 BPred2) (eq_refl _) (eq_refl _))
                   set_internalspecB.
Proof. eapply funspec_sub_trans. 2: apply set_internal_internalB. eapply binary_intersection'_sub. Qed.

Lemma binary_set_internal_subB:
       funspec_sub (binary_intersection' set_internalspec (set_spec point _ bm1 BPred2) (eq_refl _) (eq_refl _))
                   (set_spec point _ bm1 BPred2).
Proof. eapply binary_intersection'_sub. Qed.
Lemma binary_set_internal_func_ptrB p:
  func_ptr' ((binary_intersection' set_internalspec
     (set_spec point PointRep bm1 BPred2) eq_refl eq_refl)) p
|-- func_ptr' (set_spec point PointRep bm1 BPred2) p.
Proof. apply func_ptr'_mono. apply binary_set_internal_subB. Qed.

(*Note that we use  get_spec _ bm1 BPred1 rather than get_spec _m1 Pred1)*)
Definition BPoint1_Gprog : funspecs :=   ltac:(with_library prog
    ([(_get_I2, get_spec point _ bm1 BPred2);
      (_set_I2, binary_intersection' set_internalspec(*B*) (set_spec point _ bm1 BPred2)(eq_refl _) (eq_refl _));
      (_bump_I2, bump_spec bpoint _ bm1 BPred2);
    makeBPoint_I2_spec] ++ SM_ASI MF)).

Lemma body_makeBPoint_I2: semax_body Vprog BPoint1_Gprog f_makeBPoint_I2 makeBPoint_I2_spec.
Proof.
unfold makeBPoint_I2_spec, makeBPoint_spec, bptConstr1.
    (*redundant function_pointers.*)
start_function. 
forward_call surelymalloc_spec_sub (bmethods, gv).
    { simpl. split3; trivial. rep_lia. }
Intros m.
forward. forward. forward.

forward_call surelymalloc_spec_sub (bpoint2, gv).
    { simpl. split3; trivial. rep_lia. }
Intros p.
forward.
forward_call set_internal_subB (p,i,m,{| value := 3 |}).
(*needed for application of make_PointMT 2 lines below*) 
make_func_ptr _get_I2.  make_func_ptr _set_I2. sep_apply (binary_set_internal_func_ptrB). make_func_ptr _bump_I2.
forward.
Exists p. entailer!. 
sep_apply (make_BPointMT PointRep bm1 BPred2 (gv _get_I2)  (gv _set_I2) (gv _bump_I2) m).
rewrite BPoint_fold_unfold. simpl. Exists (CompSpecs, bpoint2) BPred2 m. entailer!.
unfold Init_PointRep; cancel.
rewrite sepcon_comm. apply sepcon_derives; [ | apply derives_refl]. 
unfold FldAt. Exists m; simpl. entailer!. 
clear - H0. apply field_compatible_bpoint2_bpoint_mtable in H0. unfold field_at; entailer.
Qed.

Lemma set_internal_subC:
       funspec_sub (binary_intersection' set_internalspec (set_spec point _ cm1 CPred2) (eq_refl _) (eq_refl _))
                   set_internalspecC.
Proof. eapply funspec_sub_trans. 2: apply set_internalB_internalC.
       eapply funspec_sub_trans. 2: apply set_internal_internalB.
       eapply binary_intersection'_sub. 
Qed.

Lemma binary_set_internal_subC:
       funspec_sub (binary_intersection' set_internalspec (set_spec point _ cm1 CPred2) (eq_refl _) (eq_refl _))
                   (set_spec point _ cm1 CPred2).
Proof. eapply binary_intersection'_sub. Qed.
Lemma binary_set_internal_func_ptrC p:
  func_ptr' ((binary_intersection' set_internalspec
     (set_spec point CPointRep cm1 CPred2) eq_refl eq_refl)) p
|-- func_ptr' (set_spec point CPointRep cm1 CPred2) p.
Proof. apply func_ptr'_mono. apply binary_set_internal_subC. Qed.

Definition makeCPoint_I2_spec := (_makeCPoint_I2, 
           makeCPoint_spec point bpoint cpoint cmethods _mtable _get _set _bump _getC CompSpecs MF _ (*cm1 Init_CPointRep*)cptConstr1).

Definition CPoint1_Gprog : funspecs :=   ltac:(with_library prog
    ([(_get_I2, get_spec point _ cm1 CPred2);
      (_set_I2, binary_intersection' set_internalspec(*C*) (set_spec point _ cm1 CPred2) (eq_refl _) (eq_refl _));
      (_bump_I2, bump_spec bpoint _ cm1 CPred2);
      (_getC_I2, getC_spec cpoint _ cm1 CPred2);
    makeCPoint_I2_spec] ++ SM_ASI MF)).

Lemma body_makeCPoint_I2: semax_body Vprog CPoint1_Gprog f_makeCPoint_I2 makeCPoint_I2_spec.
Proof.
unfold makeCPoint_I2_spec, makeCPoint_spec, cptConstr1.
    (*redundant function_pointers.*)
start_function. 
forward_call surelymalloc_spec_sub (cmethods, gv).
    { simpl. split3; trivial. rep_lia. }
Intros m.
forward. forward. forward. forward.
 
forward_call surelymalloc_spec_sub (cpoint2, gv).
    { simpl. split3; trivial. rep_lia. }
Intros p.
forward. forward.
forward_call set_internal_subC (p,i,m,{| pointRep := {| value := 3|}; color := c |}).
(*needed for application of make_PointMT 2 lines below*) 
make_func_ptr _get_I2.  make_func_ptr _set_I2. sep_apply (binary_set_internal_func_ptrC).
make_func_ptr _bump_I2. make_func_ptr _getC_I2.
forward.
Exists p. entailer!. 
sep_apply (make_CPointMT CPointRep cm1 CPred2 (gv _get_I2)  (gv _set_I2) (gv _bump_I2) (gv _getC_I2) m).
rewrite CPoint_fold_unfold. simpl. Exists (CompSpecs, cpoint2) CPred2 m. entailer!. cancel.
rewrite sepcon_comm. apply sepcon_derives; [ | apply derives_refl]. 
unfold FldAt. Exists m; simpl. entailer!. 
clear - H0. apply field_compatible_cpoint2_cpoint_mtable in H0. unfold field_at; entailer.
Qed.

Section Subsumption.
Lemma gsub_point_bpoint pt: funspec_sub (get_spec pt _ m1 Pred2) (get_spec pt _ bm1 BPred2).
Proof. do_funspec_sub. destruct w as [P p]; simpl; Intros. clear H.
unfold BPred2, Pred2; simpl. Intros minint z.
Exists (P, p) emp. Exists minint z. entailer!.
+ intros. Exists x0 x1. entailer!.
  apply sepcon_derives; unfold field_at; simpl; entailer!.
  (*apply field_compatible_point_bpoint_intmin; trivial.
  apply field_compatible_point_bpoint_nvalue; trivial.*)
+ (*Intros minint z. Exists minint z. entailer!.*)
  apply sepcon_derives; unfold field_at; simpl; entailer!.
  apply field_compatible_bpoint_point_intmin; trivial.
  apply field_compatible_bpoint_point_nvalue; trivial.
Qed.

Lemma gsub_bpoint_cpoint pt : funspec_sub (get_spec pt _ bm1 BPred2) (get_spec pt _ cm1 CPred2).
Proof. do_funspec_sub. destruct w as [CP p]; simpl; Intros. clear H.
destruct CP as [PT COL].
unfold CPred2, BPred2; simpl. Intros minint z.
Exists (PT, p) 
       (field_at Ews cpoint2 [StructField _color] (Vint (Int.repr (mycolor_encoding COL))) p).
simpl. apply andp_right.
+ cancel. Exists minint z. entailer!. unfold field_at; simpl.
  entailer!. split.
  apply field_compatible_cpoint_bpoint_intmin; trivial.
  apply field_compatible_cpoint_bpoint_nvalue; trivial.
+ entailer!. intros. Exists x0 x1. unfold field_at; simpl. entailer!.
Qed.

Lemma gsub_point_cpoint pt: funspec_sub (get_spec pt _ m1 Pred2) (get_spec pt _ cm1 CPred2).
Proof. eapply funspec_sub_trans. apply gsub_point_bpoint. apply gsub_bpoint_cpoint. Qed.

Lemma ssub_point_bpoint pt: funspec_sub (set_spec pt _ m1 Pred2) (set_spec pt _ bm1 BPred2).
Proof. do_funspec_sub. destruct w as [[P p] i]; simpl; Intros. clear H. rename H0 into BNDS.
unfold BPred2, Pred2; simpl.
Exists ((P, p), i) emp. Intros minint z. Exists minint z.
entailer!.
+ intros. entailer!. Exists x0 x1. unfold field_at; simpl. entailer!.
+ unfold field_at; simpl. entailer!. split.
  apply field_compatible_bpoint_point_intmin; trivial.
  apply field_compatible_bpoint_point_nvalue; trivial.
Qed.

Lemma ssub_bpoint_cpoint pt: funspec_sub (set_spec pt _ bm1 BPred2) (set_spec pt _ cm1 CPred2).
Proof. do_funspec_sub. destruct w as [[[PT COL] p] c]; simpl; Intros. clear H. rename H0 into BND.
unfold CPred2, BPred2; simpl. Intros intmin z.
Exists ((PT, p), c) 
       (field_at Ews cpoint2 [StructField _color] (Vint (Int.repr (mycolor_encoding COL))) p).
Exists intmin z.
entailer!.
+ intros. entailer!. Exists x0 x1. unfold field_at; simpl. entailer!.
+ unfold field_at; simpl. entailer!. split.
  apply field_compatible_cpoint_bpoint_intmin; trivial.
  apply field_compatible_cpoint_bpoint_nvalue; trivial.
Qed. 

Lemma ssub_point_cpoint pt: funspec_sub (set_spec pt _ m1 Pred2) (set_spec pt _ cm1 CPred2).
Proof. eapply funspec_sub_trans. apply ssub_point_bpoint. apply ssub_bpoint_cpoint. Qed.

Lemma bsub_bpoint_cpoint pt: funspec_sub (bump_spec pt _ bm1 BPred2) (bump_spec pt _ cm1 CPred2).
Proof. do_funspec_sub. destruct w as [[PT COL] p]; simpl; Intros. clear H. rename H0 into BND.
unfold CPred2, BPred2; simpl. Intros intmin z.
Exists (PT, p) 
       (field_at Ews cpoint2 [StructField _color] (Vint (Int.repr (mycolor_encoding COL))) p).
Exists intmin z. entailer!.
+ intros. Exists x0 x1. unfold field_at; simpl. entailer!.
+ unfold field_at; simpl. entailer!. split.
  apply field_compatible_cpoint_bpoint_intmin; trivial.
  apply field_compatible_cpoint_bpoint_nvalue; trivial.
Qed.

End Subsumption.

Definition gspec pt := binary_intersection' 
                   (binary_intersection' (get_spec pt _ m1 Pred2) (get_spec pt _ bm1 BPred2) (eq_refl _) (eq_refl _))
                   (get_spec pt _ cm1 CPred2) (eq_refl _) (eq_refl _).

Definition sspecB:= binary_intersection' (set_spec point _ m1 Pred2) 
                                         (set_spec point _ bm1 BPred2) (eq_refl _) (eq_refl _).
Definition sspecC:= binary_intersection' sspecB (set_spec point _ cm1 CPred2) (eq_refl _) (eq_refl _).

Definition sspec:= binary_intersection' set_internalspec sspecC (eq_refl _) (eq_refl _).

Definition bspec pt:= binary_intersection' (bump_spec pt _ bm1 BPred2) (bump_spec pt _ cm1 CPred2) (eq_refl _) (eq_refl _).

Definition commonGprog:= ltac:(with_library prog
    ([(_get_I2, gspec point);
      (_set_I2, sspec);
      (_bump_I2, bspec bpoint);
      (_getC_I2, getC_spec cpoint _ cm1 CPred2);
    makePoint_I2_spec;
    makeBPoint_I2_spec;
    makeCPoint_I2_spec] ++ (*free_spec' MF _free ::*) SM_ASI MF )).

Lemma body_makeCPoint_I2_common: semax_body Vprog commonGprog f_makeCPoint_I2 makeCPoint_I2_spec.
Proof. eapply VSU.semax_body_subsumespec. apply body_makeCPoint_I2.
  * intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id
        by (apply compute_list_norepet_e; reflexivity). (* by LNR_tac.*)
    remember (initial_world.find_id i CPoint1_Gprog) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; reflexivity |]).
    congruence.
  * intros i; red. 
    remember (initial_world.find_id i CPoint1_Gprog ) as w.
    destruct w; [symmetry in Heqw | solve [trivial]]. simpl in Heqw.
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub. }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub3. apply binary_intersection'_sub.
      eapply funspec_sub_trans with (f2:=sspecC).
      - apply binary_intersection'_sub.
      - apply binary_intersection'_sub. (*
      eapply seplog.funspec_sub_trans with (f2 := (binary_intersection' sspecB sspecC (eq_refl _) (eq_refl _)));
      apply binary_intersection'_sub.*) }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub. }
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.

Lemma body_makeBPoint_I2_common: semax_body Vprog commonGprog f_makeBPoint_I2 makeBPoint_I2_spec.
Proof. eapply VSU.semax_body_subsumespec. apply body_makeBPoint_I2.
  * intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). (* by LNR_tac.*)
    remember (initial_world.find_id i BPoint1_Gprog) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; try reflexivity |]).
    congruence.
  * intros i; red. 
    remember (initial_world.find_id i BPoint1_Gprog ) as w.
    destruct w; [symmetry in Heqw | solve [trivial]]. simpl in Heqw.
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply seplog.funspec_sub_trans
      with (f2 := binary_intersection' (get_spec point _ m1 Pred2) (get_spec point _ bm1 BPred2) (eq_refl _) (eq_refl _));
      apply (binary_intersection'_sub). }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub3. apply binary_intersection'_sub.
      eapply funspec_sub_trans with (f2:=sspecC).
      - apply binary_intersection'_sub.
      - eapply funspec_sub_trans with (f2:=sspecB). apply binary_intersection'_sub.
        apply binary_intersection'_sub. (*eapply seplog.funspec_sub_trans with 
      (f2:=binary_intersection' sspecB sspecC eq_refl eq_refl); apply (binary_intersection'_sub). *)}
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si. apply binary_intersection'_sub. }
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.

Lemma body_makePoint_I2_common: semax_body Vprog commonGprog f_makePoint_I2 makePoint_I2_spec.
Proof. eapply VSU.semax_body_subsumespec. apply body_makePoint_I2.
  * intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). (* by LNR_tac.*)
    remember (initial_world.find_id i Point2_Gprog) as w.
    destruct w; [symmetry in Heqw | simpl; trivial]. simpl in Heqw.
    repeat (if_tac in Heqw; [ subst i; inv Heqw; try reflexivity |]).
    congruence.
  * intros i; red. 
    remember (initial_world.find_id i Point2_Gprog ) as w.
    destruct w; [symmetry in Heqw | solve [trivial]]. simpl in Heqw.
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply seplog.funspec_sub_trans
      with (f2 := binary_intersection' (get_spec point _ m1 Pred2) (get_spec point _ bm1 BPred2) (eq_refl _) (eq_refl _));
      apply (binary_intersection'_sub). }
    if_tac in Heqw; [ | clear H].
    { subst i; inv Heqw.
      eexists; split; [ reflexivity | ].
      apply seplog.funspec_sub_sub_si.
      apply binary_intersection'_sub3. apply binary_intersection'_sub.
      eapply funspec_sub_trans with (f2:=sspecC).
      - apply binary_intersection'_sub.
      - eapply funspec_sub_trans with (f2:=sspecB). apply binary_intersection'_sub.
        apply binary_intersection'_sub. (*
      apply (binary_intersection'_sub).*) }
    repeat (if_tac in Heqw; [ subst i; inv Heqw;
                              eexists; split; [ reflexivity | apply funspec_sub_si_refl]
                            | ]).
    congruence.
Qed.
End ConstructorSpecs.


Require Import VST.floyd.VSU.
Section MakeVSU.
Variable M: MallocFreeAPD.


  Definition I2_Imports := (*[free_spec' M _free] ++*) SM_ASI M.
  Definition I2_InternalSpecs: funspecs:=[(_get_I2, gspec point);
      (_set_I2, sspec);
      (_bump_I2, bspec bpoint);
      (_getC_I2, getC_spec cpoint _ cm1 CPred2);
    makePoint_I2_spec M;
    makeBPoint_I2_spec M;
    makeCPoint_I2_spec M].

  Definition I2_Gprog:funspecs := I2_Imports ++ I2_InternalSpecs.
  Definition I2_ASI := SimplePointASI CompSpecs M _makePoint_I2 _makeBPoint_I2 _makeCPoint_I2.

Lemma dataverif_get_common: semax_body Vprog (commonGprog M) f_get_I2 (_get_I2, gspec point).
Proof. eapply VSU.semax_body_subsumespec.
  + eapply semax_body_funspec_sub. apply dataverif_get. 2: LNR_tac.
    apply binary_intersection'_sub3.
    - apply binary_intersection'_sub3. apply funspec_sub_refl.
      apply gsub_point_bpoint.
    - apply gsub_point_cpoint.
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.


Lemma dataverif_set_common: semax_body Vprog (commonGprog M) f_set_I2 (_set_I2, sspec).
Proof. eapply VSU.semax_body_subsumespec.
+ eapply semax_body_binaryintersection''. apply dataverif_setinternal.
  eapply semax_body_binaryintersection''.
  { eapply semax_body_binaryintersection''. apply dataverif_set.
    eapply semax_body_funspec_sub. apply dataverif_set. apply ssub_point_bpoint.
    apply compute_list_norepet_e; reflexivity. }
    eapply semax_body_funspec_sub. apply dataverif_set. apply ssub_point_cpoint.
    apply compute_list_norepet_e; reflexivity.
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.
 
Lemma dataverif_bump_common: semax_body Vprog (commonGprog M) f_bump_I2 (_bump_I2, bspec bpoint).
Proof. eapply VSU.semax_body_subsumespec.
  + eapply semax_body_funspec_sub. apply dataverif_bump. 2: LNR_tac.
    apply binary_intersection'_sub3.
    - apply funspec_sub_refl.
    - apply bsub_bpoint_cpoint.
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.

Lemma dataverif_getC_common: semax_body Vprog (commonGprog M) f_getC_I2 (_getC_I2, getC_spec cpoint _ cm1 CPred2).
Proof. eapply VSU.semax_body_subsumespec. apply dataverif_getC.
  + intros i; red.
    rewrite 2 semax_prog.make_context_g_char, 2 initial_world.make_tycontext_s_find_id 
        by (apply compute_list_norepet_e; reflexivity). simpl; trivial. (* by LNR_tac.*)
  + intros i; red. simpl; trivial.
Qed.

  Definition I2_Component: @Component NullExtension.Espec Vprog _ 
      nil I2_Imports prog I2_ASI emp I2_InternalSpecs.
  Proof. mkComponent_cautious. 
     - solve_SF_internal dataverif_get_common.
     - solve_SF_internal dataverif_set_common.
     - solve_SF_internal dataverif_bump_common.
     - solve_SF_internal dataverif_getC_common. 
     - solve_SF_internal body_makePoint_I2_common.
     - solve_SF_internal body_makeBPoint_I2_common. 
     - solve_SF_internal body_makeCPoint_I2_common.
Qed.

  Definition I2_VSU: @VSU NullExtension.Espec Vprog _ 
      nil I2_Imports prog I2_ASI emp.
  Proof. eexists; apply I2_Component. Qed.

End MakeVSU.