Require Import VST.floyd.proofauto.
Require Import VST.floyd.library.
Require Import object.

Require Import VST.msl.contractive.

Require Import VST.floyd.Funspec_old_Notation.

Instance CompSpecs : compspecs. make_compspecs prog. Defined.
Definition Vprog : varspecs. mk_varspecs prog. Defined.

Local Open Scope Z.
Local Open Scope logic.

(*Product factoring slightly easier for HOrec construction*)
Definition object_invariant := (list Z * val) -> mpred.
Definition ObjInv : Type:= (list Z * val).

Definition tobject := tptr (Tstruct _object noattr).

Definition reset_spec (instance: object_invariant) :=
  WITH hs:ObjInv
  PRE [ _self OF tobject]
          PROP (isptr (snd hs) (*NEW*))
          LOCAL (temp _self (snd hs))
          SEP (instance hs)
  POST [ tvoid ]
          PROP() LOCAL () SEP(instance (nil, snd hs)).

Definition twiddle_spec (instance: object_invariant) :=
  WITH hs: ObjInv, i: Z
  PRE [ _self OF tobject, _i OF tint]
          PROP (0 < i <= Int.max_signed / 4;
                0 <= fold_right Z.add 0 (fst hs) <= Int.max_signed / 4; 
               isptr (snd hs) (*NEW*))
          LOCAL (temp _self (snd hs); temp _i (Vint (Int.repr i)))
          SEP (instance hs)
  POST [ tint ]
      EX v: Z, 
          PROP(2* fold_right Z.add 0 (fst hs) < v <= 2* fold_right Z.add 0 (i::(fst hs)))
          LOCAL (temp ret_temp (Vint (Int.repr v))) 
          SEP(instance (i::(fst hs), snd hs)).
Eval compute in (reptype (Tstruct _methods noattr)).
Definition object_methods (instance: object_invariant) (mtable: val) : mpred :=
  EX sh: share, EX reset: val, EX twiddle: val, EX d, 
  !! readable_share sh && 
  func_ptr' (reset_spec instance) reset *
  func_ptr' (twiddle_spec instance) twiddle *
  data_at sh (Tstruct _methods noattr) (reset,(twiddle, d)) mtable.

Lemma object_methods_contr P Q x m: |> P x <=> |> Q x
|-- object_methods P m >=>
    object_methods Q m.
Proof. unfold object_methods.
apply subp_exp. intros sh.
apply subp_exp. intros r.
apply subp_exp. intros t. apply subp_sepcon.
(*Andrew's definition*)
Definition Andrews_object_mpred: object_invariant := fun (hs:ObjInv) =>
  EX instance: object_invariant, EX mtable: val, 
       (object_methods instance mtable *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs) *
     instance hs).

Require Import VST.progs.conclib.
Require Import VST.concurrency.semax_conc.
Section MkObjMpred.
Variable instance: object_invariant. (*what Andrew really meant: predicate on datarep part, ie P_pre A for the data-only APD A*)
(*should be this, ie with later
  Definition F (X: ObjInv -> mpred) (hs: ObjInv): mpred :=
   ((EX mtable: val, !!(isptr mtable) &&
     ( |> object_methods X mtable ) *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs)) *
   instance hs)%logic.
Lemma HOcontrF: (*HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => instance x)
      -> *)HOcontractive F.
Proof. 
apply prove_HOcontractive; intros ?? (?, ?).
    unfold F. simpl.
    eapply predicates_hered.derives_trans, eqp_subp.
    eapply predicates_hered.derives_trans. (*
    2:{ admit. (*apply nonexpansive_entail.
        apply sepcon_nonexpansive. apply const_nonexpansive. apply identity_nonexpansive.*) }
    apply eqp_exp. intros m. apply eqp_sepcon; [ | apply eqp_refl].
    unfold MTableEWS.
    rewrite ! later_sepcon.  apply eqp_sepcon; [ | apply eqp_refl].
    apply funcptrs_contr.*)
Admitted.*)

Definition F (X: ObjInv -> mpred) (hs: ObjInv): mpred :=
   ((EX mtable: val, 
     ((*?|>?*) object_methods X mtable) *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs)) *
   instance hs)%logic.
Lemma HOcontrG: HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => instance x)
      -> HOcontractive F.
Proof. intros InstanceContr. (*
assert (predicates_rec.HOcontractive G). 2: apply H.
unfold G.
    apply prove_HOcontractive; intros ?? (?, ?). simpl.
 (*   unfold F_RD. simpl. (*t_lock_pred_uncurry, t_lock_pred_r'.*)*)
    eapply predicates_hered.derives_trans, eqp_subp.
    eapply predicates_hered.derives_trans.
    2:{ apply nonexpansive_entail.
        apply sepcon_nonexpansive. apply const_nonexpansive. apply identity_nonexpansive. }
    (*eapply predicates_hered.derives_trans. apply predicates_hered.derives_refl.*)
    apply eqp_exp. intros m. apply eqp_sepcon; [ | apply eqp_refl].
    unfold MTableRD.
    rewrite ! later_exp'; try apply Tsh.
    apply eqp_exp; intros sh. rewrite ! later_andp.  apply eqp_andp. apply eqp_refl.
    apply funcptrs_contr.
intros InstanceContr
assert (predicates_rec.HOcontractive F). 2: apply H.
unfold F.
eapply (@Trashcan.sepcon_HOcontractive compcert_rmaps.RML.R.rmap).
1: apply compcert_rmaps.RML.R.Sep_rmap. (*or use the other lemma?*)
2: apply InstanceContr.
1:{ specialize (@Trashcan.exp_HOcontractive compcert_rmaps.RML.R.rmap
                compcert_rmaps.RML.R.ag_rmap ObjInv val). intros.
  specialize (H (fun m oi => oi)). simpl in H.
Admitted. (*
  specialize (H (fun m oi P => P)). simpl in H.  oi * 
       field_at Ews (Tstruct _object noattr) [StructField _mtable] m)). (snd x)
  remember (predicates_hered.pred compcert_rmaps.RML.R.rmap) as mp.
  admit. (*Check object_methods.
  specialize (H (fun m OI => OI) (fun m OI P => P)). object_methods OI m)).
  mpred Iveric).*)
+ unfold instance.*)*)

apply prove_HOcontractive; intros.
eapply derives_trans. apply (allp_instantiate _ x).
unfold F. apply subp_sepcon.
+ apply subp_exp. intros m. apply subp_sepcon.
  - (*eapply derives_trans. 2: apply subp_later1.
   rewrite subp_later.  unfold object_methods. simpl.*) admit. (*maybe need later here*)
  - apply subp_refl.
+ apply subp_refl.
Admitted.

Definition obj_mpred:ObjInv -> mpred := (HORec F). (*ie same type as Andrew's object_mpred.*)
(*Check (HORec_fold_unfold _ F HOcontrF).*)

Lemma ObjMpred_fold_unfold: 
HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => instance x) ->
obj_mpred = 
fun hs => 
  ((EX mtable: val, 
     object_methods obj_mpred mtable *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs)) *
   instance hs)%logic.
Proof.
  intros; unfold obj_mpred at 1.
  rewrite HORec_fold_unfold. fold obj_mpred. [ reflexivity | apply HOcontrF]; trivial.
Qed.
Lemma ObjMpred_fold_unfold' hs: 
HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => instance x) ->
obj_mpred hs = 
  ((EX mtable: val, 
     object_methods obj_mpred mtable *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs)) *
   instance hs)%logic.
Proof.
  intros. rewrite ObjMpred_fold_unfold, <- ObjMpred_fold_unfold; trivial. 
Qed.
End MkObjMpred.

Definition object_mpred: object_invariant := fun hs =>
  EX instance, !!(HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => instance x)) && obj_mpred instance hs.
(*This now plays the role of Andrews_obj_mpred*)

Definition foo_data : object_invariant :=
  (fun hs =>
    withspacer Ews (sizeof size_t + sizeof tint) (2 * sizeof size_t) (field_at Ews (Tstruct _foo_object noattr) 
            [StructField _data] (Vint (Int.repr (2*fold_right Z.add 0 (fst hs))))) (snd hs)
      *  malloc_token Ews (Tstruct _foo_object noattr) (snd hs)).
(*
Definition foo_reset_spec :=
 DECLARE _foo_reset (reset_spec foo_invariant).

Definition foo_twiddle_spec :=
 DECLARE _foo_twiddle  (twiddle_spec foo_invariant).
*)
Lemma foo_data_HOcontr: HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => foo_data x).
Proof.
  assert (predicates_rec.HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => foo_data x)). 2: apply H.
  unfold foo_data.
  unfold withspacer; simpl.
  apply Trashcan.sepcon_HOcontractive.
  apply Trashcan.const_HOcontractive.
  apply Trashcan.const_HOcontractive.
Qed.

Definition foo_obj_invariant :object_invariant := obj_mpred foo_data.

(*New lemma!*)
Lemma foo_obj_invariant_fold_unfold: foo_obj_invariant =
 fun hs =>
  ((EX mtable: val, 
     object_methods foo_obj_invariant  mtable *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs)) *
   foo_data hs)%logic.
Proof.
  unfold foo_obj_invariant.
  rewrite <- ObjMpred_fold_unfold. trivial. apply foo_data_HOcontr.
Qed.

(*Sometimes this variant is preferable, sometimes the one above*)
Lemma foo_obj_invariant_fold_unfold' hs: foo_obj_invariant hs =
  ((EX mtable: val, 
     object_methods foo_obj_invariant  mtable *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs)) *
   foo_data hs)%logic.
Proof. rewrite foo_obj_invariant_fold_unfold. rewrite <- foo_obj_invariant_fold_unfold; trivial. Qed.

Definition foo_reset_spec :=
 DECLARE _foo_reset (reset_spec foo_obj_invariant).

Definition foo_twiddle_spec :=
 DECLARE _foo_twiddle  (twiddle_spec foo_obj_invariant).

Definition make_foo_spec :=
 DECLARE _make_foo
 WITH gv: globals
 PRE [ ]
    PROP () LOCAL (gvars gv) 
    SEP (mem_mgr gv; object_methods foo_obj_invariant (gv _foo_methods))
 POST [ tobject ]
    EX p: val, PROP () LOCAL (temp ret_temp p)
     SEP (mem_mgr gv; object_mpred (nil,p); object_methods foo_obj_invariant (gv _foo_methods)).

Definition main_spec :=
 DECLARE _main
  WITH gv: globals
  PRE  [] main_pre prog tt gv
  POST [ tint ]
     EX i:Z, PROP(0<=i<=6) LOCAL (temp ret_temp (Vint (Int.repr i))) SEP(TT).

Definition Gprog : funspecs :=  ltac:(with_library prog [
    foo_reset_spec; foo_twiddle_spec; make_foo_spec; main_spec]).

Lemma object_methods_local_facts: forall instance p,
  object_methods instance p |-- !! isptr p.
Proof.
intros.
unfold object_methods.
Intros sh reset twiddle.
entailer!.
Qed.
Hint Resolve object_methods_local_facts : saturate_local.
(*
Lemma Andrews_object_mpred_i:
  forall (history: list Z) (self: val) (instance: object_invariant) (mtable: val),
    object_methods instance mtable *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable self *
     instance history self 
    |-- Andrews_object_mpred history self.
Proof.
intros. unfold Andrews_object_mpred. Exists instance mtable; auto.
Qed.
*)
Lemma foo_data_isptr hs: foo_data hs = !!(isptr (snd hs)) && foo_data hs.
apply pred_ext; entailer.
unfold foo_data. entailer!. destruct (snd hs); simpl in *; trivial;  contradiction.
Qed.

Lemma body_foo_reset: semax_body Vprog Gprog f_foo_reset foo_reset_spec.
Proof.
unfold foo_reset_spec. rewrite foo_obj_invariant_fold_unfold. (*foo_obj_invariant. rewrite ObjMpred_fold_unfold.*)
unfold (*foo_invariant,*) reset_spec. (*Intros. rewrite foo_invariant_isptr. *)
start_function. 
destruct hs as [history self]; simpl. Intros m. unfold foo_data. Intros. (*NEW*)
unfold withspacer; simpl; Intros.
forward.  (* self->data=0; *)
entailer!.
Exists m. 
unfold foo_data.
all: unfold withspacer; simpl; entailer!.  (* needed if Archi.ptr64=true *)
Qed.

Lemma body_foo_twiddle: semax_body Vprog Gprog f_foo_twiddle foo_twiddle_spec.
Proof.
unfold foo_twiddle_spec. rewrite foo_obj_invariant_fold_unfold. 
unfold (*foo_invariant,*) twiddle_spec.
start_function.
destruct hs as [history self]; simpl. Intros m. unfold foo_data. Intros. (*NEW*)
unfold withspacer; simpl.
Intros. 
forward.  (* d = self->data; *)

(*Invocation of statically-bound function call foo_reset(self)*)
(*assert_PROP (isptr self) by entailer!.*) simpl in H0, H1.
forward_call (history, self).
{ rewrite foo_obj_invariant_fold_unfold'. Exists m. cancel.
  unfold foo_data. simpl. unfold withspacer. simpl. cancel. }
simpl. clear m. 

(*Invocation of virtual method call (dynamic binding)*)
rewrite foo_obj_invariant_fold_unfold'; Intros m. 
unfold foo_data. simpl. unfold withspacer. simpl. Intros.
forward. (*  mtable = p->mtable; *)
unfold object_methods at 1.
Intros sh r0 t0.
forward. (* r = m->reset; *)
forward_call  (@nil Z, self). (* invocation r(self); *)
{ rewrite foo_obj_invariant_fold_unfold'. Exists m. cancel.
  unfold foo_data. simpl. unfold withspacer. simpl. cancel.
  unfold object_methods. Exists sh r0 t0. entailer!. }
(* WAS: (*Finish the method-call by regathering the object p back together *)
sep_apply (make_object_methods sh instance r0 t0 mtable0); auto.
sep_apply (object_mpred_i [] p instance mtable0).*)
deadvars!. clear m r0 t0 H2 sh.

rewrite foo_obj_invariant_fold_unfold'. Intros m. simpl.
unfold foo_data. Intros.
unfold withspacer; simpl.
forward.  (* self -> data = d+2*i; *) 
{ set (j:= Int.max_signed / 4) in *; compute in j; subst j.
 forget (fold_right Z.add 0 history) as h. entailer!. }
forward.  (* return d+i; *)
{ set (j:= Int.max_signed / 4) in *; compute in j; subst j.
  forget (fold_right Z.add 0 history) as h.
  entailer!. }
Exists (2 * fold_right Z.add 0 history + i).
Exists m.
simpl;
entailer!.
unfold foo_data. unfold withspacer; simpl.
entailer!.
rewrite Z.mul_add_distr_l, Z.add_comm. cancel. 
Qed.

Lemma split_object_methods:
  forall instance m, 
    object_methods instance m |-- object_methods instance m * object_methods instance m.
Proof.
intros.
unfold object_methods.
Intros sh reset twiddle.

Exists (fst (slice.cleave sh)) reset twiddle.
Exists (snd (slice.cleave sh)) reset twiddle.
rewrite (split_func_ptr' (reset_spec instance) reset) at 1.
rewrite (split_func_ptr' (twiddle_spec instance) twiddle) at 1.
entailer!.
split.
apply slice.cleave_readable1; auto.
apply slice.cleave_readable2; auto.
rewrite (data_at_share_join (fst (slice.cleave sh)) (snd (slice.cleave sh)) sh).
auto.
apply slice.cleave_join.
Qed.

Lemma body_make_foo: semax_body Vprog Gprog f_make_foo make_foo_spec.
Proof.
unfold make_foo_spec.
start_function.
forward_call (Tstruct _foo_object noattr, gv).
   split3; simpl; auto; computable.
Intros p.
forward_if
  (PROP ( )
   LOCAL (temp _p p; gvars gv)
   SEP (mem_mgr gv;
          malloc_token Ews (Tstruct _foo_object noattr) p;
          data_at_ Ews (Tstruct _foo_object noattr) p;
          object_methods foo_obj_invariant (gv _foo_methods))).
*
change (Memory.EqDec_val p nullval) with (eq_dec p nullval).
if_tac; entailer!.
*
forward_call 1.
contradiction.
*
rewrite if_false by auto.
Intros.
forward.  (*  /*skip*/;  *)
entailer!.
*
unfold data_at_, field_at_, default_val; simpl.
forward. (* p->mtable = &foo_methods; *)
forward. (* p->data = 0; *)
forward. (* return (struct object * ) p; *)
Exists p.
sep_apply (split_object_methods foo_obj_invariant (gv _foo_methods)).
entailer!.
unfold object_mpred. 
Exists foo_data. entailer!. 1: solve [apply foo_data_HOcontr].
rewrite ObjMpred_fold_unfold by (apply foo_data_HOcontr).
Exists (gv _foo_methods). simpl.
rewrite ! sepcon_assoc. apply sepcon_derives; [ apply derives_refl |].
unfold foo_data; simpl. unfold withspacer; simpl.
cancel.
unfold_data_at (field_at _ _ nil _ p).
cancel.
rewrite !field_at_data_at.
simpl.
apply derives_refl'.
rewrite <- ?sepcon_assoc. (* needed if Archi.ptr64=true *)
rewrite !field_compatible_field_address; auto with field_compatible.
clear - H.
(* TODO: simplify the following proof. *)
destruct p; try contradiction.
destruct H as [AL SZ].
repeat split; auto.
simpl in *.  unfold sizeof in *; simpl in *; lia.
eapply align_compatible_rec_Tstruct; [reflexivity |].
simpl co_members; intros.
simpl in H.
if_tac in H; [| inv H].
inv H. inv H0.
eapply align_compatible_rec_by_value.
reflexivity.
rewrite Z.add_0_r.
simpl.
unfold natural_alignment in AL.
eapply Z.divide_trans; [ | apply AL].
apply prove_Zdivide.
reflexivity.
left; auto.
Qed.


Lemma make_object_methods:
  forall sh instance reset twiddle mtable,
  readable_share sh ->
  func_ptr' (reset_spec instance) reset *
  func_ptr' (twiddle_spec instance) twiddle *
  data_at sh (Tstruct _methods noattr) (reset, twiddle) mtable
  |-- object_methods instance mtable.
Proof.
  intros.
  unfold object_methods.
  Exists sh reset twiddle.
  entailer!.
Qed.

(*TODO: adapt tactic
Ltac method_call witness hist' result :=
repeat apply seq_assoc1;
match goal with 
   |- semax _ (PROPx _ (LOCALx ?Q (SEPx ?R))) 
            (Ssequence (Sset ?mt (Efield (Ederef (Etempvar ?x _)  _) _ _))
                 _) _  =>
    match Q with context [temp ?x ?x'] =>
     match R with context [object_mpred _ x'] =>
          let instance := fresh "instance" in let mtable := fresh "mtable" in
          unfold object_mpred; Intros instance mtable;
          forward;
          unfold object_methods at 1; 
          let sh := fresh "sh" in let r := fresh "r" in let t := fresh "t" in
          Intros sh r t;
          forward;
          forward_call witness;
          [ .. | try Intros result;
                  sep_apply (make_object_methods sh instance r t mtable); [ auto .. | ];
                  sep_apply (object_mpred_i hist' x' instance mtable);
                  deadvars; try clear dependent sh; try clear r; try clear t
           ]
    end end
end.*)

Lemma body_main:  semax_body Vprog Gprog f_main main_spec.
Proof.
start_function.
sep_apply (create_mem_mgr gv).
(* assert_gvar _foo_methods. (* TODO: this is needed for a field_compatible later on *) *)
fold noattr cc_default.
(* 0. This part should be handled automatically by start_function *)
gather_SEP (mapsto _ _ _ _) (data_at _ _ _ _);
replace_SEP 0 (data_at Ews (Tstruct _methods noattr) 
   (gv _foo_reset, gv _foo_twiddle) (gv _foo_methods)). {
  entailer!.
  unfold_data_at (data_at _ (Tstruct _methods _) _ (gv _foo_methods)).
  rewrite <- mapsto_field_at with (gfs := [StructField _twiddle]) (v:= (gv _foo_twiddle))
  by  auto with field_compatible.
  rewrite field_at_data_at.  rewrite !field_compatible_field_address by auto with field_compatible.
  rewrite !isptr_offset_val_zero by auto.
  cancel.
}

(* 1. Prove that [mtable] is a proper method-table for foo-objects *)

make_func_ptr _foo_twiddle.
make_func_ptr _foo_reset.
sep_apply (make_object_methods Ews foo_obj_invariant (gv _foo_reset) (gv _foo_twiddle) (gv _foo_methods)); auto.

(* 2. Build an instance of class [foo], called [p] *)
forward_call (* p = make_foo(); *)
        gv.
Intros p.
assert_PROP (p<>Vundef) by entailer!.
(*(* Illustration of an alternate method to prove the method calls.
   Method 1:  comment out lines AA and BB and the entire range CC-DD.
   Method 2:  comment out lines AA-BB, inclusive.
*)

(* AA *) try (tryif 
  (method_call (p, @nil Z) (@nil Z) whatever;
   method_call (p, 3, @nil Z) [3%Z] i;
     [simpl; computable | ])
(* BB *)  then fail else fail 99)
  .
*)
(* CC *)
(* 4. first method-call *)
unfold object_mpred. Intros instance. rename H0 into HOContr.
(*Need HOContr several times in the rest of this proof.*)
rewrite ObjMpred_fold_unfold by trivial.
Intros mtable0. simpl.
assert_PROP (isptr p) by entailer!.
forward. (*  mtable = p->mtable; *)
unfold object_methods at 1.
Intros sh r0 t0.
forward. (* p_reset = mtable->reset; *)
forward_call (* p_reset(p); *)
      (@nil Z, p).
{ rewrite ObjMpred_fold_unfold' by trivial. cancel. simpl. Exists mtable0. cancel.
  (*unfold foo_data. simpl. unfold withspacer. simpl. cancel.*)
  unfold object_methods. Exists sh r0 t0. entailer!. }
clear t0 H1 sh.
(* Finish the method-call by regathering the object p back together *)
Intros sh r t. simpl. repeat sep_apply func_ptr'_emp.
 (*NEW: Our object-level spec already did the packing - but let's still to remove the fun_ptrs*) 

(*sep_apply (object_mpred_i [] p instance mtable0).*)
deadvars!. clear - HOContr H1 H0.

(* 5. second method-call *)
rewrite ObjMpred_fold_unfold by trivial.
Intros mtable0. simpl.
forward.  (* mtable = p->mtable; *)
unfold object_methods at 1.
Intros sh0 r0 t0.
forward.   (* p_twiddle = mtable->twiddle; *)
(*assert_PROP (p<>Vundef) by entailer!.*)
forward_call (* i = p_twiddle(p,3); *)
      ((@nil Z, p), 3).
{ rewrite ObjMpred_fold_unfold' by trivial. cancel. simpl. Exists mtable0. cancel.
  (*unfold foo_data. simpl. unfold withspacer. simpl. cancel.*)
  unfold object_methods. Exists sh0 r0 t0. entailer!. }
{ simpl. split3; trivial; computable. }
Intros i.
simpl in H2; simpl.
(*sep_apply (make_object_methods sh instance r0 t0 mtable0); auto.
sep_apply (object_mpred_i [3] p instance mtable0).*)
deadvars!.
(*simpl in H1.*)

(* DD *)

(* 6. return *)
forward.  (* return i; *)
Exists i; entailer!.
Qed.

Definition reset_APIspec (instance: object_invariant) :=
  WITH hs:ObjInv
  PRE [ _self OF tobject]
          PROP (isptr (snd hs) (*NEW*))
          LOCAL (temp _self (snd hs))
          SEP (instance hs)
  POST [ tvoid ] EX instance': object_invariant,
          PROP() LOCAL () SEP(instance' (nil, snd hs)).

Definition twiddle_APIspec (instance: object_invariant) :=
  WITH hs: ObjInv, i: Z
  PRE [ _self OF tobject, _i OF tint]
          PROP (0 < i <= Int.max_signed / 4;
                0 <= fold_right Z.add 0 (fst hs) <= Int.max_signed / 4; 
               isptr (snd hs) (*NEW*))
          LOCAL (temp _self (snd hs); temp _i (Vint (Int.repr i)))
          SEP (instance hs)
  POST [ tint ]
      EX instance': object_invariant,
      EX v: Z, 
          PROP(2* fold_right Z.add 0 (fst hs) < v <= 2* fold_right Z.add 0 (i::(fst hs)))
          LOCAL (temp ret_temp (Vint (Int.repr v))) 
          SEP(instance' (i::(fst hs), snd hs)).

Lemma fs_sub_reset inst: funspec_sub (reset_spec inst) (reset_APIspec inst).
Proof.
  do_funspec_sub. Exists w emp. unfold convertPre; Intros; entailer!.
  intros. Exists inst. entailer!.
Qed.
Lemma fs_sub_twiddle inst: funspec_sub (twiddle_spec inst) (twiddle_APIspec inst).
Proof.
  do_funspec_sub. Exists w emp. unfold convertPre; Intros; entailer!.
  intros. destruct w as [hs i]. Exists inst. entailer!. Exists x0. entailer!.
Qed.

Definition foo_reset_APIspec :=
 DECLARE _foo_reset (reset_APIspec foo_obj_invariant).

Definition foo_twiddle_APIspec :=
 DECLARE _foo_twiddle  (twiddle_APIspec foo_obj_invariant).

Lemma fs_sub_fooreset: funspec_sub (snd foo_reset_spec) (snd foo_reset_APIspec).
Proof.
  unfold foo_reset_spec, foo_reset_APIspec. simpl snd. apply fs_sub_reset.
Qed.
Lemma fs_sub_footwiddle: funspec_sub (snd foo_twiddle_spec) (snd foo_twiddle_APIspec).
Proof.
  unfold foo_twiddle_spec, foo_twiddle_APIspec. simpl snd. apply fs_sub_twiddle.
Qed.
(*
Definition object_methodsFunctor rspec tspec (mtable: val) : mpred :=
  EX sh: share, EX reset: val, EX twiddle: val,
  !! readable_share sh && 
  func_ptr' rspec reset *
  func_ptr' tspec twiddle *
  data_at sh (Tstruct _methods noattr) (reset,twiddle) mtable.

Lemma omF_mono rspec rspec' tspec tspec' (R: funspec_sub rspec rspec')  (T: funspec_sub tspec tspec') m:
object_methodsFunctor rspec tspec m |-- object_methodsFunctor rspec' tspec' m.
Proof.
  apply exp_derives; intros sh. apply exp_derives; intros r. apply exp_derives; intros t.
  entailer!. apply sepcon_derives; apply func_ptr'_mono; trivial.
Qed.*)

Definition object_methodsAPI (instance: object_invariant) (mtable: val) : mpred :=
  EX sh: share, EX reset: val, EX twiddle: val,
  !! readable_share sh && 
  func_ptr' (reset_APIspec instance) reset *
  func_ptr' (twiddle_APIspec instance) twiddle *
  data_at sh (Tstruct _methods noattr) (reset,twiddle) mtable.


Section MkObjMpredAPI.
Variable instance: object_invariant. (*what Andrew really meant: predicate on datarep part, ie P_pre A for the data-only APD A*)
Definition F_API (X: ObjInv -> mpred) (hs: ObjInv): mpred :=
   ((EX mtable: val, 
     ((*?|>?*) object_methodsAPI instance mtable) *
     field_at Ews (Tstruct _object noattr) [StructField _mtable] mtable (snd hs)) *
   instance hs)%logic.

Lemma HOcontrFAPI: HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => instance x)
      -> HOcontractive F_API.
Admitted.

Definition obj_mpredAPI:ObjInv -> mpred := (HORec F_API).
End MkObjMpredAPI.

Definition object_mpredAPI: object_invariant := fun hs =>
  EX instance, !!(HOcontractive (fun (_ : ObjInv -> mpred) (x : ObjInv) => instance x)) && obj_mpredAPI instance hs.

Definition make_foo_APIspec :=
 DECLARE _make_foo
 WITH gv: globals
 PRE [ ]
    PROP () LOCAL (gvars gv) 
    SEP (mem_mgr gv; object_methodsAPI foo_obj_invariant (gv _foo_methods))
 POST [ tobject ]
    EX p: val, PROP () LOCAL (temp ret_temp p)
     SEP (mem_mgr gv; object_mpredAPI (nil,p); object_methodsAPI foo_obj_invariant (gv _foo_methods)).


Definition GprogAPI : funspecs :=  (* ltac:(with_library prog*) ( [
    foo_reset_APIspec; foo_twiddle_APIspec; make_foo_APIspec; main_spec]).
Lemma body_main':  semax_body Vprog GprogAPI f_main main_spec.
Proof.
start_function.
sep_apply (create_mem_mgr gv).
(* assert_gvar _foo_methods. (* TODO: this is needed for a field_compatible later on *) *)
fold noattr cc_default.
(* 0. This part should be handled automatically by start_function *)
gather_SEP (mapsto _ _ _ _) (data_at _ _ _ _);
replace_SEP 0 (data_at Ews (Tstruct _methods noattr) 
   (gv _foo_reset, gv _foo_twiddle) (gv _foo_methods)). {
  entailer!.
  unfold_data_at (data_at _ (Tstruct _methods _) _ (gv _foo_methods)).
  rewrite <- mapsto_field_at with (gfs := [StructField _twiddle]) (v:= (gv _foo_twiddle))
  by  auto with field_compatible.
  rewrite field_at_data_at.  rewrite !field_compatible_field_address by auto with field_compatible.
  rewrite !isptr_offset_val_zero by auto.
  cancel.
}

(* 1. Prove that [mtable] is a proper method-table for foo-objects *)

make_func_ptr _foo_twiddle.
make_func_ptr _foo_reset.
(*sep_apply (make_object_methods Ews foo_obj_invariant (gv _foo_reset) (gv _foo_twiddle) (gv _foo_methods)); auto.*)
gather_SEP 0 1 2.
replace_SEP 0 (object_methodsAPI foo_obj_invariant (gv _foo_methods)).
{ entailer!. unfold object_methodsAPI. Exists Ews (gv _foo_reset) (gv _foo_twiddle). entailer!. }

(* 2. Build an instance of class [foo], called [p] *)
forward_call (* p = make_foo(); *)
        gv.
Intros p.
assert_PROP (p<>Vundef) by entailer!.
(*(* Illustration of an alternate method to prove the method calls.
   Method 1:  comment out lines AA and BB and the entire range CC-DD.
   Method 2:  comment out lines AA-BB, inclusive.
*)

(* AA *) try (tryif 
  (method_call (p, @nil Z) (@nil Z) whatever;
   method_call (p, 3, @nil Z) [3%Z] i;
     [simpl; computable | ])
(* BB *)  then fail else fail 99)
  .
*)
(* CC *)
(* 4. first method-call *)
unfold object_mpred. Intros instance. rename H0 into HOContr.
(*Need HOContr several times in the rest of this proof.*)
rewrite ObjMpred_fold_unfold by trivial.
Intros mtable0. simpl.
assert_PROP (isptr p) by entailer!.
forward. (*  mtable = p->mtable; *)
unfold object_methods at 1.
Intros sh r0 t0.
forward. (* p_reset = mtable->reset; *)
forward_call (* p_reset(p); *)
      (@nil Z, p).
{ rewrite ObjMpred_fold_unfold' by trivial. cancel. simpl. Exists mtable0. cancel.
  (*unfold foo_data. simpl. unfold withspacer. simpl. cancel.*)
  unfold object_methods. Exists sh r0 t0. entailer!. }
clear t0 H1 sh.
(* Finish the method-call by regathering the object p back together *)
Intros sh r t. simpl. repeat sep_apply func_ptr'_emp.
 (*NEW: Our object-level spec already did the packing - but let's still to remove the fun_ptrs*) 

(*sep_apply (object_mpred_i [] p instance mtable0).*)
deadvars!. clear - HOContr H1 H0.

(* 5. second method-call *)
rewrite ObjMpred_fold_unfold by trivial.
Intros mtable0. simpl.
forward.  (* mtable = p->mtable; *)
unfold object_methods at 1.
Intros sh0 r0 t0.
forward.   (* p_twiddle = mtable->twiddle; *)
(*assert_PROP (p<>Vundef) by entailer!.*)
forward_call (* i = p_twiddle(p,3); *)
      ((@nil Z, p), 3).
{ rewrite ObjMpred_fold_unfold' by trivial. cancel. simpl. Exists mtable0. cancel.
  (*unfold foo_data. simpl. unfold withspacer. simpl. cancel.*)
  unfold object_methods. Exists sh0 r0 t0. entailer!. }
{ simpl. split3; trivial; computable. }
Intros i.
simpl in H2; simpl.
(*sep_apply (make_object_methods sh instance r0 t0 mtable0); auto.
sep_apply (object_mpred_i [3] p instance mtable0).*)
deadvars!.
(*simpl in H1.*)

(* DD *)

(* 6. return *)
forward.  (* return i; *)
Exists i; entailer!.
Qed.
  destruct w as [hist self]; simpl in *. inv H. inv H9. simpl in *.
Print ObjInv.
 Print object_invariant. Exists ( emp.