Require Import VST.floyd.proofauto.
Require Import VST.floyd.linking.
Require Import SimplePoint_I1.
Require Import SimplePoint_I2.
(*
(*duplicate of lemma in globals_lemas*)
Lemma prog_defs_Clight_mkprogram:
 forall c g p m w,
 prog_defs (Clightdefs.mkprogram c g p m w) = g.
Proof.
intros. unfold Clightdefs.mkprogram.
destruct ( build_composite_env' c w).
reflexivity.
Qed.

Lemma prog_types_Clight_mkprogram:
  forall (c : list composite_definition) (g : list (ident * globdef Clight.fundef type)) (p : list ident) 
    (m : ident) (w : wf_composites c), prog_types (Clightdefs.mkprogram c g p m w) = c.
Proof. intros. unfold prog_types. unfold Clightdefs.mkprogram.
destruct (build_composite_env' c w ); trivial.
Qed. 
(* Now, to use NEW_LINK_PROGS, it is unfortunately necessary to do this in two steps: *)
Module NEW_LINK_PROGS.  (* Everything in this Module should perhaps be moved to floyd/linking.v *)

(* All of this complexity is because the naturally computed proof whose type is
     build_composite_env t12 = Errors.OK e12
  blows up:  the nested environments explode exponentially.
 And that's a pity, because after all we have  proof irrelevance.  But I could not
 think of a better way than this to exploit proof irrelevance.  -- Andrew, 7/24/2020
*)
Definition carefully_link_progs (prog1 prog2 : Clight.program) 
  (MAIN: prog_main prog1 = prog_main prog2)
  (d12: list (ident * globdef (fundef function) type))
  (Hd12: merge_global_definitions 
               (SortGlobdef.sort (prog_defs prog1)) (SortGlobdef.sort (prog_defs prog2)) = Errors.OK d12)
  (t12: list composite_definition) 
  (Ht12: merge_prog_types (SortComp.sort (prog_types prog1)) (SortComp.sort (prog_types prog2)) = Errors.OK t12)
  (e12: composite_env)
  (He12: build_composite_env t12 = Errors.OK e12)
  : Clight.program := 
 {| prog_defs := d12;
    prog_public := SortPos.merge (SortPos.sort (prog_public prog1)) (SortPos.sort (prog_public prog2));
    prog_main := prog_main prog2;
    prog_types := t12;
    prog_comp_env := e12;
    prog_comp_env_eq := He12|} .

Lemma Gt_neq_Lt: Gt = Lt -> False.
Proof. congruence. Qed.

Lemma prove_exists_align_attr:
  forall d, two_power_nat (Z.to_nat (Z.log2 d)) = d ->
    exists n, align_attr noattr d = two_power_nat n.
Proof.
intros.
exists (Z.to_nat (Z.log2 d)).
rewrite H. reflexivity.
Qed.

Lemma prove_align_attr:
  forall i j,  (j/i)*i=j -> (align_attr noattr i | j).
Proof. intros. exists (j/i). symmetry. apply H. Qed. 

Ltac process_composite_definitions_step := 
repeat
match goal with
|- Errors.bind match ?z with _ => _ end  _  = _ => 
  set (j := z); hnf in j; simpl in j; subst j; cbv beta iota
end;
 match goal with |- context [Ctypes.composite_of_def_obligation_1 _ _ _ _] =>
   set (x := Ctypes.composite_of_def_obligation_1 _ _ _ _);
  simpl in x;
  match type of x with ?t => 
    replace x with (Gt_neq_Lt : t) by apply proof_irr
 end; 
  clear x
end;
 match goal with |- context [Ctypes.composite_of_def_obligation_2 _ _ _] =>
   set (x := Ctypes.composite_of_def_obligation_2 _ _ _);
  simpl in x;
  match type of x with (exists i, align_attr noattr ?d = two_power_nat i) => 
    replace x with (prove_exists_align_attr d (eq_refl _)) by apply proof_irr
 end; 
  clear x
end;
repeat
  match goal with |- context [Ctypes.composite_of_def_obligation_3 _ _ _ _] =>
   set (x := Ctypes.composite_of_def_obligation_3 _ _ _ _);
  simpl in x;
  match type of x with (align_attr noattr ?i | ?j)  => 
    replace x with (prove_align_attr i j (eq_refl _)) by apply proof_irr
 end; 
  clear x
end;
change (Errors.bind (Errors.OK ?x) ?f) with (f x); cbv beta iota.

Ltac process_composite_definitions :=
 simpl;
 unfold build_composite_env; 
 unfold add_composite_definitions, composite_of_def; 
 simpl align; simpl align_attr; simpl rank_members;
 simpl PTree.set;
 repeat process_composite_definitions_step;
 reflexivity.

Ltac do_merge_global_definitions := 
match goal with |- context [SortGlobdef.sort ?x] =>
 set (j :=SortGlobdef.sort x); hnf in j; simpl in j; subst j
end;
match goal with |- context [SortGlobdef.sort ?x] =>
 set (j :=SortGlobdef.sort x); hnf in j; simpl in j; subst j
end;
match goal with |- ?A = _ =>
 set (j :=A); hnf in j; simpl in j; subst j
end; reflexivity.

Ltac do_merge_prog_types := 
match goal with |- context [SortComp.sort ?x] =>
 set (j :=SortComp.sort x); hnf in j; simpl in j; subst j
end;
match goal with |- context [SortComp.sort ?x] =>
 set (j :=SortComp.sort x); hnf in j; simpl in j; subst j
end;
match goal with |- ?A = _ =>
 set (j :=A); hnf in j; simpl in j; subst j
end; reflexivity.  

Ltac do_link_progs_step1 p1 p2 := 
  eapply (carefully_link_progs p1 p2 (eq_refl _));
  [time "merge_global" do_merge_global_definitions
  |time "merge_types" do_merge_prog_types
  |time "process_composites" process_composite_definitions].

Ltac do_merge_global_definitions_unfold p1 p2 :=
  unfold p1; try rewrite prog_defs_Clight_mkprogram;
  unfold p2; try rewrite prog_defs_Clight_mkprogram;
  do_merge_global_definitions.

Ltac do_merge_prog_types_unfold p1 p2 :=
 unfold p1; try rewrite prog_types_Clight_mkprogram;
 unfold p2; try rewrite prog_types_Clight_mkprogram;
  do_merge_prog_types.

Ltac do_link_progs_step1_unfold p1 p2 := 
  eapply (carefully_link_progs p1 p2 (eq_refl _));
  [time "merge_global" do_merge_global_definitions_unfold p1 p2
  |time "merge_types" do_merge_prog_types_unfold p1 p2
  |time "process_composites" process_composite_definitions].

Ltac do_link_progs_step2 p := 
let x := eval hnf in p in
match x with
 {| prog_defs := ?d;
    prog_public := ?p;
    prog_main := ?m;
    prog_types := ?t;
    prog_comp_env := ?e;
    prog_comp_env_eq := _ |} =>
refine  {| prog_defs := d;
    prog_public := p;
    prog_main := m;
    prog_types := t;
    prog_comp_env := e;
    prog_comp_env_eq := _ |} 
end;
abstract (exact (prog_comp_env_eq p)).

End NEW_LINK_PROGS.
 Now, to use NEW_LINK_PROGS, it is unfortunately necessary to do this in two steps *)

Time   (* 0.6s + 0.04s + 27 = 34s*)
Definition prog_I1_I2_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold SimplePoint_I1.prog SimplePoint_I2.prog).

Time  (* 0.076 seconds *)
Definition prog_I1_I2 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2  prog_I1_I2_step1).

Require Import SimpleClient.

Time
Definition with_client_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold prog_I1_I2 SimpleClient.prog).
(*4.8 + 4.5 + 27 = 43*)

Definition with_client : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2 with_client_step1).

Require Import Wrapper. 
Time
Definition with_wrapper_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold with_client Wrapper.prog).
(*5 + 4.4 + 30.9 = 47*)

Definition with_wrapper : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2 with_wrapper_step1).

Require Import PointList. 
Time
Definition with_pointlist_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold with_wrapper PointList.prog).
(*5.6 + 4.9 + 41 = 59*)

Definition with_pointlist : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2 with_pointlist_step1).

Require Import WrapperClient. 
Time
Definition with_wrapperclient_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold with_pointlist WrapperClient.prog).
(*5.7 + 4.9 + 40 = 58*)

Definition with_wrapperclient : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2 with_wrapperclient_step1).

Require Import surelyMalloc. 
Time
Definition with_sm_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold with_wrapperclient surelyMalloc.prog).
(*5.1 + 4.2 + 29 = 44*)

Definition application_prog : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2 with_sm_step1).

Require Import MMlib_prog.
Definition applib_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold mmlib_prog application_prog).
(*18+14+43*)
Definition applib : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2 applib_step1).

Require Import main.
Definition withmain_step1 : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step1_unfold applib main.prog).
(*10+9+36*)
Definition withmain : Clight.program :=
  ltac:(NEW_LINK_PROGS.do_link_progs_step2 withmain_step1).

Instance p12_cs : compspecs. make_compspecs  prog_I1_I2. Defined.
Instance with_client_cs: compspecs. make_compspecs with_client. Defined.
Instance with_wrapper_cs: compspecs. make_compspecs with_wrapper. Defined.
Instance with_wrapperclient_cs: compspecs. make_compspecs with_wrapperclient. Defined.
Instance with_pointlist_cs: compspecs. make_compspecs with_pointlist. Defined.
Instance application_cs: compspecs. make_compspecs application_prog. Defined.
Instance applib_cs: compspecs. make_compspecs applib. Defined.
Instance with_main_cs: compspecs. make_compspecs withmain. Defined.

Definition p12_Vprog:varspecs. mk_varspecs prog_I1_I2. Defined.
Definition with_client_Vprog: varspecs. mk_varspecs with_client. Defined.
Definition with_wrapper_Vprog: varspecs. mk_varspecs with_wrapper. Defined.
Definition with_wrapperclient_Vprog: varspecs. mk_varspecs with_wrapperclient. Defined.
Definition with_pointlist_Vprog: varspecs. mk_varspecs with_pointlist. Defined.
Definition application_Vprog: varspecs. mk_varspecs application_prog. Defined.
Definition applib_Vprog: varspecs. mk_varspecs applib. Defined.
Definition with_main_Vprog: varspecs. mk_varspecs withmain. Defined.

